C
C ********************************************************************
C *         S                                                        *
C *         !                                                        *
C *         !                                                        *
C * F1--<---!---<---F2    SUBROUTINE FFCS                            *
C *                                                                  *
C *     Purpose: to calculate vertex amplitudes for scalar boson-    *
C *              massive fermions vertex.                            *
C *                                                                  *
C *      -            T             T -        T                     *
C *      U(II,AAM,Q) C (AL*PL+AR*PR)  U(I,AM,P)                      *
C *                                                                  *
C *                                                                  *
C *                                                                  *
C *     I,II=: indices to specify fermion(I,II=3) or antifermion     *
C *            (I,II=1) state.                                       *
C *     AM,AAM=: masses of fermions.                                 *
C *     AL,AR=: coupling constants for vertex.                       *
C *     C,CC=: phase factors for massive fermions.                   *
C *     P1,P2,Q1,Q2=: light-like vectors decomposed by subroutine    *
C *                   SPLTQ.                                         *
C *     AALL=: calculated results of vertex amplitudes for all       *
C *            possible helicity states.                             *
C *                                                                  *
C *                                                                  *
C *                                      written by H. Tanaka        *
C ********************************************************************
C
C
C       ===========================================================
        SUBROUTINE FFCS(II,I,AAM,AM,AL,AR,CC,C,Q1,Q2,P1,P2,AALLL)
C       ===========================================================
C
        IMPLICIT REAL*8(A-H,O-Z)
        COMPLEX*16 AALL,AALLL,C,CC
        DIMENSION P1(4),P2(4),Q1(4),Q2(4),C(2),CC(2)
        DIMENSION AALL(2,2),AALLL(2,2)

        I1 = I - 2
        I1 = -I1
        I1 = I1 + 2

        CALL FFS(II,I1,AAM,AM,AL,AR,CC,C,Q1,Q2,P1,P2,AALL)

        AALLL(1,1) = -AALL(1,1)
        AALLL(1,2) = -AALL(1,2)
        AALLL(2,1) = -AALL(2,1)
        AALLL(2,2) = -AALL(2,2)

        RETURN
        END
