C ******************************************************************    
C *                                                                *    
C *                                                                *    
C *   SUBROUTINE FFFF(N:I*4, AJM1:C*16(4,2,2), AJM2:C*16(4,2,2),   *    
C *  &                EM:R*(4), AALL:C*16(2,2,2,2))                *    
C *                                                                *    
C *                                                                *    
C *       Purpose: To calculate fermion-fermion interactions       *    
C *                mediated by vector boson.                       *    
C *                                                                *    
C *                                                                *    
C *       N=: No. of polarization states for intermadiated         *    
C *           vector boson                                         *    
C *       AJM1,AJM2=: Vertex amplitudes for fermion-fermion-       *    
C *           vector boson vertices calculated by subroutine FFV.  *    
C *       EM=: Weight factors to reconstruct the numerator of      *    
C *            vector boson propagator.                            *    
C *       AALL=: Calculated results for all possible helicity      *    
C *              states.                                           *    
C *                                                                *    
C *                                                                *    
C *                                       written by H. Tanaka     *    
C ******************************************************************    
C                                                                       
C                                                                       
C              ====================================                     
               SUBROUTINE FFFF(N,AJM1,AJM2,EM,AALL)                     
C              ====================================                     
C                                                                       
C                                                                       
        IMPLICIT REAL*8(A-H,O-Z)                                        
        COMPLEX*16 AALL,AJM1,AJM2                                       
        DIMENSION AJM1(4,2,2),AJM2(4,2,2),AALL(2,2,2,2),EM(4)           
CX      REAL*16 EE                                                      
                                                                        
        DO 10 J3=1,2                                                    
        DO 10 J4=1,2                                                    
        AALL(1,1,J3,J4)=0.D0                                            
        AALL(1,2,J3,J4)=0.D0                                            
        AALL(2,1,J3,J4)=0.D0                                            
        AALL(2,2,J3,J4)=0.D0                                            
        EE=1.D0                                                         
        DO 10 L=1,N                                                     
        EE=EM(L)                                                        
        AALL(1,1,J3,J4)=AALL(1,1,J3,J4)+AJM1(L,1,1)*AJM2(L,J3,J4)*EE    
        AALL(1,2,J3,J4)=AALL(1,2,J3,J4)+AJM1(L,1,2)*AJM2(L,J3,J4)*EE    
        AALL(2,1,J3,J4)=AALL(2,1,J3,J4)+AJM1(L,2,1)*AJM2(L,J3,J4)*EE    
        AALL(2,2,J3,J4)=AALL(2,2,J3,J4)+AJM1(L,2,2)*AJM2(L,J3,J4)*EE    
   10   CONTINUE                                                        
        RETURN                                                          
        END                                                             
