C
C      ************************************************************
C      *                                                          *
C      *                                                          *
C      *                                                          *
C      *                                                          *
C      *                                                          *
C      *                                                          *
C      *             ============================                 *
C      *                CHANEL   (version. 2.X)                   *
C      *             ============================                 *
C      *                                                          *
C      *                                                          *
C      *             SUBROUTINES for SUSY processes.              *
C      *             Version. 2 should be added.                  *
C      *                                                          *
C      *                   Written by H. Tanaka                   *
C      *                                                          *
C      *                                                          *
C      *                        ---------                         *
C      *                        Reference                         *
C      *                        ---------                         *
C      *    ' Numerical Calculation of Helicity Amplitudes        *
C      *      for Processes Involving Massive Fermions'           *
C      *                                                          *
C      *      H. Tanaka, Hiroshima University preprint            *
C      *         Comput. Phys. Commun. 58(1990) 153 - 168         *
C      *                                                          *
C      *                                                          *
C      *                         and                              *
C      *                                                          *
C      *    ' Numerical Calculation of Feynman Amplitudes         *
C      *      for Electroweak Theories and Application to         *
C      *      e(-) e(+) --> W(+) W(-) photon '                    *
C      *                                                          *
C      *                                                          *
C      *          H. Tanaka, T. Kanako and Y. Shimizu             *
C      *          KEK preprint 90 - 29, May 1990.                 *
C      *          Submitted to Comput. Phys. Commun.              *
C      *                                                          *
C      *   version for real*8 & complex*16                        *
C      *                                                          *
C      *                                                          *
C      ************************************************************

C
C ********************************************************************
C *         S                                                        *
C *         !                                                        *
C *         !                                                        *
C * F1--<---!---<---F2    SUBROUTINE FFST                            *
C *                                                                  *
C *     Purpose: to calculate vertex amplitudes for scalar boson-    *
C *              massive fermions vertex.                            *
C *                                                                  *
C *                 T             T_        T                        *
C *      U(II,AAM,Q) (AL*PL+AR*PR) U(I,AM,P)                         *
C *                                                                  *
C *                                                                  *
C *                                                                  *
C *     I,II=: indices to specify fermion(I,II=3) or antifermion     *
C *            (I,II=1) state.                                       *
C *     AM,AAM=: masses of fermions.                                 *
C *     AL,AR=: coupling constants for vertex.                       *
C *     C,CC=: phase factors for massive fermions.                   *
C *     P1,P2,Q1,Q2=: light-like vectors decomposed by subroutine    *
C *                   SPLTQ.                                         *
C *     AALL=: calculated results of vertex amplitudes for all       *
C *            possible helicity states.                             *
C *                                                                  *
C *                                                                  *
C *                                      written by H. Tanaka        *
C ********************************************************************
C
C
C       ===========================================================
        SUBROUTINE FFST(II,I,AAM,AM,AL,AR,CC,C,Q1,Q2,P1,P2,AALLL)
C       ===========================================================
C
        IMPLICIT REAL*8(A-H,O-Z)
        COMPLEX*16 AALL,AALLL,C,CC
        DIMENSION P1(4),P2(4),Q1(4),Q2(4),C(2),CC(2)
        DIMENSION AALL(2,2),AALLL(2,2)

CX      CALL FFS(I,II,AM,AAM,AL,AR,C,CC,P1,P2,Q1,Q2,AALL)

CX      AALLL(1,1) = AALL(1,1)
CX      AALLL(1,2) = AALL(2,1)
CX      AALLL(2,1) = AALL(1,2)
CX      AALLL(2,2) = AALL(2,2)

        I1 = I - 2
        I1 = -I1
        I1 = I1 + 2

        I2 = II - 2
        I2 = -I2
        I2 = I2 + 2

        CALL FFS(I2,I1,AAM,AM,AL,AR,CC,C,Q1,Q2,P1,P2,AALL)

        AALLL(1,1) = -AALL(1,1)
        AALLL(1,2) = -AALL(1,2)
        AALLL(2,1) = -AALL(2,1)
        AALLL(2,2) = -AALL(2,2)

        RETURN
        END
