      SUBROUTINE KINIT
C---------------------------------------------------------------------
C   GRACE System Library File
C   KINEM No. : 3002
C   Date      : 1994.04.30
C   Author    : Y.Kurihara
C---------------------------------------------------------------------
C
      IMPLICIT REAL*8(A-H,O-Z)
 
      PARAMETER ( MXDIM = 50 )
      COMMON / LOOP0 / LOOP
      COMMON / BPARM1 / XL(MXDIM),XU(MXDIM),NDIM,NWILD,
     &                 IG(MXDIM),NCALL
      COMMON / BPARM2 / ACC1,ACC2,ITMX1,ITMX2
 
      CHARACTER XSTR*14
 
      INCLUDE 'inclk.f'
*-----------------------------------------------------------------------
      COMMON/KMCNTL/IRESNS,ICOS3
      COMMON/KINEM1/S,W,FACT
      COMMON/CUT001/COSCUT(2,3),ENGYCT(2,3),AMASCT(2),ARESNS(2)
      common /knmflg/ knmgvs, knmhst, knmsfp, knmisr
*-----------------------------------------------------------------------
* 3-body kinematics
*
*  Integration variables
*        COS(the3),Phi3(=0)
*        Q2=(p4+p5)**2
*        COS(the4^),Phi4^    in particle 4,5 CM frame
*
*                         /P4
*                        /
*           p1\         /
*              \      q/\
*               ***** /  \
*              *******    \P5
*               ***** \
*               /      \
*              /        \
*           p2/          \p3
*
*
*--- 1. Initialize constants for kinematics.
* S = (P1+P2)**2
* W = SQRT(S)
      if( knmgvs .eq. 1 ) then
      W =1000.D0
      endif
      S = W*W
      totmas=amass1(3)+amass1(4)+amass1(5)
      if(totmas.gt.w) then
	write(6,*)
     .' CM energy is less than the sum of final particle masses.'
	stop
      end if
*******************************
* Angular cuts in Lab-frame.  *
*******************************
* particle 3
* minimum cos  cut
      COSCUT(1,1)= -1
* maximum cos  cut
      COSCUT(2,1)=  1
* particle 4
* minimum cos  cut
      COSCUT(1,2)= -1
* maximum cos  cut
      COSCUT(2,2)=  1
* particle 5
* minimum cos  cut
      COSCUT(1,3)= -1
* maximum cos  cut
      COSCUT(2,3)=  1
 
*******************************
* Energy  cuts in Lab-frame.  *
*******************************
* particle 3
* minimum energy cut
      ENGYCT(1,1) = AMASS1(3)
      ENGYCT(1,1) = 0.1d-2
* maximum energy cut
      ENGYCT(2,1) = W
* particle 4
* minimum energy cut
      ENGYCT(1,2) = AMASS1(4)
* maximum energy cut
      ENGYCT(2,2) = W
* particle 5
* minimum energy cut
      ENGYCT(1,3) = AMASS1(5)
* maximum energy cut
      ENGYCT(2,3) = W
 
*******************************
* Cut on invariant mass of 4-5*
*******************************
* minimum
      AMASCT(1)= AMASS1(4)+AMASS1(5)
* maximum
      AMASCT(2)= W-AMASS1(3)
 
*******************************
* Q2 singularity treatment    *
*******************************
* no-singularity           : IRESNS= 0
* narrow resonance         : IRESNS= 1
* 1/Q2 singularity         : IRESNS=-1
      IRESNS   = 0 
      IF(IRESNS.EQ.1) THEN
C If you want treat narrow resonance, you should set resonance mass
C and width.
C       ARESNS(1)=AM**
C       ARESNS(2)=AG**
        ARESNS(1)=0
        ARESNS(2)=0
      END IF
 
*********************************
* mass singularity treatment    *
*********************************
* no-singularity           : ICOS3=0
* mass-singularity         : ICOS3=1
*
*  If ICOS3=1, a particle 3 is assumed to be an initial state
* radiated photon (amass1(3)=0) and particle1 and particle2 have
* the same mass (AMASS1(1)=AMASS1(2)>0), and IRESNS is ignored.
*
      ICOS3=1
      if(icos3.eq.1 .and. amass1(3).gt.1.d-20) then
       write(6,*)'If ICOS3=1, particle 3 should be photon.'
       stop
      endif

      WRITE(6,*)'*'
      WRITE(6,*)'* Kinematics initialization '
      WRITE(6,*)'* particle mass ',amass1
      WRITE(6,*)'* CM Energy ',W
      WRITE(6,*)'* COSCUT    '
      WRITE(6,'(2F10.3)')COSCUT
      WRITE(6,*)'* ENGYCT    '
      WRITE(6,'(2F10.3)')ENGYCT
      WRITE(6,*)'* AMASCT    '
      WRITE(6,'(2F10.3)')AMASCT
      WRITE(6,*)'* Singularity treatment '
      if(icos3.eq.1) goto 10
      WRITE(6,*)'*   Q2'
      IF     (IRESNS.EQ.-1) THEN
         WRITE(6,*)'* 1/Q2 singularity '
      ELSE IF(IRESNS.EQ. 0) THEN
         WRITE(6,*)'* NO   singularity '
      ELSE IF(IRESNS.EQ. 1) THEN
         WRITE(6,*)'* Narrow resonance '
         WRITE(6,*)'*  resonance mass and width'
         WRITE(6,*) ARESNS
      END IF
10    IF     (ICOS3.EQ.0) THEN
         WRITE(6,*)'* NO mass-singularity '
      ELSE
         WRITE(6,*)'*    mass-singularity '
      END IF
      WRITE(6,*)'*'
 
* Following flux factor is for a particle-antiparticle collision
* at s/m**2 >> 1. For an e-gamma collision or low energy interactions
* you should use appropriate formulae.
      VREL = 2
      FLUX = VREL*S
      FACT = GEVPB/FLUX
 
*----- BASES RELATED INITIALIZATIONS -----------------------------------
*--- 2. Dimension of integration variables.
 
          NDIM =   5
          NWILD=   4
 
C-----------------------------------------------------------------------
*--- 3. Region of integration.
 
        DO 1 I=1,NDIM
         XL(I)  =   0.D0
         XU(I)  =   1.D0
         IG(I)  =   1
1       CONTINUE
C-----------------------------------------------------------------------
*--- 4. Number of iterations
      ITMX1  =  5
      ITMX2  =  5
 
      NCALL  = 50000
 
*--- 5. Set MXREG : the maximum number of values which are returned
*       by FUNC for one phase space point
 
      MXREG =   1
 
*--- 6. Set histograms
      if( knmhst .eq. 1 ) then
      NX = 50
      ND = 50
 
      DO 100 I = 1, NDIM
         WRITE(XSTR, 110) I
  110    FORMAT('X(',I2,') SPECTRUM')
         CALL XHINIT(I, XL(I), XU(I), NX, XSTR)
  100 CONTINUE
         CALL XHINIT(ndim+1, 0.d0, w, NX, 'Energy of Particle 3')
         CALL XHINIT(ndim+2, 0.d0, w, NX, 'Energy of Particle 4')
         CALL XHINIT(ndim+3, 0.d0, w, NX, 'Energy of Particle 5')
         CALL XHINIT(ndim+4,-1.d0,1.d0, NX, 'cos_the of Particle 3')
         CALL XHINIT(ndim+5,-1.d0,1.d0, NX, 'cos_the of Particle 4')
         CALL XHINIT(ndim+6,-1.d0,1.d0, NX, 'cos_the of Particle 5')

      endif 
      RETURN
      END
