  
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#define MXLEN  80
#define MXVTX  100
#define MXLINE 100
#define MXGRF  10000
#define MXEXTV 100
  
/*
 ===================================================================
     Structure vertex, vtx_line, int_line, inl_vtx
 ===================================================================
*/
struct vertex {
	int     flag;
	int     cntr_flag;
	char    kind[4];       
	float   x;         
	float   y;   
	int     no_line;   
        }      vtx[MXVTX];

struct vtx_line {
	int   vtxno[10];
        int   lineno[10];
        }     vtxline[MXVTX];

struct int_line {
	int   flag;
	char  kind[4];      
        float x;
        float y;
	int   direct;   
	char  species[4];  
	char  form[4];  
	int   print;  
        }     intline[MXLINE];

struct inl_vtx {
	int   vtxno[2];
	}     inlvtx[MXLINE];

/*
  ----------------------------------
      global variable list
  ----------------------------------
*/
long addr_graph[5][MXGRF];
/*
    addr_graph[0][i] : graph number
    addr_graph[1][i] : pointer of this graph to the temp file
    addr_graph[2][i] : the number of vertecies in the graph
    addr_graph[3][i] : the array size of vtx_no[] and intline_no[]
    addr_graph[4][i] : number of internal and external lines
*/
int level[2][MXGRF];
int no_graph, graph_no;
int no_initial, no_final;
int nvtx_iniline;            /* The number of vertecies on the initial line */
int no_iniline;            /* The number of vertecies on the initial line */
int nvertex;
int no_vertex, no_line;
float xmax, ymax;
float size[2][MXGRF];                             /* relative size of graph */
int cluster_table[MXEXTV], no_cluster; /* cluster table and no. of clusters */
int vtx_cntr, adj_flag, IR_flag;

int   vtxflag[MXEXTV];
int   ngen, ngt, nf;
float xf, yf, ysum, y_adj;
float y_center, x_margin, y_margin;
float maxx, maxy, minx, miny;

/*****************************************************************************
*                      ========================                             */
                         void tree_adjust( )
/*                     ========================                              *
******************************************************************************
* (Function )                                                                *
*  To adjust the (x,y) positions of vertices of tree diagrams                *
******************************************************************************
									    */
{
                                                                            /*
-----------------
 Local variables
-----------------
									    */
float  x, y, x_shift, y_shift;
int    i, j, nx, n, line_n, nv, nl, flag, v;
float  dev, ratio;
                                                                            /*
-------------
  Functions
-------------
									    */
void adjust_final( );
void adjust_trace( );
void adjust_f0( );
void adjust_s( );
void adjust_t( );
void print_vtxtable( );
int  get_vtxno( );
                                                                            /*
============================ Entry point =====================================

==============================================================================
  (I) Initialization
==============================================================================
									    */
x_margin =     1.0,    y_margin =  1.0;
minx     = 10000.0,    maxx     =  0.0;
miny     =    minx,    maxy     = maxx;
                                                                            /*
==============================================================================
  (II) Adjust the x- position of the final state vertices
==============================================================================
									    */
      /*
       printf("** call adjust_final from tree_adjust for tree\n");
      */

   adjust_final( );
                                                                            /*
==============================================================================
  (III) For the s-channel
==============================================================================
									    */
if( no_cluster == 1 )
  {
    v = cluster_table[0];

    y_center = vtx[ v ].y;
    dev      = 0.8*( ymax - y_center);

      /*
       printf(" y_center(%3d ) =%8.1f, dev =%8.1f\n", v, y_center, dev );
      */

    maxy = 0.0;
    for( n = 0; n < nvtx_iniline; n++ )
       {
         nv   = get_vtxno( n );
         y    = vtx[nv].y;
         maxx = y - y_center;
         if( maxx < 0.0 )
             maxx = -maxx;
         if( maxx > maxy )
             maxy = maxx;
        }
         /*
          printf(" maxy = %6.1f\n", maxy );
         */
    if( dev > maxy && maxy > 0.0 )
      {
        ratio = dev / maxy;
        for( n = 0; n < nvtx_iniline; n++ )
           {
             nv   = get_vtxno( n );
             y   = vtx[nv].y;
             dev = y - y_center;
             y   = dev*ratio + y_center;
             vtx[nv].y = y;
            }
       }
   }
                                                                            /*
==============================================================================
  (IV) For the t-channel
==============================================================================
									    */
else if( no_cluster > 1 )
  {
                                                                            /*
------------------------------------------------------------------------------
   (1) Adjust (x,y)- position of those vertices, not on the initial line.
------------------------------------------------------------------------------
									    */
       /*
	print_vtxtable( -1, 0 );
       */
/*
    for( n = 0; n < no_cluster; n++ )
       {
         v  = cluster_table[n];

         if( n == 0 )                     adjust_t( n, v, v, -1 );
         else if( n == no_cluster-1 )     adjust_t( n, v, v,  1 );
         else                             adjust_t( n, v, v,  0 );
        }
*/
                                                                            /*
------------------------------------------------------------------------------
   (2) Adjust again the x- position of the final state vertices
------------------------------------------------------------------------------
									    */
      /*
       printf("** call adjust_final from tree_adjust after adjust_t\n");
      */

    adjust_final( );

   }

       /*
	print_vtxtable( -1, 0 );
       */

  size[0][no_graph-1] = xmax;
  size[1][no_graph-1] = ymax;
  addr_graph[0][no_graph-1] = graph_no;
  addr_graph[2][no_graph-1] = no_vertex;
  addr_graph[4][no_graph-1] = no_line;

}

/*****************************************************************************
*                      ========================                             */
                         void loop_adjust( )
/*                     ========================                              *
******************************************************************************
* (Function )                                                                *
*  To adjust the (x,y) positions of vertices of tree diagrams                *
******************************************************************************
									    */
{
                                                                            /*
-------------
  Functions
-------------
									    */
void adjust_final( );
                                                                            /*
============================ Entry point =====================================

==============================================================================
  (I) Initialization
==============================================================================
									    */
x_margin =     1.0,    y_margin =  1.0;
minx     = 10000.0,    maxx     =  0.0;
miny     =    minx,    maxy     = maxx;
                                                                            /*
==============================================================================
  (II) Adjust the x- position of the final state vertices
==============================================================================
									    */
   adjust_final( );

  size[0][no_graph-1] = xmax;
  size[1][no_graph-1] = ymax;
  addr_graph[0][no_graph-1] = graph_no;
  addr_graph[2][no_graph-1] = no_vertex;
  addr_graph[4][no_graph-1] = no_line;

}

/*****************************************************************************
*                      ========================                             */
                         void adjust_final( )
/*                     ========================                              *
******************************************************************************
* (Function )                                                                *
*  To adjust the (x,y) positions of vertices of tree diagrams                *
******************************************************************************
									    */
{
									    /*
-----------------
 Local variables
-----------------
									    */
float x, y, x_shift, y_shift;
char *final = "fin", *photon = "g";
int  i, j, nx, n, line_n, nv, nl, flag;
int  clstrid, fl_cluster;
									    /*
-----------------
  Functions
-----------------
									    */
void print_vtxtable( );
int  get_clID      ( );
int  get_kind      ( );
									    /*
============================= Entry point ====================================
									    */
     /*
      print_vtxtable( -1, 0 );
     */

x_margin =     1.0,    y_margin =  1.0;
minx     = 10000.0,    maxx     =  0.0;
miny     =    minx,    maxy     = maxx;

for( nv = 0; nv < no_vertex; nv++ )
   {
    if( vtx[nv].flag >= 0 )
      {
        x  = vtx[nv].x, y = vtx[nv].y;

          /*
            printf(" vertex[%d]  x = %6.1f   y = %6.1f\n", 
		  nv, x, y );
          */

       if( x > maxx )
           maxx = x;
       if( x < minx )
           minx = x;
       if( y > maxy )
           maxy = y;
       if( y < miny )
           miny = y;
      }
    }
x_shift = x_margin - minx;
y_shift = y_margin - miny;
maxx    = maxx + x_shift;
maxy    = maxy + y_shift;
xmax    = maxx + x_margin;
ymax    = maxy + y_margin;

  /*
  printf(" maxx = %6.1f,  maxy = %6.1f\n", maxx, maxy );
  */

for( nv = 0; nv < no_vertex; nv++ )
   {
     if( vtx[nv].flag >= 0 )
       {
         x  = vtx[nv].x, y = vtx[nv].y;
         vtx[nv].x = x + x_shift;
         vtx[nv].y = y + y_shift;

               /*
	       printf("** vertex id =%3d ***\n",nv );
               */
         if( strcmp( vtx[nv].kind, final ) == 0 )
           {
             line_n = vtxline[nv].lineno[0];
             nx     = vtxline[nv].vtxno[0];

	      /*
               printf(" vtx[%d].x = %6.1f   vtx[%d].y = %6.1f\n", 
		        nv, vtx[nv].x, nv, vtx[nv].y );
               printf(" vtx[%d].kind = %s intline[%d].species = %s\n", 
		nv, vtx[nv].kind, line_n, intline[line_n].species );
               printf(" vtx[%d] is connected to vtx[%d]\n", nv, nx );
              */

             if( strcmp( intline[line_n].species, photon ) == 0 )
               {
                 clstrid = get_clID( nx );

                 fl_cluster = 0;
                 if( clstrid == 0 || clstrid == no_cluster-1 ) 
		     fl_cluster = 1;

                  /*
                   printf("   cluster-id[ vtx(%d) ] =%2d",
                                nx, clstrid );
                   printf(" vtx_kind[%3d ] =%2d fl_cluster =%2d\n",
                                nx, get_kind( nx ), fl_cluster );
                  */

		 if( ( fl_cluster && get_kind( nx ) != 3 ) ||
                     ( fl_cluster && nx == cluster_table[clstrid] ) ||
                      !fl_cluster )
                       vtx[nv].x = maxx;
                }
             else
               {
                 vtx[nv].x = maxx;
                }
            }
            /*
             printf(" vtx[%d].x = %6.1f   vtx[%d].y = %6.1f\n", 
	            	  nv, vtx[nv].x, nv, vtx[nv].y );
            */
        }
    }
for( nl = 0; nl < no_line; nl++ )
   {
     if( intline[nl].x > 0.0 )
       {
         x = intline[nl].x,  y = intline[nl].y;
         intline[nl].x = x + x_shift;
         intline[nl].y = y + y_shift;
        }
    }

}

