  
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#define MXLEN  80
#define MXVTX  100
#define MXLINE 100
#define MXGRF  10000
#define MXEXTV 100
  
                                                                            /*
==============================================================================
     Structure vertex, vtx_line, int_line, inl_vtx
==============================================================================
                                                                            */
struct vertex {
	int     flag;
	int     cntr_flag;
	char    kind[4];       
	float   x;         
	float   y;   
	int     no_line;   
        }      vtx[MXVTX];

struct vtx_line {
	int   vtxno[10];
        int   lineno[10];
        }     vtxline[MXVTX];

struct int_line {
	int   flag;
	char  kind[4];      
        float x;
        float y;
	int   direct;   
	char  species[4];  
	char  form[4];  
	int   print;  
        }     intline[MXLINE];

struct inl_vtx {
	int   vtxno[2];
	}     inlvtx[MXLINE];
                                                                            /*
------------------------------------------------------------------------------
      global variable list
------------------------------------------------------------------------------
                                                                            */
long addr_graph[5][MXGRF];
                                                                            /*
    addr_graph[0][i] : graph number
    addr_graph[1][i] : pointer of this graph to the temp file
    addr_graph[2][i] : the number of vertecies in the graph
    addr_graph[3][i] : the array size of vtx_no[] and intline_no[]
    addr_graph[4][i] : number of internal and external lines
                                                                            */
int level[2][MXGRF];
int no_graph, graph_no;
int no_initial, no_final;
int nvtx_iniline;          /* The number of vertecies on the initial line   */
int no_iniline;            /* The number of vertecies on the initial line   */
int nvertex;               /* The number of vertices in the vertex table    */
int no_vertex, no_line;
float xmax, ymax;
float size[2][MXGRF];                             /* relative size of graph */
int cluster_table[MXEXTV], no_cluster; /* cluster table and no. of clusters */
int n_IRmax, n_photon[2];
int vtx_cntr, adj_flag, IR_flag;

int dfnumber[MXEXTV], count;

/*****************************************************************************
                   ============================                             */
                      void gftree(  )
/*                 =============================
******************************************************************************
*( Function )                                                                *
*     To determine the vertex positions in the tree Feynman diagram.         *
*( input )                                                                   *
*     structure vtx[i] : vertex information                                  *
*     structure intline[i] : line information                                *
*( output )                                                                  *
*     The vertex position in the graph                                       *
*         structure vtx[].x and structure vtx[].y                            *
*****************************************************************************/
{
                                                                            /*
----------------------------
  local variable list
----------------------------
                                                                            */
  int root, i, n, nv, level_no;
                                                                            /*
----------------------------
  function list
----------------------------
                                                                            */

void    clear_vtxtable( );
void    fv_inildecay  ( );
void    dfcount       ( );
void    find_line     ( );
void    fv_iniline    ( );
void    fix_iniline   ( ); 
void    fix_inirad    ( ); 
void    fix_vertex    ( ); 
void    print_vtxtable( ); 
void    print_straight( ); 


                                                                            /*
=========================== Entry point of gftree ============================
                                                                            */

 xmax = 0.0, ymax = 0.0;

   /*
   printf(" no_graph(%d) no_initial(%d)\n", no_graph, no_initial );
   tgdlist( );
   */

                                                                            /*
==============================================================================
 Initialization 
==============================================================================
                                                                            */

 nvtx_iniline = 0;           /* The number of vertecies on the initial line */
 nvertex      = nvtx_iniline;
 no_iniline   = nvtx_iniline;
                                                                            /*
 <<<<<<<<<<<>>>>>>>>>>>>>                                                   */
     clear_vtxtable( );                                                     /*
 <<<<<<<<<<<>>>>>>>>>>>>>                                                   */
                                                                            /*
==============================================================================
 Determine the initial line
==============================================================================

+----------------------------------------------------------------------------+
| (1) For the Decay process i.e. no_initial = 1                              |
+----------------------------------------------------------------------------+
                                                                            */
 if( no_initial == 1 )
   {                                                                        /*
       <<<<<<<<<<<<<<<<>>>>>>>>>                                            */
         fv_inildecay( 0, 0 );                                              /*
       <<<<<<<<<<<<<<<<>>>>>>>>>                                            */
    }
 else
   {
                                                                            /*
+----------------------------------------------------------------------------+
| (2) For the Scattering process                                             |
+----------------------------------------------------------------------------+
    +------------------------------------------------------------------------+
    |(i) calculate distances from v0 and v1                                  |
    +------------------------------------------------------------------------+
                                                                            */

        for( i = 0; i < no_vertex; i++ )
           {
             dfnumber[i] = 1000;
             if( vtx[i].flag < 0 ) dfnumber[i] = -1 ;
            }
                                                                            /*
        <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                                    */
         find_line( 0, 1, &level_no );                                      /*
        <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                                    */

                                                                            /*
    +------------------------------------------------------------------------+
    |(iii)  find the vertecies on the initial line                           |
    |       whose level number is equal to that of vertex 0.                 |
    +------------------------------------------------------------------------+
                                                                            */
                                                                            /*
        <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                                    */
          fv_iniline( 0, 0, level_no );                                     /*
        <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                                    */
    }


                                                                            /*
==============================================================================
 Fix the vertex poistions of initial line
==============================================================================
                                                                            
   <<<<<<<<<<>>>>>>>>>>                                                     */
     fix_iniline( );                                                        /*
   <<<<<<<<<<>>>>>>>>>>                                                     */

    /*
     printf("***** after fix ****\n");
     print_vtxtable( -1, 0 );
    */
                                                                            /*
==============================================================================
 Fix the vertex poistions of initial radiation
==============================================================================
                                                                            
   <<<<<<<<<<>>>>>>>>>>                                                     */
     fix_inirad( );                                                         /*
   <<<<<<<<<<>>>>>>>>>>                                                     */

    /*
     printf("***** after inirad ****\n");
     print_vtxtable( -1, 0 );
    */

                                                                            /*
==============================================================================
 Fix positions for all other vertecies
==============================================================================
                                                                            
   <<<<<<<<<<>>>>>>>>>>                                                     */
      fix_vertex(  );                                                       /*
   <<<<<<<<<<>>>>>>>>>>                                                     */

    /*
     printf("***** after fix_vertex ****\n");
     print_vtxtable( -1, 0 );
    /*
     print_straight( );
    */



}

/*****************************************************************************
                ========================================                    */
                 void fv_inildecay( parentv, currentv )
/*              ========================================
******************************************************************************
*( Function )                                                                *
*     To find the vertecies on the initial line for the decay processes.     *
*     This function is used recursively to find the initial line vertecies.  *
*( input )                                                                   *
*     parentv          : The vertex ID number of parent vertex.              *
*     currentv         : The vertex ID number of currently tested.           *
*     structure vtx[i] : vertex information                                  *
*     structure intline[i] : line information                                *
*( output )                                                                  *
*     Store the information of the initial line vertecies into vtx_table.    *
*****************************************************************************/

int parentv, currentv;
{
                                                                            /*
-----------------------
  Local variables
-----------------------
                                                                            */
  int n_vertex, vertex_n;
  int n_line,   line_n, adirect;
  int add, n, count, kind;
  char *vertex   = "vtx", *photon   = "g";
  char *initial  = "ini", *final    = "fin", *external = "ext";
                                                                            /*
------------------------
  Functions
------------------------
                                                                            */
  void add_vtxtable( );
  void set_kind    ( );
  int  iabs        ( );
                                                                            /*
=========================== Entry point ======================================
                                                                            */
  /*
    printf("fv_inildecay : parentv(%3d)  currentv(%3d)\n", 
        	  parentv, currentv );
  */
                                                                            /*
==============================================================================
  Initialization
==============================================================================
                                                                            */

  kind  = 1;
                                                                            /*
  <<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>                                    */
   add_vtxtable( 1, currentv, 0, 0, 0 );                                    /*
  <<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>                                    */

  n_line = vtx[currentv].no_line;

                                                                           /*
==============================================================================
            Determine the kind of vertex
==============================================================================
  +--------------------------------------------------------------------------+
  |(1) If the current vertex is an internal one,                             | 
  |    test for all lines connected to the current vertex                    |
  |    whether it is a radiative photon/gluon or not.                        |
  +--------------------------------------------------------------------------+
                                                                            */
  if( strcmp( vtx[currentv].kind, vertex ) == 0 )
    {                                                                       /*
      +----------------------------------------------------------------------+
      |(a) If a line is a radiative photon/gluon,                            |
      |    the kind of current vertex is categolized to kind = 3.            |
      +----------------------------------------------------------------------+
                                                                            */
      if( vtx[currentv].flag == 0 && n_line == 3 )
      for( n = 0; n < n_line; n++ )
         {
           vertex_n = vtxline[currentv].vtxno[n];

	   if( vertex_n != parentv )
             {
               line_n   = vtxline[currentv].lineno[n];
               if( strcmp( intline[line_n].species, photon ) == 0 &&
                   strcmp( intline[line_n].kind, external ) == 0     )
                 {
                   adirect = iabs( intline[line_n].direct );
                   if( adirect == 4 || adirect == 5 ) kind = 3;
                   else                               kind = 4;
                  }
              }
          }
                                                                            /*
      +----------------------------------------------------------------------+
      |(b) If the current vertex is categolized as kind = 3,                 |
      |    printing the particle names of the lines connected to the         |
      |    current vertex is supressed.                                      |
      +----------------------------------------------------------------------+
                                                                            */
      if( kind == 3 )
        {
          for( n = 0; n < n_line; n++ )
             {
               line_n = vtxline[currentv].lineno[n];

	       if( strcmp( intline[line_n].kind, external ) != 0 )
                   if( iabs( intline[line_n].direct ) < 10 )
	               intline[line_n].print = 0;
              }
         }
      else if( kind == 4 ) kind = 3;
     }
                                                                            /*
  +--------------------------------------------------------------------------+
  |(2) If the current vertex is not the initial vertex,                      |
  |    it is categolized into kind = 2.                                      |
  +--------------------------------------------------------------------------+
                                                                            */
  else if( strcmp( vtx[currentv].kind, initial ) == 0 )
      kind = 2;

       /*
       printf("fv_inildecay : n_line = %d  kind(%d)\n", 
	       n_line, kind );
       */
                                                                            /*
  +--------------------------------------------------------------------------+
  |(3) If the current vertex is the four points vertex, it is categolized    |
  |    into kind = 1 ( internal vertex ) and return.                         |
  +--------------------------------------------------------------------------+
                                                                            */

  if( n_line > 3 )
    {                                                                       /*
     <<<<<<<<<<<<<<>>>>>>>>>>>>>                                            */
      set_kind( currentv, 1 );                                              /*
     <<<<<<<<<<<<<<>>>>>>>>>>>>>                                            */
      
      return;
     }
                                                                            /*
  +--------------------------------------------------------------------------+
  |(4) If the current vertex has neither a radiative photon/gluon nor an     |
  |    initial vertex, save the kind into vtx_table and return.              |
  +--------------------------------------------------------------------------+
                                                                            */
  else if( kind != 3 && currentv != 0 )
    {                                                                       /*
     <<<<<<<<<<<<<<>>>>>>>>>>>>>                                            */
      set_kind( currentv, kind );                                           /*
     <<<<<<<<<<<<<<>>>>>>>>>>>>>                                            */

      return;
     }
                                                                            /*
  +--------------------------------------------------------------------------+
  |(5) If the current vertex has an radiative photon.gluon or is an          |
  |    initial vertex, test for all vertecies connected to the current       |
  |    vertex.                                                               |
  +--------------------------------------------------------------------------+
                                                                            */
  else
    {                                                                       /*
     <<<<<<<<<<<<<<>>>>>>>>>>>>>                                            */
      set_kind( currentv, kind );                                           /*
     <<<<<<<<<<<<<<>>>>>>>>>>>>>                                            */

      /*
        print_vtxtable( currentv, 1 );
      */

      n_vertex = vtx[currentv].no_line;
                                                                            /*
      +----------------------------------------------------------------------+
      |(a) For all the vertecies connected to the current vertex             |
      +----------------------------------------------------------------------+
                                                                            */
      for( n = 0; n < n_vertex; n++ )
         {
           vertex_n = vtxline[currentv].vtxno[n];
           line_n   = vtxline[currentv].lineno[n];
                  /*
                    printf(" vertex_n(%d) != parentv(%d)  ?\n", 
                        vertex_n, parentv );
                  */
           if( vertex_n != parentv )
             {
                 /*
                   printf(" intline[ line_n = %d].species (%s) != (%s) ?\n",
                            line_n, intline[line_n].species, photon );
                   printf(" intline[ line_n = %d].kind    (%s) != (%s) ?\n",
                            line_n, intline[line_n].kind, external );
                   printf(" currentv(%d) = (0) ?\n", currentv );
                 */
                                                                            /*
               +-------------------------------------------------------------+
	       |(i) If this vertex is not the parent and                     |
               |     the line between this and the current vertex            |
	       |       is neither a photon/gluon nor an external line        |
               |     or the current vertex is not an initial one,            |
               |    move to this vertex as a current vertex.                 |
               +-------------------------------------------------------------+
                                                                            */
               if( strcmp( intline[line_n].species, photon ) != 0 || 
                   strcmp( intline[line_n].kind, external )  != 0 ||
                   currentv == 0 )
	           {
                       /*
                       printf(" call fv_inildecay");
                       printf("( parentv = %d, currentv = %d )\n",
                                 currentv, vertex_n );
                       */
                                                                            /*
                     <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>                 */
                       fv_inildecay( currentv, vertex_n );                  /*
                     <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>                 */
                     }
               }
          }
     }
}

/*****************************************************************************
                   ======================================                   */
                      void find_line( v0, v1, level_n )
/*                 ======================================
******************************************************************************
*( Function )                                                                *
*     To determine the vertex positions in the tree Feynman diagram.         *
*( input )                                                                   *
*     structure vtx[i] : vertex information                                  *
*     structure intline[i] : line information                                *
*( output )                                                                  *
*     The vertex position in the graph                                       *
*         structure vtx[].x and structure vtx[].y                            *
*****************************************************************************/
int v0, v1, *level_n;
{
                                                                            /*
  ----------------------------------
      local variable list
  ----------------------------------
                                                                            */
int i;

void dfcount( );
                                                                            /*
====================== Entry point of find_line ==============================

==============================================================================
 calculate the distances of all vertices from the vertex v0
==============================================================================
                                                                            */
 for( i = 0; i < no_vertex; i++ )
    {
      level[0][i] = -1;
      level[1][i] = -1;
     }
                /*
                 printf("** find_line calls dfcount for v0(%3d )\n", v0);
               */
                                                                            /*
   <<<<<<<<<<<<<>>>>>>>>>>>>                                                */
     dfcount( v0, v0, 0 );                                                  /*
   <<<<<<<<<<<<<>>>>>>>>>>>>                                                */

               /*
                 printf("* vtx_id  dfnumber from v0(%3d )\n", v0);
                 for( i = 0; i < no_vertex; i++ )
                      printf("  %5d%10d\n", i, dfnumber[i] );
              */

 for( i = 0; i < no_vertex; i++ )
      level[0][i] = dfnumber[i];
                                                                            /*
==============================================================================
 calculate the distances of all vertices from the vertex v1
==============================================================================
                                                                            */
              /*
                printf("** find_line calls dfcount for v1(%3d )\n", v1);
              */
                                                                            /*
   <<<<<<<<<<<<<>>>>>>>>>>>>                                                */
     dfcount( v1, v1, 0 );                                                  /*
   <<<<<<<<<<<<<>>>>>>>>>>>>                                                */

              /*
                printf("* vtx_id  dfnumber  from v1(%3d )\n", v1);
                for( i = 0; i < no_vertex; i++ )
                     printf("  %5d%10d\n", i, dfnumber[i] );
              */

 for( i = 0; i < no_vertex; i++ )
      level[1][i] = dfnumber[i];
   
              /*
                printf("** find_line vtx# frm_v[%3d ] frm_v[%3d ]\n", 
                          v0, v1);
                for( i = 0; i < no_vertex; i++ )
                     printf("            %5d%10d%12d\n",
                             i, level[0][i], level[1][i] );
              */
                                                                            /*
==============================================================================
 calculate the level numbers of each vertex
==============================================================================
                                                                            */
 *level_n = level[0][v0] + level[1][v0];
 for( i = 0; i < no_vertex; i++ )
    {
      level[0][i] = level[0][i] + level[1][i];
     }
              /*
                printf("** find_line vtx# frm_v[%3d ] level_n =%3d\n",
                        v0, *level_n );
                 for( i = 0; i < no_vertex; i++ )
                      printf("            %5d%10d%12d\n",
                           i, level[0][i], level[1][i] );
               */
}

/*****************************************************************************
               ============================================                 */
                void dfcount( parentv, currentv, flag )
/*             ============================================
******************************************************************************
* ( Function )                                                               *
*       To determine the distance of each vertex from the root vertex.       *
*       The distance is defined  by the number of vertecies through          *
*        which the path approachs to it from the root vertex.                *
*  ( input )                                                                 *
*       parentv   : the vertex ID for the parent                             *
*       currentv  : the vertex ID for the current vertex                     *
*       flag      : ( 0 / 1 ) = ( distance / dfnumber )                      *
*       Structure vtx, vtxline                                               *
*  ( Output )                                                                *
*       level[flag][] : distance from the root to the current vertex and     *
*                   set the distances for the daughters same as the current. *
*****************************************************************************/
int parentv, currentv, flag;
{
                                                                            /*
--------------------------
  local variable list
--------------------------
                                                                            */
  int n_vertex, vertex_n;
  int n;
                                                                            /*
=========================== Entry point of dfcount ===========================

==============================================================================
 Initialization for the first call where parentv equals to currentv
==============================================================================
                                                                            */
  if( parentv == currentv )
    {
      count = 0;
      for( n = 0; n < no_vertex; n++ )
         {
           if( dfnumber[n] != -1 ) dfnumber[n] = 1000;
          }

         /*
          printf("** dfcount  starts from vertex[%3d ]\n", parentv );
          printf("            initialization has been done for flag =%2d.\n",
	        	flag );
          printf("* vtx_id  dfnumber before calling dfcount\n");
          for( n = 0; n < no_vertex; n++ )
               printf("  %5d%10d\n", n, dfnumber[n] );
         */
     }
                                                                            /*
==============================================================================
 handling for the current vertex currentv
==============================================================================
                                                                            
------------------------------------------------------------------------------
 (1) increment the counter ( dfnumber ) if parentv != currentv        
------------------------------------------------------------------------------
                                                                            */
  else
       count++;
                                                                            /*
------------------------------------------------------------------------------
 (2) save the count as the dfnumber of the current vertex
------------------------------------------------------------------------------
                                                                            */
  dfnumber[currentv] = count;

  n_vertex = vtx[currentv].no_line;

        /*
         printf("  currentv[%3d ] n_vertex =%3d  parentv[%3d ] count =%3d\n", 
	           currentv, n_vertex, parentv, count );  
         printf("  currentv[%3d ] dfnumber[%3d ] =%3d  count =%3d\n",
                   currentv, currentv, dfnumber[currentv], count );
        */
                                                                            /*
------------------------------------------------------------------------------
 (3) trace the unvisited vertices
------------------------------------------------------------------------------
                                                                            */
  for( n = 0; n < n_vertex; n++ )
     {
       vertex_n =  vtxline[currentv].vtxno[n];

       if( vertex_n != parentv )
         {
           if( dfnumber[vertex_n] == 1000 || dfnumber[vertex_n] > count )
	     {                                                              /*
               <<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                    */
                  dfcount( currentv, vertex_n, flag );                      /*
               <<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                    */
                                                                            /*
              +--------------------------------------------------------------+
              | if flag == 0, decrement the count by 1                       |
              |      to determine the distance of the current vertex from    |
              |            the root vertex.                                  |
              | if flag != 0, nothing should be done                         |
              |      to calculate the dfnumber of the current vertex         |
              +--------------------------------------------------------------+
                                                                            */
	       if( flag == 0 ) count--;

              }
         }
      }
  /*
   printf("* currentv[%3d ] dfnumber[%3d ] =%3d  count =%3d\n",
             currentv, currentv, dfnumber[currentv], count );
  */
}

/*****************************************************************************
          ===============================================                   */
           void fv_iniline( parentv, currentv, n_level )
/*        =============================================== 
******************************************************************************
* ( Function )                                                               *
*      To determine the vertecies on the initial line                        *
*            for the scattering processes.                                   *
* ( Input )                                                                  *
*      parentv     : the vertex ID of the parent                             *
*      currentv    : the vertex ID of the current                            *
*      n_level     : the level number to which the vertecies on the initial  *
*                    line have an identical level number.                    *
*      level[0][i] : level number for the i-th vertex which determined       *
*                    before calling fv_iniline in the function tgdlayout     *
*      Struct vtx                                                            *
* ( Output )                                                                 *
*      vtx_table   : store the vertex ID on the initial line                 *
*****************************************************************************/
int parentv, currentv, n_level;
{
                                                                            /*
--------------------------
  local variable list
--------------------------
                                                                            */
  int n_vertex, vertex_n;
  int n_line,   line_n;
  int add, nv, n, count, kind, adirect;
  char *vertex   = "vtx", *photon   = "g";
  char *initial = "ini", *final = "fin", *external = "ext";
                                                                            /*
----------------------------
  function list
----------------------------
                                                                            */
  void add_vtxtable( );
  void set_kind    ( );
                                                                            /*
======================== Entry point of fv_iniline ===========================
                                                                            */
  count = level[0][currentv];

    /*
    printf("parentv(%3d)  currentv(%3d)  count(%3d) level(%3d)\n", 
	  parentv, currentv, count, n_level );
    */
                                                                            /*
==============================================================================
 When the level number of the current vertex is identical to that of
    the initial line, determine the kind of vertex and test further.
==============================================================================
                                                                            */

  if( count == n_level )
    {
      kind  = 1;
                                                                            /*
      <<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>                           */
         add_vtxtable( 1, currentv, 0, 0, 0 );                              /*
      <<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>                           */

                                                                            /*
      +----------------------------------------------------------------------+
      |(a) If the current vertex is an internal vertex,                      |
      +----------------------------------------------------------------------+
                                                                            */
      if( strcmp( vtx[currentv].kind, vertex ) == 0 )
        {                                                                   /*
          +------------------------------------------------------------------+
	  |(i) If the current vertex connects to a single radiative          |
	  |    photon/gluon, this is categolized into kind = 3.              |
          +------------------------------------------------------------------+
                                                                            */
	  n_line = vtx[currentv].no_line;
          if( vtx[currentv].flag == 0 && n_line == 3 )
	  for( n = 0; n < n_line; n++ )
	     {
	       line_n = vtxline[currentv].lineno[n];
	       nv     = vtxline[currentv].vtxno[n];

	       if( strcmp( intline[line_n].species, photon ) == 0 
		   && strcmp( vtx[nv].kind, final ) == 0 )
                 {
                   adirect = iabs( intline[line_n].direct );
                   if( adirect == 4 || adirect == 5 ) kind = 3;
                   else                               kind = 4;
		 }
              }
                                                                            /*
          +------------------------------------------------------------------+
	  |(ii) If the current vertex is catgolized into kind = 3,           |
	  |     printing particle names of the lines connected to            |
	  |     this vertex is suppressed.                                   |
          +------------------------------------------------------------------+
                                                                            */
          if( kind == 3 )
            {
	      for( n = 0; n < n_line; n++ )
	         {
                   if( n_line < 4 )
                     {
	               line_n = vtxline[currentv].lineno[n];

	               if( strcmp( intline[line_n].kind, external ) != 0 )
                           if( iabs( intline[line_n].direct ) < 10 )
		               intline[line_n].print = 0;
                     }
                  }
             }
           else if( kind == 4 ) kind = 3;
         }
                                                                            /*
      +----------------------------------------------------------------------+
      |(b) If the current vertex is an initial vertex, this is categolized   |
      |     into kind = 0.                                                   |
      +----------------------------------------------------------------------+
                                                                            */
      else if( strcmp( vtx[currentv].kind, initial ) == 0 )
	       kind = 0;
                                                                            /*
      +----------------------------------------------------------------------+
      |(c) Save the kind flag in vtx_table.                                  |
      +----------------------------------------------------------------------+

      <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                                    */
         set_kind( currentv, kind );                                        /*
      <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                                    */

        /*
         print_vtxtable( currentv, 1 );
        */
                                                                            /*
      +----------------------------------------------------------------------+
      |(d) Test the further vertecies connected to the current vertex.       |
      +----------------------------------------------------------------------+
                                                                            */
      n_vertex = vtx[currentv].no_line;

      for( n = 0; n < n_vertex; n++ )
         {
           vertex_n =  vtxline[currentv].vtxno[n];

           if( vertex_n != parentv )
	     {                                                              /*
                 <<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>               */
                  fv_iniline( currentv, vertex_n, n_level );                /*
                 <<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>               */
              }
          }
     }
}

/*****************************************************************************
                         =====================                              */
                          void fix_iniline( )
/*                       =====================
******************************************************************************
* ( Function )                                                               *
*   Determine the relative position of the vertecies on the initial line.    *
* ( Input )                                                                  *
*   nvtx_iniline  : the number of vertecies on the initial line              *
*   Struct vtx, vtxline, etc.                                                *
*   vtx_table     : vertex information for the initial line vertecies        *
* ( Output )                                                                 *
*   vtx[].x, vtx[].y : vertex position of the initial line vertecies         *
*   no_cluster       : number of clusters                                    *
*   cluster_table    : the vertex ID of the root vertex of the cluster       *
*****************************************************************************/
{
                                                                            /*
--------------------------
  local variable list
--------------------------
                                                                            */
 int n, nv, nx, n_IR;
 int breakp[4], n_cvtx[2], vertex_n, vt0, vt1;
 float x, y, dx0, dy0, dx1, dy1, phmax, ph0, ph1;
                                                                            /*
----------------------------
  function list
----------------------------
                                                                            */
 void print_vtxtable( );
 int  get_vtxno     ( );
 int  get_kind      ( );
 void set_kind      ( );
 void set_clID      ( );
 void create_straight();
 void addve_straight();
 void reverse_straight();
                                                                            /*
======================== Entry point of fix_iniline ==========================
                                                                            */
 /*
   print_vtxtable( );
 */
                                                                            /*
==============================================================================
 Initialization
==============================================================================

 <<<<<<<<<<>>>>>>>>>>>>>                                                    */
   clear_straight( 0, -1 );                                                 /*
 <<<<<<<<<<<>>>>>>>>>>>>                                                    */
 vt0         = 0;
 vt1         = -1;
 no_cluster  = 0;
 for( n = 0; n < MXEXTV; n++ )
      cluster_table[n] = -1;
                                                                            /*
==============================================================================
 For the decay processes
==============================================================================
                                                                            */
 if( no_initial == 1 )
   {
     nv = 0;
                                                                            /*
   +-------------------------------------------------------------------------+
   |(1) Counting the number of the initial radiations                        |
   +-------------------------------------------------------------------------+
                                                                            */
     n_photon[0] = 0;
     n_cvtx[0]    = 0;
     for( n = 1; n < nvtx_iniline; n++ )
        {                                                                   /*
                    <<<<<<<>>>>>>>>>>>                                      */
          vertex_n = get_vtxno( n );                                        /*
                    <<<<<<<>>>>>>>>>>>                                      */
                                                                            /*
             <<<<<<<<<<<<>>>>>>>>>>                                         */
          if( get_kind( vertex_n ) == 3 )                                   /*
             <<<<<<<<<<<<>>>>>>>>>>                                         */
              n_photon[0]++;
          else if( vtx[vertex_n].cntr_flag == 1 )
              n_cvtx[0]++;
          else 
              n = nvtx_iniline;
         }
                                                                            /*
   +-------------------------------------------------------------------------+
   |(2) fix the position of initial vertex ID = 0                            |
   +-------------------------------------------------------------------------+
                                                                            */
     x        = 30.0,  y        = 30.0;
     vtx[0].x = x,     vtx[0].y = y;
                                                                            /*
   +-------------------------------------------------------------------------+
   |(3) determine the break point, the max number of the radiations,         |
   |    and the step dx0, and initialize vtx_table[3] and [4]                |
   |    break point                                                          |
   |                   IR1  IR2                                              |
   |                    /   /     /                                          |
   |                   /   /     /break[0]      In this case,                |
   |             0----+----+----+ vertex ID = i    no_cluster = 1            |
   |                  \break point      cluster_table[0] = i                 |
   |                   \                                                     |
   +-------------------------------------------------------------------------+
                                                                            */
     n_photon[0] = n_photon[0] + n_cvtx[0];
     breakp[0] = n_photon[0] + 1;
     phmax     = ( float ) n_photon[0] + 1.0;
     dx0       = 30.0; 
                                                                            /*
    <<<<<<<<<>>>>>>>>>>>                                                    */
     set_clID( 0, 0 );
     set_kind( 0, 0 );
     create_straight( vt0, vt1 );                                           /*
    <<<<<<<<<>>>>>>>>>>>                                                    */
                                                                            /*
   +-------------------------------------------------------------------------+
   |(4) Determine the x, y positions of the initial line vertecies           |
   +-------------------------------------------------------------------------+
                                                                            */
     for( n = 1; n < nvertex; n++ )
        {                                                                   /*
                     <<<<<<>>>>>>>>>>                                       */
          vertex_n  = get_vtxno( n );                                       /*
                     <<<<<<>>>>>>>>>>                                       */
                                                                            /*
          +------------------------------------------------------------------+
	  |(a) x, y positions of vertecies which radiate photon/gluon        |
          +------------------------------------------------------------------+
                                                                            */
          if( n < breakp[0] )
            {
              x  =  x + dx0;
                                                                            /*
             <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                            */
              set_clID( vertex_n, no_cluster );
              addv_straight( vt0, vt1, vertex_n );                          /*
             <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                            */

             }
                                                                            /*
          +------------------------------------------------------------------+
	  |(b) x, y position of the break point vertex                       |
          +------------------------------------------------------------------+
                                                                            */
          else if( n == breakp[0] )
            {
              vtx_cntr = vertex_n;
              if( n_photon[0] == 0 )
                  x = x + 60.0;
              else
                  x  =  x + dx0;

              cluster_table[no_cluster] = vertex_n;
                                                                            /*
             <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                            */
              set_clID( vertex_n, no_cluster );
              addve_straight( vt0, vt1, vertex_n );                         /*
             <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                            */

              no_cluster++;
             }
          vtx[vertex_n].x = x;
          vtx[vertex_n].y = y;
         }
    }
 else
                                                                            /*
==============================================================================
  For the scattering processes
==============================================================================
                                                                            */
   {                                                                        /*
   +-------------------------------------------------------------------------+
   |(1) Counting the number of the initial radiations                        |
   |                                                                         |
   |    n_photon[0] : the number of initial radiations                       |
   |                  from the line connecting to vertex ID = 0              |
   |    n_photon[1] : the number of initial radiations                       |
   |                  from the line connecting to vertex ID = 1              |
   +-------------------------------------------------------------------------+
                                                                            */
     n_photon[0] = 0;
     n_cvtx[0]   = 0;
     for( n = 1; n < nvtx_iniline; n++ )
        {                                                                   /*
                    <<<<<<<>>>>>>>>>>>                                      */
          vertex_n = get_vtxno( n );                                        /*
                    <<<<<<<>>>>>>>>>>>   
             <<<<<<<<<<<<>>>>>>>>>>                                         */
          if( get_kind( vertex_n ) == 3 )                                   /*
             <<<<<<<<<<<<>>>>>>>>>>                                         */
              n_photon[0]++;
          else if( vtx[vertex_n].cntr_flag == 1 )
              n_cvtx[0]++;
          else
              n = nvtx_iniline;
         }
     n_photon[1] = 0;
     n_cvtx[1]   = 0;
     for( n = nvtx_iniline-2; n >= 0; n-- )
        {                                                                   /*
                    <<<<<<<>>>>>>>>>>>                                      */
          vertex_n = get_vtxno( n );                                        /*
                    <<<<<<<>>>>>>>>>>>   
             <<<<<<<<<<<<>>>>>>>>>>                                         */
          if( get_kind( vertex_n ) == 3 )                                   /*
             <<<<<<<<<<<<>>>>>>>>>>                                         */
              n_photon[1]++;
          else if( vtx[vertex_n].cntr_flag == 1 )
              n_cvtx[1]++;
          else
	      n = -1;
         }
                                                                            /*
   +-------------------------------------------------------------------------+
   |  n_IR : the total number of initial radiations                          |
   |  n_IRmax : the max. number of initial radiations                        |
   |                                                                         |
   |                            /     /     /                                |
   |                           /     /     n_photon[0] = 2                   |
   |     vertex ID = 0  +-----+-----+-----+--------                          |
   |                                      |                                  |
   |                                      |         n_IR = 3, n_IRmax = 2    |
   |                                      |                                  |
   |     vertex ID = 1  +--------+--------+--------                          |
   |                              \        n_photon[1] = 1                   |
   |                               \                                         |
   +-------------------------------------------------------------------------+
                                                                            */
     n_photon[0] = n_photon[0] + n_cvtx[0];
     n_photon[1] = n_photon[1] + n_cvtx[1];
     n_IR = n_photon[0] + n_photon[1];
     n_IRmax = n_photon[0];
     if( n_IRmax < n_photon[1] )
         n_IRmax = n_photon[1];
                                                                            /*
   +-------------------------------------------------------------------------+
   |(2) fix the position of initial vertex ID = 0                            |
   +-------------------------------------------------------------------------+
                                                                            */
     x        = 30.0,  y        = 30.0;
     vtx[0].x = x,     vtx[0].y = y;

     /*
       printf(" nvtx_iniline = %d   > 2 ?\n", nvtx_iniline );
       printf(" n_photon[0] = %d  n_photon[1] = %d\n", 
                             n_photon[0], n_photon[1] );
     */
                                                                            /*
   +-------------------------------------------------------------------------+
   |(3) To classify the types of initial lines                               |
   |    the number "nv" is calculated.                                       |
   |    nv : the number of internal lines between the internal vertecies     |
   |           on the initial line                                           |
   |       +                           +-----------+---------                |
   |        \                                      \                         |
   |         \        nv = 0                        +------- nv = 3          |
   |          +---------                            |                        |
   |         /        s channel                     +-------                 |
   |        /         scattering                   /                         |
   |      +                           +-----------+---------                 |
   |    ((Caution))                                                          |
   |      gluon-gluon scattering requires a special handling ( nv < 0 )      |
   +-------------------------------------------------------------------------+
                                                                            */
     if( nvtx_iniline > 2 )
       {
         nv   = nvtx_iniline - 3 - n_IR;
         phmax = ( float ) n_IRmax + 1.0;
         ph0   = ( float ) n_photon[0] + 1.0;
         ph1   = ( float ) n_photon[1] + 1.0;

          /*
	   printf(" nv = %d = ( 0, 1, 2, more ) ?\n", nv );
          */

	  nx      = no_final - n_IR + n_cvtx[0] + n_cvtx[1];
          IR_flag = 0;
                                                                            /*
       +---------------------------------------------------------------------+
       |                               v0                                    |
       |                               +---------+--------                   |
       |                                         |                           |
       |(A) Gluon - Gluon Scattering             +--------                   |
       |                                         |                           |
       |                               +---------+--------                   |
       |                               v1                                    |
       +---------------------------------------------------------------------+
                                                                            */
	if( nv < 0 || nx < 2)
           {
             IR_flag   =  1;
             breakp[0] =  1;
             breakp[1] = nvertex - 2;
             dx0       = 60.0; 
             dy0       = 10.0;
             dx1       = dx0;
             dy1       = dy0;

                                                                            /*
            <<<<<<<<>>>>>>>>>>>                                             */
             set_clID( 0, 0 );       
             create_straight( vt0, vt1 );                                   /*
            <<<<<<<<>>>>>>>>>>>                                             */

             for( n = 1; n < nvertex; n++ )
                {                                                           /*
                           <<<<<<<>>>>>>>>>>>                               */
                  vertex_n = get_vtxno( n );                                /*
                           <<<<<<<>>>>>>>>>>>                               */
                     
                  if( n == breakp[0] )
                    {
                          x  =  x + 60.0;
                          y  =  y + 10.0;
                          if( nvertex == 3 )
                              y  = y + 20.0;
                    
  
                      cluster_table[no_cluster] = vertex_n;
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster ); 
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      vt0  = vertex_n;
                      vt1  = -1;                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      create_straight( vt0, vt1 );                          /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */

                      no_cluster++;
                     }
                  else if( n <= breakp[1] )
                    {
                      if( nv == 1 ) y = y + 40.0;
                      else          y  =  y + 50.0;

                      cluster_table[no_cluster] = vertex_n;
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster ); 
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                  
                      vt1 = vertex_n;
                      if( n == breakp[1] )
                        {
                          vt0  = vertex_n;                    
                          vt1  = -1;                                         /*
                         <<<<<<<<<<<<<<>>>>>>>>>>>                          */
                          create_straight( vt0, vt1 );                       /*
                         <<<<<<<<<<<>>>>>>>>>>>>>>                          */
                         }

                      no_cluster++;
                     }
                  else if( n < nvertex - 1 )
                    {
                      x  =  x - dx1,  y  =  y + dy1; 
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster-1 ); 
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      vt1 = vertex_n;

                     }
                  else
                    {
                      x  =  x - dx1,  y  =  y + dy1;
                      if( nvertex == 3 )
                          y = y + 20.0;
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster-1 );
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      vt1  = vertex_n;

                     }

                  vtx[vertex_n].x = x;
                  vtx[vertex_n].y = y;
                 }
                                                                           /*
               <<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                            */
                reverse_straight( vt0, vt1 );                              /*
               <<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                            */
            }
                                                                            /*
       +---------------------------------------------------------------------+
       |                                   + V0                              |
       |                                    \                                |
       |                                     \                               |
       |(B) S channel process                 +------->                      |
       |                                     /                               |
       |                                    /                                |
       |                                   + V1                              |
       +---------------------------------------------------------------------+
                                                                            */
         else if( nv == 0 )
           {
             breakp[0] = n_photon[0] + 1;
             dx0       = phmax*30.0 / ph0; 
             dy0       = phmax*15.0 / ph0;
             dy0       = 30.0 / ph0;
             dx1       = phmax*30.0 / ph1; 
             dy1       = phmax*15.0 / ph1;
             dy1       = 30.0 / ph1;
                                                                            /*
            <<<<<<<<>>>>>>>>>>>                                             */
             set_clID( 0, 0 );
             create_straight( vt0, vt1 );                                   /*
            <<<<<<<<>>>>>>>>>>>                                             */

             for( n = 1; n < nvertex; n++ )
                {                                                           /*
                           <<<<<<<>>>>>>>>>>>                               */
                  vertex_n  = get_vtxno( n );                               /*
                           <<<<<<<>>>>>>>>>>>                               */

                  if( n < breakp[0] )
                    {
                      x  =  x + dx0, y = y + dy0;
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster ); 
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      vt1  = vertex_n;

                     }
                  else if( n == breakp[0] )
                    {
                      vtx_cntr = vertex_n;
                      if( n_photon[0] == 0 )
                        {
                          if( n_IRmax == 0 )
	                      x = x + 60.0, y = y + 30.0;
                          else
                            {
                              x  =  x + phmax*30.0;
                              /*y  =  y + phmax*15.0;*/
                              y  =  y + 30.0;
                             }
                         }
                      else
                          x  =  x + dx0,  y  =  y + dy0;
  
                      cluster_table[no_cluster] = vertex_n;
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster );                     /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      vt0  = vertex_n;
                      vt1  = -1;                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      create_straight( vt0, vt1 );                          /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */

                      no_cluster++;
                     }
                  else if( n < nvertex - 1 )
                     {
                       x  =  x - dx1,  y  =  y + dy1; 
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster-1 ); 
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      vt1  = vertex_n;

                     }
                  else
                    {
                      if( n_photon[1] == 0 )
                        {
                          if( n_IRmax == 0 )
	                      x = x - 60.0, y = y + 30.0;
                          else
                            {
                              x  =  x - phmax*30.0;
                              /*y  =  y + phmax*15.0;*/
                              y  =  y + 30.0;
                             }
                         }
                      else
                          x  =  x - dx1,  y  =  y + dy1;
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster-1 ); 
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      vt1  = vertex_n;

                     }

                  vtx[vertex_n].x = x;
                  vtx[vertex_n].y = y;
                 }
                                                                           /*
               <<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                            */
                reverse_straight( vt0, vt1 );                              /*
               <<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                            */
            }
                                                                            /*
       +---------------------------------------------------------------------+
       |                               v0                                    |
       |                               +---------+--------                   |
       |                                         |                           |
       |(C) Pure t channel process               |                           |
       |                                         |                           |
       |                               +---------+--------                   |
       |                               v1                                    |
       +---------------------------------------------------------------------+
                                                                            */
         else if( nv == 1 )
           {
             breakp[0] = n_photon[0] + 1;
             breakp[1] = nvertex - n_photon[1] - 2;
             dx0       = phmax*30.0 / ph0; 
             dy0       = 10.0 / ph0;
             dx1       = phmax*30 / ph1;
             dy1       = 10.0 / ph1;
                                                                            /*
            <<<<<<<<>>>>>>>>>>>                                             */
             set_clID( 0, 0 );
             create_straight( vt0, vt1 );                                   /*
            <<<<<<<<>>>>>>>>>>>                                             */

             for( n = 1; n < nvertex; n++ )
                {                                                           /*
                           <<<<<<<>>>>>>>>>>>                               */
                  vertex_n = get_vtxno( n );                                /*
                           <<<<<<<>>>>>>>>>>>                               */

                  if( n < breakp[0] )
                    {
                      x  =  x + dx0,   y = y + dy0;
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster );
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      vt1  = vertex_n;

                     }
                  else if( n == breakp[0] )
                    {
                      if( n_photon[0] == 0 )
                        {
                          x  =  x + phmax*30.0;
                          y  =  y + 10.0;
                          if( n_IRmax == 0 )
	                      x = x + 30.0;
                         } 
                      else
                          x  =  x + dx0,  y  =  y + dy0;
  
                      cluster_table[no_cluster] = vertex_n;
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster );
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      vt0  = vertex_n;
                      vt1  = -1;                                            /*
                     <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                       */
                      create_straight( vt0, vt1 );                          /*
                     <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                       */

                      no_cluster++;
                     }
                  else if( n <= breakp[1] )
                    {
                      y  =  y + 40.0;

                      cluster_table[no_cluster] = vertex_n;
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster );
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      vt1  = vertex_n;

                      if( n == breakp[1] )
                        {
                          vt0  = vertex_n;
                          vt1  = -1;                                        /*
                          <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                 */
                           create_straight( vt0, vt1 );                     /*
                          <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>                 */
                         }

                      no_cluster++;
                     }
                  else if( n < nvertex - 1 )
                    {
                      x  =  x - dx1,  y  =  y + dy1; 
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster-1 );
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      vt1  = vertex_n;

                     }
                  else
                    {
                      if( n_photon[1] == 0 )
                        {
                          x  =  x - phmax*30.0;
                          y  =  y + 10.0;
                          if( n_IRmax == 0 )
	                      x = x - 30.0;
                         }
                      else
                          x  =  x - dx1,  y  =  y + dy1;
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster-1 );  
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      vt1  = vertex_n;

                     }

                  vtx[vertex_n].x = x;
                  vtx[vertex_n].y = y;
                 }
                                                                           /*
               <<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                            */
                reverse_straight( vt0, vt1 );                              /*
               <<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                            */
            }
                                                                            /*
       +---------------------------------------------------------------------+
       |                                              break point[0]         |
       |                                 V0 +--------+-------                |
       |                                              \  break point[1]      |
       |                                               +------               |
       |                                               |                     |
       |(D) general t channel process                  :                     |
       |                                               | break point[2]      |
       |                                               +------               |
       |                                              /  break point[3]      |
       |                                 V1 +--------+-------                |
       +---------------------------------------------------------------------+
                                                                            */
         else
           {
             breakp[0] = n_photon[0] + 1;
             breakp[1] = breakp[0] + 1;
             breakp[3] = nvertex - n_photon[1] - 2;
             breakp[2] = breakp[3] - 1;
             dx0       = phmax*30.0 / ph0; 
             dy0       = 10.0 / ph0;
             dx1       = phmax*30 / ph1;
             dy1       = 10.0 / ph1;
                                                                            /*
            <<<<<<<<>>>>>>>>>>>                                             */
             set_clID( 0, 0 );
             create_straight( vt0, vt1 );                                   /*
            <<<<<<<<>>>>>>>>>>>                                             */

             for( n = 1; n < nvertex; n++ )
                {                                                           /*
                           <<<<<<<>>>>>>>>>>>                               */
                  vertex_n = get_vtxno( n );                                /*
                           <<<<<<<>>>>>>>>>>>                               */

                  if( n < breakp[0] )
                    {
                      x  =  x + dx0,    y = y + dy0;
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster );
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      vt1  = vertex_n;
                     }
                  else if( n == breakp[0] )
                    {
                      if( n_photon[0] == 0 )
                        {
                          x  =  x + phmax*30.0;
                          y  =  y + 10.0;
                          if( n_IRmax == 0 )
	                      x = x + 30.0;
                         }
                      else
                          x  =  x + dx0,  y  =  y + dy0;

                      cluster_table[no_cluster] = vertex_n;
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster );
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      vt0  = vertex_n;
                      vt1  = -1;                                            /*
                     <<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                         */
                      create_straight( vt0, vt1 );                          /*
                     <<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                         */

                      no_cluster++;
                     }
                  else if( n <= breakp[1] )
                    {
                      x  =  x +  5.0;
                      y  =  y + 50.0;

                      cluster_table[no_cluster] = vertex_n;
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster ); 
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      vt1 = vertex_n;
                      if( n == breakp[1] )
                        {
                          vt0  = vertex_n;
                          vt1  = -1;                                        /*
                          <<<<<<<<<<<<<<<>>>>>>>>>>>>>                      */
                           create_straight( vt0, vt1 );                     /*
                          <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                    */
                         }

                      no_cluster++;
                     }
                  else if( n <= breakp[2] )
                    {
                      y  =  y + 60.0;
 
                      cluster_table[no_cluster] = vertex_n;
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster );
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      vt1 = vertex_n;
                      if( n == breakp[2] )
                        {
                          vt0  = vertex_n;
                          vt1  = -1;                                        /*
                          <<<<<<<<<<<<<<<>>>>>>>>>>>>>                      */
                           create_straight( vt0, vt1 );                     /*
                          <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                    */
                         }

                      no_cluster++;
                     }
                  else if( n <= breakp[3] )
                    {
                      x  =  x -  5.0;
                      y  =  y + 50.0;

                      cluster_table[no_cluster] = vertex_n;
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster );  
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                                                                            /*
                     <<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                       */
                      reverse_straight( vt0, vt1 );                         /*
                     <<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                       */

                      vt1 = vertex_n;
                      if( n == breakp[3] )
                        {
                          vt0  = vertex_n;
                          vt1  = -1;                                        /*
                          <<<<<<<<<<<<<<<>>>>>>>>>>>>>                      */
                           create_straight( vt0, vt1 );                     /*
                          <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                    */
                         }

                      no_cluster++;
                     }
                  else if( n < nvertex - 1 )
                    {
                      x  =  x - dx1,  y  =  y + dy1; 
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster-1 );
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      vt1 = vertex_n;

                     }
                  else
                    {
                      if( n_photon[1] == 0 )
                        {
                          x  =  x - phmax*30.0;
                          y  =  y + 10.0;
                          if( n_IRmax == 0 )
	                      x = x - 30.0;
                         }
                      else
                          x  =  x - dx1,  y  =  y + dy1;
                                                                            /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      set_clID( vertex_n, no_cluster-1 );
                      addve_straight( vt0, vt1, vertex_n );                 /*
                     <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                    */
                      vt1 = vertex_n;

                     }

                  vtx[vertex_n].x = x;
                  vtx[vertex_n].y = y;
                 }
                                                                           /*
               <<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                            */
                reverse_straight( vt0, vt1 );                              /*
               <<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                            */
            }

/*************************/
	 set_kind( 0, 0 );
	 set_kind( 1, 0 );
       }
    }
/*
   print_vtxtable( );
*/

}

/*****************************************************************************
                     =======================                                */
                       void fix_inirad(  )
/*                   =======================
******************************************************************************
* ( Function )                                                               *
*   To determin the x, y position of the external vertecies for initial      *
*      radiation.                                                            *
*****************************************************************************/
{
                                                                            /*
--------------------------
  local variable list
--------------------------
                                                                            */
 int currentv, n_vertex, line_no, n_extern, n_intern, vertex_n;
 int nv, nx_vtx, nc;
 char *internal = "int", *external = "ext", *photon = "g";
 float ycenter, x, y, yt, y_new, ylower, yupper;
 int i, j, k, flag, clstrid, fl_cluster;
 int add;

                                                                            /*
----------------------------
  function list
----------------------------
                                                                            */
 int get_vtxno( );
 int get_clID ( );
 int get_kind ( );
 int  addr_vtxtable( );
 void add_vtxtable ( );
 void create_straight ( );
                                                                            /*
========================= Entry point of fix_iniline =========================
                                                                            */
for( add = 0; add < nvtx_iniline; add++ )
   {                                                                        /*
             <<<<<<<>>>>>>>>>>>                                             */
   currentv = get_vtxno( add );                                             /*
             <<<<<<<<<<<<>>>>>>>>>>>>>                                      */
   clstrid  = get_clID( currentv );                                         /*
             <<<<<<<<<<<<>>>>>>>>>>>>>                                      */

   fl_cluster = 0;
   if( clstrid == 0 || clstrid == no_cluster-1 ) fl_cluster = 1;
                                                                            /*
==============================================================================
 Consider only those vertices which locate on the real initial lines
==============================================================================
                                                                            */
   if( fl_cluster && currentv != cluster_table[clstrid] )
     {

           /*
           printf("*** fix_inirad ***\n");
           printf(" currentv(%4d) clstrid(%d)\n",
                    currentv, clstrid);
           print_vtxtable( currentv, 1 );
           */

       nx_vtx = nvertex;
                                                                            /*
     +-----------------------------------------------------------------------+
     |  Consider only those vertices which radiate a single photon/gluon     |
     +-----------------------------------------------------------------------+
                                       
          <<<<<<<<<<>>>>>>>>>>>>                                            */
       if( get_kind( currentv ) == 3 && IR_flag == 0 )                      /*
          <<<<<<<<<<>>>>>>>>>>>>                                            */
         {
           n_vertex = vtx[currentv].no_line;

           if( n_vertex == 3 )
             {
               x        = vtx[currentv].x;
               y        = vtx[currentv].y;
                    /*
                     printf(" n_vertex = %d  ( x, y ) = (%6.1f, %6.1f)\n", 
                              n_vertex, x, y );
                    */
              
               for( i = 0; i < n_vertex; i++ )
                  {
                                                                            /*
                   +---------------------------------------------------------+
	           |(1) Select a vertex connecting to the current vertex     |
                   +---------------------------------------------------------+
                                                                            */
                    vertex_n = vtxline[currentv].vtxno[i];

                          /*
                            printf(" i = %2d   vertex_n(%4d)", i, vertex_n );
                          */
                                                                            /*
                   +---------------------------------------------------------+
                   |(2) if the vertex connecting to the current vertex       |
                   |      is not on the vertex_table,                        |
                   |      save the new vertex on the vertex_table.           |
                   +---------------------------------------------------------+

                       <<<<<<<<<<<<>>>>>>>>>>>>>>>                          */
                    if( addr_vtxtable( vertex_n ) == -1 )                   /*
                       <<<<<<<<<<<<>>>>>>>>>>>>>>>                          */
                      {
                        line_no = vtxline[currentv].lineno[i];
                                                                            /*
                       <<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>        */
                        add_vtxtable( 1, vertex_n, 2, 0, clstrid );
                        create_straight( currentv, vertex_n );              /*
                       <<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>        */

                            /*
                             printf("  clstrid(%2d)\n", clstrid );
                             printf("                  ");
                             printf(" intline[line_no = %d].kind = %s\n", 
                                  line_no, intline[line_no].kind );
                            */
                                                                            /*
                       +-----------------------------------------------------+
                       |(i) for s-channel interactions :                     |
		       |      Determine the x-y position                     |
                       +-----------------------------------------------------+
                                                                            */
                        if( no_cluster  == 1 )
                          {
                            ycenter = vtx[cluster_table[0]].y;
	                    x  = x + 40.0;
                            if( y > ycenter )
                                y  = y + 20.0;
                            else
                                y  = y - 20.0;

                            vtx[vertex_n].x = x;
                            vtx[vertex_n].y = y;

                           }
                                                                            /*
                       +-----------------------------------------------------+
                       |(ii) for t-channel interactions :                    |
		       |      Determine the x-y position                     |
                       +-----------------------------------------------------+
                                                                            */
                        else 
                          {
                            ycenter = vtx[cluster_table[clstrid]].y;
	                    x  = x + 40.0;
                            if( y < ycenter )
		                y = y - 20.0;
                            else 
		                y = y + 20.0;
                   
                            vtx[vertex_n].x = x;
                            vtx[vertex_n].y = y;
                           }
                       }
                   }

              }
            }
        }
    }
      

}
int ngt, ngen, nf, nfset, rootv;
float xrootv, yrootv, ysetmin, xsetmax;

/*****************************************************************************
                ========================================                    */
                      void fix_vertex(  )
/*              ========================================
******************************************************************************
*( Function )                                                                *
*   To fix the x-y positions of vertices for tree diagram                    *
******************************************************************************
                                                                            */
{
                                                                            /*
 ---------------------
  Local variables
 ---------------------
                                                                            */
 int i, j, n, nv, v, v0, v1;
 int ngen0, nf0;
 float ylow, yhigh;
 int cvtx;
                                                                            /*
 ---------------------
  Functions
 ---------------------
                                                                            */
 void print_vtxtable( );
 int  addr_vtxtable ( );
 void add_vtxtable  ( );
 void get_nofinal( );
 void fix_schannel( );
 void fix_tchannel( );
 void create_straight( );
                                                                            /*
============================ Entry point of fix_vertex =======================
                                                                            */
  /*
    print_vtxtable( );
  */
                                                                            /*
==============================================================================
(I) for s-channel processes
==============================================================================
									    */
if( no_cluster == 1 )
  {
                                                                            /*
------------------------------------------------------------------------------
   (1) Get the numbers of final vertices and of generations
------------------------------------------------------------------------------
									    */
    v  = cluster_table[0];
           /*
            printf(" cluster vertex [%d] = (%8.1f,%8.1f )\n", 
	             v, vtx[v].x, vtx[v].y );
           */
    nv = vtx[v].no_line;
    n  = 0;
    for( i = 0; i < nv; i++ )
       {
         v1 = vtxline[v].vtxno[i];
                                                                            /*
                       <<<<<<<<<<>>>>>>>>>>>                                */
         if( v1 != v && addr_vtxtable( v1 ) == -1)                          /*
                       <<<<<<<<<<>>>>>>>>>>>                                */
           {
             v0 = v1;
             n++;
            }
        }
    if( n == 1 )
      {
        cluster_table[0] = v0;
                                                                            /*
       <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                                   */
        add_vtxtable( 0, v0, 1, 0, 0 );                                     /*
       <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                                   */
        create_straight( v, v0 );                                           /*
       <<<<<<<<<<<<<<<>>>>>>>>>>>>>                                         */

	vtx[v0].x = vtx[v].x + 60.0;
	vtx[v0].y = vtx[v].y;
        v = v0;
       }
     /*
      printf(" cluster vertex [%d] = (%8.1f,%8.1f )\n", 
	       v, vtx[v].x, vtx[v].y );
      printf("**** call get_nofinal [%d] for the first\n", v );
     */
    ngt  = 0;
    ngen = 0;
    nf   = 0;
                                                                            /*
   <<<<<<<<<>>>>>>>>>>>>>>                                                  */
    get_nofinal( v, v );                                                    /*
   <<<<<<<<<>>>>>>>>>>>>>>                                                  */
              /*
               printf("**** nf =%3d  ngen = %d\n", nf, ngen );
              */
                                                                            /*
------------------------------------------------------------------------------
   (2) Determine the (x,y) positions of vertices
------------------------------------------------------------------------------
									    */
    xrootv  = vtx[v].x;
    yrootv  = vtx[v].y;
    xsetmax = xrootv + 60.0*( ngen - 1 );
    ysetmin = yrootv - 30.0*( nf - 1 );
    nfset   = 0;
                                                                            /*
    <<<<<<<<<<<<>>>>>>>>>>>>>>                                              */
     fix_schannel( v, v, 0 );                                               /*
    <<<<<<<<<<<<>>>>>>>>>>>>>>                                              */

   }
                                                                            /*
==============================================================================
(I) for t-channel processes
==============================================================================
									    */
else if( no_cluster > 1 )
  {
                                                                            /*
------------------------------------------------------------------------------
   (1) Get the numbers of final vertices and of generations
------------------------------------------------------------------------------
									    */
    ngen0 =  0;
    nf0   =  0;
    for( i = 0; i < no_cluster; i++ )
       {
         v  = cluster_table[i];

         ngt  = 0;
         ngen = 0;
         nf   = 0;
                                                                            /*
        <<<<<<<<>>>>>>>>>>>>>>                                              */
         get_nofinal( v, v );                                               /*
        <<<<<<<<>>>>>>>>>>>>>>                                              */

         nf0 += nf;
         if( ngen > ngen0 ) ngen0 = ngen;

            /*
             printf("**** cluster_ID =%3d  root vertex =%3d\n", i, v );
             printf("            nf  =%3d        ngen  = %d\n", nf, ngen );
             printf("            nf0 =%3d        ngen0 = %d\n", nf0, ngen0 );
            */
        }

    if( no_cluster > 2 )
        xrootv = vtx[cluster_table[1]].x;
    else
        xrootv = vtx[cluster_table[0]].x;
    yrootv = vtx[cluster_table[0]].y;

    xsetmax = xrootv + 60.0*( ngen0 - 1 );
    ysetmin = yrootv - 10.0;
    nfset   = 0;
                                                                            /*
------------------------------------------------------------------------------
   (2) Determine the (x,y) positions of vertices
------------------------------------------------------------------------------
                                                                            */
    cvtx  = -1;
    for( i = 0; i < no_cluster; i++ )
       {
         v  = cluster_table[i];
            
         if( i == 0 )
           {                                                                /*
               <<<<<<<<<<<<>>>>>>>>>>>>>>>>>                                */
                fix_tchannel( i, v, v, -2 );                                /*
               <<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                               */

              ylow = vtx[v].y;
            }
         else if( i == no_cluster-1 )
           {                                                                /*
               <<<<<<<<<<<<<<>>>>>>>>>>>>>>>                                */
                fix_tchannel( i, v, v,  2 );                                /*
               <<<<<<<<<<<<<<>>>>>>>>>>>>>>>                                */

               if( cvtx >= 0 )
                 {
                   vtx[cvtx].y = 0.5*( ylow + vtx[v].y );
                  }
            }
         else  
           {
              if( vtx[v].cntr_flag == 1 )   cvtx = v;
              else
                {                                                           /*
                  <<<<<<<<<<<<<>>>>>>>>>>>>>>>>>                            */
                    fix_tchannel( i, v, v,  0 );                            /*
                  <<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                            */

                  if( cvtx >= 0 )
                    {
                      vtx[cvtx].y = 0.5*( ylow + vtx[v].y );
                      cvtx        = -1;
                     }
                    
                  ylow = vtx[v].y;
                 }
              
             }
        }
   }
}

/*****************************************************************************
*                      ================================                     */
                       void get_nofinal(parentv, currentv )
/*                     ================================                      *
******************************************************************************
* ( Function )                                                               *
*   To get the number of final vertices and the generation numbers to        *
*      the final vertices connected to the parentv.                          *
******************************************************************************
                                                                            */
int parentv, currentv;
{
									    /*
-------------------
 Local variables
-------------------
									    */
 int    n, nv,  v;
									    /*
-------------------
 Local variables
-------------------
									    */
 void get_nofonal  ( );
 int  addr_vtxtable( );
									    /*
============================= Entry point ====================================
									    */
          /*
           printf("get_nofinal parentv[%d] currentv[%d]\n", 
             	      parentv, currentv );
          */
 ngt++;
 nv = vtx[currentv].no_line;

 if( nv == 1 )
   {
         if( ngt > ngen ) ngen = ngt;
         nf++;
            /*
	    printf("  get_nofinal has found a final vertex {%3d ]\n", 
		      currentv );
            printf("   generation_depth =%3d, number of finals =%3d\n", 
		      ngen, nf);
            */
    }
 else
   {
     for( n = 0; n < nv; n++ )
        {
          v  = vtxline[currentv].vtxno[n];
                                                                            /*
                             <<<<<<<<<>>>>>>>>>>>                           */
          if( v != parentv && addr_vtxtable( v ) == -1)                     /*
                             <<<<<<<<<>>>>>>>>>>>                           */
            {                                                               /*
             <<<<<<<<<<<<<<>>>>>>>>>>>>>>                                   */
              get_nofinal( currentv, v );                                   /*
             <<<<<<<<<<<<<<>>>>>>>>>>>>>>                                   */
             }
         }
    }

 ngt--;
}

float xf, yf, ysum, y_adj;
int vtxflag[100];
/*****************************************************************************
*                =============================================              */
                  void fix_schannel(parentv, currentv, flag )
/*               =============================================
******************************************************************************
* (Function )                                                                *
*  To fix the x-y positions of vertices for s channel processes              *
*****************************************************************************/
int parentv, currentv, flag;
{
									    /*
-------------------
 Local variables
-------------------
									    */
 float x0, y0, xc, yc, dydx, b;
 int   i, exchange, flag1, nov, n, nv, v, vt;
 int   next[3][10];
 char *photon = "g";
									    /*
-------------------
  Functions
-------------------
									    */

 void fix_schannel( );
 void fixvtx_trace( );
 int  addr_vtxtable( );
 void add_vtxtable( );
 void addve_straight( );
 void create_straight( );
 int  find_straight( );
									    /*
============================= Entry point ====================================
									    */
if( parentv != currentv )
  {
      /*
       printf("**** call fixvtx_trace [%d]\n", parentv, currentv );
      */

    ngen = 0;
    ngt  = 0;
    nf   = 0;
    ysum = 0.0;
                                                                            /*
    <<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>                              */
     fixvtx_trace( parentv, currentv, flag );                               /*
    <<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>                              */

    if( nf > 0 )
      {
        ngen--;
        xc   = vtx[parentv].x;
        if(      ngen == 0 ) x0 = xsetmax;
        else if( ngen == 1 ) x0 = 0.5*( xc + xsetmax );
        else                 x0 = xc + 60.0;

        if( flag != 0 )
          {
            yc   = vtx[parentv].y;
            dydx = ( yf - yc )/( xf - xc );
            b    = yc - dydx*xc;
            y0   = dydx*x0 + b;
           }
        else if( nf == 1 )
            y0   = yf;
        else
            y0   = ysum / nf;
                                                                            /*
          <<<<<<<<<<<<<<<>>>>>>>>>>>>>                                      */
        if( addr_vtxtable( currentv ) == -1 )                               /*
          <<<<<<<<<<<<<<<>>>>>>>>>>>>>                                      */
          {
            vtx[currentv].x = x0;
            vtx[currentv].y = y0;
            if( nf == 1 ) 
	      {                                                             /*
               <<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                     */
                add_vtxtable( -1, currentv, 2, 0, 0 );                      /*
               <<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                     */
		nfset++;
               }
            else                                                            /*
               <<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                     */
                add_vtxtable( -1, currentv, 1, 0, 0 );                      /*
               <<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                     */
           }
       }
   }


 nv = vtx[currentv].no_line;
 n  = 0;
 for( i = 0; i < nv; i++ )
    {
      v   = vtxline[currentv].vtxno[i];
                                                                            /*
                         <<<<<<<<<>>>>>>>>>>>                               */
      if( v != parentv && addr_vtxtable( v ) == -1 )                        /*
                         <<<<<<<<<>>>>>>>>>>>                               */
        {
          next[0][n] = v;
          next[1][n] = vtxline[currentv].lineno[i];
          next[2][n] = vtx[v].no_line;
          n++;
         }
     }
 nov = n;

     /*
       printf(" %d vertices connected to  v[%d] are found\n", 
                nov, currentv );
       printf(" vtx_no line_n  no_v\n");
       for( i = 0; i < nov; i++ )
            printf(" %6d%6d%6d\n", next[0][i], next[1][i], next[2][i] );
     */

 if( vtx[currentv].flag == 0 && nov == 2 )
   {
     exchange = 0;
     if( flag == -1 )
       {
         if( next[2][1] == 1 && next[2][0] == 1 )
             if( strcmp( intline[next[1][0]].species, photon ) == 0 )
                 exchange = 1;
        }
     else if( flag == 1 )
       {
         if( next[2][0] == 1 && next[2][1] == 1 )
             if( strcmp( intline[next[1][1]].species, photon ) == 0 )
                 exchange = 1;
        }
     if( exchange == 1 )
       {
         n          = next[0][0];
         next[0][0] = next[0][1];
         next[0][1] = n;
          /*
            printf(" %d vertices connected to  v[%d] are reordered\n", 
                     nov, currentv );
            for( i = 0; i < nov; i++ )
                 printf("%6d", next[0][i] );
            printf("\n");
          */
        }
    }

 nv = nov / 2 ;
 for( n = 0; n < nov; n++ )
    {
      v = next[0][n];

      if( nov == 1 )
          if( vtx[currentv].cntr_flag == 1 )
                              flag1 = flag;
          else                flag1 =    0;
      else if( n <     nv   ) flag1 =   -1;
      else if( n > nov-nv-1 ) flag1 =    1;
      else                    flag1 =    0;
             
           /*
             printf("**** call fix_sch  " );
             printf("pv[%3d ]  cv[%3d ]  v[%3d]  flag[%3d ] flag1[%3d ]\n", 
                      parentv, currentv, v, flag, flag1 );
           */

      if( flag == flag1 )
        {
          if( parentv == currentv )
            {                                                               /*
              <<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                            */
                create_straight( currentv, v );                             /*
              <<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                            */
             }
          else if( n == 0 || n == nov-1 )
            {
              vt = find_straight( currentv );                               /*
              <<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>                        */
               addve_straight( vt, currentv, v );                           /*
              <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>                        */
             }
         }
      else
        {                                                                   /*
         <<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                                */
          create_straight( currentv, v );                                   /*
         <<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                                */
         }
            
                                                                            /*
         <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                               */
	  fix_schannel( currentv, v, flag1 );                               /*
         <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                               */

     }
}

/*****************************************************************************
*           =======================================================         */
             void fix_tchannel( clstrid, parentv, currentv, flag )
/*          =======================================================
******************************************************************************
* (Function )                                                                *
*  To fix the x-y positions of vertices for t-channel processes              *
*****************************************************************************/
int clstrid, parentv, currentv, flag;
{
									    /*
-------------------
 Local variables
-------------------
									    */
 float x0, y0, xc, yc, yt, dy, dydx, b;
 int   i, exchange, flag1, nov, n, nv, v, vt;
 int   next[3][10];
 char *photon = "g";
									    /*
-------------------
  Functions
-------------------
									    */
 int  addr_vtxtable( );
 void add_vtxtable( );
 void fixvtx_trace ( );
 void fix_tchannel ( );
 void create_straight ( );
 void addve_straight ( );
 int iabs( );
									    /*
============================= Entry point ====================================
									    */
                                                                            /*
+----------------------------------------------------------------------------+
|(1)  Preparation                                                            |
+----------------------------------------------------------------------------+
                                                                            */
if( parentv == currentv )
  {
    rootv  = currentv;
    xrootv = vtx[currentv].x;
    yrootv = vtx[currentv].y;
   }
 flag1  = flag;

     /*
      printf("** fix_tch : currentv[%3d] rootv[%3d](%8.2f,%8.2f) flag(%2d)\n",
                  currentv, rootv, xrootv, yrootv, flag );
     */
                                                                            /*
+----------------------------------------------------------------------------+
|(2) count the number of generation at the down stream of the current vtx    |
+----------------------------------------------------------------------------+
                                                                            */
 ysum = 0.0;
 ngen = 0;
 ngt  = 0;
 nf   = 0;
                                                                            /*
 <<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>                                 */
  fixvtx_trace( parentv, currentv, flag );                                  /*
 <<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>                                 */

       /*
       printf("  fix_tch : ngen =%3d  nf =%3d, ysum =%8.2f\n",
                 ngen, nf, ysum );
       */
                                                                            /*
+----------------------------------------------------------------------------+
|(3) Determine the x-y position of the current vtx                           |
+----------------------------------------------------------------------------+
                                                                            */
 if( nf > 0 )
   {
     ngen--;
     xc   = vtx[parentv].x;
     yc   = vtx[parentv].y;
     if(      ngen == 0 ) x0 = xsetmax;
     else if( ngen == 1 ) x0 = 0.5*( xc + xsetmax );
     else                 x0 = xc + 60.0;

     y0 = ysum/(float) nf;
     dy = 0.0;
                                                                            /*
    +------------------------------------------------------------------------+
    |(a) current vtx is neither on a upper nor lower line ( a middle line )  |
    +------------------------------------------------------------------------+
                                                                            */
     if( flag  == 0 )
       {                                                                    /*
         +-------------------------------------------------------------------+
         |(i) The current vtx is a root vertex of a cluster                  |
         +-------------------------------------------------------------------+
                                                                            */
         if( currentv == cluster_table[clstrid] )
           {
             x0 = vtx[currentv].x;
             yt = vtx[currentv].y;
             dy = y0 - yt;
            }
                                                                            /*
         +-------------------------------------------------------------------+
         |(ii) The current vtx is a counter term on a propagator             |
         +-------------------------------------------------------------------+
                                                                            */
         else if( vtx[currentv].cntr_flag == 1 )
           {
             y0 = yc + 0.5 * ( y0 - yc );
            }
        }
                                                                            /*
    +------------------------------------------------------------------------+
    |(b) current vtx is on a upper or lower line                             |
    +------------------------------------------------------------------------+
                                                                            */
     else
       {
         if( parentv == currentv )
           {
             if(      flag == -2 ) yc = yf + 10.0;
             else if( flag ==  2 ) yc = yf - 10.0;
            }
         dydx = ( yf - yc )/( xf - xc );
         b    = yc - dydx*xc;
         y0   = dydx*x0 + b;

         if( iabs( flag ) == 2 )
           {
             if( nf == 1 && ngen == 0 ) y0 = yf;
             else if( flag == -2 )      y0 = yf + 10.0;
             else                       y0 = yf - 10.0;

             if( currentv == cluster_table[clstrid] )
               {
                 dy = y0 - vtx[currentv].y;
                 x0 = vtx[currentv].x;
                }
            }
        }
                                                                            /*
    +------------------------------------------------------------------------+
    |(c) Save the position                                                   |
    +------------------------------------------------------------------------+
                                                                            */
     vtx[currentv].x = x0;
     vtx[currentv].y = y0;
                                                                            /*
+----------------------------------------------------------------------------+
|(4) Store the current vtx on the vertex table                               |
+----------------------------------------------------------------------------+

        <<<<<<<<<<<<<<>>>>>>>>>>>>>                                         */
     if( addr_vtxtable( currentv ) == -1 )                                  /*
        <<<<<<<<<<<<<<>>>>>>>>>>>>>                                         */
       {
         if( nf == 1 ) 
           {                                                                /*
            <<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                        */
             add_vtxtable( -1, currentv, 2, 0, 0 );                         /*
            <<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                        */
             nfset++;
            }
         else
                                                                            /*
            <<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                        */
             add_vtxtable( -1, currentv, 1, 0, 0 );                         /*
            <<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                        */
        }
    }
 if( dy != 0.0 && currentv == cluster_table[clstrid] )
   {
     for( i = 0; i < nvertex; i++ )
        {                                                                   /*
             <<<<<<<>>>>>>>>>>>                                             */
          v  = get_vtxno( i );                                              /*
             <<<<<<<>>>>>>>>>>>                                             */
          if( get_clID( v ) >= clstrid )                                    /*
             <<<<<<<>>>>>>>>>>>                                             */
            {
              if( v != currentv ) vtx[v].y += dy;
             }
         }
    }

 nv = vtx[currentv].no_line;
 n  = 0;
 for( i = 0; i < nv; i++ )
    {
      v   = vtxline[currentv].vtxno[i];                                     /*
                         <<<<<<<<<<>>>>>>>>>>                               */
      if( v != parentv && addr_vtxtable( v ) == -1 )                        /*
                         <<<<<<<<<<>>>>>>>>>>                               */
        {
          next[0][n] = v;
          next[1][n] = vtxline[currentv].lineno[i];
          next[2][n] = vtx[v].no_line;
          n++;
         }
     }
 nov = n;

      /*
       printf(" %d vertices connected to  v[%d] are found\n", 
                nov, currentv );
       printf(" vtx_no line\#  no_v\n");
       for( i = 0; i < nov; i++ )
            printf(" %6d%6d%6d\n", next[0][i], next[1][i], next[2][i] );
      */

 if( vtx[currentv].flag == 0 && nov == 2 )
   {
     exchange = 0;
     if( flag == -2 )
       {
         if( next[2][1] == 1 )
           {
             if( next[2][0] == 1 )
               {
                 if( strcmp( intline[next[1][0]].species, photon ) == 0 )
                     exchange = 1;
                }
             else if( next[2][0] > 1 )
               {
                 exchange = 1;
                }
            }
        }
     else if( flag == -1 )
       {
             if( next[2][0] == 1 )
                 if( strcmp( intline[next[1][0]].species, photon ) == 0 )
                     exchange = 1;
        }
     else if( flag == 1 )
       {
             if( next[2][1] == 1 )
                 if( strcmp( intline[next[1][1]].species, photon ) == 0 )
                     exchange = 1;
        }
     else if( flag == 2 )
       {
         if( next[2][0] == 1 )
           {
             if( next[2][1] == 1 )
               {
                 if( strcmp( intline[next[1][1]].species, photon ) == 0 )
                     exchange = 1;
                }
             else if( next[2][1] > 1 )
               {
                 exchange = 1;
                }
            }
        }
     if( exchange == 1 )
       {
         n          = next[0][0];
         next[0][0] = next[0][1];
         next[0][1] = n;

	   /*
            printf(" %d vertices connected to  v[%d] are reordered\n", 
                     nov, currentv );
            for( i = 0; i < nov; i++ )
                 printf("%6d", next[0][i] );
            printf("\n");
           */
        }
    }

 nv = nov / 2 ;
 for( i = 0; i < nov; i++ )
    {

      v   = next[0][i];

      if( vtx[currentv].cntr_flag == 1 )
                                     flag1 = flag;
      else if( flag == -2 )
        {
           if( i ==     0 )          flag1 = -2;
           else                      flag1 =  0;
         }
      else if( flag ==  2 )
        {
           if( i == nov-1 )          flag1 =  2;
           else                      flag1 =  0;
         }
      else
        {
           if(      i <     nv )     flag1 = -1;
           else if( i > nov-nv-1 )   flag1 =  1;
           else                      flag1 =  0;
         }

             /*
              printf("**** call fix_tch  cl[%2d] ", clstrid );
              printf("pv[%3d ]  cv[%3d ]  v[%3d]  flag[%3d ] flag1[%3d ]\n", 
                      parentv, currentv, v, flag, flag1 );
             */

      if( flag == flag1 )
        {
          if( parentv == currentv )
            {                                                               /*
              <<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                            */
                create_straight( currentv, v );                             /*
              <<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                            */
             }
          else
            {
              vt = find_straight( currentv );                               /*
              <<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>                        */
               addve_straight( vt, currentv, v );                           /*
              <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>                        */
             }
         }
      else
        {                                                                   /*
         <<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                                */
          create_straight( currentv, v );                                   /*
         <<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                                */
         }
            
                                                                            /*
      <<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>                         */
       fix_tchannel( clstrid, currentv, v, flag1 );                         /*
      <<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>                         */

     }
}

/*****************************************************************************
*            ==============================================                 */
               void fixvtx_trace(parentv, currentv, flag )
/*           ==============================================
******************************************************************************
* (Function )                                                                *
*  To adjust the (x,y) positions of vertices of tree diagrams                *
*****************************************************************************/
int parentv, currentv;
{
									    /*
-------------------
 Local variables
-------------------
									    */
 float x0, y0, x1;
 int    n, nv,  v;
									    /*
-------------------
 Local variables
-------------------
									    */
 void fixvtx_trace( );
 int  addr_vtxtable( );
									    /*
============================= Entry point ====================================
									    */
 ngt++;

 nv = vtx[currentv].no_line;
 if( nv == 1 )
   {
         if( ngt > ngen ) ngen = ngt;
         xf = xsetmax;
         y0 = ysetmin + 60.0*nfset + 60.0* nf;
         if( nf == 0 ) 
           {
             ysum = 0.0;
             yf = y0;
            }
         ysum += y0;
         nf++;

         if( flag < 0 )
           {
             if( y0 < yf ) yf = y0;
            }
         else if( flag > 0 )
           {
             if( y0 > yf ) yf = y0;
            }
           /*
	    printf("  fixvtx_trace has found a final vertex {%3d ]\n", 
		      currentv );
            printf("   generation_depth =%3d, number of finals =%3d\n", 
		      ngen, nf);
            printf("   yf =%8.2f  y0 =%8.2f  ysum =%10.2f\n", 
		      yf, y0, ysum );
           */
    }
 else
   {
     for( n = 0; n < nv; n++ )
        {
          v  = vtxline[currentv].vtxno[n];
          x1 = vtx[v].x;

           /*
           printf("parentv[%d] currentv[%d] next[%d]_x =%8.2f >? x0 =%8.2f\n", 
             	      parentv, currentv, v, v, x1, x0 );
           */
                                                                             /*
                             <<<<<<<<<<>>>>>>>>>>                            */
          if( v != parentv && addr_vtxtable( v ) == -1 )                     /*
                             <<<<<<<<<<>>>>>>>>>>                            */
            {                                                                /*
             <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                            */
              fixvtx_trace( currentv, v, flag );                             /*
             <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                            */
             }
         }
    }

 ngt--;
}
