  
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>

#define MXLEN  80
#define MXVTX  100
#define MXLINE 100
#define MXGRF  10000
#define MXEXTV 100
                                                                      /*
 ===================================================================
     Structure vertex, vtx_line, int_line, inl_vtx
 ===================================================================
                                                                      */
struct vertex {
        int     flag;
        int     cntr_flag;
	char    kind[4];       
	float   x;         
	float   y;   
	int     no_line;   
        }      vtx[MXVTX];

struct vtx_line {
	int   vtxno[10];
        int   lineno[10];
        }     vtxline[MXVTX];

struct int_line {
	int   flag;
	char  kind[4];      
        float x;
        float y;
	int   direct;   
	char  species[4];  
	char  form[4];  
	int   print;  
        }     intline[MXLINE];

struct inl_vtx {
	int   vtxno[2];
	}     inlvtx[MXLINE];
                                                                      /*
 ====================================================================
     Structure self_energy,
 ====================================================================
                                                                      */
int no_self;
struct self_energy {
	int   edge[2];
        int   selfv[2];
        int   edge_line[2];
        int   self_line[2];
       }      self[3];
                                                                      /*
 ====================================================================
     Structure vertex_correction,
 ====================================================================
                                                                      */
int no_vcor;
struct vertex_correction {
	int   edge[3];
	float edgex[3];
	float edgey[3];
        int   vcorv[3];
        int   edge_line[3];
        int   edge_flag[3];
        int   vcor_line[3];
       }      vtxcor[3];
                                                                      /*
 ====================================================================
     Structure loop_correction,
 ====================================================================
                                                                      */
int no_multi;
struct multiloop_correction {
	int   no_edge;
	int   edge[10];
	float edgex[10];
	float edgey[10];
        int   artv[10];
        int   artv_flag[10];
        int   edge_line[10];
        int   edge_flag[10];
	int   no_vtx;
        int   vtxno[10];
       }      multi[3];
                                                                      /*
=====================================================================
     Structure for keeping original information of graphs
=====================================================================
                                                                      */
int no_loop, cflag[10];
struct keep_vertex {
        int   flag;
        int   no_line;
	int   vtxno[10];
        int   lineno[10];
        }     keepvtx[500];

struct keep_line {
	int   flag;
	int   vtxno[2];
	}     keepline[500];
                                                                      /*
  ----------------------------------
      global variable list
  ----------------------------------
                                                                      */
long addr_graph[5][MXGRF];
                                                                      /*
    addr_graph[0][i] : graph number
    addr_graph[1][i] : pointer of this graph to the temp file
    addr_graph[2][i] : the number of vertecies in the graph
    addr_graph[3][i] : the array size of vtx_no[] and intline_no[]
    addr_graph[4][i] : number of internal and external lines
                                                                      */
int level[2][MXGRF];
int no_graph, graph_no;
int no_initial, no_final;
int nvtx_iniline;            /* The number of vertecies on the initial line */
int no_iniline;            /* The number of vertecies on the initial line */
int nvertex;
int vtx_table[4][MXEXTV];

int no_vertex, no_line;
float xmax, ymax;
float size[2][MXGRF];                             /* relative size of graph */
int cluster_table[MXEXTV], no_cluster; /* cluster table and no. of clusters */
int n_IRmax, n_photon[2];

int add1, add2, rm_flag;
int   vc0, vj; 
float xc0, yc0;
int fermiflag, fermi[10]; 
int horizflag, horiz[10]; 
int vertflag,  vert[10];
int dupvflag,  dupv[10];
int tmpbuf[3][10]; 
int no_art, no_inside;
int no_in, v_in[2],   j_in[2];
int n_out, v_out[10], j_out[10];
int count, vtxbuf[100];

int dfnumber[MXEXTV];


/*****************************************************************************
                   ============================                             */
                      void put_loop(  )
/*                 =============================
******************************************************************************
*( Function )                                                                *
*     To put the loop parts in the tree Feynman diagram,  and                *
*(input)                                                                     *
*     structure vtx[i] : vertex information                                  *
*     structure intline[i] : line information                                *
*( output )                                                                  *
*     The vertex position in the graph                                       * 
*         structure vtx[].x and structure vtx[].y                            *
******************************************************************************
*/
{
                                                                            /*
-----------------------------
     List of Functions
-----------------------------
                                                                            */
void put_self( );
void put_vcor( );
void put_multi( );
void recover( );

void tgdlist( );
void print_straight( );
/*
================= Entry point of putloop ====================================
*/

int l, n, nv;

void print_vtxtable( );


while( no_loop > 0 ) 
  {

/********* For debug purpose *****
   if( no_loop != 1 ){ 
/*********************************/
   /*
     printf("*** putloop starts *****\n" );
     printf("*** graph number %d *****\n", graph_no );
     printf(" number of loops: no_loop =%d\n", no_loop );
     printf(" List of rm_flags: cflag[]\n");
     for( n = 0; n < no_loop; n++ )
          printf("%5d", cflag[n] );
     printf("\n");

    
     print_vtxtable();
   
     printf("****** list of initial radiations ***************\n" );
     printf("   number of max. IRs =%4d\n", n_IRmax );
     for( n = 0; n < no_initial; n++ )
        {
          printf("   initial vertex =%4d  numbers of IRs =%4d\n",
	  n, n_photon[n] );
         }
   */
                                                                             /*
===============================================================================
           Recover the information of the loop part
===============================================================================
                                                                             */
    recover( );

   /*
    printf(" after recover:   rm_flag =%d\n", rm_flag );
   */
                                                                             /*
===============================================================================
           Put the loop vertices into the skeleton graph
===============================================================================

  +---------------------------------------------------------------------------+
  |(1)  Self energy correction part                                           |
  +---------------------------------------------------------------------------+
                                                                             */
    if(      rm_flag == 1 )
      {
                           /*
                            printf("****** start put_self **********\n");
                           */
                                                                             /*
       <<<<<<<>>>>>>>>                                                       */
         put_self( );                                                        /*
       <<<<<<<>>>>>>>>                                                       */
                          /*
                            printf("****** after put_self ***********\n");
                          */
       }
                                                                             /*
  +---------------------------------------------------------------------------+
  |(2)  Vertex correction part                                                |
  +---------------------------------------------------------------------------+
                                                                             */
    else if( rm_flag == 2 )
      {
                         /*
                            printf("****** start put_vcor **********\n");
                         */
                                                                             /*
       <<<<<<<>>>>>>>>                                                       */
         put_vcor( );                                                        /*
       <<<<<<<>>>>>>>>                                                       */
                          /*
                            printf("****** after put_vcor ***********\n");
                          */
       }
                                                                             /*
  +---------------------------------------------------------------------------+
  |(3)  Multi-ponits loop correction part                                     |
  +---------------------------------------------------------------------------+
                                                                             */
    else if( rm_flag == 3 ) 
      {
                          /*
                            printf("****** start put_multi **********\n");
                          */
                                                                             /*
       <<<<<<<>>>>>>>>                                                       */
         put_multi( );                                                       /*
       <<<<<<<>>>>>>>>                                                       */
                          /*
                            printf("****** after put_multi ***********\n");
                          */
       }


/********  for debug purpose ***********
       }
   else if( no_loop == 1 ) no_loop--;
/***************************************/
       /*
        print_straight( );
       */
   }
}

/*****************************************************************************
                   ============================                             */
                      void recover(  )
/*                 =============================
******************************************************************************
*( Function )                                                                *
*     To recover the content of graph information before removing            **
*        loop correction part considered.                                    *
*( input )                                                                   *
*     structure keepvtx[i] : vertex information                              *
*     structure keepline[i] : line information                               *
*( output )                                                                  *
*     The vertex position in the graph                                       *
*         structure vtx[].xxx and structure intline[].xxx                    *
******************************************************************************
*/
{
                                                                            /*
---------------------
   Local variables
---------------------
                                                                            */
int j, l, m, nv;
                                                                            /*
================= Entry point of recover ====================================
                                                                            */
 no_loop--;
 rm_flag = cflag[no_loop];

 if(      rm_flag == 1 ) no_self--;
 else if( rm_flag == 2 ) no_vcor--;
 else if( rm_flag == 3 ) no_multi--;

 for( j = 0; j < no_vertex; j++ )
    {
      m              = j + add1;
      vtx[j].flag    = keepvtx[m].flag;
      nv             = keepvtx[m].no_line;
      vtx[j].no_line = nv;
      for( l = 0; l < nv; l++ )
         {
           vtxline[j].vtxno[l]  = keepvtx[m].vtxno[l];
           vtxline[j].lineno[l] = keepvtx[m].lineno[l];
          }
     }
 for( j = 0; j < no_line; j++ )
    {
      m                   = j + add2;
      intline[j].flag     = keepline[m].flag;
      inlvtx[j].vtxno[0]  = keepline[m].vtxno[0];
      inlvtx[j].vtxno[1]  = keepline[m].vtxno[1];
     }

 add1 = add1 - no_vertex;
 add2 = add2 - no_line;

}

int vtxtype, vymax;
/*****************************************************************************
                   ============================                             */
                      void put_multi(  )
/*                 =============================
******************************************************************************
*( Function )                                                                *
*     To put the multi-points loop part in the tree Feynman diagram,  and    *
*     to determine their vertex positions.                                   *
*( input )                                                                   *
*     structure vtx[i] : vertex information                                  *
*     structure intline[i] : line information                                *
*( output )                                                                  *
*     The vertex position in the graph                                       *
*         structure vtx[].x and structure vtx[].y                            *
******************************************************************************
*/
{
                                                                            /*
  ----------------------------------
      local variable list
  ----------------------------------
                                                                            */
  int i, j, k, l, n, nl, nv, v, vc, v0, v1, v2, v3, typef;
  int ngflag, xflag, yflag;
  float x0, y0, x1, y1, x2, y2, x3, y3, xx, yy, dd, dd1;
                                                                            /*
  ----------------------------------
      function list
  ----------------------------------
                                                                            */
  int mode( );
  void medge_coord( );
  void put_3point( );
  void put_4point( );
  void put_5point( );
  void put_6point( );
                                                                            /*
================== Entry point of put_multi ================================

============================================================================
      Initialization
============================================================================
                                                                            */
 dd1 = 11.0;
   
 no_art     = multi[no_multi].no_edge;
 no_inside  = multi[no_multi].no_vtx;
                                                                            /*
 +--------------------------------------------------------------------------+
 |(1) The number of initial vertices of the loop : no_in                    |
 +--------------------------------------------------------------------------+
                                                                            */
 no_in = 0;
 for( j = 0; j < no_art; j++ )
    {
      if( multi[no_multi].artv_flag[j] == 1 )
        {
          v_in[no_in] = multi[no_multi].artv[j];
          j_in[no_in] = j;
          no_in++;
         }
     }

 vc0 = multi[no_multi].artv[0];
 xc0 = vtx[vc0].x,        yc0 = vtx[vc0].y;
 vtxtype  = 0;
 vertflag = 0,            horizflag = 0,     fermiflag = 0;
                                                                            /*
 +--------------------------------------------------------------------------+
 |(2) Determine the x-y positions of the edge vertices                      |
 +--------------------------------------------------------------------------+

  <<<<<<<<>>>>>>>>>>                                                        */
   medge_coord( 0 );                                                        /*
  <<<<<<<<>>>>>>>>>>                                                        */
                                                                            /*
 +--------------------------------------------------------------------------+
 |(3) Determine the vertical order of the initial vertices                  |
 +--------------------------------------------------------------------------+
                                                                            */

 if( no_in > 1 && 
     multi[no_multi].edgey[j_in[0]] > multi[no_multi].edgey[j_in[1]] )
   {
     k       = j_in[0];
     j_in[0] = j_in[1];
     j_in[1] = k;
     v_in[0] = multi[no_multi].artv[j_in[0]];
     v_in[1] = multi[no_multi].artv[j_in[1]];
    }

      /*
       printf(" central vertex vc0[ %3d ] ( xc0, yc0 ) = ( %6.1f, %6.1f )\n",
                vc0, xc0, yc0 );
       printf(" no_art =%3d,  no_inside =%3d\n", no_art, no_inside );
       printf(" The vertices of initial side in the loop\n");
       for( j = 0; j < no_in; j++ )
       printf("    v_in[%2d] =%3d", j, v_in[j] );
       printf("\n");
       printf("*** loop information *****\n");
       for( j = 0; j < no_art; j++ )
          {
            printf("%2d edge[%3d ] === line[%3d ] === artv[%3d ] flag[%3d ]\n",
                   j, multi[no_multi].edge[j], multi[no_multi].edge_line[j],
                      multi[no_multi].artv[j], multi[no_multi].artv_flag[j] );
            printf("  (%8.2f,%8.2f)\n", 
		      multi[no_multi].edgex[j], multi[no_multi].edgey[j] );
           }
     */
                                                                            /*
============================================================================
      Call relevant function to the loop type
============================================================================
                                                                            */
 dd1 = 11.0;
                                                                            /*
                       <<<<<<<<>>>>>>>>>                                    */
 if( no_art == 3 )       put_3point( );                                     /*
                       <<<<<<<<>>>>>>>>>                                    */
                                                                            /*
                       <<<<<<<<>>>>>>>>>                                    */
 else if( no_art == 4 )  put_4point( );                                     /*
                       <<<<<<<<>>>>>>>>>                                    */
                                                                            /*
                       <<<<<<<<>>>>>>>>>                                    */
 else if( no_art == 5 )  put_5point( );                                     /*
                       <<<<<<<<>>>>>>>>>                                    */
                                                                            /*
                       <<<<<<<<>>>>>>>>>                                    */
 else if( no_art == 6 )  put_5point( );                                     /*
                       <<<<<<<<>>>>>>>>>                                    */

 else
   {
     printf("**********************************************\n");
     printf("*     Warning !!                             *\n");
     printf("*                                            *\n");
     printf("*   %3d-points loop is not available yet     *\n", no_art);
     printf("*                                            *\n");
     printf("**********************************************\n");
    }
}


/*****************************************************************************
                   ============================                             */
                      void put_3point(  )
/*                 =============================
******************************************************************************
*( Function )                                                                *
*     To put the multi-points loop part in the tree Feynman diagram,  and    *
*     to determine their vertex positions.                                   *
*( input )                                                                   *
*     structure vtx[i] : vertex information                                  *
*     structure intline[i] : line information                                *
*( output )                                                                  *
*     The vertex position in the graph                                       *
*         structure vtx[].x and structure vtx[].y                            *
******************************************************************************
*/
{
                                                                            /*
  ----------------------------------
      local variable list
  ----------------------------------
                                                                            */
  int i, j, k, l, n, nl, nv, v, vc, v0, v1, v2, v3; 
  int vx0, vx1, looptype;
  int flag;
  float x0, y0, x1, y1, xx, yy, dydx, b, dis;
  int addr, clcenter, cl;
                                                                            /*
  ----------------------------------
      function list
  ----------------------------------
                                                                            */
  int mode( );
  void chngv_straight( );
  void create_straight( );
  void put_straight( );
  int  chk_straight( );
  void medge_coord( );
  void move_others( );
  void config_3point1( );
  void config_3point2( );
  void add_vtxtable( );
  void rm_vtxtable( );
  int  addr_vtxtable( );
  void vtx_trace( );
                                                                            /*
================== Entry point of put_3point ================================
                                                                            
=============================================================================
   (I)  Initialization
=============================================================================
                                                                            */
                                                                            /*
=============================================================================
   (II)   For one particle input
=============================================================================
                                                                            */
if( no_in == 1 )
  {                                                                         /*
   +------------------------------------------------------------------------+
   |(0) determine the central vertices vc0                                  |
   +------------------------------------------------------------------------+
                                                                            */
    vc = v_in[0];
    if( vc != vc0 )
      {                                                                     /*
       <<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>                                 */
        chngv_straight( -1, -1, vc0, vc );                                  /*
       <<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>                                 */

        vtx[vc].x  = xc0,     vtx[vc].y  = yc0;
        vtx[vc0].x = 0.0,     vtx[vc0].y = 0.0;
        vc0        = vc;
       }
                                                                            /*
   +------------------------------------------------------------------------+
   |(1) determine the output vertices v_out[ ]                              |
   +------------------------------------------------------------------------+
                                                                            */
    n = 0;
    for( j = 0; j < no_art; j++ )
       {
         v  = multi[no_multi].artv[j];
         if( v != vc0 )
           {
             v_out[n] = v;
             n++;
            }
         }
                                                                            /*
   +------------------------------------------------------------------------+
   |(2) determine the type of loop                                          |
   +------------------------------------------------------------------------+
                                                                            */

    looptype = 0;

    if( no_inside > 0 )                                                     /*
      <<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                            */
        config_3point1( &looptype, &v0, &v1 );                              /*
      <<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                            */

                                                                            /*
  ...........................................................................

 looptype = 1              looptype = 2               looptype = 3
                       -+                        -+                       -+
                     --                        --                       --
                   --                        --                       --
                 -*                        -*                       -*
          v0   -- |                      -- |                     -- |
             **   |                    --   |                   --   |
           --     |                  --     |                 --     |
 +--------*       |         +-------*       |         +------*       * v0
           --     |                  --     |                 --     |
             --   |                    **   |                   --   |
               -- |                 v0   -- |                     -- |
                 -*                        -*                       -*
                   --                        --                       --
                     --                        --                       --
                       -+                        -+                       -+


 looptype = 4              
                       -+  
                     --    
                   **      
                 - /       
          v0   -- /        
             **  /         
           -- \ /          
 +--------*                
           -- / \          
             **  \         
          v1   -- \        
                 - \       
                   **      
                     --    
                       -+  

  ............................................................................
                                                                            */
      /*
       printf(" put_3point  looptype =%3d, v0(%3d )  v1(%3d )\n",
                looptype, v0, v1 );
      */


   if( looptype < 4 ) dis = 60.0;

   else               dis = 120.0;
                                                                            /*
   +------------------------------------------------------------------------+
   |(3) move those vertices which located downstream of vc0                 |
   +------------------------------------------------------------------------+
                                                                            
    <<<<<<<<<<<<<<>>>>>>>>>>>>>>                                           */
      move_others( vc0, dis );                                              /*
    <<<<<<<<<<<<<<>>>>>>>>>>>>>>                                           */
                                                                            /*
   +------------------------------------------------------------------------+
   |(4) re-calculate the x-y position of edge vertices                      |
   +------------------------------------------------------------------------+

    <<<<<<<<<>>>>>>>>>>>>>                                                  */
      medge_coord( 1 );                                                     /*
    <<<<<<<<<>>>>>>>>>>>>>                                                  */
                                                                            /*
   +------------------------------------------------------------------------+
   |(5) calculate the x-y position of loop vertices and internal vertices   |
   +------------------------------------------------------------------------+
                                                                            */
    for( j = 0; j < no_art; j++ )
       {
         v  = multi[no_multi].artv[j];
         if( v != vc0 )
           {                                                                /*
            +---------------------------------------------------------------+
            |(a) Loop vertices                                              |
            +---------------------------------------------------------------+
                                                                            */
             v2   = multi[no_multi].edge[j];
             x0   = multi[no_multi].edgex[j];
             y0   = multi[no_multi].edgey[j];

             if( y0 < yc0 ) n = 0;
             else           n = 1;

             dydx = ( y0 - yc0 )/( x0 - xc0 );
             b    = yc0 - dydx*xc0;
             xx   = xc0 + dis;
             yy   = dydx*xx + b;
	     vtx[v].x = xx,    vtx[v].y = yy;
                                                                            /*
           <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                                   */
             put_straight( vc0, v2, v );                                    /*
           <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                                   */
                                                                            /*
            +---------------------------------------------------------------+
            |(b) Internal vertices                                          |
            +---------------------------------------------------------------+
                                                                            */
	     if( n == 0 )
               {
	         if( looptype == 1  )
                   {
                     xx = xc0 + 30.0;
                     yy = dydx*xx + b;
                     vtx[v0].x = xx;
                     vtx[v0].y = yy;
                                                                            /*
                     <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                         */
                       put_straight( vc0, v, v0 );                          /*
                     <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                         */

                    }
	         else if( looptype == 4 )
                   {
                     xx = xc0 + 60.0;
                     yy = dydx*xx + b;
                     vtx[v0].x = xx;
                     vtx[v0].y = yy;
                                                                            /*
                     <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                         */
                       put_straight( vc0, v, v0 );                          /*
                     <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                         */

                    }
                }
	     else if( n == 1 )
               {
	         if( looptype == 2 )
                   {
                     xx = xc0 + 30.0;
                     yy = dydx*xx + b;
	             vtx[v0].x = xx;
	             vtx[v0].y = yy;
                                                                            /*
                     <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                         */
                       put_straight( vc0, v, v0 );                          /*
                     <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                         */

                    }
	         else if( looptype == 4 )
                   {
                     xx = xc0 + 60.0;
                     yy = dydx*xx + b;
	             vtx[v1].x = xx;
	             vtx[v1].y = yy;
                                                                            /*
                     <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                         */
                       put_straight( vc0, v, v1 );                          /*
                     <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                         */

                    }
                }
            }
        }

     if( looptype == 3 )
       {
          x0 = vtx[v_out[0]].x,  y0 = vtx[v_out[0]].y;
          x1 = vtx[v_out[1]].x,  y1 = vtx[v_out[1]].y;
          xx = 0.5*( x0 + x1 );  yy = 0.5*( y0 + y1 );
          vtx[v0].x = xx;
          vtx[v0].y = yy;
                                                                            /*
         <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                            */
          create_straight( v_out[0], v_out[1] );                            /*
         <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                            */
                                                                            /*
         <<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                         */
          put_straight( v_out[0], v_out[1], v0 );                           /*
         <<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                         */
        }
   }
                                                                            /*
=============================================================================
  (II)  For two particle input
=============================================================================
                                                                            */
else
  {                                                                         /*
   +------------------------------------------------------------------------+
   |(1) determine the output vertices v_out[ ]                              |
   +------------------------------------------------------------------------+
                                                                            */
    n = 0;
    for( j = 0; j < no_art; j++ )
       {
         v  = multi[no_multi].artv[j];
         if( v != v_in[0] && v != v_in[1] )
           {
             v_out[n] = v;
             n++;
            }
         }
                                                                            /*
   +------------------------------------------------------------------------+
   |(0) determine the central vertices vc0                                  |
   +------------------------------------------------------------------------+
                                                                            */
    vc = v_out[0];
    if( vc != vc0 )
      {                                                                     /*
       <<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>                                 */
        chngv_straight( -1, -1, vc0, vc );                                  /*
       <<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>                                 */

        vtx[vc].x  = xc0,     vtx[vc].y  = yc0;
        vtx[vc0].x = 0.0,     vtx[vc0].y = 0.0;
        vc0        = vc;
       }
                                                                            /*
   +------------------------------------------------------------------------+
   |(2) determine the type of loop                                          |
   +------------------------------------------------------------------------+
                                                                            */

    looptype = 0;

    if( no_inside > 0 )                                                     /*
      <<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                            */
        config_3point2( &looptype, &v0, &v1 );                              /*
      <<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                            */

                                                                            /*
  ...........................................................................

 looptype = 1              looptype = 2               looptype = 3
   +-                        +-                       +-
     --                        --                       --
       --                        --                       --
         *-                        *-                       *-
         | --   v0                 | --                     | --
         |   **                    |   --                   |   --
         |     --                  |     --                 |     --
         |       *--------+        |       *-------+    v0  *       +------+
         |     --                  |     --                 |     --
         |   --                    |   **                   |   --
         | --                      | --   v0                | --
         *-                        *-                       *-
       --                        --                       --
     --                        --                       --
   +-                        +-                       +-


 looptype = 4              
    +-
      --
        **
         \ - 
          \ --
           \  **
            \ / --
                  *----------+
            / \ --
           /  ** 
          / --   v1
         / -
        **
      --
    +-

  ............................................................................
                                                                            */
      /*
       printf(" config_3point2  looptype =%3d, v0(%3d )  v1(%3d )\n",
                looptype, v0, v1 );
      */

   if( looptype < 4 ) dis = -60.0;

   else               dis = -120.0;
                                                                            /*
   +------------------------------------------------------------------------+
   |(3) move those vertices which located downstream of vc0                 |
   +------------------------------------------------------------------------+
                                                                            
    <<<<<<<<<<<<<<>>>>>>>>>>>>>>                                           */
      move_others( vc0, dis );                                              /*
    <<<<<<<<<<<<<<>>>>>>>>>>>>>>                                           */
                                                                            /*
   +------------------------------------------------------------------------+
   |(4) re-calculate the x-y position of edge vertices                      |
   +------------------------------------------------------------------------+

    <<<<<<<<<>>>>>>>>>>>>>                                                  */
      medge_coord( 1 );                                                     /*
    <<<<<<<<<>>>>>>>>>>>>>                                                  */
                                                                            /*
   +------------------------------------------------------------------------+
   |(5) calculate the x-y position of loop vertices and internal vertices   |
   +------------------------------------------------------------------------+
                                                                            */
    for( j = 0; j < no_art; j++ )
       {
         v  = multi[no_multi].artv[j];
         if( v != vc0 )
           {                                                                /*
            +---------------------------------------------------------------+
            |(a) Loop vertices                                              |
            +---------------------------------------------------------------+
                                                                            */
             v2   = multi[no_multi].edge[j];
             x0   = multi[no_multi].edgex[j];
             y0   = multi[no_multi].edgey[j];

             if( y0 < yc0 ) n = 0;
             else           n = 1;

             dydx = ( y0 - yc0 )/( x0 - xc0 );
             b    = yc0 - dydx*xc0;
             xx   = xc0 + dis;
             yy   = dydx*xx + b;
	     vtx[v].x = xx,    vtx[v].y = yy;
                                                                            /*
           <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                                   */
             put_straight( vc0, v2, v );                                    /*
           <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                                   */
                                                                            /*
            +---------------------------------------------------------------+
            |(b) Internal vertices                                          |
            +---------------------------------------------------------------+
                                                                            */
	     if( n == 0 )
               {
	         if( looptype == 1  )
                   {
                     xx = xc0 - 30.0;
                     yy = dydx*xx + b;
                     vtx[v0].x = xx;
                     vtx[v0].y = yy;
                                                                            /*
                     <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                         */
                       put_straight( vc0, v, v0 );                          /*
                     <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                         */

                    }
	         else if( looptype == 4 )
                   {
                     xx = xc0 - 60.0;
                     yy = dydx*xx + b;
                     vtx[v0].x = xx;
                     vtx[v0].y = yy;
                                                                            /*
                     <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                         */
                       put_straight( vc0, v, v0 );                          /*
                     <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                         */

                    }
                }
	     else if( n == 1 )
               {
	         if( looptype == 2 )
                   {
                     xx = xc0 - 30.0;
                     yy = dydx*xx + b;
	             vtx[v0].x = xx;
	             vtx[v0].y = yy;
                                                                            /*
                     <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                         */
                       put_straight( vc0, v, v0 );                          /*
                     <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                         */

                    }
	         else if( looptype == 4 )
                   {
                     xx = xc0 - 60.0;
                     yy = dydx*xx + b;
	             vtx[v1].x = xx;
	             vtx[v1].y = yy;
                                                                            /*
                     <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                         */
                       put_straight( vc0, v, v1 );                          /*
                     <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                         */

                    }
                }
            }
        }

     if( looptype == 3 )
       {
          x0 = vtx[v_in[0]].x,  y0 = vtx[v_in[0]].y;
          x1 = vtx[v_in[1]].x,  y1 = vtx[v_in[1]].y;
          xx = 0.5*( x0 + x1 );  yy = 0.5*( y0 + y1 );
          vtx[v0].x = xx;
          vtx[v0].y = yy;
                                                                            /*
         <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                            */
          create_straight( v_in[0], v_in[1] );                            /*
         <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                            */
                                                                            /*
         <<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                         */
          put_straight( v_in[0], v_in[1], v0 );                           /*
         <<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>                         */
        }
   }
                                                                            /*
=============================================================================
  (III)  Save the internal loop lines in the straight line table
=============================================================================
                                                                            */
 for( j = 0; j < 3; j++ )
    {
      v0 = multi[no_multi].artv[j];
      nl = vtx[v0].no_line;
      for( k = 0; k < nl; k++ )
         {
           v1  = vtxline[v0].vtxno[k];
                                                                            /*
              <<<<<<<<<<<<>>>>>>>>>>>>>                                     */
           if( chck_straight( v0, v1 ) < 0 )                                /*
              <<<<<<<<<<<<<<>>>>>>>>>>>>>>>                                 */
               create_straight( v0, v1 );                                   /*
              <<<<<<<<<<<<<<<<<>>>>>>>>>>>                                  */
          }
     }
 nv = multi[no_multi].no_vtx;
 for( j = 0; j < nv; j++ )
    {
      v0 = multi[no_multi].vtxno[j];
      nl = vtx[v0].no_line;
      for( k = 0; k < nl; k++ )
         {
           v1  = vtxline[v0].vtxno[k];
                                                                            /*
              <<<<<<<<<<<<>>>>>>>>>>>>>                                     */
           if( chck_straight( v0, v1 ) < 0 )                                /*
              <<<<<<<<<<<<<<>>>>>>>>>>>>>>>                                 */
               create_straight( v0, v1 );                                   /*
              <<<<<<<<<<<<<<<<<>>>>>>>>>>>                                  */
          }
     }
}  

/*****************************************************************************
                   ============================                             */
                      void put_4point(  )
/*                 =============================
******************************************************************************
*( Function )                                                                *
*     To put the multi-points loop part in the tree Feynman diagram,  and    *
*     to determine their vertex positions.                                   *
*( input )                                                                   *
*     structure vtx[i] : vertex information                                  *
*     structure intline[i] : line information                                *
*( output )                                                                  *
*     The vertex position in the graph                                       *
*         structure vtx[].x and structure vtx[].y                            *
******************************************************************************
*/
{
                                                                            /*
  ----------------------------------
      local variable list
  ----------------------------------
                                                                            */
  int i, j, k, l, n, nl, nv, v, vc, v0, v1, v2, v3; 
  int vx0, vx1, looptype;
  int flag;
  float x0, y0, x1, y1, y2, xx, yy, dydx, b;
  int addr, clcenter, cl;
                                                                            /*
  ----------------------------------
      function list
  ----------------------------------
                                                                            */
  int mode( );
  void create_straight( );
  void chngv_straight( );
  void changey_straight( );
  void medge_coord( );
  void move_others( );
  void put_straight( );
  void config_4point( );
  void add_vtxtable( );
  void rm_vtxtable( );
  int  addr_vtxtable( );
  void set_clID ( );
  int  get_kind ( );
  void vtx_trace( );
                                                                            /*
================== Entry point of put_4point ================================
   
=============================================================================
  Initialization
=============================================================================
                                                                            */
 vc = v_in[0];
 if( vc != vc0 )
   {                                                                        /*
    <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                                    */
     chngv_straight( -1, -1, vc0, vc );                                     /*
    <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                                    */

     vtx[vc].x  = xc0,     vtx[vc].y  = yc0;
     vtx[vc0].x = 0.0,     vtx[vc0].y = 0.0;
     vc0        = vc;
    }
                                                                            /*
=============================================================================
         For one particle input
=============================================================================
                                                                            */
if( no_in == 1 )
  {                                                                         /*
   +------------------------------------------------------------------------+
   |(1) move those vertices which locates in the downstream of vtx vc0      |
   +------------------------------------------------------------------------+

   <<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                                           */
    move_others( vc0, 120.0 );                                              /*
   <<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                                           */
                                                                            /*
   +------------------------------------------------------------------------+
   |(2) re-calculate the x-y positions of the edge vertices                 |
   +------------------------------------------------------------------------+

   <<<<<<<<<>>>>>>>>>                                                       */
    medge_coord( 0 );                                                       /*
   <<<<<<<<<>>>>>>>>>                                                       */
                                                                            /*
   +------------------------------------------------------------------------+
   |(3) calculate the x-y positions of the upper and lower vertices         |
   +------------------------------------------------------------------------+
                                                                            */
    xc0 = vtx[vc0].x,    yc0 = vtx[vc0].y;

    n = 0;
    for( j = 0; j < no_art; j++ )
       {
         v0 = multi[no_multi].artv[j];
         if( v0 != vc0 )
           {
             v1   = multi[no_multi].edge[j];
             x0   = multi[no_multi].edgex[j];
             y0   = multi[no_multi].edgey[j];
             dydx = ( y0 - yc0 )/( x0 - xc0 );
             b    = yc0 - dydx*xc0;
             xx   = xc0 + 60.0;
             yy   = dydx*xx + b;

             if( n == 0 )
               {
                 y1  = yy;
                 vx0 = v0;
                }
             else if( n == 2 )
               {
                 y2  = yy;
                 vx1 = v0;
                }
                                                                            /*
                         <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                    */
             if( n != 1 )  put_straight( vc0, v1, v0 );                     /*
                         <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                    */
             n++;

	     vtx[v0].x = xx,    vtx[v0].y = yy;
            }
        }
                                                                            /*
   +------------------------------------------------------------------------+
   |(4) calculate the x-y positions of the middle vertex                    |
   +------------------------------------------------------------------------+
                                                                            */
    n = 0;
    for( j = 0; j < no_art; j++ )
       {
         v0 = multi[no_multi].artv[j];
         v1   = multi[no_multi].edge[j];
         if( v0 != vc0 )
           {
             if( n == 1 )
               {
                 xx   = xc0 + 120.0;
                 yy   = y1 + y2 - yc0;

	         vtx[v0].x = xx,    vtx[v0].y = yy;

                 flag = 0;
                 if( v0 != v1 )
                   {
                     if( get_kind( v1 ) == 2 )
		         flag = 1 ;
                                                                            /*
                     <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                   */
                      chngv_straight( vc0, v1, vc0, v0 );                   /*
                     <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                   */
                    }
                 else
                   {
                     nl = vtx[v1].no_line;
                     for( k = 0; k < nl; k++ )
                        {
                          vx1  = vtxline[v1].vtxno[k];
                          if( get_kind( vx1 ) == 2 )
                            {
                              flag = 1;
                              k    = nl;
                             }
                                                                            /*
                         <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>               */
                          chngv_straight( vc0, vx1, vc0, v0 );              /*
                         <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>               */
                         }
                    }
                }

             n++;
            }
        }
   }
                                                                            /*
=============================================================================
 (II)   For two particle input
=============================================================================
                                                                            */
else
  {
                                                                            /*
   +------------------------------------------------------------------------+
   |(1) move those vertices which locates in the downstream of vtx vc0      |
   +------------------------------------------------------------------------+

   <<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                                           */
    move_others( vc0, 60.0 );                                               /*
   <<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                                           */
                                                                            /*
   +------------------------------------------------------------------------+
   |(2) re-calculate the x-y positions of the edge vertices                 |
   +------------------------------------------------------------------------+

   <<<<<<<<<>>>>>>>>>                                                       */
    medge_coord( 1 );                                                       /*
   <<<<<<<<<>>>>>>>>>                                                       */
                                                                            /*
   +------------------------------------------------------------------------+
   | (3) Calculate the positions of output vertices v_out[i], j_out[i]      |
   +------------------------------------------------------------------------+
                                                                            */
    n = 0;
    for( j = 0; j < no_art; j++ )
       {
         x0 = multi[no_multi].edgex[j];
         if( x0 > xc0 + 6.0 )
           {
             v0   = multi[no_multi].artv[j];
             xx   = xc0 + 60.0;
             yy   = multi[no_multi].edgey[j];

	      /*
		printf(" artv[%d] =%3d  edge[%d] =%3d edgey =%8.1f\n",
                         j, v0, j, multi[no_multi].edge[j], yy );
              */

             v1 = multi[no_multi].edge[j];
             flag = 0;
             if( v0 != v1 )
               {
                 if( get_kind( v1 ) == 2 )
		     flag = 1 ;
                                                                            /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                       */
                  chngv_straight( vc0, v1, vc0, v0 );                       /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                       */
                }
             else
               {
                 nl = vtx[v1].no_line;
                 for( k = 0; k < nl; k++ )
                    {
                      vx1  = vtxline[v1].vtxno[k];
                      if( get_kind( vx1 ) == 2 )
                        {
                          flag = 1;
                          k    = nl;
                         }
                                                                            /*
                     <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                   */
                      chngv_straight( vc0, vx1, vc0, v0 );                  /*
                     <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                   */
                     }
                }

             if( yy < 40.0 && flag )
                 yy = yy + 10.0;
             else if( yy > ymax - 40.0 && flag )
                 yy = yy - 10.0;

	     v_out[n] = v0;
	     j_out[n] = j;
             n++;

	     vtx[v0].x = xx,    vtx[v0].y = yy;
              
            }
        }
                                                                            /*
   +------------------------------------------------------------------------+
   | (4) if no_inside > 0, determine the structure of the loop              |
   +------------------------------------------------------------------------+
                                                                            */
   looptype = 0;

   if( no_inside > 0 )
     {
                                                                           /*
               ..............................................................
               ==============================================
                void config_4point( &looptype, &vx0, &vx1 )
               ==============================================
                (( Function ))
                   To determine the configuration of vertices
		      in the 4-point two loop.
                (( input ))
                   v_in[i], v_out[i] through the global variables
                (( output ))
                   looptype : type of 4-point two loop
                   vx0      : vertex ID of the vertex located at v0
                   vx1      : vertex ID of the vertex located at v1

                << Type of 4-point two loop >>

                (( Type 0 ))
                    looptype = 0

                    v_in[0]           v_out[0]
		       ---+-----------+--- 
                          |           |    
                          |           |    
                          |           |    
                          |           |    
                          |           |   
                       ---+-----------+---
                    v_in[1]           v_out[1]

                (( Type 1 ))
                    looptype = 1

                    v_in[0]    vx0    v_out[0]        vx0
		       ---+-----+-----+---    ---+-----+-----+---
                          |     |     |           \   /      |
                          |     |     |            \ /       |
                          |     |     |                      |
                          |     |     |            / \       |
                          |     |     |           /   \      |
                       ---+-----+-----+---    ---+-----+-----+---
                    v_in[1]    vx1    v_out[1]        vx1

                (( Type 2 ))
                    looptype = 2

                    v_in[0]           v_out[0]
		       ---+-----------+---    ---+-----------+---
                          |           |          |           |
                          |           |      vx0 +--       --+ vx1
                      vx0 +-----------+ vx1      |  --   --  |
                          |           |          |    ---    |
                          |           |          |  --   --  |
                       ---+-----------+---    ---+--       --+---
                    v_in[1]           v_out[1]
               ..............................................................
                                                                            */

       config_4point( &looptype, &vx0, &vx1 );

               /*
                 if( looptype == 1 )
                   {
                     printf("looptype =%3d\n", looptype );
                     printf(" v(%3d )  vx0(%3d )  v_out(%3d )\n", 
			v_in[0], vx0, v_out[0]);
                     printf(" ----+---------+-----------+----\n");
                     printf("     |         |           |    \n");
                     printf("     |         |           |    \n");
                     printf("     |         |           |    \n");
                     printf(" ----+---------+-----------+----\n");
                     printf(" v(%3d )  vx1(%3d )  v_out(%3d )\n", 
			v_in[1], vx1, v_out[1]);
                    }
                */
           if( looptype == 2 )
             {
                /*
                 printf("looptype =%3d\n", looptype );
                 printf("      v_in(%3d )  v_out(%3d )\n", 
	        	  	     v_in[0], v_out[0]);
                 printf("      ----+-----------+----\n");
                 printf("          |           |    \n");
                 printf("          |           |    \n");
                 printf("vx0(%3d ) +-----------+ vx1(%3d )\n", vx0, vx1);
                 printf("          |           |    \n");
                 printf("          |           |    \n");
                 printf("      ----+-----------+----\n");
                 printf("      v_in(%3d )  v_out(%3d )\n", 
	        		     v_in[1],  v_out[1]);
                */
	       x0 = vtx[v_out[0]].x, y0 = vtx[v_out[0]].y; 
	       x1 = vtx[v_out[1]].x, y1 = vtx[v_out[1]].y; 

	       xx = 0.5*( x0 + x1 ), yy = 0.5*( y0 + y1 );
               vtx[vx1].x = xx;
               vtx[vx1].y = yy;
              }
           if( looptype != 1 && looptype != 2 )
               printf(" other type of looptype is appeared !!\n" );

      }
                                                                            /*
   +------------------------------------------------------------------------+
   | (5) Change the cluster numbers on the vertex table vtx_table[2][i]     |
   +------------------------------------------------------------------------+
                                                                            */
    addr     = addr_vtxtable( vc0 );
    clcenter = vtx_table[2][addr];
    if( addr <= no_iniline )
      {
        if( no_cluster > 1 )
          {
            for( n = 0; n < nvertex; n++ )
               {
                 cl = vtx_table[2][n];
                 if( cl > clcenter )
                     vtx_table[2][n] = cl + 1;
                }
            for( n = no_cluster ; n > clcenter+1; n-- )
               {
                 cluster_table[n] = cluster_table[n-1];
                }
           }
         no_cluster++;
       }

    rm_vtxtable( vc0 );

                                                                            /*
   +------------------------------------------------------------------------+
   | (6) Calculate the positions of input vertices v_in[i], j_in[i]         |
   |                  and the initial lines if they are on those lines      |
   +------------------------------------------------------------------------+
                                                                            */
    for( j = 0; j < 2; j++ )
       {                                                                    /*
         +------------------------------------------------------------------+
         |(a) Calculate the positions of input vertices v_in[i], j_in[i]    |
         +------------------------------------------------------------------+
                                                                            */
         v0 = v_in[j];
         v1 = v_out[j];
         if( j == 0 ) cl = clcenter;
         else         cl = clcenter+1;
           
         add_vtxtable( 1, v0, 0, 0, cl );
         add_vtxtable( 0, v1, 0, 0, cl );

         x0 = xc0;
         y0 = vtx[v1].y;

         vtx[v0].x = x0,    vtx[v0].y = y0;
                                                                            /*
         +------------------------------------------------------------------+
         |(b) register the input lines in the straight line table           |
         +------------------------------------------------------------------+
                                                                            */
	 k  = j_in[j];
	 v2 = multi[no_multi].edge[k];
           /*
            printf(" edge[%2d]  v-in[%2d] =%3d  v_out[%2d] =%3d\n",
                     v2, j, v_in[j], j, v_out[j] );
           */
         if( v2 != v0 )
           {                                                                /*
            <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                           */
             chngv_straight( v2, vc0, vc0, v0 );                            /*
            <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                           */
            }
         else
           {
             nl = vtx[v1].no_line;
             for( k = 0; k < nl; k++ )
                {
                  vx1  = vtxline[v1].vtxno[k];
                                                                            /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                     */
                  chngv_straight( vx1, vc0, vc0, v0 );                      /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                     */
                 }
            }
                                                                            /*
         +------------------------------------------------------------------+
         |(c) If the input line is the initial line,                        |
         |      re-determine the vertex poisitons on the initial line       |
         +------------------------------------------------------------------+
                                                                            */
         for( l = 0; l < no_initial; l++ )
            {                                                               /*
                  <<<<<<<<<<<<<>>>>>>>>>>>>>                                */
              nl  = chck_straight( l, v0 );                                 /*
                  <<<<<<<<<<<<<>>>>>>>>>>>>>                                */

              if( nl >= 0 )
                {
                  x1 = vtx[l].x;
                  y1 = y0 - 10.0;
                  if( l == 1 ) y1 = y0 + 10.0;
                                                                            /*
                  <<<<<<<<<<<<<<>>>>>>>>>>>>>>                              */
                   changey_straight( l, y1 );                               /*
                  <<<<<<<<<<<<<<>>>>>>>>>>>>>>                              */
                 }
             }
        }
                                                                            /*
   +------------------------------------------------------------------------+
   | (7) If looptype = 1, shift the downstream vertices                     |
   |         and determine the x-y positions of internal vertices           |
   +------------------------------------------------------------------------+
                                                                            */
     if( looptype == 1 )
       {                                                                    /*
         +------------------------------------------------------------------+
         |(a) move those vertices                                           |
         |            which locates in the downstream of vtx vc0            |
         +------------------------------------------------------------------+

         <<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>                                    */
          move_others( v_in[0], 60.0 );                                     /*
         <<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>                                    */
                                                                            /*
         +------------------------------------------------------------------+
         |(b) determine the x-y positions of internal vertices              |
         +------------------------------------------------------------------+
                                                                            */
         x0 = vtx[v_in[0]].x,  y0 = vtx[v_in[0]].y; 
         x1 = vtx[v_out[0]].x, y1 = vtx[v_out[0]].y; 

         xx = 0.5*( x0 + x1 ), yy = 0.5*( y0 + y1 );
         vtx[vx0].x = xx;
         vtx[vx0].y = yy;

         x0 = vtx[v_in[1]].x,  y0 = vtx[v_in[1]].y; 
         x1 = vtx[v_out[1]].x, y1 = vtx[v_out[1]].y; 

         xx = 0.5*( x0 + x1 ), yy = 0.5*( y0 + y1 );
         vtx[vx1].x = xx;
         vtx[vx1].y = yy;
        }
                                                                            /*
   +------------------------------------------------------------------------+
   | (8) If looptype = 2, determine the x-y positions of internal vertices  |
   +------------------------------------------------------------------------+
                                                                            */
     else if( looptype == 2 )
       {
         x0 = vtx[v_in[0]].x, y0 = vtx[v_in[0]].y; 
         x1 = vtx[v_in[1]].x, y1 = vtx[v_in[1]].y; 

         xx = 0.5*( x0 + x1 ), yy = 0.5*( y0 + y1 );
         vtx[vx0].x = xx;
         vtx[vx0].y = yy;
        }
                                                                            /*
   +------------------------------------------------------------------------+
   | (9) Update the vertex table                                            |
   +------------------------------------------------------------------------+
                                                                            */
    for( j = 0; j < 2; j++ )
       {
         v1 = v_out[j];
         if( j == 0 ) cl = clcenter;
         else         cl = clcenter+1;

	 cluster_table[cl] = v1;

	 vtx_trace( v1, v1, 1 );
               /*
                printf("*** vtx_trace from v[%d] ***** count =%4d\n", 
			 v1, count );
               */
         for( k = 1; k < count; k++ )
            {
              nv   = vtxbuf[k];
                                                                            /*
             <<<<<<<<<<>>>>>>>>>>>                                          */
              set_clID( nv, cl );                                           /*
             <<<<<<<<<<>>>>>>>>>>>                                          */
               /*
                printf("%5d", nv );
               */
             }
              /*  printf("\n");  */
        }
   }
                                                                            /*
=============================================================================
  (III)  Save the internal loop lines in the straight line table
=============================================================================
                                                                            */
    for( j = 0; j < 4; j++ )
       {
         v0 = multi[no_multi].artv[j];
         nl = vtx[v0].no_line;
         for( k = 0; k < nl; k++ )
            {
              v1  = vtxline[v0].vtxno[k];
                                                                            /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                     */
              if( chck_straight( v0, v1 ) < 0 )                             /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                     */
                  create_straight( v0, v1 );                                /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>                               */
                 }
        }
    nv = multi[no_multi].no_vtx;
    for( j = 0; j < nv; j++ )
       {
         v0 = multi[no_multi].vtxno[j];
         nl = vtx[v0].no_line;
         for( k = 0; k < nl; k++ )
            {
              v1  = vtxline[v0].vtxno[k];
                                                                            /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                     */
              if( chck_straight( v0, v1 ) < 0 )                             /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                     */
                  create_straight( v0, v1 );                                /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>                               */
                 }
        }
                                                                            /*
   <<<<<<<<<<<>>>>>>>>>>>>>>>>                                              */
    move_others( 0, 0.001 );                                                /*
   <<<<<<<<<<<>>>>>>>>>>>>>>>>                                              */
}  



/*****************************************************************************
               ==========================================                   */
                 void config_3point1( looptype, v0, v1 )
/*             ==========================================
******************************************************************************
*( Function )                                                                *
*     To determine the configuration of vertices in the 4-point two loop     *
*( input )                                                                   *
*     v_in[i]  : vertex numbers of input vertices                            *
*     v_out[i] : vertex numbers of output vertices                           *
*( output )                                                                  *
*     looptype : looptype of the 4-point two loop                            *
*     v0       : vertex number of the vertex located at v0 position          *
*     v1       : vertex number of the vertex located at v1 position          *
******************************************************************************/

int *looptype, *v0, *v1;
{
                                                                            /*
  ----------------------------------
      local variable list
  ----------------------------------
                                                                            */
  int level_no, flag;
  int i, n, v, n_line;
  int nv0[2], v_table0[2][10], nv1[2], v_table1[2][10];
  int nv2[2], v_table2[2][10], nv3[2], v_table3[2][10];
                                                                            /*
  ----------------------------------
      function list
  ----------------------------------
                                                                            */
  int  mode( );
  void find_line( );
                                                                            /*
================== Entry point of config_4point =============================
                                                                            */
                                                                            /*
=============================================================================
      Initialization
=============================================================================
                                                                            */
 for( i = 0; i < no_vertex; i++ )
      dfnumber[i] = -1;
 for( i = 0; i < no_art; i++ )
    {
      v = multi[no_multi].artv[i];
      dfnumber[v] = 1000;
     }
 for( i = 0; i < no_inside; i++ )
    {
      v = multi[no_multi].vtxno[i];
      dfnumber[v] = 1000;
     }
                                                                            /*
=============================================================================
  Find the structure of loop ( Case looptype 2 ) 
=============================================================================
                                                                            */
 if( no_inside == 1 )
   {
                                                                            /*
-----------------------------------------------------------------------------
   count the number of vertices on the line from v_in[0] to v_out[0]
-----------------------------------------------------------------------------
                                                                            */

   find_line( v_in[0], v_out[0], &level_no );

   n = 0;
   for( i = 0; i < no_vertex; i++ )
      {
        if( level[0][i] == level_no )
          {
            v_table0[0][n] = i;
            n++;
           }
       }

       /*
	printf(" %2d vertices on the line from v_in[0](%3d )",
                 n, v_in[0] );
	printf(" to v_out[0](%3d )\n",
                 v_out[0] );
        for( i = 0; i < n; i++ )
             printf("%5d", v_table0[0][i] );
        printf("\n");
       */

   if( n != 2 )
     {
       *looptype = 1;
       for( i = 0; i < n; i++ )
          {
            v = v_table0[0][i];
            if( v != v_in[0] && v != v_out[0] && v != v_out[1] )
                *v0       = v;
           }
       return;
      }
                                                                            /*
-----------------------------------------------------------------------------
   count the number of vertices on the line from v_in[0] to v_out[1]
-----------------------------------------------------------------------------
                                                                            */

   find_line( v_in[0], v_out[1], &level_no );

   n = 0;
   for( i = 0; i < no_vertex; i++ )
      {
        if( level[0][i] == level_no )
          {
            v_table0[0][n] = i;
            n++;
           }
       }

       /*
	printf(" %2d vertices on the line from v_in[0](%3d )",
                 n, v_in[0] );
	printf(" to v_out[1](%3d )\n",
                 v_out[1] );
        for( i = 0; i < n; i++ )
             printf("%5d", v_table0[0][i] );
        printf("\n");
       */

   if( n != 2 )
     {
       *looptype = 2;
       for( i = 0; i < n; i++ )
          {
            v = v_table0[0][i];
            if( v != v_in[0] && v != v_out[0] && v != v_out[1] )
                *v0       = v;
           }
       return;
      }
                                                                            /*
-----------------------------------------------------------------------------
   count the number of vertices on the line from v_out[0] to v_out[1]
-----------------------------------------------------------------------------
                                                                            */
   find_line( v_out[0], v_out[1], &level_no );

   n = 0;
   for( i = 0; i < no_vertex; i++ )
      {
        if( level[0][i] == level_no )
          {
            v_table0[0][n] = i;
            n++;
           }
       }

      /*
	printf(" %2d vertices on the line from v_out[0](%3d )",
                 n, v_out[0] );
	printf(" to v_out[1](%3d )\n",
                 v_out[1] );
        for( i = 0; i < n; i++ )
             printf("%5d", v_table0[0][i] );
        printf("\n");
      */

   if( n != 2 )
     {
       *looptype = 3;
       for( i = 0; i < n; i++ )
          {
            v = v_table0[0][i];
            if( v != v_in[0] && v != v_out[0] && v != v_out[1] )
                *v0       = v;
           }
       return;
      }
   }
                                                                            /*
=============================================================================
  Find the structure of loop ( Case looptype 4 ) 
=============================================================================
                                                                            */
else if( no_inside == 2 )
  {
     *looptype = 4;
      v = multi[no_multi].vtxno[0];
     *v0 = v;
      v = multi[no_multi].vtxno[1];
     *v1 = v;

      return;
   }

  printf("************* from config_3point1 *********************\n");
  printf("*                                                     *\n");
  printf("*  Unexpected configuration at 3-point loop           *\n");
  printf("*                                                     *\n");
  printf("************* from config_3point1 *********************\n");

exit( 0 );
}

/*****************************************************************************
               ==========================================                   */
                 void config_3point2( looptype, v0, v1 )
/*             ==========================================
******************************************************************************
*( Function )                                                                *
*     To determine the configuration of vertices in the 4-point two loop     *
*( input )                                                                   *
*     v_in[i]  : vertex numbers of input vertices                            *
*     v_out[i] : vertex numbers of output vertices                           *
*( output )                                                                  *
*     looptype : looptype of the 4-point two loop                            *
*     v0       : vertex number of the vertex located at v0 position          *
*     v1       : vertex number of the vertex located at v1 position          *
******************************************************************************/

int *looptype, *v0, *v1;
{
                                                                            /*
  ----------------------------------
      local variable list
  ----------------------------------
                                                                            */
  int level_no, flag;
  int i, n, v, n_line;
  int nv0[2], v_table0[2][10], nv1[2], v_table1[2][10];
  int nv2[2], v_table2[2][10], nv3[2], v_table3[2][10];
                                                                            /*
  ----------------------------------
      function list
  ----------------------------------
                                                                            */
  int  mode( );
  void find_line( );
                                                                            /*
================== Entry point of config_4point =============================
                                                                            */
                                                                            /*
=============================================================================
      Initialization
=============================================================================
                                                                            */
 for( i = 0; i < no_vertex; i++ )
      dfnumber[i] = -1;
 for( i = 0; i < no_art; i++ )
    {
      v = multi[no_multi].artv[i];
      dfnumber[v] = 1000;
     }
 for( i = 0; i < no_inside; i++ )
    {
      v = multi[no_multi].vtxno[i];
      dfnumber[v] = 1000;
     }
                                                                            /*
=============================================================================
  Find the structure of loop ( Case looptype 2 ) 
=============================================================================
                                                                            */
 if( no_inside == 1 )
   {
                                                                            /*
-----------------------------------------------------------------------------
   count the number of vertices on the line from v_in[0] to v_out[0]
-----------------------------------------------------------------------------
                                                                            */

   find_line( v_in[0], v_out[0], &level_no );

   n = 0;
   for( i = 0; i < no_vertex; i++ )
      {
        if( level[0][i] == level_no )
          {
            v_table0[0][n] = i;
            n++;
           }
       }

       /*
	printf(" %2d vertices on the line from v_in[0](%3d )",
                 n, v_in[0] );
	printf(" to v_out[0](%3d )\n",
                 v_out[0] );
        for( i = 0; i < n; i++ )
             printf("%5d", v_table0[0][i] );
        printf("\n");
       */

   if( n != 2 )
     {
       *looptype = 1;
       for( i = 0; i < n; i++ )
          {
            v = v_table0[0][i];
            if( v != v_in[0] && v != v_in[1] && v != v_out[0] )
                *v0       = v;
           }
       return;
      }
                                                                            /*
-----------------------------------------------------------------------------
   count the number of vertices on the line from v_in[1] to v_out[0]
-----------------------------------------------------------------------------
                                                                            */

   find_line( v_in[1], v_out[0], &level_no );

   n = 0;
   for( i = 0; i < no_vertex; i++ )
      {
        if( level[0][i] == level_no )
          {
            v_table0[0][n] = i;
            n++;
           }
       }

       /*
	printf(" %2d vertices on the line from v_in[1](%3d )",
                 n, v_in[1] );
	printf(" to v_out[0](%3d )\n",
                 v_out[0] );
        for( i = 0; i < n; i++ )
             printf("%5d", v_table0[0][i] );
        printf("\n");
       */

   if( n != 2 )
     {
       *looptype = 2;
       for( i = 0; i < n; i++ )
          {
            v = v_table0[0][i];
            if( v != v_in[0] && v != v_in[1] && v != v_out[0] )
                *v0       = v;
           }
       return;
      }
                                                                            /*
-----------------------------------------------------------------------------
   count the number of vertices on the line from v_in[0] to v_in[1]
-----------------------------------------------------------------------------
                                                                            */
   find_line( v_in[0], v_in[1], &level_no );

   n = 0;
   for( i = 0; i < no_vertex; i++ )
      {
        if( level[0][i] == level_no )
          {
            v_table0[0][n] = i;
            n++;
           }
       }

      /*
	printf(" %2d vertices on the line from v_in[0](%3d )",
                 n, v_in[0] );
	printf(" to v_in[1](%3d )\n",
                 v_in[1] );
        for( i = 0; i < n; i++ )
             printf("%5d", v_table0[0][i] );
        printf("\n");
      */

   if( n != 2 )
     {
       *looptype = 3;
       for( i = 0; i < n; i++ )
          {
            v = v_table0[0][i];
            if( v != v_in[0] && v != v_in[1] && v != v_out[0] )
                *v0       = v;
           }
       return;
      }
   }
                                                                            /*
=============================================================================
  Find the structure of loop ( Case looptype 4 ) 
=============================================================================
                                                                            */
else if( no_inside == 2 )
  {
     *looptype = 4;
      v = multi[no_multi].vtxno[0];
     *v0 = v;
      v = multi[no_multi].vtxno[1];
     *v1 = v;

      return;
   }

  printf("************* from config_3point2 *********************\n");
  printf("*                                                     *\n");
  printf("*  Unexpected configuration at 3-point loop           *\n");
  printf("*                                                     *\n");
  printf("************* from config_3point1 *********************\n");

exit( 0 );
}

/*****************************************************************************
               ==========================================                   */
                 void config_4point( looptype, v0, v1 )
/*             ==========================================
******************************************************************************
*( Function )                                                                *
*     To determine the configuration of vertices in the 4-point two loop     *
*( input )                                                                   *
*     v_in[i]  : vertex numbers of input vertices                            *
*     v_out[i] : vertex numbers of output vertices                           *
*( output )                                                                  *
*     looptype : looptype of the 4-point two loop                            *
*     v0       : vertex number of the vertex located at v0 position          *
*     v1       : vertex number of the vertex located at v1 position          *
******************************************************************************/

int *looptype, *v0, *v1;
{
                                                                            /*
  ----------------------------------
      local variable list
  ----------------------------------
                                                                            */
  int level_no, flag;
  int i, n, v, n_line;
  int nv0[2], v_table0[2][10], nv1[2], v_table1[2][10];
  int nv2[2], v_table2[2][10], nv3[2], v_table3[2][10];
                                                                            /*
  ----------------------------------
      function list
  ----------------------------------
                                                                            */
  int  mode( );
  void find_line( );
  int trace_fermion( );
                                                                            /*
================== Entry point of config_4point =============================
                                                                            */
                                                                            /*
=============================================================================
      Initialization
=============================================================================
                                                                            */
 for( i = 0; i < no_vertex; i++ )
      dfnumber[i] = -1;
 for( i = 0; i < no_art; i++ )
    {
      v = multi[no_multi].artv[i];
      dfnumber[v] = 1000;
     }
 for( i = 0; i < no_inside; i++ )
    {
      v = multi[no_multi].vtxno[i];
      dfnumber[v] = 1000;
     }
                                                                            /*
=============================================================================
    Trace the fermion line 
=============================================================================
                                                                            */
 flag = 0;

 count = 0;

 v = trace_fermion( v_in[0], v_in[0] );

 if( v == v_out[0] ) flag = 1;
 if( v == v_out[1] ) flag = 2;
                                                                            /*
=============================================================================
  Find the structure of loop ( Case looptype 2 ) 
=============================================================================

-----------------------------------------------------------------------------
   count the number of vertices on the line from v_in[0] to v_in[1]
-----------------------------------------------------------------------------
                                                                            */
      /*
       printf("** config_4point start **\n");
      */

   find_line( v_in[0], v_in[1], &level_no );

   n = 0;
   for( i = 0; i < no_vertex; i++ )
      {
        if( level[0][i] == level_no )
          {
            v_table0[0][n] = i;
            n++;
           }
       }
    nv0[0] = n;

      /*
	printf(" %2d vertices on the line from v_in[0](%3d )",
                 nv0[0], v_in[0] );
	printf(" to v_in[1](%3d )\n",
                 v_in[1] );
        for( i = 0; i < nv0[0]; i++ )
             printf("%5d", v_table0[0][i] );
        printf("\n");
      */
                                                                            /*
-----------------------------------------------------------------------------
   count the number of vertices on the line from v_out[0] to v_out[1]
-----------------------------------------------------------------------------
                                                                            */
   find_line( v_out[0], v_out[1], &level_no );

   n = 0;
   for( i = 0; i < no_vertex; i++ )
      {
        if( level[0][i] == level_no )
          {
            v_table0[1][n] = i;
            n++;
           }
       }
    nv0[1] = n;

       /*
	printf(" %2d vertices on the line from v_out[0](%3d )",
                 nv0[1], v_out[0] );
	printf(" to v_out[1](%3d )\n",
                 v_out[1] );
        for( i = 0; i < nv0[1]; i++ )
             printf("%5d", v_table0[1][i] );
        printf("\n");
       */

   if( nv0[0] == nv0[1] && nv0[0] == 3 && flag == 2 )
     {
       *looptype = 2;
       for( i = 0; i < 3; i++ )
          {
            v = v_table0[0][i];
            if( v != v_in[0] && v != v_in[1] )
                *v0       = v;
           }
       for( i = 0; i < 3; i++ )
          {
            v = v_table0[1][i];
            if( v != v_out[0] && v != v_out[1] )
                *v1       = v;
           }
       return;
      }
                                                                            /*
=============================================================================
  Find the structure of loop ( Case looptype 1 ) 
=============================================================================

-----------------------------------------------------------------------------
   count the number of vertices on the line from v_in[0] to v_out[0]
-----------------------------------------------------------------------------
                                                                            */
   find_line( v_in[0], v_out[0], &level_no );

   n = 0;
   for( i = 0; i < no_vertex; i++ )
      {
        if( level[0][i] == level_no )
          {
            v_table1[0][n] = i;
            n++;
           }
       }
    nv1[0] = n;

       /*
	printf(" %2d vertices on the line from v_in[0](%3d )",
                 nv1[0], v_in[0] );
	printf(" to v_out[0](%3d )\n",
                 v_out[0] );
        for( i = 0; i < nv1[0]; i++ )
             printf("%5d", v_table1[0][i] );
        printf("\n");
       */
                                                                            /*
-----------------------------------------------------------------------------
   count the number of vertices on the line from v_in[1] to v_out[1]
-----------------------------------------------------------------------------
                                                                            */
   find_line( v_in[1], v_out[1], &level_no );

   n = 0;
   for( i = 0; i < no_vertex; i++ )
      {
        if( level[0][i] == level_no )
          {
            v_table1[1][n] = i;
            n++;
           }
       }
    nv1[1] = n;

       /*
	printf(" %2d vertices on the line from v_in[1](%3d )",
                 nv1[1], v_in[1] );
	printf(" to v_out[1](%3d )\n",
                 v_out[1] );
        for( i = 0; i < nv1[1]; i++ )
             printf("%5d", v_table1[1][i] );
        printf("\n");
       */

   if( nv1[0] == nv1[1] && nv1[0] == 3 && flag == 1 )
     {
       *looptype = 1;
       for( i = 0; i < 3; i++ )
          {
            v = v_table1[0][i];
            if( v != v_in[0] && v != v_out[0] )
                *v0       = v;
           }
       for( i = 0; i < 3; i++ )
          {
            v = v_table1[1][i];
            if( v != v_in[1] && v != v_out[1] )
                *v1       = v;
           }
       return;
      }
                                                                            /*
=============================================================================
  Find the structure of loop ( Case looptype 1 special ) 
=============================================================================
                                                                            */

   if( nv0[0] == nv0[1] && nv0[0] == 3 )
     {
       *looptype = 2;
       for( i = 0; i < 3; i++ )
          {
            v = v_table0[0][i];
            if( v != v_in[0] && v != v_in[1] )
                *v0       = v;
           }
       for( i = 0; i < 3; i++ )
          {
            v = v_table0[1][i];
            if( v != v_out[0] && v != v_out[1] )
                *v1       = v;
           }
       return;
      }

   if( nv1[0] == nv1[1] && nv1[0] == 3 )
     {
       *looptype = 1;
       for( i = 0; i < 3; i++ )
          {
            v = v_table1[0][i];
            if( v != v_in[0] && v != v_out[0] )
                *v0       = v;
           }
       for( i = 0; i < 3; i++ )
          {
            v = v_table1[1][i];
            if( v != v_in[1] && v != v_out[1] )
                *v1       = v;
           }
       return;
      }

   flag = 0;

   if( nv0[0] == nv0[1] && nv0[0] == 2 )
     {
       *looptype = 1;
       n_line    = vtx[v_out[0]].no_line;
       for( i = 0; i < n_line; i++ )
          {
            v = vtxline[v_out[0]].vtxno[i];
            n = vtxline[v_out[0]].lineno[i];
            if( v != v_out[1] && dfnumber[v] >= 0 )
              {
                if( iabs( intline[n].direct ) == 1 )
                  {
                    *v0 = v;
                    flag = 1;
                   }
                else
                    *v1 = v;
               }
           }
       n_line    = vtx[v_out[1]].no_line;
       for( i = 0; i < n_line; i++ )
          {
            v = vtxline[v_out[1]].vtxno[i];
            if( v != v_out[0] && dfnumber[v] >= 0 )
              {
                if( flag == 1 )
                  {
                    *v1 = v;
                   }
                else
                    *v0 = v;
               }
           }
       return;
      }
                                                                            /*
=============================================================================
  Find the structure of loop ( Case looptype 2 special ) 
=============================================================================
                                                                            */
   if( nv1[0] == nv1[1] && nv1[0] == 2 )
     {
       *looptype = 2;
       n_line    = vtx[v_in[1]].no_line;
       for( i = 0; i < n_line; i++ )
          {
            v = vtxline[v_in[1]].vtxno[i];
            n = vtxline[v_in[1]].lineno[i];
            if( v != v_out[1] && dfnumber[v] >= 0 )
              {
                if( iabs( intline[n].direct ) == 1 )
                  {
                    *v0 = v;
                    flag = 1;
                   }
                else
                    *v1 = v;
               }
           }
       n_line    = vtx[v_out[1]].no_line;
       for( i = 0; i < n_line; i++ )
          {
            v = vtxline[v_out[1]].vtxno[i];
            if( v != v_in[1] && dfnumber[v] >= 0 )
              {
                if( flag == 1 )
                  {
                    *v1 = v;
                   }
                else
                    *v0 = v;
               }
           }
       return;
      }

exit( 0 );
}

/*****************************************************************************
                   ============================                             */
                      void put_5point(  )
/*                 =============================
******************************************************************************
*( Function )                                                                *
*     To put the multi-points loop part in the tree Feynman diagram,  and    *
*     to determine their vertex positions.                                   *
*( input )                                                                   *
*     structure vtx[i] : vertex information                                  *
*     structure intline[i] : line information                                *
*( output )                                                                  *
*     The vertex position in the graph                                       *
*         structure vtx[].x and structure vtx[].y                            *
******************************************************************************
*/
{
                                                                            /*
  ----------------------------------
      local variable list
  ----------------------------------
                                                                            */
  int i, j, k, l, n, nl, nv, v, vc, v0, v1, v2, v3, typef;
  int vx0, vx1, flag;
  float x0, y0, x1, y1, y2, xx, yy, dydx, b;
                                                                            /*
  ----------------------------------
      function list
  ----------------------------------
                                                                            */
  int mode( );
  void medge_coord( );
  int  chck_straight( );
  void create_straight( );
  void chngv_straight( );
  void put_straight( );
  void print_straight( );
  void changey_straight( );
  void move_others( );
                                                                            /*
================== Entry point of put_5point ================================

=============================================================================
      Initialization
=============================================================================
                                                                            */
 vc = v_in[0];
 if( vc != vc0 )
   {                                                                        /*
    <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                                    */
     chngv_straight( -1, -1, vc0, vc );                                     /*
    <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                                    */

     vtx[vc].x  = xc0,     vtx[vc].y  = yc0;
     vtx[vc0].x = 0.0,     vtx[vc0].y = 0.0;
     vc0        = vc;
    }

                                                                            /*
=============================================================================
 (I)    For one particle input
=============================================================================
                                                                            */
if( no_in == 1 )
  {                                                                         /*
   +------------------------------------------------------------------------+
   |(1) move those vertices which locate in the downstream                  |
   +------------------------------------------------------------------------+
                                  
   <<<<<<<<<<<<>>>>>>>>>>>>>>>>                                             */
    move_others( vc0, 120.0 );                                              /*
   <<<<<<<<<<<<>>>>>>>>>>>>>>>>                                             */
                                                                            /*
   +------------------------------------------------------------------------+
   |(2) re-calculate the x-y positions of the edge vertices                 |
   +------------------------------------------------------------------------+
                                  
   <<<<<<<<>>>>>>>>>                                                        */
    medge_coord( 0 );                                                       /*
   <<<<<<<<>>>>>>>>>                                                        */
      
                                                                            /*
   +------------------------------------------------------------------------+
   |(3) calculate the x-y positions of the upper and lower vertices         |
   +------------------------------------------------------------------------+
                                                                            */
    xc0 = vtx[vc0].x,    yc0 = vtx[vc0].y;

    n = 0;
    for( j = 0; j < no_art; j++ )
       {
         v0 = multi[no_multi].artv[j];
         if( v0 != vc0 )
           {
             v1   = multi[no_multi].edge[j];
             x0   = multi[no_multi].edgex[j];
             y0   = multi[no_multi].edgey[j];
             dydx = ( y0 - yc0 )/( x0 - xc0 );
             b    = yc0 - dydx*xc0;
             xx   = xc0 + 60.0;
             yy   = dydx*xx + b;

             if( n == 0 )
               {
                 vtx[v0].x = xx,    vtx[v0].y = yy;

                 y1 = yy;                                                   /*
                <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                             */
                 put_straight( vc0, v1, v0 );                               /*
                <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                             */
                }
             else if( n == no_art - 2 )
               {
                 vtx[v0].x = xx,    vtx[v0].y = yy;

                 y2 = yy;                                                   /*
                <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                             */
                 put_straight( vc0, v1, v0 );                               /*
                <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                             */
                }

             n++;
            }
        }
                                                                            /*
   +------------------------------------------------------------------------+
   |(4) calculate the x-y positions of the middle vertices                  |
   +------------------------------------------------------------------------+
                                                                            */
    n    = 0;
    b    = ( y2 -y1 )/(float)( no_art - 2 );
    xx   = xc0 + 120.0;
    yy   = y1;
    for( j = 0; j < no_art; j++ )
       {
         v0 = multi[no_multi].artv[j];
         v1 = multi[no_multi].edge[j];
         if( v0 != vc0 )
           {
             if( n > 0 && n < no_art - 2 )
               {
                 yy   = yy  + b;

	         vtx[v0].x = xx,    vtx[v0].y = yy;
                                                                            /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                       */
                  chngv_straight( vc0, v1, vc0, v0 );                        /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                       */
                }

             n++;
            }
        }
   }
                                                                            /*
=============================================================================
 (II)    For two particles input
=============================================================================
                                                                            */
else
  {                                                                         /*
   +------------------------------------------------------------------------+
   |(1) move those vertices which locate in the downstream                  |
   +------------------------------------------------------------------------+
                                  
   <<<<<<<<<<<<>>>>>>>>>>>>>>>>                                             */
    move_others( vc0, 60.0 );                                              /*
   <<<<<<<<<<<<>>>>>>>>>>>>>>>>                                             */
                                                                            /*
   +------------------------------------------------------------------------+
   |(2) re-calculate the x-y positions of the edge vertices                 |
   +------------------------------------------------------------------------+
                                  
   <<<<<<<<>>>>>>>>>                                                        */
    medge_coord( 1 );                                                       /*
   <<<<<<<<>>>>>>>>>                                                        */
                                                                            /*
   +------------------------------------------------------------------------+
   |(3) calculate the x-y positions of the upper and lower vertices         |
   +------------------------------------------------------------------------+
                                                                            */
    xc0 = vtx[vc0].x,    yc0 = vtx[vc0].y;

    n_out = 0;
    for( j = 0; j < no_art; j++ )
       {
         x0 = multi[no_multi].edgex[j];
         if( x0 > xc0 + 6.0 )
           {
             v0   = multi[no_multi].artv[j];
             xx   = xc0 + 60.0;
             yy   = multi[no_multi].edgey[j];

             v1 = multi[no_multi].edge[j];
             flag = 0;
             if( v0 != v1 )
               {
                 if( get_kind( v1 ) == 2 )
		     flag = 1 ;
                                                                            /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                       */
                  chngv_straight( vc0, v1, vc0, v0 );                       /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                       */
                }
             else
               {
                 nl = vtx[v1].no_line;
                 for( k = 0; k < nl; k++ )
                    {
                      vx1  = vtxline[v1].vtxno[k];
                      if( get_kind( vx1 ) == 2 )
                        {
                          flag = 1;
                          k    = nl;
                         }
                                                                            /*
                     <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                   */
                      chngv_straight( vc0, vx1, vc0, v0 );                  /*
                     <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>                   */
                     }
                }
             if( yy < 40.0 && flag )
                 yy = yy + 10.0;
             else if( yy > ymax - 40.0 && flag )
                 yy = yy - 10.0;
             if( multi[no_multi].artv_flag[j] == -1 )
		  xx = xc0;

	     v_out[n_out] = v0;
	     j_out[n_out] = j;
             n_out++;

	     vtx[v0].x = xx,    vtx[v0].y = yy;

              /*
		printf(" 5point  v =%3d  ( x, y ) = (%8.2f,%8.2f)\n",
                         v0, xx, yy );
                printf("        v_out[%d] =%3d\n", n_out-1, v_out[n_out-1] );
              */
            }
        }

                                                                            /*
   +------------------------------------------------------------------------+
   | (4) Calculate the positions of input vertices v_in[i], j_in[i]         |
   |                  and the initial lines if they are on those lines      |
   +------------------------------------------------------------------------+
                                                                            */
    for( j = 0; j < 2; j++ )
       {                                                                    /*
         +------------------------------------------------------------------+
         |(a) Calculate the positions of input vertices v_in[i], j_in[i]    |
         +------------------------------------------------------------------+
                                                                            */
         v0 = v_in[j];
         if( j == 0 )
             v1 = v_out[j];
         else
             v1 = v_out[n_out-1];
         x0 = xc0;
         y0 = vtx[v1].y;

         vtx[v0].x = x0,    vtx[v0].y = y0;

		/*
		printf(" 5point j =%3d v1 =%3d  ( x, y ) = (%8.2f,%8.2f)\n",
                         j, v1, x0, y0 );
                */
                                                                            /*
         +------------------------------------------------------------------+
         |(b) register the input lines in the straight line table           |
         +------------------------------------------------------------------+
                                                                            */
	 k  = j_in[j];
	 v2 = multi[no_multi].edge[k];
           /*
            printf(" edge[%2d]  v-in[%2d] =%3d  v_out[%2d] =%3d\n",
                     v2, j, v_in[j], j, v_out[j] );
           */
         if( v2 != v0 )
           {                                                                /*
            <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                           */
             chngv_straight( v2, vc0, vc0, v0 );                            /*
            <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>                           */
            }
         else
           {
             nl = vtx[v1].no_line;
             for( k = 0; k < nl; k++ )
                {
                  vx1  = vtxline[v1].vtxno[k];
                                                                            /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                     */
                  chngv_straight( vx1, vc0, vc0, v0 );                      /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                     */
                 }
            }
                                                                            /*
         +------------------------------------------------------------------+
         |(c) If the input line is the initial line,                        |
         |      re-determine the vertex poisitons on the initial line       |
         +------------------------------------------------------------------+
                                                                            */
         for( l = 0; l < no_initial; l++ )
            {                                                               /*
                   <<<<<<<<<<<<<>>>>>>>>>>>>>>                              */
              nl  = chck_straight( l, v0 );                                 /*
                   <<<<<<<<<<<<<>>>>>>>>>>>>>>                              */

              if( nl >= 0 )
                {
                  x1 = vtx[l].x;
                  y1 = y0 - 10.0;
                  if( l == 1 ) y1 = y0 + 10.0;
                                                                            /*
                  <<<<<<<<<<<<<<>>>>>>>>>>>>>>                              */
                   changey_straight( l, y1 );                               /*
                  <<<<<<<<<<<<<<>>>>>>>>>>>>>>                              */
                 }
             }

        }
   }
                                                                            /*
=============================================================================
  Save the internal loop lines in the straight line table
=============================================================================
                                                                            */
    for( j = 0; j < 4; j++ )
       {
         v0 = multi[no_multi].artv[j];
         nl = vtx[v0].no_line;
         for( k = 0; k < nl; k++ )
            {
              v1  = vtxline[v0].vtxno[k];
                                                                            /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                     */
              if( chck_straight( v0, v1 ) < 0 )                             /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                     */
                  create_straight( v0, v1 );                                /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>                               */
                 }
        }
    nv = multi[no_multi].no_vtx;
    for( j = 0; j < nv; j++ )
       {
         v0 = multi[no_multi].vtxno[j];
         nl = vtx[v0].no_line;
         for( k = 0; k < nl; k++ )
            {
              v1  = vtxline[v0].vtxno[k];
                                                                            /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                     */
              if( chck_straight( v0, v1 ) < 0 )                             /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                     */
                  create_straight( v0, v1 );                                /*
                 <<<<<<<<<<<<<<<<<>>>>>>>>>>>                               */
                 }
        }
                                                                            /*
   <<<<<<<<<<<>>>>>>>>>>>>>>>>                                              */
    move_others( 0, 0.001 );                                                /*
   <<<<<<<<<<<>>>>>>>>>>>>>>>>                                              */
}  


/*****************************************************************************
                   ============================                             */
                      void medge_coord( type )
/*                 =============================
******************************************************************************
*( Function )                                                                *
*     To calculate the x, y coordinates of edge_vertices for the loop        *
*( input )                                                                   *
*     structure vtx[i] : vertex information                                  *
*     structure vtxcor[i] : vertex part nformation                           *
*( output )                                                                  *
*     structure multi[i].edgex : x position of edge                          *
*     structure multi[i].edgey : y position of edge                          *
******************************************************************************
                                                                           */
int type;
{
                                                                            /*
 ------------------
  local variables
 ------------------
                                                                            */
int flag, left, right, fermion;
int i, j, k, l, n, nl, nx, v, ve, v0, v1;
float x, y, x1, y1;
float max_y, min_y;
int   maxv[3], minv[3];
int   nclustr;
                                                                            /*
 ------------------
   Functions
 ------------------
                                                                            */
 int iabs( );
 int chck_straight( );
 int getv_straight( );
                                                                            /*
================================ Entry point =================================
                                                                            */

max_y   = -1000.0, min_y   = 1000.0;
maxv[0] =      -1, minv[0] = -1;

no_art = multi[no_multi].no_edge;
                                                                            /*
==============================================================================
 I. Determine the virtual position of edge by the average of y values
==============================================================================
                                                                            */
for( j = 0; j < no_art; j++ )
   {
       /*
	printf(" no_art =%3d  j =%3d\n", no_art, j );
       */

     ve = multi[no_multi].edge[j];
     v  = multi[no_multi].artv[j];
                                                                            /*
     +-----------------------------------------------------------------------+
     | [1] vertex v has two or more edge vertices                            |
     +-----------------------------------------------------------------------+
                                                                            */
     if( v == ve )
       {                                                                    /*
         +-------------------------------------------------------------------+
         | (1) determine the numbers of left and right vertices              |
         |        left  = the number of vertices in the upstream of v_c      |
         |        right = the number of vertices in the downstream of v_c    |
         +-------------------------------------------------------------------+
                                                                            */
         left = 0,  right = 0;
         n = vtx[v].no_line;
         for( l = 0; l < n; l++ )
            {
              v1 = vtxline[v].vtxno[l];                                     /*
              +--------------------------------------------------------------+
              | (a) check whether the vertex v1 is an edge vertex or art_v   |
              +--------------------------------------------------------------+
                                                                            */
              flag = 0;
              for( k = 0; k < no_art; k++ )
                 {
                   if( v1 == multi[no_multi].artv[k] )
                     {
                       flag = 1;
                       k    = no_art;
                      }
                  }
                                                                            /*
              +--------------------------------------------------------------+
              | (b) if v1 is an edge vertex, judge whether is left or right. |
              +--------------------------------------------------------------+
                                                                            */
              if( flag == 0 )
                {
                  x = vtx[v1].x;
                  if( x > xc0 )
                      right++;
                  else
                    {
                      v0 = v1;
                      nl = vtxline[v].lineno[l];
                      left++;
                     }
                 }
             }
                                                                            /*
         +-------------------------------------------------------------------+
         | (2) Only one vertex is in the upstream                            |
         +-------------------------------------------------------------------+
                                                                            */
         if( left == 1 )
           {
             y1 = vtx[v0].y;
             multi[no_multi].edgex[j] = vtx[v0].x;
             multi[no_multi].edgey[j] = y1;
             k  = intline[nl].direct;
             if( iabs( k ) == 1 )
                 multi[no_multi].edge_flag[j] = 1;
            }
                                                                            /*
         +-------------------------------------------------------------------+
         | (3) Two or more vertices are in the dowstream                     |
         +-------------------------------------------------------------------+
                                                                            */
         else if( right > 1 )
           {
             multi[no_multi].edge_flag[j] = 0;
             x  = 0.0,  y = 0.0;
             nx = 0;
             for( l = 0; l < n; l++ )
                {
                  v1 = vtxline[v].vtxno[l];
                                                                            /*
                  +----------------------------------------------------------+
                  | (a) chck if the vertex v1 is an edge vertex.             |
                  +----------------------------------------------------------+
                                                                            */
                  flag = 0;
                  for( k = 0; k < 3; k++ )
                     {
                       if( v1 == multi[no_multi].artv[k] )
                         {
                           flag = 1;
                           k    = 3;
                          }
                      }
                                                                            /*
                  +----------------------------------------------------------+
                  | (b) if v1 is an edge vertex of the vertex v,             |
                  |       take each sum for x and y values of right vertices |
                  +----------------------------------------------------------+
                                                                            */
                  if( flag == 0 )
                    {
                      if( vtx[v1].x > xc0 )
                        {
                          nx++;
                            /*
                              printf(" nx =%2d vtx[%2d] = (%8.1f,%8.1f)\n",
                                       nx, v1, vtx[v1].x, vtx[v1].y );
                            */
                          y1 = vtx[v1].y;
                          x  = x + vtx[v1].x;
                          y  = y + y1;
                          nl = vtxline[v].lineno[l];
                          k  = intline[nl].direct;
                          if( iabs( k ) == 1 )
                              multi[no_multi].edge_flag[j] = 1;
                          if( y1 > max_y )
                            {
                              max_y   = y1;
                              maxv[0] = j;
                             }
                          if( y1 < min_y )
                            {
                              min_y   = y1;
                              minv[0] = j;
                             }
                         }
                     }
                 }
                                                                            /*
              +--------------------------------------------------------------+
              | (c) take the averages of x and y coordinates as the edge     |
              +--------------------------------------------------------------+
                                                                            */
             multi[no_multi].edgex[j] = x / nx;
             multi[no_multi].edgey[j] = y / nx;
                  /*
		   printf(" edge[%d] = (%8.1f, %8.1f)\n",
                            j, multi[no_multi].edgex[j], 
			       multi[no_multi].edgey[j] );
                  */
            }
                                                                            /*
         +-------------------------------------------------------------------+
         | (4) Two or more vertices are in the upstream                      |
         +-------------------------------------------------------------------+
                                                                            */
         else
           {
             multi[no_multi].edge_flag[j] = 0;
             x  = 0.0,  y = 0.0;
             nx = 0;
             for( l = 0; l < n; l++ )
                {
                  v1 = vtxline[v].vtxno[l];
                                                                            /*
                  +----------------------------------------------------------+
                  | (a) chck if the vertex v1 is an edge vertex.             |
                  +----------------------------------------------------------+
                                                                            */
                  flag = 0;
                  for( k = 0; k < 3; k++ )
                     {
                       if( v1 == multi[no_multi].artv[k] )
                         {
                           flag = 1;
                           k    = 3;
                          }
                      }
                                                                            /*
                  +----------------------------------------------------------+
                  | (b) if v1 is an edge vertex of the vertex v,             |
                  |       take each sum for x and y values of left vertices  |
                  +----------------------------------------------------------+
                                                                            */
                  if( flag == 0 )
                    {
                      if( vtx[v1].x < xc0 )
                        {
                          nx++;
                          x = x + vtx[v1].x;
                          y = y + vtx[v1].y;
                          nl = vtxline[v].lineno[l];
                          k  = intline[nl].direct;
                          if( iabs( k ) == 1 )
                              multi[no_multi].edge_flag[j] = 1;
                         }
                     }
                 }
                                                                            /*
              +--------------------------------------------------------------+
              | (c) take the averages of x and y coordinates as the edge     |
              +--------------------------------------------------------------+
                                                                            */
             multi[no_multi].edgex[j] = x / nx;
             multi[no_multi].edgey[j] = y / nx;
            }
        }
                                                                            /*
     +-----------------------------------------------------------------------+
     | [2] vertex v has only one edge vertex                                 |
     +-----------------------------------------------------------------------+
                                                                            */
     else
       {                                                                    /*
              <<<<<<<<<<<<<>>>>>>>>>>>>>                                    */
          nl  = chck_straight( vc0, ve );                                   /*
              <<<<<<<<<<<<<>>>>>>>>>>>>>                                    */

          v0 = ve;
          if( nl >= 0 )
            {
              x  = vtx[ve].x;
              if( vtx[ve].x < xc0 ) v0  = getv_straight( nl, 0 );
              else if( x > xc0 )    v0  = getv_straight( nl, 1 );
             }

          x  = vtx[v0].x,     y = vtx[v0].y;
          multi[no_multi].edgex[j] = x;
          multi[no_multi].edgey[j] = y;
          nl = multi[no_multi].edge_line[j];
          k  = intline[nl].direct;
          if( iabs( k ) == 1 )
              multi[no_multi].edge_flag[j] = 1;
          else
              multi[no_multi].edge_flag[j] = 0;

          if( x > xc0 )
            {
              if( y > max_y )
                {
                  max_y   = y;
                  maxv[0] = j;
                 }
              if( y < min_y )
                {
                  min_y   = y;
                  minv[0] = j;
                 }
             }
        }
    }
   /*
    printf(" maxv[0] = %5d  max_y = %6.1f\n", maxv[0], max_y );
    printf(" vtx_ID  = %5d\n", multi[no_multi].edge[maxv[0]] );
    printf(" minv[0] = %5d  min_y = %6.1f\n", minv[0], min_y );
    printf(" vtx_ID  = %5d\n", multi[no_multi].edge[minv[0]] );
    printf(" no_cluster = %d\n", no_cluster );

    printf("   j    artv    edge  edge-x  edge-y\n" );
    for( i = 0; i < no_art; i++ )
         printf(" %3d %7d %7d %7.1f %7.1f\n",
                   i, multi[no_multi].artv[i], multi[no_multi].edge[i],
                      multi[no_multi].edgex[i], multi[no_multi].edgey[i] );
   */
                                                                            /*
==============================================================================
 II. Adjust the virtual position of edge by the average of y values
==============================================================================
                                                                            */
if( type > 0 )
  {
 if( no_cluster == 1 )
   {
     for( i = 0; i < 2; i++ )
        {
          if( i == 0 ) j  = maxv[0];
          else        j  = minv[0];

          if( j >= 0 )
            {
              ve = multi[no_multi].edge[j];
              v  = multi[no_multi].artv[j];

              if( ve == v ) 
                {
                  if( i == 0 ) multi[no_multi].edgey[j] = max_y;
                  if( i == 1 ) multi[no_multi].edgey[j] = min_y;
                 }
              else if( get_kind( ve ) != 2 && vtx[ve].x > xc0 )
                {
                  nl      = vtx[ve].no_line;
                  if( nl > 2 )
                    {
                      if( i == 0 )
                        {
                          for( k = 0; k < nl; k ++ )
                             {
                               v1  = vtxline[ve].vtxno[k];
                               x   = vtx[v1].x,    y = vtx[v1].y;
                               if( vtx[v1].x > xc0 && y > max_y )
                                   max_y = y;
                              }
                          y1 = max_y - 10.0;
                         }
                      if( i == 1 )
                        {
                          for( k = 0; k < nl; k ++ )
                             {
                               v1  = vtxline[ve].vtxno[k];
                               x   = vtx[v1].x,    y = vtx[v1].y;
                               if( vtx[v1].x > xc0 && y < min_y )
                                   min_y = y;
                              }
                          y1 = min_y + 10.0;
                         }
                      multi[no_multi].edgey[j] = y1;
                      vtx[ve].y                = y1;
                     }
                 }
             }
         }
    }
 else
   {
     for( i = 0; i < 2; i++ )
        {
          if( i == 0 ) j  = maxv[0];
          else        j  = minv[0];

	  if( j >= 0 )
            {
              ve = multi[no_multi].edge[j];
              v  = multi[no_multi].artv[j];

              if( ve == v ) 
                {
                  nclustr = get_clID( v );

                  if( nclustr < 0 )
                    {
                      nl = vtx[ve].no_line;
                      for( k = 0; k < nl; k++ )
                         {
                           v0 = vtxline[ve].vtxno[k];
                           if( vtx[v0].x > xc0 + 5.0 )
                             {
                                nclustr = get_clID( v0 );
                                if( nclustr >= 0 )
                                    k = nl;
                              }
                          }
                     }

                  if(  nclustr == 0 )
                    {
                      if( i == 1 ) 
                        {
                          multi[no_multi].edgey[j] = min_y;
                         }
                     }
                  else if( nclustr == no_cluster-1 )
                    {
                      if( i == 0 )
                        {
                          multi[no_multi].edgey[j] = max_y;
                         }
                    }
                 }
              else if( get_kind( ve ) != 2 )
                {
                  nclustr = get_clID( ve );

                  if( nclustr < 0 )
                    {
                      nl = vtx[ve].no_line;
                      for( k = 0; k < nl; k++ )
                         {
                           v0 = vtxline[ve].vtxno[k];
                           if( vtx[v0].x > xc0 + 5.0 )
                             {
                                nclustr = get_clID( v0 );
                                if( nclustr >= 0 )
                                    k = nl;
                              }
                          }
                     }

                  nl      = vtx[ve].no_line;
                  if( i == 0 && nclustr == no_cluster-1 )
                    {
                      for( j = 0; j < nl; j ++ )
                         {
                            v1  = vtxline[ve].vtxno[j];
                            x   = vtx[v1].x,    y = vtx[v1].y;
                            if( vtx[v1].x > xc0 && y > max_y )
                                max_y = y;
                          }
                      multi[no_multi].edgey[j] = max_y;
                     }
                  if( i == 1 && nclustr == 0 )
                    {
                      for( j = 0; j < nl; j ++ )
                         {
                            v1  = vtxline[ve].vtxno[j];
                            x   = vtx[v1].x,    y = vtx[v1].y;
                            if( vtx[v1].x > xc0 && y < min_y )
                                min_y = y;
                          }
                      multi[no_multi].edgey[j] = min_y;
                     }
                 }
             }
         }
    }
  }
   /*
    printf("   j    artv    edge  edge-x  edge-y\n" );
    for( i = 0; i < no_art; i++ )
         printf(" %3d %7d %7d %7.1f %7.1f\n",
                   i, multi[no_multi].artv[i], multi[no_multi].edge[i],
                      multi[no_multi].edgex[i], multi[no_multi].edgey[i] );
   */

}

/*****************************************************************************
                   ============================                             */
                      void put_vcor(  )
/*                 =============================
******************************************************************************
*( Function )                                                                *
*     To put the vertex correction part in the tree Feynman diagram,  and    *
*     to determine their vertex positions.                                   *
*( input )                                                                   *
*     structure vtx[i] : vertex information                                  *
*     structure intline[i] : line information                                *
*( output )                                                                  *
*     The vertex position in the graph                                       *
*         structure vtx[].x and structure vtx[].y                            *
******************************************************************************
*/
{
                                                                            /*
  ----------------------------------
      local variable list
  ----------------------------------
                                                                            */
  int i, j, k, l, m, n, nl, v0, v1, v2, v3, typef;
  int ne, nv, times, kind;
  int triangle, left, right;
  float dydx, b, xwidthp, xwidthm;
  float x0, y0, x1, y1, x2, y2, xx, yy, dd, dd1;
  int flag, IRflag;
  double aa, bb, cc, a, xx0, yy0, ratio;
                                                                            /*
  ----------------------------------
      function list
  ----------------------------------
                                                                            */
  void  vedge_coord( );
  void  config_vcor( );
  void  move_others( );
  void  chngv_straight( );
  void  create_straight( );
  void  put_straight( );
  int   chck_straight( );
  int   mode( );
  int   iabs( );
  void  print_vcor( );

                                                                            /*
================== Entry point of put_vcor ================================
                                                                            */
  vtxtype = 0;

  vc0     = vtxcor[no_vcor].vcorv[0];
  vj      = 0;
  v3      = vc0;
  xc0     = vtx[vc0].x,            yc0      = vtx[vc0].y;

                                                                            /*
=============================================================================
(I)  Calculate the x-y position of the edge vertices
=============================================================================

  <<<<<<<<<>>>>>>>>>                                                       */
    vedge_coord( );                                                        /*
  <<<<<<<<<>>>>>>>>>                                                       */

        /*
         printf("======= determine the vertex position ============\n" );
         printf(" temporary central vertex number =%3d\n", vc0 );
         printf("  x  =%8.3f,   y  =%8.3f\n", xc0, yc0 );

	 print_vcor( 1 );
  
         printf(" central vertex is %3d\n", vc0 );
        */
                                                                            /*
=============================================================================
(II)  Determine the type of vertex correction part
=============================================================================

  <<<<<<<<<>>>>>>>>>                                                        */
   config_vcor( );                                                          /*
  <<<<<<<<<>>>>>>>>>                                                        */
                                                                            /*
=============================================================================
(III)  Determine the x-y positions of loop vertices
=============================================================================
 +---------------------------------------------------------------------------+
 |(1) Check whether the loop correlates with the initial lines or not.       |
 +---------------------------------------------------------------------------+
                                                                            */
 flag = 0;
 for( l = vj+1; l < vj+3; l++ )
    {
      k  = mode( l, 3 );
      x1 = vtxcor[no_vcor].edgex[k];
      v2 = vtxcor[no_vcor].edge[k];
      dd = x1 - xc0;
                      /*
		       printf("  Edge vertex v2 =%3d    x1 =%8.2f\n",
                                 v2, x1 );
                      */
      if( dd < 0.0 && dd > -110.0 )
        {
          for( m = 0; m < no_initial; m++ )
             {                                                              /*
                  <<<<<<<<<<<<<<>>>>>>>>>>>                                 */
               if( chck_straight( m, vc0 ) >= 0 )                           /*
                  <<<<<<<<<<<<<<>>>>>>>>>>>                                 */
                        if( n_photon[m] > 0 ) flag = 3;
              }
         }
     }
         /*
          printf("  flag =%3d  ( 1/ 0) = ( initial lines are moved/ no)\n",
                    flag );
         */

                                                                            /*
 +---------------------------------------------------------------------------+
 |(2) Shift the x-positions of vertices if there is no enough length of      |
 |    distance between the central vertex and others.                        |
 +---------------------------------------------------------------------------+
                                                                           */
 xwidthp = 120.0;
 xwidthm = 120.0;
 if( flag > 2 ) xwidthm = 90.0;

 for( l = vj+1; l < vj+3; l++ )
    {
      k  = mode( l, 3 );
      v2 = vtxcor[no_vcor].edge[k];
      if( vertflag == 1 && k == vert[0] )
                   ;
      else if(  vertflag == 2 && 
             ( k == vert[0] || k == vert[1] ) )
                   ;
      else
        {
          x1 = vtxcor[no_vcor].edgex[k];
          dd = x1 - xc0;
                 /*
                  printf("** flag =%3d  v2  =%3d  x1  =%8.2f  dd =%8.2f\n",
                                   flag, v2, x1, dd );
		  printf("              vc0 =%3d  xc0 =%8.2f\n",
                                   vc0, xc0 );
                 */
          if( dd < 0.0 ) dd = -dd;

          if( x1 < xc0 )
            {
              if( dd < xwidthm )
                {
                  dd = dd - xwidthm;
                                                                           /*
                  <<<<<<<<<<<<>>>>>>>>>>>>>                                */
                   move_others( vc0, dd );                                 /*
                  <<<<<<<<<<<<>>>>>>>>>>>>>                                */

                   xc0 = vtx[vc0].x;
                   yc0 = vtx[vc0].y;
                                                                           /*
                  <<<<<<<<<<>>>>>>>>>                                      */
                    vedge_coord(  );                                       /*
                  <<<<<<<<<<>>>>>>>>>                                      */
                       /*
                        printf("  All vertices were moved by dd = %8.3f\n",
                                  dd );
                       */
                 }
             }
          else if( x1 > xc0 )
            {
              if( dd < xwidthp )
                {
                  dd = xwidthp - dd;
                                                                           /*
                 <<<<<<<<<<<<>>>>>>>>>>>>>                                 */
                   move_others( vc0, dd );                                 /*
                 <<<<<<<<<<<<>>>>>>>>>>>>>                                 */
 
                  xc0 = vtx[vc0].x;
                  yc0 = vtx[vc0].y;
                                                                           /*
                  <<<<<<<<<>>>>>>>>>                                       */
                   vedge_coord(  );                                        /*
                  <<<<<<<<<>>>>>>>>>                                       */
                        /*
                          printf("  All vertices except for ");
        	          printf("initial lines were moved");
                          printf("  by dd = %8.3f\n", dd );
                        */
                 }
             }
                /*
		  printf("   changed    v2  =%3d  x1  =%8.2f\n",
                                    v2, vtx[v2].x );
                */
         }
    }
                                                                            /*
 +---------------------------------------------------------------------------+
 |(3) Check the shape of the vertex part   triangle = ( 0 / 1 )              |
 +---------------------------------------------------------------------------+
                                                                            */
 k  = mode( vj+1, 3 );
 l  = mode( vj+2, 3 );
 x1 = vtxcor[no_vcor].edgex[k],     y1 = vtxcor[no_vcor].edgey[k];

 x2 = vtxcor[no_vcor].edgex[l],     y2 = vtxcor[no_vcor].edgey[l];

 dd1 = ( x1 - x2 )*( x1 - x2 ) + ( y1 - y2 )*( y1 - y2 );
 dd1 = 0.25*sqrt( dd1 );
 xx  = 0.5*( x1 + x2 ), yy = 0.5*( y1 + y2 );
 dd  = ( xx - xc0 )*( xx - xc0 ) + ( yy - yc0 )*( yy - yc0 );
 dd  = sqrt( dd );

 triangle = 1;
 if( dd < dd1 ) triangle = 0;

          /*
           printf(" triangle =%3d\n", triangle );
           printf(" central vertex vc0 =%3d  ( x, y) = (%8.2f,%8.2f)\n",
                    vc0, xc0, yc0 );
          */
                                                                            /*
 +---------------------------------------------------------------------------+
 |(4) Calculate the x-y positions of loop vertices                           |
 +---------------------------------------------------------------------------+
                                                                            */
 for( k = vj+1; k < vj+3; k++ )
    {
      j  = mode( k, 3 );

      v1 = vtxcor[no_vcor].vcorv[j];
      x1 = vtxcor[no_vcor].edgex[j],     y1 = vtxcor[no_vcor].edgey[j];

           /*
             printf(" vc0  [%3d ]( xc0 =%8.1f, yc0 =%8.1f )\n", 
                          vc0, xc0, yc0 );
             printf(" edgev[%3d ](  x1 =%8.1f,  y1 =%8.1f )\n", 
                          vtxcor[no_vcor].edge[j], x1, y1 );
           */

      if( x1 > xc0 + 6.0 || x1 < xc0 - 6.0 )
        {
          dydx = ( y1 - yc0 )/( x1 - xc0 );
          b    = yc0 - dydx*xc0;
          dd   = 60.0;

          if( x1 > xc0 )
            {
              xx = xc0 + dd;
             }
          if( x1 < xc0 )
            {
              xx = xc0 - dd;
             }
          yy = dydx*xx + b;

                   /*
                     printf(" dydx =%8.2f\n", dydx );
                   */
         }
      else
        {
          if( triangle > 0 )
            {
              xx = 0.5*( x1 + xc0 );
              yy = 0.5*( y1 + yc0 );
             }
          else
            {
              xx = 0.4*( x1 - xc0 ) + xc0;
              yy = 0.4*( y1 - yc0 ) + yc0;
             }
         }

      vtx[v1].x = xx,  vtx[v1].y = yy;

      v2 = vtxcor[no_vcor].edge[j];
                                                                            /*
      <<<<<<<<<<<<<>>>>>>>>>>>>>>>>>                                        */
       put_straight( vc0, v2, v1 );                                         /*
      <<<<<<<<<<<<<>>>>>>>>>>>>>>>>>                                        */

               /*
                printf(" edge v2 =%3d,  (x, y ) = (%8.2f,%8.2f)\n",
                              v2, x1, y1 );
                printf(" loop v1 =%3d,  (x, y ) = (%8.2f,%8.2f)\n",
			    v1, xx, yy );
               */
     }
      /*
      printf(" v0 ( x, y ) = (%8.3f,%8.3f)\n", vtx[0].x, vtx[0].y );
      */

                                                                            /*
=============================================================================
(IV)  Calculate the middle point of the internal line
=============================================================================
                                                                            */
 k  = mode( vj+1, 3 );
 l  = mode( vj+2, 3 );
 v2 = vtxcor[no_vcor].vcorv[k];
 x1 = vtx[v2].x,       y1 = vtx[v2].y;
 v3 = vtxcor[no_vcor].vcorv[l];
 x2 = vtx[v3].x,       y2 = vtx[v3].y;

 dd1 = ( x1 - x2 )*( x1 - x2 ) + ( y1 - y2 )*( y1 - y2 );
 dd1 = 0.25*sqrt( dd1 );
 xx = 0.5*( x1 + x2 ), yy = 0.5*( y1 + y2 );
 dd = ( xx - xc0 )*( xx - xc0 ) + ( yy - yc0 )*( yy - yc0 );
 dd = sqrt( dd );

              /*
               printf(" center vertex[%3d ], (%8.2f, %8.2f)\n",
                        vc0, xc0, yc0 );
               printf(" side   vertex[%3d ], (%8.2f, %8.2f)\n",
                        v2, x1, y1 );
               printf(" side   vertex[%3d ], (%8.2f, %8.2f)\n",
                        v3, x2, y2 );
               printf(" dd[%6.1f] < dd1[%6.1f]\n", dd, dd1 );
               if( dd < dd1 )
               printf(" midle point          (%8.2f, %8.2f)\n",
                            xx, yy );
              */

 if( dd < dd1 )
   {
     if( ( xx - xc0 > -5.0 && xx - xc0 < 5.0 ) )
       {
         if( yy > yc0 )
             yy = yy + dd1;
         else if( yy < yc0 )
             yy = yy - dd1;
         else
           {
             if( yc0 < 20.0 )
                 yy = yc0 - dd1;
             else
                 yy = yc0 + dd1;
            }
        }
     else if( ( yy - yc0 > -5.0 && yy - yc0 < 5.0 ) )
       {
         if( xx > xc0 )
             xx = xx + dd1;
         else if( xx < xc0 )
             xx = xx - dd1;
         else
           {
             if( xc0 < 20.0 )
                 xx = xx - dd1;
             else
                 xx = xx + dd1;
            }
        }
     else
       {
         aa    = (y2 - y1)/(x2 - x1);
         bb    = yy - aa*xx;

         a     = (x2 - x1)/(y1 - y2);
         b     = yy - a*xx;
         ratio = a*a + 1.0;
         ratio = sqrt( ratio );
         ratio = dd1/ratio;
         /*
          printf("  xx   = %6.1f   -a   = %6.1f\n", xx, -a );
          printf("  yy   = %6.1f    b   = %6.1f\n", yy,  b );
          printf("  yc0  = %6.1f\n", yc0 );
         */
         if( aa*xc0 + bb > yc0 )
           {
             if( -a > 0.0 ) xx  = xx0 + ratio;
             else           xx  = xx0 - ratio;
            }
         else
           {
             if( -a > 0.0 ) xx  = xx0 - ratio;
             else           xx  = xx0 + ratio;
            }
         yy = a*xx + b;
        }

     nl = vtxcor[no_vcor].vcor_line[k];
     intline[nl].x = xx;
     intline[nl].y = yy;

         /*
          printf(" final midle point    (%8.2f, %8.2f)\n",
                   xx, yy );
         */
    }
 else
                                                                            /*
  <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                                            */
    create_straight( v2, v3 );                                              /*
  <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>                                            */

}

/*****************************************************************************
                   ============================                             */
                      void print_vcor( flag )
/*                 =============================
******************************************************************************
*( Function )                                                                *
*     To print the vertex correction part                                    *
******************************************************************************
									     */
int flag;
{
    printf(" ====== vertex correction part %d ====\n", no_vcor );
    printf("  edge0 ln[%2d]     v0   l0[%2d]      v1 ln[%2d]  edge1\n",
              vtxcor[no_vcor].edge_line[0], vtxcor[no_vcor].vcor_line[0],
              vtxcor[no_vcor].edge_line[1] );
    printf(" v [%2d] ----- v0 [%2d] ========= v1 [%2d] ------ v [%2d]\n",
              vtxcor[no_vcor].edge[0], vtxcor[no_vcor].vcorv[0],
              vtxcor[no_vcor].vcorv[1], vtxcor[no_vcor].edge[1] );
   if( flag == 1 )
     {
       printf("%8.2f                                 %8.2f\n",
              vtxcor[no_vcor].edgex[0], vtxcor[no_vcor].edgex[1] );
       printf("%8.2f                                 %8.2f\n",
              vtxcor[no_vcor].edgey[0], vtxcor[no_vcor].edgey[1] );
      }
    printf("                  \\\\             // \n" );
    printf("               l2 [%2d]       l1 [%2d] \n",
              vtxcor[no_vcor].vcor_line[2], vtxcor[no_vcor].vcor_line[1] );
    printf("                     \\\\        // \n" );
    printf("                        v2 [%2d] \n",
              vtxcor[no_vcor].vcorv[2] );
    printf("                           ||\n" );
    printf("                        ln [%2d]\n",
              vtxcor[no_vcor].edge_line[2] );
    printf("                           ||\n" );
    printf("                  edge2  v [%2d]\n", vtxcor[no_vcor].edge[2] );
   if( flag == 1 )
     {
       printf("                        %8.2f\n", vtxcor[no_vcor].edgex[2] );
       printf("                        %8.2f\n", vtxcor[no_vcor].edgey[2] );
      }
}


/*****************************************************************************
                   ============================                             */
                      void vedge_coord(  )
/*                 =============================
******************************************************************************
*( Function )                                                                *
*     To calculate the x, y coordinates of edge_vertices for the loop        *
*( input )                                                                   *
*     structure vtx[i] : vertex information                                  *
*     structure vtxcor[i] : vertex part nformation                           *
*( output )                                                                  *
*     structure vtxcor[i].edgex : x position of edge                         *
*     structure vtxcor[i].edgey : y position of edge                         *
******************************************************************************
*/
{
                                                                            /*
 ------------------
  local variables
 ------------------
                                                                            */
int flag, left, right, fermion;
int i, j, k, l, n, nl, nx, v, ve, v0, v1;
int rlflag[3], nleft[3], vleft[3][3], nright[3], vright[3][3];
float x, y, x1, y1;
float max_y, min_y;
int   maxv[3], minv[3];
int   nclustr;
                                                                            /*
 ------------------
   Functions
 ------------------
                                                                            */
 int iabs( );
                                                                            /*
================================ Entry point =================================
         
+----------------------------------------------------------------------------+
|(1)  Count the numbers of left side and right side vertices                 |
+----------------------------------------------------------------------------+
                                                                            */
for( j = 0; j < 3; j++ )
   {
     ve = vtxcor[no_vcor].edge[j];
     v  = vtxcor[no_vcor].vcorv[j];
       /*
        printf("  j =%3d  ve =%3d  v =%3d\n", j, ve, v );
       */
     left = 0;  right = 0;
                                                                            /*
     +-----------------------------------------------------------------------+
     |(a) If vertex v is equal to edge vertex                                |
     +-----------------------------------------------------------------------+
                                                                            */
     if( v == ve )
       {
         n = vtx[v].no_line;
         for( l = 0; l < n; l++ )
            {
              v1 = vtxline[v].vtxno[l];
                                                                            /*
              +--------------------------------------------------------------+
              | (i) check whether the vertex v1 is an edge vertex or art_v   |
              +--------------------------------------------------------------+
                                                                            */
              flag = 0;
              for( k = 0; k < 3; k++ )
                 {
                   if( v1 == vtxcor[no_vcor].vcorv[k] )
                     {
                       flag = 1;
                       k    = 3;
                      }
                  }
                                                                            /*
              +--------------------------------------------------------------+
              | (ii) if v1 is an edge vertex, judge whether is left or right |
              +--------------------------------------------------------------+
                                                                            */
              if( flag == 0 )
                {
                  x = vtx[v1].x;
                  if( x > xc0 )
                    {
                      vright[right][j] = v1;
                      right++;
                     }
                  else
                    {
                      vleft[left][j]   = v1;
                      left++;
                     }
                 }
             }
        }
                                                                            /*
     +-----------------------------------------------------------------------+
     |(b) If vertex v has only one edge vertex                               |
     +-----------------------------------------------------------------------+
                                                                            */
     else
       {
          x  = vtx[ve].x;
          if( x > xc0 )
            {
              vright[right][j] = ve;
              right++;
             }
          else
            {
              vleft[left][j]   = ve;
              left++;
             }
        }
                                                                            /*
     +-----------------------------------------------------------------------+
     |(c) Store the numbers of left side and right side vertices             |
     +-----------------------------------------------------------------------+
                                                                            */
     nleft[j]  = left;
     nright[j] = right;
    }
       /*
        printf("**** Left-right configuration of vertices ******\n");
        for( j = 0; j < 3; j++ )
           {
             printf("  Loop vertex[%3d ]  %3d  left_vtx  ",
                       vtxcor[no_vcor].vcorv[j], nleft[j] );
             for( i = 0; i < nleft[j]; i++ ) printf("%4d", vleft[i][j] );
             printf("\n");
             printf("                     %3d  right_vtx ", nright[j] );
             for( i = 0; i < nright[j]; i++ ) printf("%4d", vright[i][j] );
             printf("\n");
            }
       */
                                                                            /*
+----------------------------------------------------------------------------+
|(2) Select left or right side vertex for connection                         |
+----------------------------------------------------------------------------+
                                                                            */
left = 0;
for( j = 0; j < 3; j++ )
   {
     if( nright[j] == 0 ) 
       {
         rlflag[j] = -1;
         left++;
        }
     else
         rlflag[j] = 1;
    }
if( left < 2 )
  {
    right = 0;
    for( j = 0; j < 3; j++ )
       {
         if( nleft[j] == 0 ) 
           {
             rlflag[j] = 1;
             right++;
            }
         else
             rlflag[j] = -1;
        }
    if( right < 2 )
      {
        for( j = 0; j < 3; j++ )
           {
             if( nleft[j] == 0 )       rlflag[j] =  1;
             else if( nright[j] == 0 ) rlflag[j] = -1;
             else                      rlflag[j] = -1;
            }
       }
   }
                                                                          /*
+--------------------------------------------------------------------------+
|(3) Determine the x-y positions of edge vertices                          |
+--------------------------------------------------------------------------+
                                                                          */
max_y   = -1000.0, min_y   = 1000.0;
maxv[0] =      -1, minv[0] = -1;


for( j = 0; j < 3; j++ )
   {                                                                      /*
     +---------------------------------------------------------------------+
     |(a) If the vertex has only one edge                                  |
     +---------------------------------------------------------------------+
                                                                          */
     if( nleft[j]+nright[j] == 1 )
       {
          ve = vtxcor[no_vcor].edge[j];
          x  = vtx[ve].x,     y = vtx[ve].y;
          vtxcor[no_vcor].edgex[j] = x;
          vtxcor[no_vcor].edgey[j] = y;
          nl = vtxcor[no_vcor].edge_line[j];
          k  = intline[nl].direct;
          if( iabs( k ) == 1 )
              vtxcor[no_vcor].edge_flag[j] = 1;
          else
              vtxcor[no_vcor].edge_flag[j] = 0;

          if( x > xc0 + 5.0 )
            {
              if( y > max_y )
                {
                  max_y   = y;
                  maxv[0] = j;
                 }
              if( y < min_y )
                {
                  min_y   = y;
                  minv[0] = j;
                 }
             }
        }
                                                                          /*
     +---------------------------------------------------------------------+
     |(b) If the vertex has more than one edge                             |
     +---------------------------------------------------------------------+
                                                                          */
     else
       {
         if( rlflag[j] == -1 ) 
           {
             n    = nleft[j];
             left = -1;
            }
         else if( rlflag[j] ==  1 )
           {
             n    = nright[j];
             left =  1;
            }
         else
           {
             n    = nleft[j] + nright[j];
             left =  0;
            }

         vtxcor[no_vcor].edge_flag[j] = 0;
         x  = 0.0,  y = 0.0;
         nx = 0;

         for( l = 0; l < n; l++ )
            {
              if(      left == -1 ) v1  = vleft[l][j];
              else if( left ==  1 ) v1  = vright[l][j];
              else 
                {
                  if( l < nleft[j] ) v1 = vleft[l][j];
                  else               v1 = vright[l-nleft[j]][j];
                 }
                                                                            /*
              +----------------------------------------------------------+
              | (i) if v1 is an edge vertex of the vertex v,             |
              |       take each sum for x and y values of right vertices |
              +----------------------------------------------------------+
                                                                           */
                  nx++;
                         /*
                              printf(" nx =%2d vtx[%2d] = (%8.1f,%8.1f)\n",
                                       nx, v1, vtx[v1].x, vtx[v1].y );
                         */
                  y1 = vtx[v1].y;
                  x  = x + vtx[v1].x;
                  y  = y + y1;
                  nl = vtxline[v].lineno[l];
                  k  = intline[nl].direct;
                  if( iabs( k ) == 1 )
                      vtxcor[no_vcor].edge_flag[j] = 1;
                  if( y1 > max_y )
                    {
                      max_y   = y1;
                      maxv[0] = j;
                     }
                  if( y1 < min_y )
                    {
                      min_y   = y1;
                      minv[0] = j;
                     }
            }
                                                                            /*
         +--------------------------------------------------------------+
         | (ii) take the averages of x and y coordinates as the edge    |
         +--------------------------------------------------------------+
                                                                            */
         vtxcor[no_vcor].edgex[j] = x / nx;
         vtxcor[no_vcor].edgey[j] = y / nx;
                  /*
		   printf(" edge[%d] = (%8.1f, %8.1f)\n",
                            j, vtxcor[no_vcor].edgex[j],
                               vtxcor[no_vcor].edgey[j] );
                  */
        }
    }

                                                                            /*
==============================================================================
 I. Determine the virtual position of edge by the average of y values
==============================================================================
                                                                            
for( j = 0; j < 3; j++ )
   {
     ve = vtxcor[no_vcor].edge[j];
     v  = vtxcor[no_vcor].vcorv[j];
      /*
        printf("  j =%3d  ve =%3d  v =%3d\n", j, ve, v );
      
                                                                            /*
     +-----------------------------------------------------------------------+
     | [1] vertex v has two or more edge vertices                            |
     +-----------------------------------------------------------------------+
                                                                            
     if( v == ve )
       {                                                                    /*
         +-------------------------------------------------------------------+
         | (1) determine the numbers of left and right vertices              |
         |        left  = the number of vertices in the upstream of v_c      |
         |        right = the number of vertices in the downstream of v_c    |
         +-------------------------------------------------------------------+
                                                                            
         left = 0,  right = 0;
         n = vtx[v].no_line;
         for( l = 0; l < n; l++ )
            {
              v1 = vtxline[v].vtxno[l];
                                                                            /*
              +--------------------------------------------------------------+
              | (a) check whether the vertex v1 is an edge vertex or art_v   |
              +--------------------------------------------------------------+
                                                                            
              flag = 0;
              for( k = 0; k < 3; k++ )
                 {
                   if( v1 == vtxcor[no_vcor].vcorv[k] )
                     {
                       flag = 1;
                       k    = 3;
                      }
                  }
                                                                            /*
              +--------------------------------------------------------------+
              | (b) if v1 is an edge vertex, judge whether is left or right. |
              +--------------------------------------------------------------+
                                                                            
              if( flag == 0 )
                {
                  x = vtx[v1].x;
                  if( x > xc0 )
                      right++;
                  else
                    {
                      v0 = v1;
                      nl = vtxline[v].lineno[l];
                      left++;
                     }
                 }
             }
              /*
               printf(" vcorv[ %d ]  left[ %d ] right[ %d ]\n",
                        v, left, right );

                                                                            /*
         +-------------------------------------------------------------------+
         | (2) Only one vertex is in the upstream                            |
         +-------------------------------------------------------------------+
                                                                            
         if( left == 1 )
           {
             y1 = vtx[v0].y;
             vtxcor[no_vcor].edgex[j] = vtx[v0].x;
             vtxcor[no_vcor].edgey[j] = y1;
             k  = intline[nl].direct;
             if( iabs( k ) == 1 )
                 vtxcor[no_vcor].edge_flag[j] = 1;
            }
                                                                            /*
         +-------------------------------------------------------------------+
         | (3) Two or more vertices are in the dowstream                     |
         +-------------------------------------------------------------------+
                                                                            
         else if( right > 1 )
           {
             vtxcor[no_vcor].edge_flag[j] = 0;
             x  = 0.0,  y = 0.0;
             nx = 0;
             for( l = 0; l < n; l++ )
                {
                  v1 = vtxline[v].vtxno[l];
                                                                            /*
                  +----------------------------------------------------------+
                  | (a) chck if the vertex v1 is an edge vertex.             |
                  +----------------------------------------------------------+
                                                                            
                  flag = 0;
                  for( k = 0; k < 3; k++ )
                     {
                       if( v1 == vtxcor[no_vcor].vcorv[k] )
                         {
                           flag = 1;
                           k    = 3;
                          }
                      }
                                                                            /*
                  +----------------------------------------------------------+
                  | (b) if v1 is an edge vertex of the vertex v,             |
                  |       take each sum for x and y values of right vertices |
                  +----------------------------------------------------------+
                                                                            
                  if( flag == 0 )
                    {
                      if( vtx[v1].x > xc0 + 5.0 )
                        {
                          nx++;
                            /*
                              printf(" nx =%2d vtx[%2d] = (%8.1f,%8.1f)\n",
                                       nx, v1, vtx[v1].x, vtx[v1].y );
                            
                          y1 = vtx[v1].y;
                          x  = x + vtx[v1].x;
                          y  = y + y1;
                          nl = vtxline[v].lineno[l];
                          k  = intline[nl].direct;
                          if( iabs( k ) == 1 )
                              vtxcor[no_vcor].edge_flag[j] = 1;
                          if( y1 > max_y )
                            {
                              max_y   = y1;
                              maxv[0] = j;
                             }
                          if( y1 < min_y )
                            {
                              min_y   = y1;
                              minv[0] = j;
                             }
                         }
                     }
                 }
                                                                            /*
              +--------------------------------------------------------------+
              | (c) take the averages of x and y coordinates as the edge     |
              +--------------------------------------------------------------+
                                                                            
             vtxcor[no_vcor].edgex[j] = x / nx;
             vtxcor[no_vcor].edgey[j] = y / nx;
                  /*
		   printf(" edge[%d] = (%8.1f, %8.1f)\n",
                            j, vtxcor[no_vcor].edgex[j],
                               vtxcor[no_vcor].edgey[j] );
                  
            }
                                                                            /*
         +-------------------------------------------------------------------+
         | (4) Two or more vertices are in the upstream                      |
         +-------------------------------------------------------------------+
                                                                            
         else
           {
             vtxcor[no_vcor].edge_flag[j] = 0;
             x  = 0.0,  y = 0.0;
             nx = 0;
             for( l = 0; l < n; l++ )
                {
                  v1 = vtxline[v].vtxno[l];
                                                                            /*
                  +----------------------------------------------------------+
                  | (a) chck if the vertex v1 is an edge vertex.             |
                  +----------------------------------------------------------+
                                                                            
                  flag = 0;
                  for( k = 0; k < 3; k++ )
                     {
                       if( v1 == vtxcor[no_vcor].vcorv[k] )
                         {
                           flag = 1;
                           k    = 3;
                          }
                      }
                                                                            /*
                  +----------------------------------------------------------+
                  | (b) if v1 is an edge vertex of the vertex v,             |
                  |       take each sum for x and y values of left vertices  |
                  +----------------------------------------------------------+
                                                                            
                  if( flag == 0 )
                    {
                      if( vtx[v1].x < xc0 )
                        {
                          nx++;
                          x = x + vtx[v1].x;
                          y = y + vtx[v1].y;
                          nl = vtxline[v].lineno[l];
                          k  = intline[nl].direct;
                          if( iabs( k ) == 1 )
                              vtxcor[no_vcor].edge_flag[j] = 1;
                         }
                     }
                 }
                                                                            /*
              +--------------------------------------------------------------+
              | (c) take the averages of x and y coordinates as the edge     |
              +--------------------------------------------------------------+
                                                                            
             vtxcor[no_vcor].edgex[j] = x / nx;
             vtxcor[no_vcor].edgey[j] = y / nx;
            }
        }
                                                                            /*
     +-----------------------------------------------------------------------+
     | [2] vertex v has only one edge vertex                                 |
     +-----------------------------------------------------------------------+
                                                                            
     else
       {
          x  = vtx[ve].x,     y = vtx[ve].y;
          vtxcor[no_vcor].edgex[j] = x;
          vtxcor[no_vcor].edgey[j] = y;
          nl = vtxcor[no_vcor].edge_line[j];
          k  = intline[nl].direct;
          if( iabs( k ) == 1 )
              vtxcor[no_vcor].edge_flag[j] = 1;
          else
              vtxcor[no_vcor].edge_flag[j] = 0;

          if( x > xc0 + 5.0 )
            {
              if( y > max_y )
                {
                  max_y   = y;
                  maxv[0] = j;
                 }
              if( y < min_y )
                {
                  min_y   = y;
                  minv[0] = j;
                 }
             }
        }
        /*
         printf(" j[%3d]   vcorv[%3d]   edge[%3d]",
                  j, vtxcor[no_vcor].vcorv[j], vtxcor[no_vcor].edge[j] );
         printf("   maxv[%2d, %6.1f],   minv[%2d, %6.1f]\n",
                  maxv[0], max_y, minv[0], min_y );
        
    }
   /*
    printf(" no_cluster = %d\n", no_cluster );
    printf("   j   vcorv    edge  edge-x  edge-y\n" );
    for( i = 0; i < 3; i++ )
         printf(" %3d %7d %7d %7.1f %7.1f\n",
                   i, vtxcor[no_vcor].vcorv[i], vtxcor[no_vcor].edge[i],
                      vtxcor[no_vcor].edgex[i], vtxcor[no_vcor].edgey[i] );
   */
                                                                            /*
==============================================================================
 II. Adjust the virtual position of edge by the average of y values
==============================================================================
                                                                            */
 if( no_cluster == 1 )
   {
     for( i = 0; i < 2; i++ )
        {
          if( i == 0 ) j  = maxv[0];
          else        j  = minv[0];

          if( j >= 0 )
            {
              ve = vtxcor[no_vcor].edge[j];
              v  = vtxcor[no_vcor].vcorv[j];

              if( ve == v ) 
                {
                  if( i == 0 ) vtxcor[no_vcor].edgey[j] = max_y;
                  if( i == 1 ) vtxcor[no_vcor].edgey[j] = min_y;
                 }
             }
         }
    }
 else
   {
     for( i = 0; i < 2; i++ )
        {
          if( i == 0 ) j  = maxv[0];
          else         j  = minv[0];

	  if( j >= 0 )
            {
              ve = vtxcor[no_vcor].edge[j];
              v  = vtxcor[no_vcor].vcorv[j];

              if( ve == v ) 
                {
                  nclustr = get_clID( vc0 );

                  if( nclustr < 0 )
                    {
                      nl = vtx[ve].no_line;
                      for( k = 0; k < nl; k++ )
                         {
                           v0 = vtxline[ve].vtxno[k];
                           if( vtx[v0].x > xc0 + 5.0 )
                             {
                                nclustr = get_clID( v0 );
                                if( nclustr >= 0 )
                                    k = nl;
                              }
                          }
                     }


                  if(  nclustr == 0 )
                    {
                      if( i == 1 ) 
                        {
                          vtxcor[no_vcor].edgey[j] = min_y;
                         }
                     }
                  else if( nclustr == no_cluster-1 )
                    {
                      if( i == 0 )
                        {
                          vtxcor[no_vcor].edgey[j] = max_y;
                         }
                    }
                 }
             }
         }
    }
   /*
    printf("   j   vcorv    edge  edge-x  edge-y\n" );
    for( i = 0; i < 3; i++ )
         printf(" %3d %7d %7d %7.1f %7.1f\n",
                   i, vtxcor[no_vcor].vcorv[i], vtxcor[no_vcor].edge[i],
                      vtxcor[no_vcor].edgex[i], vtxcor[no_vcor].edgey[i] );
   */

}

/*****************************************************************************
                   ============================                             */
                      void config_vcor(  )
/*                 =============================
******************************************************************************
*( Function )                                                                *
*     To determine the type of vertex correction                             *
*( input )                                                                   *
*     structure vtx[i] : vertex information                                  *
*     structure intline[i] : line information                                *
*( output )                                                                  *
*     vc0   : the centeral vertex ID                                         *
*     vj    : address of the central vtx in the vcor table                   *
******************************************************************************
*/
{
                                                                            /*
  ----------------------------------
      local variable list
  ----------------------------------
                                                                            */
  int i, j, k, l, m, n, nl, v0, v1, v2, v3, typef;
  int ne, nv, times, kind;
  int triangle, left, right;
  float dydx, b, xwidthp, xwidthm;
  float x0, y0, x1, y1, x2, y2, xx, yy, dd, dd1;
  int flag, IRflag;
  double aa, bb, cc;
                                                                            /*
  ----------------------------------
      function list
  ----------------------------------
                                                                            */
  int   mode( );
  void  move_others( );
  void  chngv_straight( );
  void  vedge_coord( );
  int   iabs( );

                                                                            /*
================== Entry point of config_vcor ================================
                                                                            */

 fermiflag = 0, horizflag = 0, vertflag  = 0;
 for( j = 0; j < 3; j++ )
    {
      fermi[j] = 0;
      horiz[j] = 0;
      vert[j]  = 0;
     }


 dd1      = 15.0;
 triangle = 1;
 v3       = vc0;
  
                                                                            /*
=============================================================================
(I)  Determine several flags for distingushing type of vertex
=============================================================================
+----------------------------------------------------------------------------+
| (( flags ))                                                                |
|   (1) flag for the internal loop line                                      |
|          kind  : particle kind of the most popular particels in the loop   |
|          times : the number of those kind of particle with "kind"          |
|   (2) flag for the edge lines                                              |
|          fermiflag : the number of fermions in the edge lines              |
|          vertflag  : the number of vertical edge line                      |
|          horizflag : the number of horizontal edge lines                   |
+----------------------------------------------------------------------------+
                                                                            */

 for( j = 0; j < 5; j++ )
      tmpbuf[0][j] = 0;

 for( j = 0; j < 3; j++ )
    {
      nl    = vtxcor[no_vcor].vcor_line[j];
      kind  = intline[nl].direct;
      if( kind < 0 )
          kind = -kind;
      kind--;
      tmpbuf[0][kind] = tmpbuf[0][kind] + 1;

      x1  = vtxcor[no_vcor].edgex[j],  y1  = vtxcor[no_vcor].edgey[j];

      if( ( x1 < xc0 + dd1 ) && ( x1 > xc0 - dd1 ) )
        {
          vert[vertflag] = j;
          vertflag++;
         }
      if( ( y1 < yc0 + dd1 ) && ( y1 > yc0 - dd1 ) )
        {
          horiz[horizflag] = j;
          horizflag++;
         }
      if( vtxcor[no_vcor].edge_flag[j] == 1 )
        {
          fermi[fermiflag] = j;
          fermiflag++;
         }
     }
     /*
       printf("======== list of particle kind ======\n");
       printf("    1    2    3    4    5\n");
       printf("%5d%5d%5d%5d%5d\n", tmpbuf[0][0],tmpbuf[0][1],
	       tmpbuf[0][2],tmpbuf[0][3],tmpbuf[0][4] );
     */

 times  = 0, kind = 0;
 for( j = 0; j < 5; j++ )
    {
      if( tmpbuf[0][j] > times )
        {
          times = tmpbuf[0][j];
          kind = j + 1;
         }
     }
     /*
       printf("  kind =%3d,  times =%3d\n",
                 kind, times );
       printf("  fermiflag =%3d\n", fermiflag );
       printf("    0    1    2\n");
       for( j = 0; j < fermiflag; j++ )
          {
            printf("%5d",fermi[j] );
           }
       printf("\n");
       printf("  horizflag =%3d\n", horizflag );
       printf("    0    1    2\n");
       for( j = 0; j < horizflag; j++ )
          {
            printf("%5d",horiz[j] );
           }
       printf("\n");
       printf("  vertflag  =%3d\n", vertflag );
       printf("    0    1    2\n");
       for( j = 0; j < vertflag; j++ )
          {
            printf("%5d",vert[j] );
           }
       printf("\n");
       printf(" vc0[%d] ( x, y ) = (%8.3f,%8.3f)\n", 
	       vc0, vtx[vc0].x, vtx[vc0].y );
     */

                                                                            /*
=============================================================================
(II)  Determine the type and the central vertex of three-points loop
=============================================================================

 +-------------------------------------------------------------------------+
 |                                                               |         |
 | (I) Only one edge line is vertical.             Type T0       +         |
 |                                                               |\        |
 +---------------------------------------------+                 | \       |
                                               |            -----+--+--    |*/
 if( vertflag == 1 )                         /*+---------------------------+*/
   {
     if( times == 2 && kind == 1 )
       {
         for( j = 1; j < 4; j++ )
            {
              l = vtxcor[no_vcor].vcor_line[mode( j-1, 3 )];
              m = vtxcor[no_vcor].vcor_line[mode( j, 3 )];
              ne = intline[l].direct;
              nv = intline[m].direct;
                      /*
		       printf(" j =%2d  l[%2d,%2d] m[%2d,%2d]\n",
                                j, l, ne, m, nv );
                      */
              if( iabs( ne ) == 1 && iabs( nv ) == 1 )	
                {
                  vj  = mode( j, 3);
                  vc0 = vtxcor[no_vcor].vcorv[vj];
                        /*
			 printf("**** vertex part type T0-0 ***\n");
			 printf("     central vertex =%3d\n", vc0);
                         */
                 }
             }
        }
     else if( fermiflag == 2 )
       {
         flag = 0;
         for( j = 0; j < fermiflag; j++ )
            {
              if( fermi[j] == vert[0] )
                {
                  vc0 = vtxcor[no_vcor].vcorv[fermi[j]];
                  vj  = fermi[j];
                  flag = 1;
                         /*
			 printf("**** vertex part type T0-1 ***\n");
                         */
                 }
             }
          if( flag == 0 )
            {
              vc0 = vtxcor[no_vcor].vcorv[horiz[0]];
              vj  = horiz[0];
                        /*
			 printf("**** vertex part type T0-2 ***\n");
                        */
             }
        }
     else
       {
          vc0 = vtxcor[no_vcor].vcorv[horiz[0]];
          vj  = horiz[0];
                        /*
			 printf("**** vertex part type T0-3 ***\n");
                        */
        }
    }
                                                                            /*
 +-------------------------------------------------------------------------+
 |                                                                |        |
 | (II-1) Two edge_lines are vertical.              Type T0-4     +        |
 |                                                                |\       |
 +------------------------------------------------+               | \      |
                                                  |               |   \    |*/
 else if( vertflag == 2 )                       /*|               +----+-- |*/
   {                                            /*|               |        |*/
     if( times == 2 && kind == 1 )              /*+------------------------+*/
       {
         for( j = 1; j < 4; j++ )
            {
              l = vtxcor[no_vcor].vcor_line[mode( j-1, 3 )];
              m = vtxcor[no_vcor].vcor_line[mode( j, 3 )];
              ne = intline[l].direct;
              nv = intline[m].direct;
                      /*
		       printf(" j =%2d  l[%2d,%2d] m[%2d,%2d]\n",
                                j, l, ne, m, nv );
                      */
              if( iabs( ne ) == 1 && iabs( nv ) == 1 )	
                {
                  vj  = mode( j, 3);
                  vc0 = vtxcor[no_vcor].vcorv[vj];
                        /*
			 printf("**** vertex part type T0-4 ***\n");
			 printf("     central vertex =%3d\n", vc0);
                         */
                 }
             }
        }
                                                                            /*
 +-------------------------------------------------------------------------+
 |                                                                |        |
 | (II-2) Two edge_lines are vertical.              Type T1       +        |
 |                                                                |\       |
 +------------------------------------------------+           ----+ |      |
                                                  |               |/       |*/
     else                                       /*|               +        |*/
       {                                        /*|               |        |*/
         for( j = 0; j < 3; j++ )               /*+------------------------+*/
            {
              if( j != vert[0] && j != vert[1] )
                {
                  vc0 = vtxcor[no_vcor].vcorv[j];
                  vj  = j;
                        /*
			 printf("**** vertex part type T1-0 ***\n");
                        */
                  }
             }
        }
    }
                                                                            /*
 +-------------------------------------------------------------------------+
 | (III) There is no vertical line:  Generally Y type                      |
 +-------------------------------------------------------------------------+
                                                                            */
 else 
   {                                                                        /*
     +---------------------------------------------------------------------+
     |                                                                ..   |
     |(III-0)  Two loop lines are fermions     Type Y0              ..     |
     |                                                            ..  |    |
     +--------------------------------------------+       -------*    |    |
                                                  |               ..  |    |*/
     if( times == 2 && kind == 1 )              /*|                 ..|    |*/
       {                                        /*|                   ..   |*/
         for( j = 1; j < 4; j++ )               /*+------------------------+*/
            {
              l = vtxcor[no_vcor].vcor_line[mode( j-1, 3 )];
              m = vtxcor[no_vcor].vcor_line[mode( j, 3 )];
              ne = intline[l].direct;
              nv = intline[m].direct;
                       /*
		       printf(" j =%2d  l[%2d,%2d] m[%2d,%2d]\n",
                                j, l, ne, m, nv );
                       */

              if( iabs( ne ) == 1 && iabs( nv ) == 1 )	
                {
                  vj  = mode( j, 3);
                  vc0 = vtxcor[no_vcor].vcorv[vj];
                       /*
			 printf("**** vertex part type Y0 ***\n");
			 printf("     central vertex =%3d\n", vc0);
                       */
                 }
             }
        }
                                                                            /*
     +---------------------------------------------------------------------+
     |(III-1)  All oop lines are not fermions             Type Y1 or  Y2   |
     +---------------------------------------------------------------------+
                                                                            */
     else
       {
         y2   = 10000.0;
         left = 0,    right = 0;
         for( j = 0; j < 3; j++ )
            {
              x1  = vtxcor[no_vcor].edgex[j];
              y1  = vtxcor[no_vcor].edgey[j];

              if( x1 > xc0 + 6.0 )
                {
                  k  =  j;
                  if( y2 < y1 )  
                    {
	              y2 = y1;
	              vymax = j;
                     }
                  right++;
                 }
              else if( x1 < xc0 - 6.0 )
                {
                  l  =  j;
                  left++;
                 }
             }
         if( right > 1 ) vtxtype = 1;
                                                                            /*
         +-----------------------------------------------------------------+
         |                                                           /     |
         | (1) There is one horizontal line.     Type Y1            +      |
         |                                                         /|      |
         +----------------------------------------------+     ----+ |      |
                                                        |          \|      |*/
         if( horizflag == 1 )                         /*|           +      |*/
           {                                          /*|            \     |*/
             vc0 = vtxcor[no_vcor].vcorv[horiz[0]];   /*+------------------+*/
             vj  = horiz[0];
                        /*
			 printf("**** vertex part type Y1 ***\n");
                        */
            }
                                                                            /*
         +-----------------------------------------------------------------+
         |                                                           /     |
         | (2) There is no horizontal line.      Type Y2            +      |
         |                                                         /|      |
         +----------------------------------------------+         + |      |
                                                        |        / \|      |*/
         else                                         /*|       /   +      |*/
           {                                          /*|            \     |*/
             vj = -1;                                 /*+------------------+*/
             if( left == 1 )                   
                 vj = l;
             else if( right == 1 )
                 vj = k;

	     if( vj != -1 )
                 vc0 = vtxcor[no_vcor].vcorv[vj];
             else
                 vj = 0;
                        /*
			 printf("**** vertex part type Y2 ***\n");
			 printf(" left =%2d right =%2d  vj =%2d\n",
				  left, right, vj );
                        */
            }
        }
    }

 if( v3 != vc0 ) 
   {                                                                        /*
    <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                                   */
     chngv_straight( -1, -1, v3, vc0 );                                     /*
    <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>                                   */

     vtx[v3].x = 0.0, vtx[v3].y = 0.0;
    }

 vtx[vc0].x = xc0,  vtx[vc0].y = yc0;
     
      /*
      printf("======= the determined central vertex ============\n" );
      printf(" central vertex number =%3d\n", vc0 );
      printf("  x  =%8.3f,   y  =%8.3f\n", vtx[vc0].x, vtx[vc0].y );
      */
}

/*****************************************************************************
                   ============================                             */
                      void put_self(  )
/*                 =============================
******************************************************************************
*( Function )                                                                *
*     To put the self energy and vacuum polarization part                    *
*        in the tree Feynman diagram,  and                                   *
*     to determine their vertex positions.                                   *
*( input )                                                                   *
*     structure vtx[i] : vertex information                                  *
*     structure intline[i] : line information                                *
*( output )                                                                  *
*     The vertex position in the graph                                       *
*         structure vtx[].x and structure vtx[].y                            *
******************************************************************************
*/
{
                                                                            /*
  ----------------------------------
      local variable list
  ----------------------------------
                                                                            */
  int i, j, k, nl, v0, v1, vc, typef;
  int flag, flagf[2];
  float x0, y0, x1, y1, xx, yy, ratio, a, b, d;
                                                                            /*
  ----------------------------------
      function list
  ----------------------------------
                                                                            */
  int  mode( );
  void move_others( );
  void put_straight( );
                                                                            /*
================== Entry point of put_self ================================
                                                                            */
        
      /*
      printf("\n");
      printf("====  self energy part %d ====\n", no_self+1 );
      printf("  edge0 ln[%2d]     v0 l0[%2d]     v1 ln[%2d]  edge1\n", 
               self[no_self].edge_line[0], self[no_self].self_line[0],
               self[no_self].edge_line[1] );
      printf(" v [%2d] ------ v [%2d] ======= v [%2d] ------ v [%2d]\n", 
               self[no_self].edge[0], self[no_self].selfv[0], 
               self[no_self].selfv[1], self[no_self].edge[1] );
      printf("                      l1[%2d]\n", self[no_self].self_line[1] );
      */
                                                                            /*
==============================================================================
   Determine the type of self energy part  
==============================================================================
                                                                            */
  typef  = 0;
  for( j = 0; j < 2; j++ )
     {
       v0 = self[no_self].selfv[j];
       if( v0 != self[no_self].edge[j] )
           typef++;
      }
          /*
           printf(" typef = %d\n", typef );
          */
                                                                            /*
==============================================================================
   Determine the x-y position of self energy vertices
==============================================================================
 +---------------------------------------------------------------------------+
 |                                              \       /~~~~~\       /      |
 |(I) Both edges are isolated                    +-----+       +-----+       |
 |                                              /       \_____/       \      |
 +---------------------------------------------------------------------------+
                                                                            */
  if( typef == 2 )
    {
      ratio     = 0.25;
      v0        = self[no_self].edge[0],   v1  = self[no_self].edge[1];
      x0        = vtx[v0].x,         x1  =  vtx[v1].x;

      d    = x1 - x0;
      if( d < 0.0 ) d = -d;
      if( d < 120.0 )
        {
          d  = 120.0 - d;
          if( x0 > x1 )
            {                                                           /*
             <<<<<<<<<<>>>>>>>>>>>>>                                    */
	      move_others( v1, d );                                     /*
             <<<<<<<<<<>>>>>>>>>>>>>                                    */
             }
          else if( x1 > x0 )
	    {                                                           /*
             <<<<<<<<<<>>>>>>>>>>>>>                                    */
	      move_others( v0, d );                                     /*
             <<<<<<<<<<>>>>>>>>>>>>>                                    */
             }
         }

      x0        = vtx[v0].x,         x1  =  vtx[v1].x;
      y0        = vtx[v0].y,         y1  =  vtx[v1].y;

      xx        = ( x1 - x0 )*ratio + x0;
      yy        = ( y1 - y0 )*ratio + y0;
      vc        = self[no_self].selfv[0];
      vtx[vc].x = xx,                 vtx[vc].y = yy;
                                                                            /*
          <<<<<<<<<<<<<>>>>>>>>>>>>>>>>                                     */
           put_straight( v0, v1, vc );                                      /*
          <<<<<<<<<<<<<>>>>>>>>>>>>>>>>                                     */

      v0        = vc;
      xx        = ( x0 - x1 )*ratio + x1;
      yy        = ( y0 - y1 )*ratio + y1;
      vc        = self[no_self].selfv[1];
      vtx[vc].x = xx,                 vtx[vc].y = yy;
                                                                            /*
          <<<<<<<<<<<<<>>>>>>>>>>>>>>>>                                     */
           put_straight( v0, v1, vc );                                      /*
          <<<<<<<<<<<<<>>>>>>>>>>>>>>>>                                     */
     }
                                                                            /*
 +---------------------------------------------------------------------------+
 |                                                    \ /~~~~~\       /      |
 |(II) One edge has more than 3 legs                   +       +-----+       |
 |                                                    / \_____/       \      |
 +---------------------------------------------------------------------------+
                                                                            */
  else if( typef == 1 )
    {
      ratio     = 0.6;
      if( self[no_self].edge[0] == self[no_self].selfv[0] )
        {
          v0        = self[no_self].edge[0],    v1  = self[no_self].edge[1];
          x0        = vtx[v0].x,          x1  = vtx[v1].x;
          flag      = 0;
         }
      else if( self[no_self].edge[1] == self[no_self].selfv[1] )
        {
          v0        = self[no_self].edge[1],    v1  = self[no_self].edge[0];
          x0        = vtx[v0].x,          x1  = vtx[v1].x;
          flag      = 1;
         }

      d    = x1 - x0;
      if( d < 0.0 ) d = -d;
      if( d < 90.0 )
        {
          d  = 90.0 - d;
          if( x0 > x1 )
            {                                                           /*
                 <<<<<<<<<<>>>>>>>>>>>>>                                    */
	          move_others( v1, d );                                     /*
                 <<<<<<<<<<>>>>>>>>>>>>>                                    */
             }
          else if( x1 > x0 )
	    {                                                           /*
                 <<<<<<<<<<>>>>>>>>>>>>>                                    */
	          move_others( v0, d );                                     /*
                 <<<<<<<<<<>>>>>>>>>>>>>                                    */
             }
         }

      x0        = vtx[v0].x,          x1  = vtx[v1].x;
      y0        = vtx[v0].y,          y1  = vtx[v1].y;
                 /*
	          printf(" x0  = %6.1f    x1  = %6.1f\n", x0, x1 );
	          printf(" y0  = %6.1f    y1  = %6.1f\n", y0, y1 );
                 */
      xx        = ( x1 - x0 )*ratio;
      yy        = ( y1 - y0 )*ratio;
      if( xx > 60.0 )
        {
          xx = 60.0;
          yy = ( y1 - y0 )*xx/(x1 - x0 );
         }
      else if( xx < -60.0 )
        {
          xx = -60.0;
          yy = ( y1 - y0 )*xx/(x1 - x0 );
         }

      xx        = xx + x0;
      yy        = yy + y0;
                 /*
	          printf(" xx  = %6.1f\n", xx );
	          printf(" yy  = %6.1f\n", yy );
                 */
      if( flag == 0 )
          vc        = self[no_self].selfv[1];
      else
          vc        = self[no_self].selfv[0];

      vtx[vc].x = xx,                 vtx[vc].y = yy;
                                                                            /*
          <<<<<<<<<<<<<>>>>>>>>>>>>>>>>                                     */
           put_straight( v0, v1, vc );                                      /*
          <<<<<<<<<<<<<>>>>>>>>>>>>>>>>                                     */
     }
                                                                            /*
==============================================================================
   Determine the x-y position of the middle point
==============================================================================
                                                                            */
   v0 = self[no_self].selfv[0],    v1 = self[no_self].selfv[1];
   x0 = vtx[v0].x,           x1 = vtx[v1].x;
   y0 = vtx[v0].y,           y1 = vtx[v1].y;
   d  = (x1 - x0)*(x1 - x0) + (y1 -y0)*(y1-y0);
   d  = sqrt( d )*0.25;
           /*
            printf(" x0  = %6.1f    x1  = %6.1f\n", x0, x1 );
	    printf(" y0  = %6.1f    y1  = %6.1f\n", y0, y1 );
	    printf(" d   = %6.1f\n", d );
           */

                                                                            /*
 +---------------------------------------------------------------------------+
 |(1) count the number of fermion lines                                      |
 +---------------------------------------------------------------------------+
                                                                            */
   flag = 0;
   for( j = 0; j < 2; j++ )
      {
        flagf[j] = 0;
        nl = self[no_self].self_line[j];
        k  = intline[nl].direct;
        if( k == 1 || k == -1 )
          {
            flagf[j] = 1;
            flag++;
           }
       }
                                                                            /*
 +---------------------------------------------------------------------------+
 |(2) Self energy part runs vertically                                       |
 +---------------------------------------------------------------------------+
                                                                            */
   if( y0 == y1 )
     {
       xx = 0.5*(x1 -x0) + x0;
       if( flag == 1 )
         {
           if( flagf[0] == 1 )
             {
               nl = self[no_self].self_line[1];
               intline[nl].x = xx;
               intline[nl].y = y0 - d;
               intline[nl].flag = 1;
              }
           else
             {
               nl = self[no_self].self_line[0];
               intline[nl].x = xx;
               intline[nl].y = y0 - d;
               intline[nl].flag = 1;
              }
          }
       else
         {
           nl = self[no_self].self_line[0];
           intline[nl].x = xx;
           intline[nl].y = y0 + d;
           nl = self[no_self].self_line[1];
           intline[nl].x = xx;
           intline[nl].y = y0 - d;
          }
      }
   else
                                                                            /*
 +---------------------------------------------------------------------------+
 |(3) Self energy part runs horizontally                                     |
 +---------------------------------------------------------------------------+
                                                                            */
   if( x0 == x1 )
     {
       yy = 0.5*(y1 -y0) + y0;
       if( flag == 1 )
         {
           if( flagf[0] == 1 )
             {
               nl = self[no_self].self_line[1];
               intline[nl].x = x0 - d;
               intline[nl].y = yy;
               intline[nl].flag = 1;
              }
           else
             {
               nl = self[no_self].self_line[0];
               intline[nl].x = x0 - d;
               intline[nl].y = yy;
               intline[nl].flag = 1;
              }
          }
       else
         {
           nl = self[no_self].self_line[0];
           intline[nl].x = x0 + d;
           intline[nl].y = yy;
           nl = self[no_self].self_line[1];
           intline[nl].x = x0 - d;
           intline[nl].y = yy;
          }
      }
   else
                                                                            /*
 +---------------------------------------------------------------------------+
 |(4) Self energy part runs any other direction                              |
 +---------------------------------------------------------------------------+
                                                                            */
     {
       a     = (x0 - x1)/(y1 - y0);
       x0    = 0.5*(x1 - x0) + x0;
       y0    = 0.5*(y1 - y0) + y0;
       b     = y0 - a*x0;
       ratio = a*a + 1.0;
       ratio = sqrt( ratio );
       ratio = d/ratio;

       xx           = x0 + ratio;
       yy           = a*xx + b;

       if( flag == 1 )
         {
           if( flagf[0] == 1 )
             {
               j            = self[no_self].self_line[1];
               intline[j].x = xx;
               intline[j].y = yy;
               intline[j].flag = 1;
              }
           else
             {
               j            = self[no_self].self_line[0];
               intline[j].x = xx;
               intline[j].y = yy;
               intline[j].flag = 1;
              }
          }
       else
         {
               j            = self[no_self].self_line[0];
               intline[j].x = xx;
               intline[j].y = yy;
                   /*
                    printf(" j   = %d\n", j );
                    printf(" xx  = %6.1f\n", xx );
                    printf(" yy  = %6.1f\n", yy );
                   */
               xx           = x0 - ratio;
               yy           = a*xx + b;
               j            = self[no_self].self_line[1];
               intline[j].x = xx;
               intline[j].y = yy;
                   /*
                    printf(" j   = %d\n", j );
                    printf(" xx  = %6.1f\n", xx );
                    printf(" yy  = %6.1f\n", yy );
                   */
          }
      }

}

