OPTC = $(COPT)

EXEC=	grc

OGRC=   alist.o assign.o elem.o gnode.o scalar.o iline.o lclass.o \
	print.o outf.o ergrf.o biconn.o flgrf.o matchv.o perm.o ilgrf.o \
	divide.o args.o path.o \
	rdptbl.o token.o grc.o rdprocs.o

SGRC=   $(OGRC:.o=.c) 
SDRW=   $(ODRW:.o=.c)
SFRT=   $(OFRT:.o=.c)
SFRM=   $(OFRM:.o=.c)
SRED=   $(ORED:.o=.c)
SPLT=   $(OPLT:.o=.c)

#---------------------------------------------------------------

.SUFFIXES: .c .o .hd

.c.o:
	$(CC) $(OPTC) -c $*.c

.c.hd:
	grep 'Public' $*.c | grep 'ARG' | sed 's/^.*://' > $*.hd

#===============================================================

all: $(EXEC)

${EXEC}: $(OGRC)
	$(CC) $(OPTC) -o ${EXEC} $(OGRC)

mgrcreda:
	make -f grcreda.mak

install:all
	cp $(EXEC) $(BINDIR)

version: mkver
	./mkver

mkver: mkver.c
	$(CC) -o mkver mkver.c

clean: 
	rm -f *.o $(EXEC) core tmp* junk*
#===============================================================

lint: lgrc

lgrc:
	lint $(SGRC)

ldrw:
	lint $(XOPTI) $(SDRW)

lfrt:
	lint $(SFRT)

lfrm:
	lint $(SFRM)

lred:
	lred $(SRED)

lplt:
	lint $(OPTI) $(SPLT)

#================================================================

HGRCH = grch.h common.h dbgh.h ptable.h ptable.h
GRDT = grdata.h ptable.h
HVRH  = vrh.h common.h

dcx.o: dcx.c common.h dbgh.h dcx.h vrcolor.h $(HVRH)
	$(CC) $(OPTC) $(XCOPT) $(XOPTI) -c dcx.c

vrx.o: vrx.c common.h dcx.h vrcolor.h $(HVRH) vrxh.h
	$(CC) $(OPTC) $(XCOPT) $(XOPTI) -c vrx.c

alist.o: alist.c $(HGRCH)

args.o: args.c args.h common.h version.h

assign.o: assign.c $(HGRCH)

biconn.o: biconn.c $(HGRCH)

cfgrf.o: cfgrf.c common.h dbgh.h glue4.h $(GRDT) ptable.h rational.h

cmgrf.o: cmgrf.c common.h config.h dbgh.h gensc.h $(GRDT) ptable.h

cmfin.o: cmfin.c common.h config.h dbgh.h gensc.h $(GRDT) ptable.h putidh.h

dagrf.o: dagrf.c common.h dbgh.h $(GRDT) ptable.h

divide.o: divide.c common.h

draw.o: draw.c args.h common.h config.h dbgh.h $(GRDT) ptable.h tokenh.h \
	version.h vrcolor.h $(HVRH)

drgrf.o: drgrf.c common.h dbgh.h $(GRDT) ptable.h vrcolor.h windowh.h

elem.o: elem.c $(HGRCH)

ergrf.o: ergrf.c common.h dbgh.h

flgrf.o: flgrf.c common.h dbgh.h $(GRDT) ptable.h

formgrf.o: formgrf.c common.h config.h dbgh.h gensc.h $(GRDT) ptable.h putidh.h

ftgrf.o: ftgrf.c common.h config.h dbgh.h gensc.h glue4.h $(GRDT) ptable.h putidh.h

genform.o: genform.c args.h common.h config.h dbgh.h gensc.h glue4.h $(GRDT) \
	ptable.h tokenh.h version.h

genfort.o: genfort.c args.h common.h config.h dbgh.h gensc.h glue4.h $(GRDT) \
	ptable.h tokenh.h version.h

genred.o: genred.c args.h common.h config.h dbgh.h gensc.h glue4.h $(GRDT) \
	ptable.h tokenh.h version.h

gnode.o: gnode.c $(HGRCH)

grc.o: grc.c args.h config.h $(HGRCH) tokenh.h version.h

ilgrf.o: ilgrf.c common.h dbgh.h $(GRDT) ptable.h

iline.o: iline.c $(HGRCH)

lclass.o: lclass.c $(HGRCH)

matchv.o: matchv.c common.h dbgh.h $(GRDT) ptable.h tokenh.h

mkver.o: mkver.c common.h

outf.o: outf.c $(HGRCH)

path.o: path.c common.h

perm.o: perm.c common.h

plot.o: plot.c args.h common.h config.h dbgh.h $(GRDT) ptable.h tokenh.h \
	version.h vrcolor.h $(HVRH) windowh.h


pread.o: pread.c args.h common.h config.h dbgh.h $(GRDT) ptable.h tokenh.h \
	version.h

print.o: print.c $(HGRCH)

putid.o: putid.c common.h putidh.h

rational.o: rational.c common.h rational.h

rdgrf.o: rdgrf.c common.h dbgh.h $(GRDT) ptable.h tokenh.h

rdprocs.o: rdprocs.c $(HGRCH) tokenh.h

rdptbl.o: rdptbl.c common.h config.h dbgh.h $(GRDT) ptable.h tokenh.h

redgrf.o: redgrf.c common.h config.h dbgh.h gensc.h $(GRDT) ptable.h putidh.h

scalar.o: scalar.c $(HGRCH)

tlgrf.o: tlgrf.c common.h dbgh.h $(GRDT) ptable.h

token.o: token.c common.h dbgh.h escchar.h tokenh.h

tread.o: tread.c common.h config.h dbgh.h $(GRDT) ptable.h tokenh.h

vr.o: vr.c common.h vrcolor.h vrfh.h $(HVRH) vrxh.h

vrplot.o: vrplot.c common.h vrcolor.h vrfh.h $(HVRH)

vrps.o: vrps.c common.h vrcolor.h vrfh.h $(HVRH)

vrxfig.o: vrxfig.c common.h vrcolor.h $(HVRH) vrxh.h

window.o: window.c common.h vrcolor.h $(HVRH) windowh.h

