/*******************************************************************
 *    File : args.c
 *
 *    Analyze arguments
 *                                                 by Kaneko Toshiaki
 *
 * sample program
 *
 * #include <...>
 * 
 * Local int    a1 = 0;
 * Local int    a2 = 0;
 * Local int    a3 = 0;
 * Local int    i1 = 0;
 * Local int    i2 = 0;
 * Local string s1 = NULL;
 * Local string s2 = NULL;
 * 
 * Local topt options[] = {
 *     {"a", ARFLAG, (int*) &a1},
 *     {"b", ARFLAG, (int*) &a2},
 *     {"c", ARFLAG, (int*) &a3},
 *     {"i", ARINT,  (int*) &i1},
 *     {"j", ARINT,  (int*) &i2},
 *     {"s", ARSTR,  (int*) &s1},
 *     {"t", ARSTR,  (int*) &s2},
 * };
 * 
 * Local int noptions = sizeof(options)/sizeof(topt);
 * 
 * void main(argc, argv)
 * int    argc;
 * char **argv;
 * {
 *     int i;
 *     int pos;
 *     string str;
 * 
 *     printf("\tatgc = %d\n", argc);
 *     for(i = 0; i < argc; i++)
 *         printf("\t%d\t%s\n", i, argv[i]);
 * 
 *     argc = aropt(argc, argv, noptions, options);
 * 
 *     ardump(noptions, options);
 *     
 *     printf("\tatgc = \t%d\n", argc);
 *     for(i = 0; i < argc; i++)
 *         printf("\t%d\t%s\n", i, argv[i]);
 * 
 *     while(True) {
 *         pos = 1;
 *         if((str = arnext(argc, argv, &pos)) == NULL)
 *             break;
 *         pos--;
 *         argc = ardel(argc, argv, pos, 1);
 *         printf("\tpos = \t%d, \tstr = \t%s\n", pos, str);
 *     }
 * 
 *     printf("\tatgc = \t%d\n", argc);
 *     for(i = 0; i < argc; i++)
 *         printf("\t%d\t%s\n", i, argv[i]);
 * 
 * }
 * 
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "common.h"
#include "args.h"
#include "version.h"

/* Entry points */
/*     see args.h */

/* External references */

/* Private functions */

/* Local variable */

/*==================================================================*/
Public  int     aropt(argc, argv, nopt, opt)
int     argc;
char    **argv;
int     nopt;
topt    opt[];
/*--------------------------------------------------------------------
 */
{
    int     argp, c, b, i;
    string *sval;
    bool    ok;

    argp = 1;
    for(argp = 1; argp < argc; ) {
        ok = False;
        if((b = argv[argp][0]) == '-' || b == '+') {
            c = argv[argp][1];
            for(i = 0; i < nopt; i++) {
                if(c == opt[i].sw[0]) {
                    switch (opt[i].atype) {
                    case ARFLAG :
                        arflag(argv[argp], nopt, opt);
                        argc = ardel(argc, argv, argp, 1);
                        ok   = True;
                        break;
                    case ARINT :
                        if(b == '+')
                            continue;
                        if(strcmp(argv[argp]+1, opt[i].sw) != 0)
                            break;
                        if(argc < argp - 2)
                            break;
                        *(opt[i].val) = atoi(argv[argp+1]);
                        argc = ardel(argc, argv, argp, 2);
                        ok   = True;
                        break;
                    case ARSTR :
                        if(b == '+')
                            continue;
                        if(strcmp(argv[argp]+1, opt[i].sw) != 0)
                            break;
                        if(argc < argp - 2)
                            break;
                        sval = ((string*)(opt[i].val));
                        *sval = argv[argp+1];
                        argc = ardel(argc, argv, argp, 2);
                        ok   = True;
                        break;
                    }
                }
                if(ok)
                    continue;
            }
        }
        if(!ok)
            argp++;
    }
    return argc;
}

/*==================================================================*/
Public  Void    arflag(arg, nopt, opt)
char    *arg;
int     nopt;
topt    opt[];
/*--------------------------------------------------------------------
 */
{
    int len = strlen(arg);
    int h, i, c, j;
    int *v;
    bool ok;

    if(arg[0] == '-')
        h = ARMINUS;
    else if(arg[0] == '+')
        h = ARPLUS;
    else
        h = ARNUL;

    for(i = 1; i < len; i++) {
        c = arg[i];
        ok = False;
        for(j = 0; j < nopt; j++) {
            if(c == opt[j].sw[0] && opt[j].atype == ARFLAG) {
                *(opt[j].val) = h;
                ok = True;
                break;
            }
        }
        if(!ok)
            fprintf(Stderr, "*** unknown option flag -%c\n", c);
    }
}

/*==================================================================*/
Public  int     ardel(argc, argv, argp, n)
int     argc;
char    **argv;
int     n;
/*--------------------------------------------------------------------
 */
{
    int i;

    for(i = argp; i < argc - n; i++)
        argv[i] = argv[i+n];
    return argc - n;
}

/*==================================================================*/
Public  string  arnext(argc, argv, pos)
int     argc;
char    **argv;
int     *pos;
/*--------------------------------------------------------------------
 */
{
    int c;

    for( ; *pos < argc; (*pos)++) {
        if((c = argv[*pos][0]) != '-' && c != '+')
            return argv[(*pos)++];
    }

    return NULL;
}

/*==================================================================*/
Public  Void    ardump(nopt, opt)
int     nopt;
topt    opt[];
/*--------------------------------------------------------------------
 */
{
    int i;
    string s;

    for(i = 0; i < nopt; i++) {
        printf("%d\t%s", i, opt[i].sw);
        switch (opt[i].atype) {
        case ARFLAG:
            printf("\tFlag\t%d\n", *(opt[i].val));
            break;
        case ARINT:
            printf("\tInt\t%d\n", *(opt[i].val));
            break;
        case ARSTR:
            if((s = *((string*)(opt[i].val))) == NULL)
                printf("\tStr\tNULL\n");
            else
                printf("\tStr\t%s\n", s);
            break;
        default:
            printf("\t???\n");
        }
    }
}

/*==================================================================*/
Public  Void   grcmsg(str)
string  str;
/*--------------------------------------------------------------------
*/
{
    printf("***************************************************\n");
    printf("*                                                 *\n");
    printf("*   grc                                           *\n");
    printf("*                                                 *\n");
    printf("*   %-32s              *\n", str);
    printf("*                                                 *\n");
    if(GRCVER2 == 0) {
        printf("*                                     Ver.%2d.%02d   *\n",
               GRCVER0, GRCVER1);
    } else {
        printf("*                                   Ver.%2d.%02d(%02d) *\n",
               GRCVER0, GRCVER1, GRCVER2);
    }
    printf("*               %+32s  *\n", GRCDATE);
    printf("*                                                 *\n");
    if(strncmp(str, "Feynman graph ",14) == 0) {
        printf("*                              KANEKO Toshiaki    *\n");
        printf("*                           kaneko@minami.kek.jp  *\n");
    } else {
        printf("*                               Minami-Tateya     *\n");
        printf("*                            grace@minami.kek.jp  *\n");
    }
    printf("*                                                 *\n");
    printf("***************************************************\n");
}

