/*********************************************************************
 *  File : gnode.c
 *
 *  Generate vertices.
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "grch.h"

/* Entry points */
Public  Void gvertex         ARG((NoArg));
/*           Generate vertices
*/
Public  int  gvcount         ARG((NoArg));
/*           Returns the number of configurations of nodes.
*/

/* External refernces */
/* elem.c */
Extern  Void    elgenv       ARG((int, int, int));
Extern  Void    eldelv       ARG((int, int, int));
/* ergrf.c */
Public  Void    ercont       ARG((NoArg));
Public  Void    erenter      ARG((string));
Public  Void    erfail       ARG((NoArg));
Public  Void    ermon        ARG((string, string));
Public  Void    erreturn     ARG((string));
Public  Void    error        ARG((string, string));
/* scalar.c */
Extern  Void    gscalar      ARG((NoArg));
/* print.c */
Extern  Void    prdump       ARG((NoArg));
Extern  Void    prtype       ARG((int));

/* Local functions */
Private Void gvtree          ARG((int, int));
Private Void gvloop          ARG((int, int, int, int, int));

/*
 * #define MAXTF     (MAXTL - 2)
 * #define MINTF     (MINTL - 2)
 */

static int  gvnc;
/*          The number of configurations
*/
static int  gvflp;
/*          = for looped_vertices sum (legs - 2).
*/
static int  gvlin;
/*          The total number of loops of looped vertices.
*/
static int  gvminlg;
/*          The minimum possible value of looped vertices.
*/

static int  maxtf, mintf;

/*********************************************************************
   Module of graph generation

   What should be input?
     flags ?
*/

/*==================================================================*/
Public Void gvertex(NoArg)
/*--------------------------------------------------------------------
    Entry of the module of generating vertices.
--------------------------------------------------------------------*/
{
    int l, lin;
    int f, fmin, fmax, ftot, gvlinl, gvlinh;

#if DBG
    erenter("gvertex");
#endif

    maxtf = maxtl - 2;
    mintf = mintl - 2;
#if DBG
    printf("maxtf = %d, mintf = %d\n", maxtf, mintf);
#endif

    /* minimum number of legs of looped vertices */
    gvminlg   = (tadpole) ? 1 : 2;

    /* expand loops ? */
    if(countert) {
        gvlinl = 0;
        gvlinh = loop;
    } else if(expand) {
        gvlinl = 0;
        gvlinh = 0;
    } else {
        /* generate looped vertex with "gvlin" loops */
        gvlinl = loop;
        gvlinh = loop;
    }

    for(lin = gvlinl; lin <= gvlinh; lin++) {
        gvlin = lin;
        if(loop == 0 && nextern < mintl)
            return;
    
        /*----------------------------------
           inputs to the following part
    
               loop
               nextern
               gvminlg
               gvlin
        */
    
        /* count the number configurations */
        gvnc = 0;
    
        /* total number of sum(legs - 2) */
        ftot = nextern + 2*(loop - gvlin - 1);
    
        /* determine sum(legs - 2) for tree and looped vertices */
        fmin   = (gvlin > 0 && gvminlg == 1) ? -gvlin : 0;
        fmax   = (gvlin > 0)                 ? ftot   : 0;
        for (f = fmax; f >= fmin; f--) {
            gvflp = f;
            gvtree(maxtf, ftot - f);
        }
    
#if DBG
        if(check) {
            l = gvcount();
            if(l != gvnc) {
                error("gnvertex", "different number of confs ");
                fprintf(Stderr, "ex=%d, loop=%d, gen=%d != %d\n",
                                nextern, loop, gvnc, l);
                erfail();
            } else if(display) {
                printf("gnvertex : O.K.\n");
            }
        }
#endif
    }

#if DBG
    erreturn("gvertex");
#endif
}

/*==================================================================*/
Private Void gvtree(ft, fleft)
int ft;
int fleft;
/*--------------------------------------------------------------------
    Generate tree vertices.

    Arguments:
      ft    : input : maximum number of legs for the next vertex.
      fleft : input : sum of number of legs for the rest of
                      tree vertics.
--------------------------------------------------------------------*/
{
    int i, fl;
    int minv, v;

#if DBG
    erenter("gvtree");
#endif

    /* the last call */
    if(ft == mintf || fleft == 0) {
        fl = fleft/ft;
        /* generate rest of tree vertices */
        if(fl > 0) {
            elgenv(mintl, 0, fl);
        }

        /* if no looped vertices then
           generation of vertices is completed.
        */
        if(gvlin < 1) {
            gvnc++;
#if DBG
/*
            if(trace) {
                printf("gvtree : \n");
                prdump();
            }
 */
#endif
            if(skipasp)
                prtype(1);
            gscalar();
#if DBG
            if(trace) {
                printf("end of gscalar : \n");
                prdump();
            }
#endif

        /* generate looped vertices */
        } else {
            /* if sum(legs - 2) < 0 then generate at least
               (2 - legs) tadpoles
            */
            minv = max(1, -gvflp);

            /* determine the number of looped vertices */
            for(v = minv; v <= gvlin; v++) {
/*
                gvloop(gvlin - v, v + gvflp, v,
                      gvlin - v, v*(2 - gvminlg) + gvflp);
*/
                gvloop(gvlin - v, v*(2 - gvminlg) + gvflp, v,
                       gvlin - v, v*(2 - gvminlg) + gvflp);
                }
        }

        /* restore before return */
        if(fl > 0) {
            eldelv(mintl, 0, fl);
        }

    } else if(fleft < 0) {
        error("gvtree", "fleft < 0");
        fprintf(Stderr," fleft = %d\n", fleft);
        ercont();

    /* more tree vertices. */
    } else {
        /* case of no generation for the current value of "ft" */
        gvtree(ft - 1, fleft);

        /* generate for the current value of "ft" varying the number
           of vertices.
        */
        for(i = 1; i <= ((int) fleft / ft); i++) {
            elgenv(ft + 2, 0, i);     /* may be i == 0 */
            gvtree(ft - 1, fleft - i*ft);
            eldelv(ft + 2, 0, i);
        }
    }
#if DBG
    erreturn("gvtree");
#endif
}

/*==================================================================*/
Private Void gvloop(mleft, eleft, vleft, mlast, elast)
int mleft;
int eleft;
int vleft;
int mlast;
int elast;
/*--------------------------------------------------------------------
    Generate looped vertices in discending order of the pair
    (number of legs, number of loops)

    Arguments:
      mleft : input : the number of loops* left, where
                      loops* = for vertices sum (loops - 1).
      eleft : input : the number of legs* left, where
                      legs* = for vertices sum (legs - gvminlg).
      vleft : input : the number of vertices left to generate.
      mlast : input : the maximum possible number of loops* to be
                      allocated for the next vertex.
      elast : input : the maximum possible number of legs* to be
                      allocated for the next vertex.

    The should satisfy the following relations:
      elast*vleft >= eleft >= elast >= 0
      mleft >= mlast >= 0
      mlast*vleft >= mleft,   if elast*vleft = eleft.
      vleft >= 1
--------------------------------------------------------------------*/
{
    int l, en, mn, el, elmin, hl, ll;

#if DBG
    erenter("gvloop");
#endif

#if DBG
    if(eleft < 0 || mleft < 0 || vleft < 0) {
        error("gvloop", "illegal value of eleft, mleft, vleft");
        fprintf(Stderr, " melft = %d >= 0",
                        mleft);
        fprintf(Stderr, " eleft = %d >= 0, vleft = %d >= 0\n",
                        eleft, vleft);
        erfail();
    } else if(eleft < elast || mleft < mlast) {
        error("gvloop", "illegal value of eleft, mleft, vleft");
        fprintf(Stderr, " melft = %d >= mlast = %d",
                        mleft, mlast);
        fprintf(Stderr, " eleft = %d >= elast = %d\n",
                        eleft, elast);
        erfail();
    } else
#endif

    /* the last vertex */
    if (vleft == 1) {
#if DBG
        if(mleft == mlast && eleft == elast) {
#endif
            elgenv(eleft+gvminlg, mleft+1, 1);

            /* vertex generation is completed. */
            gvnc++;

#if DBG
            if(trace) {
                printf("gvloop : \n");
                prdump();
            }
#endif
            if(skipasp)
                prtype(1);
            gscalar();

#if DBG
            if(trace) {
                printf("end of gscalar : \n");
                prdump();
            }
#endif

            eldelv(eleft+gvminlg, mleft+1, 1);
#if DBG
        } else if(mon) {
            ermon("gvloop", "fail-2");
            fprintf(Stderr, " mleft=%4d eleft=%4d",
                            mleft, eleft);
            fprintf(Stderr, " vleft=%4d mlast=%4d elast=%4d\n",
                            vleft, mlast, elast);
        }
#endif
    } else {
        hl = mlast;
        /* = mleft,  for the next value of el */
        elmin = ((int) (eleft+vleft-1)/vleft);
        for(el = elast; el >= elmin; el--) {
            en = min(eleft-el, el);
            /* = the next value of "eleft" */

            /* if(el = the value for the last vertex) then
                 minimum number of value "ll" is limitted by
                   ll * vleft = mleft
            */
            if(el*vleft == eleft) {
                ll = ((int) (mleft+vleft-1)/vleft);
            } else {
                ll =  0;
            }

            for(l = hl; l >= ll; l--) {
                /* the next value of mleft */
                mn = min(mleft-l, l);
                /* if the "en" is changed, ordering in loops is
                   reset.
                */
                if(en != el)
                    mn = mleft - l;
                elgenv(el+gvminlg, l+1, 1);
                /* next vertex generation */
                gvloop(mleft-l, eleft-el, vleft-1, mn, en);
                eldelv(el+gvminlg, l+1, 1);
            }
            hl = mleft;
        }
    }

#if DBG
    erreturn("gvloop");
#endif
}


/*==================================================================*/
Public int gvcount(NoArg)
/*--------------------------------------------------------------------
    Returns : the number of configurations of generated nodes.
--------------------------------------------------------------------*/
{
    return gvnc;
}
