/*********************************************************************
 *  File : grc.c
 *
 *  Entry for graph generation
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <stdio.h>

#define  DEFGLOBAL
#define  MAIN_GRC
#include "grch.h"
#include "config.h"
#include "tokenh.h"
#include "args.h"
#include "version.h"

/* Entry points */
Public  int     main         ARG((int, char**));

/* External refernces */
/* outf.c */
Extern  Void    foclose      ARG((NoArg));
Extern  bool    foinit       ARG((string));
Extern  Void    fopend       ARG((NoArg));
Extern  Void    foproces     ARG((NoArg));
/* gnode.c */
Extern  Void    gvertex      ARG((NoArg));
/* assign.c */
Extern  Void    padump       ARG((NoArg));
/* print.c */
Extern  Void    prfirst      ARG((NoArg));
Extern  Void    prinit       ARG((NoArg));
Extern  Void    prlast       ARG((NoArg));
Extern  Void    prtype       ARG((int));
/* rdptbl.c */
Extern  Void    pvdump       ARG((NoArg));
Extern  Void    pvhelp       ARG((FILE*));
Extern  Void    pvprint      ARG((string));
/* rdprocs.c */
Extern  Void    rphelp       ARG((FILE*));
Extern  bool    rpinit       ARG((string));
Extern  int     rpread       ARG((NoArg));
Extern  Void    rprin1       ARG((FILE*, int, bool));
Extern  Void    rprint       ARG((string));

/* Local functions */
Private int     grcarg       ARG((int, char**));
Private Void    gropth       ARG((string));

/* Local variables */
Local   int     nprc = 0;
Local   char    finame[] = "in.prc";
Local   char    foname[] = "out.grf";

Local   int     dskasp  = 0;
Local   int     doutf   = 0;
Local   int     dtype   = 0;
Local   int     dtone   = 0;
Local   int     dgone   = 0;
Local   int     dhlp    = 0;
#if DBG
Local   int     dprt    = 0;
Local   int     ddbg    = 0;
Local   int     dmon    = 0;
Local   int     dtrc    = 0;
Local   int     dsel    = -1;
Local   int     dcpt    = 0;
#endif

Local   topt    options[] = {
    {"o", ARSTR,  (int*) &fngrf},
    {"a", ARFLAG, (int*) &dskasp},
    {"c", ARFLAG, (int*) &doutf},
    {"g", ARINT,  (int*) &dgone},
    {"t", ARINT,  (int*) &dtype},
    {"T", ARFLAG, (int*) &dtone},
    {"h", ARFLAG, (int*) &dhlp},
    {"?", ARFLAG, (int*) &dhlp},
#if DBG
    {"p", ARINT,  (int*) &dprt},
    {"d", ARINT,  (int*) &ddbg},
    {"m", ARFLAG, (int*) &dmon},
    {"r", ARFLAG, (int*) &dtrc},
    {"s", ARINT,  (int*) &dsel},
    {"i", ARINT,  (int*) &dcpt},
#endif
};

Local int noptions = sizeof(options)/sizeof(topt);

/*==================================================================*/
Public  int    main(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
    Main routine for graph drawin
--------------------------------------------------------------------*/
{

    grcmsg("Feynman graph generator");          /* First message */

    argc = grcarg(argc, argv);   /* analyze arguments */
    if(argc != 1) {
        gropth(argv[0]);
        exit(1);
    }

#if DBG
    /* debugging flags for modules.
       if(dbg >= xxdbg) then enable debugging
    */
    aldbg = 2;  /* debug flag for alist.c */
    bidbg = 2;  /* debug flag for biconn.c */
    eldbg = 2;  /* debug flag for elem.c */
    fldbg = 2;  /* debug flag for fline.c */
    fodbg = 2;  /* debug flag for out.c */
    gcdbg = 1;  /* debug flag for grc.c */
    gndbg = 2;  /* debug flag for gnode.c */
    gsdbg = 2;  /* debug flag for scalar.c */
    ildbg = 2;  /* debug flag for iline.c */
    lcdbg = 2;  /* debug flag for lclass.c */
    padbg = 2;  /* debug flag for assign.c */
    pmdbg = 2;  /* debug flag for matchv.c */
    prdbg = 2;  /* debug flag for print.c */
    pvdbg = 2;  /* debug flag for rdptbl.c */
    rddbg = 2;  /* debug flag for token.c */
    rgdbg = 2;  /* debug flag for rdgraph.c */
    rpdbg = 2;  /* debug flag for rdproces.c */
    sfdbg = 2;  /* debug flag for sfgrf.c */
#endif

    testfv = False;

    /* read process */
    if(!rpinit(fnprc))
        exit(16);

    /* open output file of graph structure */
    if(outf) {
        Discard foinit(fngrf);
    }

#if DBG
    if(dbgst > 0) {
        pvprint("out.mdl");  /* normalized model file */
        rprint("out.prc");  /* normalized process file */
        if(!outf)
            Discard foinit(fngrf);   /* open output file of graph structure */
    }
    if(dbg >= gcdbg)
        pvdump();  /* normalized model file */

#endif

    prinit();           /* initialize print routine */

    /* generate graphs for processes */
    nprocess = 0;


    /* multi processes */
    while((nprc= rpread()) > 0) {
#if DBG
        if(dcpt != 0)
            cptest = dcpt;
#endif
        nprocess++;
/*
        duptest = (loop > 5);
 */
        duptest = True;
        sfnum   = 0;
        sfden   = 1;

#if DBG
        if(dbg >= gcdbg)
            rprin1(stdout, nprc-1, True);
#endif

        prfirst();        /* print header */
        if(outf)
            foproces();

        prtype(-1);       /* initialize type counter */

#if DBG
        if(dbg >= gcdbg)
            padump();
#endif

        gvertex();        /* gnerate graph */

        prtype(0);        /* print the last type */

        if(outf)          /* print ending message */
            fopend();     /* end of graph structure output */
        prlast();         /* end of printing */
    }
    if(outf)
        foclose();        /* close file of graph structure */

    return 0;
}

/*==================================================================*/
Private int    grcarg(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
*/
{
    string s;
    int    pos;
    string cmd;

    cmd     = argv[0];
    skipasp = False;
    outf    = True;
    fngrf   = NULL;
    display = False;

#if DBG
    dispno  = -1;
    ddbg    = 0;
    dprt    = 0;
    trace   = False;
    mon     = False;
#endif

    argc = aropt(argc, argv, noptions, options);

    if(dhlp != 0) {
        gropth(cmd);
        exit(0);
    }

    if(fngrf == NULL)
        fngrf = foname;

    if(dskasp != 0) {
        skipasp = True;
    }

    switch (doutf) {
        case ARPLUS:
            outf = True;
            break;
        case ARMINUS:
            outf = False;
            break;
    }

    seltype = dtype;

    switch (dtone) {
        case ARPLUS:
            onetype = True;
            break;
        case ARMINUS:
            onetype = True;
            break;
    }

    onegraph = dgone;

    /* input process file name */
    pos = 1;
    if((fnprc = arnext(argc, argv, &pos)) != NULL) {
        pos--;
        argc = ardel(argc, argv, pos, 1);
    } else {
        fnprc = finame;
    }

#if DBG
    /* debugging flags */
    dbg   = 0;
    dbgst = -1;
    selprop  = 0;

    if(ddbg > 0) {
        dbg   = ddbg;
    }
    if(ddbg < 0) {
        dbgst = -ddbg;
    }

    if(dprt > 0) {
        display = dprt;
        dispno  = -1;
    } else {
        display = False;
        dispno  = -dprt;
    }

    if(dtrc != 0) {
        trace = True;
    }

    if(dmon != 0) {
        mon = True;
    }

    if(dsel > 0) {
        selprop = dsel;
    } else {
        selprop = 0;
    }
#endif

    return argc;
}

/*==================================================================*/
Private Void   gropth(cmd)
string  cmd;
/*--------------------------------------------------------------------
*/
{
    printf("Syntax %s [options] [Input_File_name]\n", cmd);
    printf("\nOption\n");
    printf("    -a                  ");
    printf(": skip particle assignment\n");
    printf("    -l int              ");
    printf(": with detailed message \n");
    printf("    -o Output_File_name ");
    printf(": specify output file\n");
    printf("    -t int              ");
    printf(": output for specified type\n");
    printf("    -T                  ");
    printf(": output one graph from each type\n");
    printf("    -c                  ");
    printf(": only count\n");
#if DBG
    printf("    -p int              ");
    printf(": print level\n");
    printf("    -d int              ");
    printf(": debug level\n");
    printf("    -m                  ");
    printf(": enable monitor\n");
    printf("    -i int              ");
    printf(": cptest flag\n");
    printf("    -r                  ");
    printf(": enable trace\n");
#endif
    printf("    -h                  ");
    printf(": print this message\n");
    printf("\nDefault file names:\n");
    printf("     Input_File_name = %s\n", finame);
    printf("    Output_File_name = %s\n", foname);
}
