/*********************************************************************
 *  File : iline.c
 *
 *  Construct table of internal lines.
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "grch.h"

/* Entry points */
Public  Void    ilsetup      ARG((NoArg));
Public  Void    illdump      ARG((NoArg));

/* External references */
/* ergrf.c */
Extern  Void    erdbg        ARG((string, string, int));
Extern  Void    erfail       ARG((NoArg));
Extern  Void    error        ARG((string, string));

/* local functions */
Private Void    ilinit       ARG((NoArg));
Private Void    ilend        ARG((NoArg));
Private Void    iladd        ARG((node,node,int));

/* local variables */

/*==================================================================*/
Private Void    ilinit(NoArg)
/*--------------------------------------------------------------------
-------------------------------------------------------------------*/
{
    nilines = 0;
}
/*==================================================================*/
Private Void    ilend(NoArg)
/*--------------------------------------------------------------------
-------------------------------------------------------------------*/
{
    cgraph->grid    = nconfig;
    cgraph->gtype   = gntype;
    cgraph->nnodes  = nnodes;
    cgraph->nilines = nilines;
}

/*==================================================================*/
Private Void    iladd(n1, n2, ng)
node    n1;
node    n2;
int     ng;
/*--------------------------------------------------------------------
 *    Connect n1 and n1 with ng legs.
 *    If n1 = n2 then ng represents twice of legs.
 *    If they are already connected, nothing is changed.
-------------------------------------------------------------------*/
{
    int i1, i2, lg1, lg2, i, ngs;

    if(ng < 1)
        return;

#if DBG
    if(dbg >= ildbg)
        printf("iladd(%d,%d,%d)\n", n1, n2, ng);
#endif

    /* ngs = required number of ilines */
    if(n1 == n2) {
        if(ng % 2 != 0) {
            error("iladd", "# connection should be even for self-loop");
            fprintf(Stderr, "node = %d,  connection = %d\n", n1, ng);
            erfail();
        }
        ngs = ng / 2;
    } else {
        ngs = ng;
    }

    /* i1, i2 = the position of iline to be connected */
    lg1 = 0;
    for(i = 0; i < ndlegs(n1); i++) {
        if(ndiline(n1, i) < 0) {
            i1 = i;
            break;
        } else if(ndnextn(n1, i) == n2)
            lg1++;
    }

    if(n1 != n2) {
        lg2 = 0;
        for(i = 0; i < ndlegs(n2); i++) {
            if(ndiline(n2, i) < 0) {
                i2 = i;
                break;
            } else if(ndnextn(n2, i) == n2)
                lg2++;
        }
    } else {
        lg2 = lg1;
    }

    if(lg1 != lg2 || (lg1 > 0 && lg1 != ng)) {
        error("iladd", "illegal i-lines");
        fprintf(Stderr,"nodes: %d, %d\n", n1, n2);
        fprintf(Stderr,"  old: %d, %d,    new: %d\n", lg1, lg2, ng);
        erfail();
    }

    /* already connected */
    if(lg1 != 0)
        return;

    if(i1 > ndlegs(n1)) {
        error("iladd", "cannnot connect");
        fprintf(Stderr, "i1 = %d > %d = ndlegs(%d)\n",
                i1, ndlegs(n1), n1);
        erfail();
    }
    if(n1 == n2) {
#if DBG
        if(dbg >= ildbg)
            printf("iladd: i1=%d\n", i1);
#endif
        for(i = 0; i < ngs; i++, i1 += 2) {
            i2 = i1 + 1;
            nilines++;
            ndnextn(n1, i1) = n1;
            ndnextn(n1, i2) = n1;
            ndiline(n1, i1) = nilines; /* >= 1 cf. fline.c */
            ndiline(n1, i2) = nilines;
        }
    } else {
        if(i2 > ndlegs(n2)) {
            error("iladd", "cannnot connect");
            fprintf(Stderr, "i2 = %d > %d = ndlegs(%d)\n",
                    i2, ndlegs(n2), n2);
            erfail();
        }
#if DBG
        if(dbg >= ildbg)
            printf("iladd: i1=%d, i2=%d\n", i1, i2);
#endif
        for(i = 0; i < ngs; i++, i1++, i2++) {
            nilines++;
            ndnextn(n1, i1) = n2;
            ndnextn(n2, i2) = n1;
            ndiline(n1, i1) = nilines;
            ndiline(n2, i2) = nilines;
        }
    }
}

/*==================================================================*/
Public  Void    ilsetup(NoArg)
/*--------------------------------------------------------------------
-------------------------------------------------------------------*/
{
    node n, nt;
    int  i;

#if DBG
    erdbg("ilsetup", "begin", ildbg);
#endif

    ilinit();
    for(n = 0; n < nnodes; n++) {
        for(i = 0; i < ndlegs(n); i++)
            ndiline(n, i) = -1;
    }

    for(n = 0; n < nnodes; n++) {
        /* ndnextn is compressed storage of edgetbl */
        for(nt = n; nt <= nnodes; nt++) {
            iladd(n, nt, edgetbl[n][nt]);
        }
    }

#if DBG
    erdbg("ilsetup", "end True", ildbg);
#endif

    ilend();
}

/*==================================================================*/
Public  Void    illdump(NoArg)
/*--------------------------------------------------------------------
-------------------------------------------------------------------*/
{
    int n, l;

    fprintf(Stderr,"node[#leg]: il[nd,p] ... \n");
    for(n = 0; n < nnodes; n++) {
        fprintf(Stderr,"%d[ndlegs]:", n, ndlegs(n));
        for(l = 0; l < ndlegs(n); l++) {
            fprintf(Stderr,"%d[%d, %d] ", ndiline(n,l), ndnextn(n,l),
                    ndparticle(n, l));
        }
        fprintf(Stderr,"\n");
    }
}
