/*********************************************************************
 *  File : outf.c        Prifix = "fo".
 *
 *  Output to a file
 *
 *                                                 by Kaneko Toshiaki
 *  calling sequence:
 *
 *    1. Initialization:
 *
 *           if(!foinit("<output file name>"))
 *               exit(FAIL);
 *
 *    2. Output a process:
 *
 *           for(processes) {
 *             foproces();
 *
 *    3. Output a graph:
 *
 *             if(!passg)
 *                 painit();  ???
 *                 fograph();
 *
 *       the data of graph is given through global variables.
 *
 *    4. Close process:
 *
 *             fopend();
 *           }
 *
 *    5. Close package:
 *
 *           foclose();
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include "grch.h"


/* Entry points */
Public  bool    foinit     ARG((string));
Public  Void    foproces   ARG((NoArg));
Public  Void    fopend     ARG((NoArg));
Public  Void    fograph    ARG((NoArg));
Public  Void    foclose    ARG((NoArg));

Public  Void    fofinit    ARG((FILE*));
Public  Void    fofproc    ARG((FILE*,tproc*,tgraph*));
Public  Void    fofpend    ARG((FILE*,int));
Public  Void    fofgraph   ARG((FILE*,tproc*,tgraph*));
Public  Void    fofend     ARG((FILE*,int));
Public  Void    foflines   ARG((FILE*));

/* External references */
/* ergrf.c */
Extern  Void    ercont       ARG((NoArg));
Extern  Void    error        ARG((string, string));
/* rdptbl.c */
Extern  string  pvname   ARG((pindex));
Extern  string  pvename  ARG((int,pindex));

/* Private functions */
Private Void    foscale  ARG((FILE*,int));

#define LLEN    40

/* Local variable */
Local   FILE    *fofp;  /* pointer to the output file */
Local   bool    fopro=False;   /* process is open or not */

/*==================================================================*/
Public  bool    foinit(fn)
string  fn;
/*--------------------------------------------------------------------
    Initialize file out package.

    Argument :
        fn : file name to output.
--------------------------------------------------------------------*/
{
    fofp = fopen(fn, "w");
    if(fofp == NULL) {
        error("foinit", "open err");
        fprintf(Stderr, "fname = %s\n", fn);
        ercont();
        return False;
    }

    fofinit(fofp);
    return True;
}

/*==================================================================*/
Public  Void    foclose(NoArg)
/*--------------------------------------------------------------------
    Close output package.
--------------------------------------------------------------------*/
{
    if(fofp != NULL) {
        if(fopro)
            fopend();

        fofend(fofp, cprocess->procid);

        fclose(fofp);
    }
}

/*==================================================================*/
Public  Void    foproces(NoArg)
/*--------------------------------------------------------------------
    output process parameter.
--------------------------------------------------------------------*/
{
    if(fopro)
        fopend();

    fopro = True;

    fofproc(fofp, cprocess, cgraph);
}

/*==================================================================*/
Public  Void    fopend(NoArg)
/*--------------------------------------------------------------------
    end of the process
--------------------------------------------------------------------*/
{
    if(fopro) {
        fofpend(fofp, cprocess->graphs);
    }
    fopro = False;
}

/*==================================================================*/
Public  Void    fograph(NoArg)
/*--------------------------------------------------------------------
    outputs current graph.
--------------------------------------------------------------------*/
{
    if(!fopro)
        foproces();

    fofgraph(fofp, cprocess, cgraph);

    /* print fermion lines */
    foflines(fofp);
}

/*==================================================================*/
Private Void    foscale(rgf, ch)
FILE    *rgf;
int     ch;
/*--------------------------------------------------------------------
    outputs scale line.
--------------------------------------------------------------------*/
{
    int i;

    putc('%', rgf);
    for(i = 1; i < LLEN; i++)
        putc(ch, rgf);
    putc(EOL, rgf);
}

/*==================================================================*/
Public  Void    fofproc(rgf, proc, grf)
FILE    *rgf;
tproc   *proc;
tgraph  *grf;
/*--------------------------------------------------------------------
    output process parameter.
--------------------------------------------------------------------*/
{
    int   n;
    tnode *nd;
    tleg  *lg;

    cprocess = proc; /* for "tovert" */
    cgraph   = grf;

    foscale(rgf,'%');

    if(proc->mult)
        fprintf(rgf, "Process=%d; %% %d x %d\n", 
            proc->procid, proc->proc1, proc->proc2);
    else
        fprintf(rgf, "Process=%d;\n", proc->procid);

    /* external particles */
    fprintf(rgf, "External=%d;\n", proc->nextern);
    for(n = 0; n < proc->nextern; n++) {
        nd = ndnode(n);
        lg = ndleg(n, 0);
        fprintf(rgf, "%4d= %s %s;\n",
                tovert(n), eliorf(nd->nvkind),
                pvename(nd->nvkind, lg->lgpart));
    }
    fprintf(rgf, "Eend;\n");

    for(n = 0; n < nmordtbl; n++) {
        if(proc->morder[n] != 0)
            fprintf(rgf, "%s=%d;", mordtbl[n], proc->morder[n]);
    }
    fprintf(rgf, "Loop=%d;\n", proc->loop);

    fprintf(rgf, "OPI=%s;", boolstr(proc->onepi));
    fprintf(rgf, "Expand=%s;\n", boolstr(proc->expand));
    fprintf(rgf, "Gm2=%s;\n", boolstr(proc->gm2));
    if(proc->kinem != NULL)
        fprintf(rgf, "Kinem=\"%s\";\n", proc->kinem);
}

/*==================================================================*/
Public  Void    fofgraph(rgf, proc, grf)
FILE    *rgf;
tproc   *proc;
tgraph  *grf;
/*--------------------------------------------------------------------
    outputs current graph.
--------------------------------------------------------------------*/
{
    int i, j, v, io, black;
    tnode *nd;
    tleg  *lg;

    cprocess = proc;
    cgraph   = grf;

    foscale(rgf,'-');

    if(grf==NULL) {
        fprintf(rgf, "%% Null Graph\n");
        return;
    }

    cgraph = grf;

    /* parameters */
    fprintf(rgf, "Graph=%d;\n", grf->grid);
    if((nd = ndnode(0)) == NULL) {
        fprintf(rgf, "NULL_NODE_TABLE;\n");
        return;
    }
    fprintf(rgf, "Gtype=%ld;\n", grf->gtype);
    fprintf(rgf, "Sfactor=%ld;\n", grf->sfactor);
    fprintf(rgf, "Vertex=%d;\n", grf->nnodes - proc->nextern);

    /* print edge table in the form of list. */
    for(i = 0; i < grf->nnodes; i++) {
        nd = ndnode(i);
        if(nd->nvtype == NDEXTERN)
            v = VNULL;
        else
            v  = nd->nvkind;
        fprintf(rgf, "%4d", tovert(i));     /* vertex number */
        if(nd == NULL) {
            fprintf(rgf, "[NULL_NODE]={};\n");
            break;
        }

        /* options */
        black = False;
        if(nd->nvtype == NDLOOPEDV) {
            fprintf(rgf, "[loop=%d", nd->nvloop);
            black = True;
        }

        if(v != VNULL && v > 0 && nmordtbl > 1) {
            if(!black) {
                fprintf(rgf, "[");
                black = True;
            } else {
                fprintf(rgf, ";");
            }
            fprintf(rgf, "order={");
            for(io = 0; io < nmordtbl; io++) {
                if(io > 0)
                    fprintf(rgf, ",");
                fprintf(rgf, "%d", vtable[v]->morder[io]);
            }
            fprintf(rgf, "}");
        }

        if(black)
            fprintf(rgf, "]");

        /* list of legs */
        if((lg = ndleg(i, 0)) == NULL) {
            fprintf(rgf, "=NULL_LEGS;\n");
            break;
        }
        fprintf(rgf, "={");

        for(j = 0; j < nd->nvlegs; j++) {
            if (j != 0)
                putc(',', rgf);
            fprintf(rgf, " %4d[%s]", lg->lgiline,
                                     pvname(lg->lgpart));
            lg++;
        }

        fprintf(rgf, "};\n");
    }
    fprintf(rgf, "Vend;\n");

    /* mark of the end of a graph */
    fprintf(rgf, "Gend;\n");
}
/*==================================================================*/
Public  Void    fofinit(rgf)
FILE    *rgf;
/*--------------------------------------------------------------------
    outputs header of the output.
--------------------------------------------------------------------*/
{
    foscale(rgf, '%');

    fprintf(rgf,"Model=\"%s\";\n", fnmdl);
}

/*==================================================================*/
Public  Void    fofpend(rgf, n)
FILE    *rgf;
int     n;
/*--------------------------------------------------------------------
    outputs header of the output.
--------------------------------------------------------------------*/
{
    foscale(rgf,'-');
    fprintf(rgf,"Pend=%d;\n", n);
}

/*==================================================================*/
Public  Void    fofend(rgf, n)
FILE    *rgf;
int     n;
/*--------------------------------------------------------------------
    outputs header of the output.
--------------------------------------------------------------------*/
{
    foscale(rgf,'%');
    fprintf(rgf,"End=%d;\n", n);
}

/*==================================================================*/
Public  Void    foflines(rgf)
FILE    *rgf;
/*--------------------------------------------------------------------
    outputs line.
--------------------------------------------------------------------*/

{
    int i, j;

    fprintf(rgf, "%% Fline=%d;\n", nflines);
    for(i = 0; i < nflines; i++) {
        fprintf(rgf, "%% %4d[type=%d;length=%d]",
                i+1, fltable[i].type, fltable[i].length);
        fprintf(rgf, "={");

        for(j = 0; j < fltable[i].length; j++) {
            if(j != 0)
                putc(',', rgf);
            if((j+2)%6 == 0)
                fprintf(rgf,"\n%%       ");
            fprintf(rgf, " %4d[%d]", 
                tovert(fltable[i].ndlist[j]), fltable[i].lglist[j]);
        }
        fprintf(rgf, "};\n");
    }
    fprintf(rgf, "%% Fend;\n");
    
#if DBG
    if(dbg > 0) {
        printf("Sfactor=%ld;\n", cgraph->sfactor);
        printf("graph %6d fline length: ", cgraph->grid);
        for(i = 0; i < nflines; i++)
            printf(" %4d", fltable[i].length);
        printf("\n");
    }
#endif
}
