/*********************************************************************
 *  File : perm.c
 *
 *  Generate permutations for given partition.
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdio.h>
#include <stdlib.h>
#include "common.h"

/* Entry points */
Public  int  *pminit  ARG((int,int*));
Public  bool  pmnext  ARG((NoArg));
Public  Void  pmdump  ARG((NoArg));

/* External references */

/* local functions */
Private Void pmdmp0  ARG((int,int*));
Private Void pmswp2  ARG((int,int,int*));
Private Void pmswp3  ARG((int,int,int*));

/* local variables */
Local int *wrk=NULL, *prm=NULL, *bnd=NULL;
Local int nelem, nelem0=0;

/*==================================================================*/
Private Void pmdmp0(n, a)
int  n;
int  a[];
/*--------------------------------------------------------------------
    Dump a list.
--------------------------------------------------------------------*/
{
    int j;

    printf("(");
    for(j = 0; j < n; j++)
        printf(" %d", a[j]);
    printf(")");
}

/*==================================================================*/
Public  int *pminit(n, p)
int n;
int p[];
/*--------------------------------------------------------------------
    Initialize the generation of permutations.

    arguments:
        n : size of array p.
        p : partition.  List of numbers of elements in classes.

    returns:
        the pointer of the first permutation, which area is kept 
        within this package.
--------------------------------------------------------------------*/
{
    int j, k, ns, ne;

    nelem = p[n];

    if(nelem > nelem0) {
        free(prm);
        free(wrk);
        free(bnd);
        prm = (int*) malloc(nelem*sizeof(int));
        wrk = (int*) malloc(nelem*sizeof(int));
        bnd = (int*) malloc(nelem*sizeof(int));
        nelem0 = nelem;
    }

    if(prm == NULL || wrk == NULL || bnd == NULL) {
        fprintf(Stderr,"pminit: memory allocation error\n");
        exit(1);
    }

    ne = 0;
    for(j = 0; j < n; j++) {
        ns  = ne;
        ne  = p[j+1];
        for(k = ns; k < ne; k++)
            bnd[k] = ne - 1;
    }

    for(j = 0; j < nelem; j++) {
        prm[j] = j;
        wrk[j] = j;
    }

/*
    printf("nelem = %d :", nelem);
    pmdmp0(n, p);
    pmdmp0(n, bnd);
    printf("\n");
 */

    return prm;
}

/*==================================================================*/
Private Void pmswp2(k, l, a)
int  k, l;
int  a[];
/*--------------------------------------------------------------------
    Swap a[k] and a[l].
--------------------------------------------------------------------*/
{
    int t;

    t    = a[k];
    a[k] = a[l];
    a[l] = t;
}

/*==================================================================*/
Private Void pmswp3(k, l, a)
int  k, l;
int  a[];
/*--------------------------------------------------------------------
    Swap a[k] and a[l] and then a[k] and a[l+1].
--------------------------------------------------------------------*/
{
    int t;

    t      = a[k];
    a[k]   = a[l+1];
    a[l+1] = a[l];
    a[l]   = t;
}

/*==================================================================*/
Public bool pmnext(NoArg)
/*--------------------------------------------------------------------
    Get next permutation.
    This function generate the second or after permutation.
    The first one is obtained by pminit.
    The array, which represents the permutation, is the same one
    returned by pminit.
--------------------------------------------------------------------*/
{
    int j;

    j = nelem-1;
    while(wrk[j] == bnd[j]) {
        if(j == 0)
            return 0;
        pmswp2(j, wrk[j], prm);
        wrk[j] = j;
        j--;
    }
    if(wrk[j] == j)
        pmswp2(j, j+1, prm);
    else
        pmswp3(j, wrk[j], prm);
    wrk[j]++;

/*
    pmdmp0(nelem, prm);
    putchar(EOL);
 */
    return 1;
}

/*==================================================================*/
Public Void pmdump(NoArg)
/*--------------------------------------------------------------------
    Dump generatied permutation.
--------------------------------------------------------------------*/
{
    static long c=0;

    c++;
    printf("c=%8ld: nelem=%3d: ", c, nelem);
    pmdmp0(nelem, prm);
    printf("\n");
}
