/*********************************************************************
 *  File : print.c
 *
 *  Print graphs
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "grch.h"

/* Entry points */
Public  Void    prdump       ARG((NoArg));
Public  Void    presult      ARG((NoArg));
Public  Void    prfirst      ARG((NoArg));
Public  Void    prgraph      ARG((NoArg));
Public  Void    prgtype      ARG((NoArg));
Public  Void    prinit       ARG((NoArg));
Public  Void    prlast       ARG((NoArg));
Public  Void    prnodes      ARG((NoArg));
Public  Void    prparm       ARG((NoArg));
Public  Void    prprn        ARG((long,long));
Public  Void    prtype       ARG((int));

/* External references */
/* elem.c */
Extern  Void    eldump       ARG((NoArg));
/* ergrf.c */
Extern  Void    erfail       ARG((NoArg));
Extern  Void    error        ARG((string, string));
/* outf.c */
Extern  Void    fograph      ARG((NoArg));
/* iline.c */
Extern  Void    illdump      ARG((NoArg));
/* lclass.c */
Extern  Void    lcdump       ARG((int));
/* assign.c */
Extern  Void    padump       ARG((NoArg));
Extern  bool    painit       ARG((NoArg));
/* rdptbl.c */
Extern  string  pvename      ARG((int,pindex));

/* Local functions */
Private Void    prdinc       ARG((long));
Private long    prgcd        ARG((long,long));

Private bool    prsel        ARG((NoArg));

/* Local variables */
#define pvyesno(x)      ((x)? "Yes" : "No")
Local  double   prclk1, prclk2;
Local  long     pr1pi;
Local  long     lastn;

/*==================================================================*/
Public  Void    prinit(NoArg)
/*--------------------------------------------------------------------
    Print initial message
*/
{
    printf("\n");
    printf("    Model  file name = \"%s\"\n", fnmdl);
    printf("    Input  file name = \"%s\"\n", fnprc);

    if(outf)
    printf("    Output file name = \"%s\"\n", fngrf);
}

/*==================================================================*/
Public  Void    presult(NoArg)
/*--------------------------------------------------------------------
    Print resulting graph.
*/
{
#if DBG
/*
    static int dbgon = 0;
    static int outff;
 */
#endif

    nconfig++;
    if(nconfig < 2)
        pr1pi = 0;
    if(bi1pi)
        pr1pi++;

    cprocess->graphs = nconfig;
    cgraph->grid     = nconfig;

    if(!skipasp)
        prdinc(cgraph->sfactor);

#if DBG
    if(nconfig == dbgst) {
        dbg     = 2;
        display = 15;
        trace   = True;
/* 
        dbgon   = 1;
        outff   = outf;
 */
        outf    = True;
    }
#endif

    if(display > 0) {
        printf("\n Graph No. %8d", nconfig);
        prgtype();
    }

    if(display > 1) {
        prgraph();
#if DBG
    } else if(dispno == nconfig) {
        prgraph();
        error("presult", "dispno == nconfig, stop by option");
        erfail();
#endif
    }

    if(outf && prsel()) {
        if(skipasp)
            Discard painit();
        if((skipasp || prsel()) && cptest < 100)
            fograph();
    }

#if DBG
/*
    if(dbgon > 0 && nconfig != dbgst) {
        foclose();
        error("presult", "nconfig != dbgst");
        erfail();

        dbg     = 0;
        trace   = False;
        display = False;
        dbgon   = 0;
        outf    = outff;
    }
*/
#endif
}

/*==================================================================*/
Public  Void    prgraph(NoArg)
/*--------------------------------------------------------------------
    Print the current graph.
*/
{
    printf("Graph No. %ld (%ld)\n", nconfig, ++ndump);
    prparm();
    eldump();
    lcdump(0);
    padump();
    illdump();
    putchar(EOL);

    if(outf) {
        if(skipasp)
            Discard painit();
        fograph();
    }
}


/*==================================================================*/
Public  Void    prdump(NoArg)
/*--------------------------------------------------------------------
    Dump current graph.
*/
{
    printf("Graph No. %ld (%ld)\n", nconfig, ++ndump);
    prparm();
    eldump();
    lcdump(1);
    padump();
    illdump();

    if(outf) {
        if(skipasp)
            Discard painit();
        fograph();
    }
}


/*==================================================================*/
Public  Void    prparm(NoArg)
/*--------------------------------------------------------------------
    Print parameters
*/
{
    printf("Parameter\n");
    printf("nextern = %4d, loop = %4d, order = %4d, nnodes = %4d\n",
           nextern, loop, order, nnodes);
    printf("nlclass = %4d\n", nlclass);
}

/*==================================================================*/
Public  Void    prtype(mode)
int     mode;
/*--------------------------------------------------------------------
    Print type
        mode < 0 : initialization
             = 0 : ending message
             > 0 : next type
*/
{
    long  genn;

    if(mode < 0) {
        gntype  = 0;
        ndump = 0;
        return;
    }
    if(gntype > 0 && display >= 0) {
        genn = nconfig-lastn;
        if(skipasp || genn > 0)
            printf("*** Graph type %4d Graphs = %6ld  Total = %6ld\n",
                   gntype, genn, nconfig);
    }
    if(mode > 0) {
        gntype++;
        if(skipasp) {
            printf("\n*** Graph type %d:  nodes = %d\n\n", 
                   gntype, nnodes);
            prnodes();
            putchar(EOL);
        }
        lastn = nconfig;
    }
}

/*==================================================================*/
Public  Void    prnodes(NoArg)
/*--------------------------------------------------------------------
    Print parameters
*/
{
    node        v;

    printf("node :  typ : legs : loop \n");
    for(v = 0; v < nnodes; v++) {
        printf("%4d : %4d : %4d : %4d\n",
               v, ndtype(v), ndlegs(v), ndloop(v));
    }
}

/*==================================================================*/
Public  Void    prfirst(NoArg)
/*--------------------------------------------------------------------
    Print the first message.

--------------------------------------------------------------------*/
{
    node n;

    printf("*************************************************\n");
    printf("Order = %d (", order);
    for(n = 0; n < nmordtbl; n++) {
        if(n > 0)
            printf(", ");
        printf("%s = %d", mordtbl[n], morder[n]);
    }
    printf(") loop = %d\n", loop);
    printf("\n%4d external lines\n\n", nextern);

    for(n = 0; n < nextern; n++) {
        printf("%4d %s %s\n",
                tovert(n), eliorf(ndiorf(n)),
                pvename(ndiorf(n), ndname(n)));
    }

    printf("\nConditions of generationg graphs:\n");
    printf("    Gnenerate expanded graph = %s\n", pvyesno(expand));
    printf("    Only 1PI graph           = %s\n", pvyesno(onepi));
    printf("    Gnenerate tadpole        = %s\n", pvyesno(tadpole));
    printf("    Assign particle          = %s\n", pvyesno(!skipasp));
    printf("    Block                    = %s\n", pvyesno(block));
    printf("    Self energy              = %s\n", pvyesno(selfe));
    printf("    Counter terms            = %s\n", pvyesno(countert));

    printf("\nConditions for outputs:\n");
    printf("    Display graph data       = %s\n", pvyesno(display));
    printf("    Write graph data to file = %s\n", pvyesno(outf));

#if DBG
    printf("\nConditions for debugging:\n");
    printf("    Debug                    = %s\n", pvyesno(dbg));
    printf("    Monitor performance      = %s\n", pvyesno(mon));
    printf("    Trace functions          = %s\n", pvyesno(trace));
    printf("    Check the result         = %s\n", pvyesno(check));

    if(selprop > 0) {
        printf("\nSelect graph with propagator of %s\n",
               ntable[ptable[selprop]->pname[0]].name);
    }
#endif
    printf("=================================================\n");
    second(&prclk1);
}

/*==================================================================*/
Public  Void    prlast(NoArg)
/*--------------------------------------------------------------------
    Print the last message.
--------------------------------------------------------------------*/
{
    double sec;

    second(&prclk2);
    sec = prclk2 - prclk1;
    printf("\nExt = %d, Loop = %d : Total = %ld",
           nextern, loop, nconfig);
    if(!skipasp) {
        printf(", Weighted sum = ");
        prprn(sfnum, sfden);
    }
    printf("(%8.2lf sec)\n", sec);
    putchar(EOL);
    printf("type = %d, 1PI = %ld\n", gntype, pr1pi);
#if DBG
    printf("\nE = %d, L = %d : T = %ld ",
           nextern, loop, nconfig);
    printf("orda1 = %ld, orda0 = %ld ", orda1, orda0);
    printf("(%8.2lf sec)\n", sec);
#endif
    printf("*************************************************\n");
}

/*==================================================================*/
Public  Void    prgtype(NoArg)
/*--------------------------------------------------------------------
    Print type of the graph.
--------------------------------------------------------------------*/
{
    if(loop < 1) {
        printf(" tree   ");
    } else if(!expand) {
        printf(" skelton");
    } else if(biblock) {
        printf(" block  ");
    } else if(bi1pi) {
        printf(" 1PI    ");
    } else {
        printf("        ");
    }

    if(extself && biexts)
        printf(" e-self ");
    else
        printf("        ");

    if(tadpole && bitadp)
        printf(" tadpole");
    else
        printf("        ");

    putchar(EOL);
}

/*==================================================================*/
Public  Void   prprn(num, den)
long    num, den;
/*--------------------------------------------------------------------
*/
{
    int sgn = 1;

    if(num < 0) {
        sgn = - 1;
        num = - num;
    }
    if(den < 0) {
        sgn = - sgn;
        den = - den;
    }
    if(sgn < 0)
        putchar('-');
    if(num == 0 || den == 1)
        printf("%ld\n", num);
    else
        printf("%ld/%ld\n", num, den);

}

/*==================================================================*/
Private Void   prdinc(den)
long    den;
/*--------------------------------------------------------------------
*/
{
    long  gcd;
    long  den1;
    int   sgn;

    if(den > 0) {
        den1 = den;
        sgn = 1;
    } else {
        den1 = - den;
        sgn = - 1;
    }
    gcd    = prgcd(sfden, den1);
    den1  /= gcd;
    sfnum  = sfnum * den1 +  sgn*(sfden / gcd);
    sfden *= den1;
    gcd    = prgcd(sfnum, sfden);
    sfnum /= gcd;
    sfden /= gcd;
}

/*==================================================================*/
Private long   prgcd(x, y)
long    x, y;
/*--------------------------------------------------------------------
*/
{
    long t;

    while (y != 0) {
        t = x % y;
        x = y;
        y = t;
    }

    return x;
}

/*==================================================================*/
Private bool   prsel(NoArg)
/*--------------------------------------------------------------------
*/
{
    int  l;
    node n;

    if(onegraph && onegraph != nconfig)
        return False;

    if(onetype && lastn + 1 != nconfig)
        return False;

    if(seltype != 0 && gntype != seltype)
        return False;

    if(selprop <= 0)
        return True;
    for(n = nextern; n < nnodes; n++) {
        for(l = 0; l < ndlegs(n); l++) {
            if(ndnextn(n,l) < nextern)
                continue;
            if(selprop == abs(ndparticle(n,l)))
                return True;
        }
    }
    return False;
}
