/*********************************************************************
 *  File : Scalar.c
 *
 *  Scalar graph generation
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "grch.h"

/* Entry points */
Public  Void    gscalar      ARG((NoArg));

/* External references */
/* biconn.c */
Extern  bool    biconn       ARG((NoArg));
/* elem.c */
Extern  bool    elcable      ARG((NoArg));
Extern  Void    elconn       ARG((node, node, int));
Extern  Void    eldisc       ARG((node, node, int));
Extern  bool    elleft       ARG((NoArg));
Extern  int     elroot       ARG((NoArg));
/* ergrf.c */
Extern  Void    erenter      ARG((string));
Extern  Void    ermon        ARG((string, string));
Extern  Void    erreturn     ARG((string));
/* iline.c */
Extern  Void    ilsetup      ARG((NoArg));
/* lclass.c */
Extern  node    lcfind       ARG((int, node));
Extern  bool    lcftest      ARG((bool));
Extern  Void    lcinit       ARG((NoArg));
Extern  Void    lcreset      ARG((NoArg));
Extern  node    lctrgt       ARG((node, node));
Extern  bool    lcupdate     ARG((NoArg));
/* assign.c */
Extern  Void    passign      ARG((NoArg));
/* print.c */
Extern  Void    presult      ARG((NoArg));
Extern  Void    prtype       ARG((int));

/* local functions */
Private Void    gslegs       ARG((int, node, int, node, int));
Private Void    gsconn       ARG((int, int, node));

/* local variables */
Local   int     gsloop;

Local   int     iself, fself;

/*==================================================================*/
Public  Void    gscalar(NoArg)
/*--------------------------------------------------------------------
    Entry point of scalar graph generation.
*/
{
    int f;

    lcinit();

    iself = cptest & 0x0f;
    fself = cptest >> 4;

    gsloop = 0;

    Discard elroot();   /* node number of the root is 0 */
    gsconn(0, -1, -1);
}

/*==================================================================*/
Private Void    gsconn(level, lastl, lnd)
int     level, lastl;
node    lnd;
/*--------------------------------------------------------------------
    Connect nodes.

    Arguments :
        level : level of the source node.
        lastl : level of the last source node.
        lnd   : the last node.
*/
{
    node        ln;

#if DBG
    erenter("gsconn");
#endif

    /* find a node with free legs and a class */
    ln = lcfind(level, lnd);
    if(ln == NDNULL) {  /* no more nodes in this level */
        if(lastl != level) {    /* no nodes in this level */
        } else if(elcable()) {  /* connected graph */
            if(iself == 0 && !lcupdate()) {   /* update l-class */
#if DBG
                if(mon) {
                    ermon("gsconn", "not in order");
                }
#endif
            } else if(elleft()) {       /* next level */
/*
                if((!onepi) || level < 2 || biconn())
*/
                if(level < 2 || biconn()) {
                    gsconn(level+1, level, -1);
                }
            } else {                        /* end */
                if(iself == 1 && !lcupdate()) {   /* update l-class */
#if DBG
                    if(mon) {
                        ermon("gsconn", "not in order");
                    }
#endif
                } else if(biconn()) {
                    if(skipasp) {
                        if(lcftest(False)) {
                            ilsetup();
                            presult();
                        }
                    } else {
                        ilsetup();
                        prtype(1);
                        passign();
                    }
                }
                if(iself == 1)
                    lcreset();      /* reset l-class */
            }
            if(iself == 0)
                lcreset();      /* reset l-class */
#if DBG
        } else {
            if(mon) {
                ermon("gsconn", "disconnected graph");
            }
#endif
        }
    } else {            /* connect legs */
        gslegs(level, ln, ndfree(ln), -1, -1);
    }
#if DBG
    erreturn("gsconn");
#endif
}

/*==================================================================*/
Private Void    gslegs(level, sn, fl, tn, ll)
int     level;
node    sn;
int     fl;
node    tn;
int     ll;
/*--------------------------------------------------------------------
    Connect legs.
    Arguments :
      level : level of the source node.
      sn    : source node.
      fl    : the number of free legs of the source node.
      tn    : the last target node.
      ll    : the multiplicity of the last connection.
*/
{
    int         lg, maxfl, minfl, sfl;
    int         luup0, luup;
    node        nn;

#if DBG
    static int trgslevel=0;
    if(trace) {
        erenter("gslegs");
        printf("       %d (l:%d) %d %d %d %d\n",
               trgslevel, level, sn, fl, tn, ll);
    }
#endif

    /* select node */
    for(nn = lctrgt(sn, tn); nn != NDNULL; nn = lctrgt(sn, nn)) {
        if(!tadpole && sn == nn)
            continue;

        if(expand) {
            minfl = 1;
            if(sn == nn)
                maxfl = ((int) fl/2);
            else
                maxfl = fl;
            if(loop < 1)        /* tree */
                maxfl = 1;
            maxfl = min(maxfl, ndfree(nn));
        } else {
            maxfl = minfl = 1;
        }
        luup0 = gsloop;
        for(lg = minfl; lg <= maxfl; lg++) {  /* vary multiplicity */
            luup = gsloop + lg - 1;
            if(ndfree(nn) < ndlegs(nn)) {
                luup++;
            }
            if(luup > loop) {
                continue;
            }
            elconn(sn, nn, lg);         /* connect nodes */
            gsloop = luup;
            sfl = (sn == nn) ? fl - 2*lg : fl - lg;
            if(sfl < 1) {              /* free legs left ? */
                if(level < 2 || biconn())
                    gsconn(level, level, sn);   /* next node */
            } else {
                gslegs(level, sn, sfl, nn, lg);
            }
            gsloop = luup0;
            eldisc(sn, nn, lg);
        }
    }
#if DBG
    if(trace)
        erreturn("gslegs");
#endif
}

