/*********************************************************************
 *    File : cfbase.h
 *
 *    Include file for color base.
 *
 *    Add prefix "cf" for names which can be referred from outside.
 */

/* type of particles:
 *     these values are used in "cftbase.cptype" and "cftgraph.tedge".
 */
#define CFNPUNDEF  0      /* undefined particle */
#define CFNOCOLOR  1      /* not colored particle */
#define CFQUARK    2      /* quark */
#define CFAQUARK   3      /* anti-quark */
#define CFGLUON    4      /* gluon */
#define CFGHOST    5      /* ghost */
#define CFAGHOST   6      /* anti-ghost */
#define CFPKIND    7      /* the number of kinds of particles */

/* names of above values */
#ifdef DEFGLOBAL
Global char *cfparticle[] = {
    "undef",
    "no_color",
    "quark",
    "anti-quark",
    "gluon",
    "ghost",
    "anti-ghost",
};
#else
Global char *cfparticle[];
#endif

/* type of vertices :
 *     these values are used in "cftgraph.lvtx".
 */
#define CFVDEL     0      /* deleted vertex */
#define CFVQQG     1      /* quark-quark-gluon vertex */
#define CFVGGG     2      /* three gluon vertex */
#define CFVHHG     3      /* gluon-ghost-ghost vertex */
#define CFVQQ      4      /* quark-quark-others vertex */
#define CFVGG      5      /* gluon-gluon vertex : counter term ? */
#define CFVHH      6      /* ghost-ghost vertex : counter term ? */
#define CFVNONQCD  7      /* no colored particle */

/* names of above values */
#ifdef DEFGLOBAL
Global char *cfvertex[] = {
    "deleted",
    "qqg",
    "ggg",
    "hhg",
    "qq",
    "gg",
    "hh",
    "non QCD",
};
#else
Global char *cfvertex[];
#endif

/* interface of "cfbase" */
typedef struct {
    /* input for cbase */
    int     nex;          /* the number of external particles
                           */
    int    *cptype;       /* cptype[0..nex] : type of the particles.
                           * cptype[i] : for i-th external particle.
                           * 0th component is used as dummy.
                           */

    /* output from cbase and input for cfproj */
    int      nbase;       /* the number of color base
                           */
    int      nquarkp;     /* the number of quark pairs
                           */
    int      ngluon;      /* the number of gluons
                           */
    int     *quark;       /* constr[0..nquarkp-1] : address of quark 
                           */
    int     *antiq;       /* constr[0..nquarkp-1] : address of antiquark 
                           */
    int     *gluon;       /* constr[0..ngluon-1] : address of gluon
                           */
    int     *constr;      /* constr[0..nbase-1] : number of strings
                           */
    int     *nconid;      /* nconid[0..nbase-1] : length of conid[i]
                           */
    int    **conid;       /* conid[0..nbase-1][0..nconid[i]-1] : 
                           * ID number particles 
                           */
    double **cwgt;        /* cwgt[0..nbase-1][0..nbase-1] : weight for 
                           * the color base
                           */
} cftbase;

/* interface to cfproj */
typedef struct {
    /* input for cfproj */
    int  nvtx;           /* the number of vertices
                          */ 
    int *nleg;           /* nleg[0..nvtx] : number of legs 
                          * when nleg[i] = 3, the vertex is 3 point vertex.
                          * 0th component is a dummy.
                          */
    int *lvtx;           /* lvtx[0..nvtx] : type of the vertices
                          * 0th component is a dummy.
                          */
    int  nedge;          /* the number of edges
                          */
    int *tedge;          /* nedge[0..nedge] : the type of edges
                          * 0th component is a dummy.
                          */
    int *edge[2];        /* starting    edge[0][0..nedge] and 
                          * terminating edge[1][0..nedge] vertices.
                          * Value:
                          *     -n < 0 : external particle corresponds to 
                          *              cftbase.cptype[n]
                          *      n > 0 : vertex corresponds to
                          *              cftgraph.nleg[n], lvtx[n]
                          *      n = 0 : not possible.
                          */

    /* output from cfproj */
    double *wgt;          /* wgt[0..nbase-1] : weight factor of the graph
                           */
} cftgraph;

/* working structure for a graph in cfproj */
typedef struct {
    /* input for cfproj */
                         /* 0th component is a dummy for all. */
    int  nvtx;           /* the number of vertices
                          */ 
    int  nedge;          /* the number of edges
                          */
    int *mvtx;           /* mvtx[0..nvtx] : address of vertex
                          *  negative = internal vertex
                          *  positive = external particle number
                          * 0th component is a dummy.
                          */
    int *lvtx;           /* lvtx[0..nvtx] : type of the vertex
                          * 0th component is a dummy.
                          */
    int *tedge;          /* nedge[0..nedge] : type of edge
                          * 0th component is a dummy.
                          */
    int *edge[2];        /* starting edge[0][0..nedge] and 
                          * terminating edge[1][0..nedge] (=vertex)
                          */

    /* output from cfproj */
    double wgtx;          /*  weight factor of the graph
                           */
} cftxgraph;

/* TI */
Global cftbase  *cbase;
Global cftgraph *colgr;
/*
#endif
*/
/********************* entry points ******************************/

/* Calculate color base */
Public int cfbase        ARG((cftbase*, int, int));
/*
 * status = cfbase(cbase, ibase, iprnt);
 *
 * cbase : pointer to cftbase : color base data structure
 * ibase : input flag for the selection of bases
 * iprnt : input flag for printing information
 *
 * returns status
 */

/* Calculate weight to color base */
Public int cfproj       ARG((cftbase *, cftgraph *, int, int));
/*
 * status = cfproj(cbase, cgraph, ibase, iprnt));
 *
 * cbase  : pointer to cftbase : color base data structure
 * cgraph : pointer to cftgraph : information about a graph and its weight
 * ibase  : input flag for the selection of bases
 * iprnt  : input flag for printing information
 */
 
/* Dump data for debugging */
Public Void cfbdump      ARG((cftbase*));
Public Void cfgdump      ARG((cftgraph*));
Public Void cfdump       ARG((cftbase *, cftgraph *));
