/********************************************************************
 *  File : cmfin.c
 *
 *  Process *.fin file for extended format.
 *
 *                                                 by Kaneko Toshiaki
 *  1995/05/30  Created.
 *      1. Function "cmamparm" is moved from "cmgrf.c".
 *      2. Function "cmamparm" should be called after 
 *         "incl1.h/incomm.h" is created.
 *      3. Function "cmfobj" should be called after "cmamparm".
 *  
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "common.h"
#include "config.h"
#include "dbgh.h"
#include "ptable.h"
#include "grdata.h"
#include "putidh.h"
#include "gensc.h"
#include "cfbase.h"

/*------------------------------------------------------------------*/
/* Entry points */
Public  Void    cmamparm   ARG((NoArg));
Public  Void    cmfobj     ARG((FILE*));

/*------------------------------------------------------------------*/
/* External references */
/* rdptbl.c */
/* cfconv.c */
Extern  int  cfnoind        ARG((int));
/*------------------------------------------------------------------*/
/* Private functions */
Public  Void    cmfin       ARG((string));
Private Void    cmprline    ARG((string,FILE*,bool));
Private string  cmfcmd      ARG((string));
Private Void    cmdexec     ARG((int,string*,FILE*,FILE*));
Private string  cmgetline   ARG((FILE*));
Private Void    cmkmass     ARG((FILE*));

/*------------------------------------------------------------------*/
/* Local variable */

Local string ficmd[] = 
{"common", "objs", "declare", "fcode", "global", "commonk"};
#define CMD_NULL      -1
#define CMD_COMMON     0
#define CMD_OBJS       1
#define CMD_DECLARE    2
#define CMD_FCODE      3
#define CMD_GLOBAL     4
#define CMD_COMMONK    5
#define FICMD     (sizeof(ficmd)/sizeof(string))

#define CMDMAXOBJS  10
Local string cmobjs[CMDMAXOBJS];
Local int    cmnobjs = 0;


#define MAXLINE     256
#define MAXCOLR     16
/*==================================================================*/
Public  Void    cmamparm(NoArg)
/*------------------------------------------------------------------*
 *------------------------------------------------------------------*/
{
    string p;
    FILE   *fin;
    int    c, n, j, df, ix, ct;
    int    ic1,ic2;
    int    i, mxlen;
    int    kol[MAXCOLR],ktmp,k,pa;

    cmopen(ROFORT, "amparm.f");

    cmsepar('*');
    fprintf(redfp, "      subroutine amparm\n");
    fprintf(redfp, "      implicit real*8(a-h,o-z)\n");
    putc(EOL, redfp);

    if( vcode )
        fprintf(redfp, "      include 'inclvs.f'\n");
    if(lang == ROFORT || lang == ROREDA) {
        fprintf(redfp, "      include 'incl1.f'\n");
    } else {
        fprintf(redfp, "      include 'incomm.f'\n");
    }

    fprintf(redfp, "      include 'inclk.f'\n");

    pvfenam(momstr2, fnmdl, ".fin", MAXPATHLEN);
    cmfin(momstr2);

    for(n = nvstart; n < nvtable; n++) {
        if(vtable[n]->phase == V_PUNDEF)
            continue;    
        df = cmdfcpl(cmttype(n), vtable[n]->nvlegs, 
             vtable[n]->loop, vtable[n]->ctype);
        df = max(df, 1);
        for(j = 0; j < df; j++) {
            if(df > 1)
                ix = j + 1;
            else
                ix = 0;
            Discard cmcoupl(n, 0, vtable[n]->loop, momstr);
            if(j >= MAXFVAL ||
               (lang != ROFORT && vtable[n]->fvalue[j] == NULL)) {
                fprintf(redfp,"      %6s    = ???\n", momstr);
            } else {
                if(ix == 0) {
                    fprintf(redfp,"      %6s    = ", momstr);
                } else {
                    fprintf(redfp,"      %6s(%d) = ", momstr, ix);
                }

                if(lang == ROFORT) {
                    if(vtable[n]->phase == V_REAL)
                        fprintf(redfp,"zr*"); 
                    else if(vtable[n]->phase == V_IMAG)
                        fprintf(redfp,"zi*"); 
    
                    ct = vtable[n]->ctype;
                    if(ct == V_V) {
                        fprintf(redfp,"(%s)\n", 
                                vtable[n]->fvalue[0]);
    
                    } else if(ct == V_A) {
                        if(j == 0) {
                            fprintf(redfp,"(-(%s))\n",
                                    vtable[n]->fvalue[0]);
                        } else {
                            fprintf(redfp,"(%s)\n", 
                                    vtable[n]->fvalue[0]);
                        }
                    } else if(ct == V_VminusA) {
                        if(j == 0) {
                            fprintf(redfp,"(%s)\n", 
                                    vtable[n]->fvalue[0]);
                        } else {
                            fprintf(redfp,"0\n");
                        }
                    } else if(ct == V_VplusA) {
                        if(j == 0) {
                            fprintf(redfp,"0\n");
                        } else {
                            fprintf(redfp,"(%s)\n", 
                                    vtable[n]->fvalue[0]);
                        }
                    } else {
                        fprintf(redfp,"(%s)\n", 
                                vtable[n]->fvalue[j]);
                    }
                } else {  /* lang != ROFORT */
                    fprintf(redfp,"%s\n", 
                            vtable[n]->fvalue[j]);
                }
            }
        } /* for */
        if(lang != ROFORT){
            Discard cmcoupl(n, -2, 0, momstr);
            if(vtable[n]->phase == V_REAL)
                fprintf(redfp,"      %6s    = zr\n\n", momstr); 
            else
                fprintf(redfp,"      %6s    = zi\n\n", momstr); 
        }
    }

    fprintf(redfp,"\n* particle name\n");
    for(n = 0; n < nextern; n++) {
        pa  = ndname(n);
        fprintf(redfp, "      kmprtc(%d) = '%s'\n",
                n+1,pvename(ndiorf(n),pa));
        fprintf(redfp, "      kmprtl(%d) = %d\n",
                n+1,strlen(pvename(ndiorf(n),pa)));
    }
    cmkmass(redfp);

/* CF */
    if( colorv == 0 ) {
    if(lang == ROFORT || lang == ROREDA) {
        if(cprocess->qcd != 0) {
            fprintf(redfp,"\n* read color factor\n");
            fprintf(redfp,"      open(22, file='cf0x0.dat', status='old')\n");
            fprintf(redfp,"\n");
            fprintf(redfp,"      do 140 i = 1, 4\n");
            fprintf(redfp,"        read(22,*)\n");
            fprintf(redfp,"  140 continue\n");
            fprintf(redfp,"      do 100 i = 1, ngraph\n");
            fprintf(redfp,"        igraph(i) = 0\n");
            fprintf(redfp,"        do 100 j = 1, ngraph\n");
            fprintf(redfp,"          read(22,*) ig, jg, nn, nd\n");
            fprintf(redfp,"          if(ig.ne.i .or. jg.ne.j) then\n");
            fprintf(redfp,"            write(6,*) 'illegal file of color ");
            fprintf(redfp,"factors (colorf)' \n");
            fprintf(redfp,"            write(6,*) ig, jg, nn, nd\n");
            fprintf(redfp,"            stop\n");
            fprintf(redfp,"          endif\n");
            fprintf(redfp,"          cf(j, i) = nn/dble(nd)\n");
            fprintf(redfp,"  100 continue\n");
            fprintf(redfp,"\n");
        } else if(cprocess->extcol != 0) {
            fprintf(redfp,"      call cfcalc\n");
        }
    }
    } else if( colorv == 1 ) {
      if( cbase->nbase > 1 ) {
        fprintf(redfp,"* weight for the color base\n");
        for(ic1=0;ic1<cbase->nbase;ic1++){
          for(ic2=0;ic2<cbase->nbase;ic2++){
/*
             fprintf(redfp, "      cfmtx(%d,%d)= %lfd0\n",
                ic1,ic2,cbase->cwgt[ic1][ic2]);
*/
             if( (cbase->cwgt[ic1][ic2]).den == 1 )
               fprintf(redfp, "      cfmtx(%d,%d)= %d.d0\n",
                       ic1,ic2,(cbase->cwgt[ic1][ic2]).num);
             else
               fprintf(redfp, "      cfmtx(%d,%d)= %d.d0/%d.d0\n",
                       ic1,ic2,(cbase->cwgt[ic1][ic2]).num,
                               (cbase->cwgt[ic1][ic2]).den);
          }
        }
      } else if( cbase->nbase == 1 && cbase->nexcolor == 1 ) {
        fprintf(redfp,"* weight for the color base\n");
        if( (cbase->cwgt[0][0]).den == 1 )
            fprintf(redfp, "      cfmtx = %d.d0\n",
               (cbase->cwgt[0][0]).num);
        else 
            fprintf(redfp, "      cfmtx = %d.d0/%d.d0\n",
               (cbase->cwgt[0][0]).num,(cbase->cwgt[0][0]).den);
      }
      fprintf(redfp, "\n");
    }
/* color string */
    if( colorv == 1 && cbase->nbase > 0 ) {
        fprintf(redfp,"\n* Color string information\n");
        if( cbase->nbase == 1 ) {

            fprintf(redfp,"      kmcbas    = %d\n",cbase->nbase);
            fprintf(redfp,"      kmcbmx    = %d\n",cbase->nconid[0]);
            for(j = 0; j < cbase->nconid[0]; j++)
                        kol[j] = cfnoind(-(cbase->conid[0][j]))+1;
/*    lower - higher */
            for(i=0;i< cbase->nconid[0];i=i+2){
                if( kol[i] > kol[i+1] ) {
                    ktmp = kol[i];
                    kol[i] = kol[i+1];
                    kol[i+1] = ktmp;
                }
            }

/*    ordering sets */
            for(i=0;i< cbase->nconid[0];i=i+2){
                for(j=i+2;j<cbase->nconid[0];j=j+2){
                    if( kol[i] > kol[j] ) {
                        ktmp = kol[i];
                        kol[i] = kol[j];
                        kol[j] = ktmp;
                        ktmp = kol[i+1];
                        kol[i+1] = kol[j+1];
                        kol[j+1] = ktmp;
                    }
                }
            }

            for(j = 0; j < cbase->nconid[0]; j++)
                fprintf(redfp,"      kmcstr(%d) = %d\n",j+1,kol[j]);
        } else if( cbase->nbase > 1 ) {
            mxlen = cbase->nconid[0];
            for( i = 1; i < cbase->nbase; i++) {
               if( cbase->nconid[i] > mxlen )
                   mxlen = cbase->nconid[i];
            }
            fprintf(redfp,"      kmcbas    = %d\n",cbase->nbase);
            fprintf(redfp,"      kmcbmx    = %d\n",mxlen);

            for(i = 0; i < cbase->nbase; i++) {
                for(j = 0; j < cbase->nconid[i]; j++)
                        kol[j] = cfnoind(-(cbase->conid[i][j]))+1;
/*    lower - higher */
                for(k=0;k< cbase->nconid[i];k=k+2){
                    if( kol[k] > kol[k+1] ) {
                        ktmp = kol[k];
                        kol[k] = kol[k+1];
                        kol[k+1] = ktmp;
                    }
                }
/*    ordering sets */
                for(j=0;j<cbase->nconid[i];j=j+2){
                    for(k=j+2;k<cbase->nconid[i];k=k+2){
                        if( kol[j] > kol[k] ) {
                            ktmp = kol[j];
                            kol[j] = kol[k];
                            kol[k] = ktmp;
                            ktmp = kol[j+1];
                            kol[j+1] = kol[k+1];
                            kol[k+1] = ktmp;
                        }
                    }
                }

                for(j = 0; j < cbase->nconid[i]; j++)
                    fprintf(redfp,"      kmcstr(%d,%d) = %d\n",j+1,
                          i,kol[j]);
            }
        }
    }
 
    fprintf(redfp, "      return\n");
    fprintf(redfp, "      end\n");
}

/*==================================================================*/
Public  Void    cmampa1(NoArg)
/*------------------------------------------------------------------*
 */
{
    fprintf(redfp, "      complex*16 zr, zi\n");
    fprintf(redfp, "      save       zr, zi\n");
    fprintf(redfp, "      zr = dcmplx(1.0d0, 0.0d0)\n");
    fprintf(redfp, "      zi = dcmplx(0.0d0, 1.0d0)\n");
    cmsepar('-');
    fprintf(redfp, "*     call sminit(   1,   1)\n");
    putc(EOL, redfp);

    fprintf(redfp, "* coupling constants\n");
    cmsepar('-');
}

/*==================================================================*/
Public  Void    cmfin(fname)
string  fname;
/*------------------------------------------------------------------*
 * Process *.fin file.
 */
{
    int    head, cmd, cmdold = CMD_NULL;
    string line, cmdstr;
    FILE   *fin;

    if((fin = fopen(fname, "r")) == NULL) {
        fprintf(Stderr, "*** cmfin: open error '%s'\n", fname);
        exit(1);
    }

    line = cmgetline(fin);
    while(line != NULL) {
        if(line[0] == '+' || line[0] == '-') {
            cmdstr = cmfcmd(line);
            for(cmd = 0; cmd < FICMD; cmd++) {
                if(strcmp(cmdstr, ficmd[cmd]) == 0)
                    break;
            }
            if(cmd >= FICMD) {
                fprintf(Stderr, "*** command '%s' is not defined\n",
                        cmdstr);
                cmprline(line, redfp, False);
                line = cmgetline(fin);
                continue;
            }
            if(line[0] == '+') {
                if(cmdold != CMD_NULL) {
                    fprintf(Stderr, "*** '- %s' is expected\n",
                                    ficmd[cmdold]);
                    fprintf(Stderr, "in %s\n", line);
                    cmprline(line, redfp, False);
                    line = cmgetline(fin);
                    continue;
                }
                cmdold = cmd;
                cmdexec(cmd, &line, fin, redfp);
            } else {
                if(cmd != cmdold) {
                    if(cmdold != CMD_NULL) {
                        fprintf(Stderr, "*** '- %s' is expected\n",
                                        ficmd[cmdold]);
                        fprintf(Stderr, "in %s\n", line);
                    } else {
                        fprintf(Stderr, "*** %s without '+'\n", line);
                    }
                    cmprline(line, redfp, False);
                } else {
                    cmdold = CMD_NULL;
                }
                line = cmgetline(fin);
                continue;
            }
        } else {
            cmdexec(CMD_FCODE, &line, fin, redfp);
        }
    }
}

/*==================================================================*/
Private string  cmgetline(fin)
FILE    *fin;
/*------------------------------------------------------------------*
 * Read a line
 */
{
    static int c = ' ';
    static char line[MAXLINE];
    int i, toolong = False;

    if(c == EOF)
        return NULL;
    i = 0;
    while((c = getc(fin)) != EOF && c != '\n') {
        if(i >= MAXLINE - 1)
            toolong = True;
        else
            line[i++] = c;
    }
    line[i] = '\0';
    if(toolong)
        fprintf(Stderr, "*** too long line %s...\n", line);
    return line;
}
            
/*==================================================================*/
Private Void    cmprline(line, redfp, tolow)
char    *line;
FILE    *redfp;
bool    tolow;
/*------------------------------------------------------------------*
 * Process *.fin file.
 */
{
    while(*line != '\0') {
        if(tolow)
            putc(tolower(*line++), redfp);
        else
            putc(*line++, redfp);
    }
    putc('\n', redfp);
}

/*==================================================================*/
Private string  cmfcmd(line)
string  line;
/*------------------------------------------------------------------*
 */
{
    int i;
    static char str[MAXLINE];
    string s;
    int len = strlen(line);

    s = line + 1;
    while(isspace(*s))
        s++;
    for(i = 0; i < MAXLINE - 1; i++) {
        if(*s == '\0' || isspace(*s))
            break;
        else
            str[i] = tolower(*s++);
    }
    str[i] = '\0';
    return str;
}

/*==================================================================*/
Private Void    cmdexec(cmd, line, fin, redfp)
int     cmd;
string  *line;
FILE    *fin;
FILE    *redfp;
/*------------------------------------------------------------------*
 */
{
    FILE *outfp;
    string s, t;
    bool static fstat=False;

    switch(cmd) {
        case CMD_COMMON:
            if(lang == ROFORT || lang == ROREDA) {
                if((outfp = fopen("incl1.f", "a")) == NULL) {
                    fprintf(Stderr, "*** incl1.f : open error\n");
                    exit(FAIL);
                }
            } else {
                if((outfp = fopen("incomm.f", "a")) == NULL) {
                    fprintf(Stderr, "*** incomm.f : open error\n");
                    exit(FAIL);
                }
            }
            *line = cmgetline(fin);
            while(*line != NULL && **line != '+' && **line != '-') {
                 cmprline(*line, outfp, True);
                 *line = cmgetline(fin);
            }
            fclose(outfp);
            break;

        case CMD_COMMONK:
            if((outfp = fopen("inclk.f", "a")) == NULL) {
                fprintf(Stderr, "*** usrprm.f : open error\n");
                exit(FAIL);
            }
            *line = cmgetline(fin);
            while(*line != NULL && **line != '+' && **line != '-') {
                 cmprline(*line, outfp, True);
                 *line = cmgetline(fin);
            }
            fclose(outfp);
            break;

        case CMD_GLOBAL:
            if((outfp = fopen("usrprm.f", "a")) == NULL) {
                fprintf(Stderr, "*** usrprm.f : open error\n");
                exit(FAIL);
            }
            *line = cmgetline(fin);
            while(*line != NULL && **line != '+' && **line != '-') {
                 cmprline(*line, outfp, True);
                 *line = cmgetline(fin);
            }
            fclose(outfp);
            break;

        case CMD_OBJS:
            *line = s = cmgetline(fin);
            while(*line != NULL && **line != '+' && **line != '-') {
                 while(True) {
                     while(isspace(*s))
                         s++;
                     if(*s == '\0')
                         break;
                     if(cmnobjs >= CMDMAXOBJS) {
                         fprintf(Stderr, "*** too many objs\n");
                         exit(1);
                     }
                     t = s;
                     while(!isspace(*s) && *s != '\0')
                         s++;
                     if(*s == '\0') {
                         cmobjs[cmnobjs++] = (string) strdup(t);
                         break;
                     } else {
                         *s = '\0';
                         cmobjs[cmnobjs++] = (string) strdup(t);
                         s++;
                     }
                 }

                 *line = s = cmgetline(fin);
            }
            break;
            
        case CMD_DECLARE:
            if(fstat) {
                fprintf(Stderr, "*** cmdexec : '+DECLARE' should be ");
                fprintf(Stderr, "should be placed before FORTRAN code\n");
            }
            *line = cmgetline(fin);
            while(*line != NULL && **line != '+' && **line != '-') {
                 cmprline(*line, redfp, True);
                 *line = cmgetline(fin);
            }
            break;

        case CMD_FCODE:
            *line = cmgetline(fin);
            while(*line != NULL && **line != '+' && **line != '-') {
                if(!fstat && **line != '*') {
                    fstat = True;
                    cmampa1();
                }
                cmprline(*line, redfp, True);
                *line = cmgetline(fin);
            }
            break;

        defalt:
            fprintf(Stderr, "*** cmdexec : illegal control\n");
            exit(FAIL);
    }
}

/*==================================================================*/
Public  Void    cmfobj(redfp)
FILE    *redfp;
/*------------------------------------------------------------------*
 */
{
    int i;

    for(i = 0; i < cmnobjs; i++) {
        putid(cmobjs[i]);
    }
    putend();
    fprintf(redfp, "\n");
}

/*==================================================================*/
Private Void    cmkmass(redfp)
FILE    *redfp;
/*------------------------------------------------------------------*
 */
{
    int n, c, il, tp;
    pindex p, pa;
    tnode  *nd;
    tleg   *leg;

    fprintf(redfp, "* masses of external particles\n");
    for(n = 0; n < nextern; n++) {
        p   = abs(ndname(n));
        cmmom(momstr, n, 0, False);

        fprintf(redfp, "      amass1(%d) = %s\n",
                n+1, ptable[p]->mass);
    }
    for(n = 0; n < nextern; n++) {
        fprintf(redfp, "      amass2(%d) = amass1(%d)**2\n",
                n+1, n+1);
    }
/* charge */
    fprintf(redfp,"\n* Charge*3\n");
    for(n = 0; n < nextern; n++) {
        pa  = ndname(n);
        p   = abs(pa);
        if(pa > 0)
            c =   ptable[p]->charge3;
        else
            c = - ptable[p]->charge3;
        if(ndiorf(n) == ELINITIAL)
            c *= ptable[p]->antip;
        fprintf(redfp, "      kcharg(%d) = %3d\n", n+1, c);
    }
/* fkcode */
    fprintf(redfp,"\n* KFcode\n");
    for(n = 0; n < nextern; n++) {
        pa  = ndname(n);
        p   = abs(pa);
        if(pa > 0)
            c =   ptable[p]->kfcode;
        else
            c = - ptable[p]->kfcode;
        if(ndiorf(n) == ELINITIAL)
            c *= ptable[p]->antip;
        fprintf(redfp, "      kfcode(%d) = %3d\n", n+1, c);
    }
/* gauge parameters should be set in 'setmas.f' */
/*
    if(lang == ROFORT || lang == ROREDA){
        if(lang == ROFORT) {
        fprintf(redfp, "\n* Gauge parametes (default is unitary gauge)\n");
        for(n = 1; n < 5; n++)
            fprintf(redfp, "      igau%cb = 0\n", gchr[n]);
        for(n = 1; n < 5; n++)
            fprintf(redfp, "      agauge(igau%cb) = 1.0d0\n", gchr[n]);
        fprintf(redfp, "      agauge(igau00) = 1.0d0\n", gchr[n]);
        } else if(lang == ROREDA) {
        fprintf(redfp, "\n* Gauge parametes (default is Feynman gauge)\n");
        for(n = 1; n < 5; n++)
            fprintf(redfp, "      igau%cb = %d\n", gchr[n],n);
        for(n = 1; n < 5; n++)
            fprintf(redfp, "      agauge(igau%cb) = 1.0d0\n", gchr[n]);
        fprintf(redfp, "      agauge(igau00) = 1.0d0\n", gchr[n]);
        }
        fprintf(redfp, "\n* Spin average\n");
        fprintf(redfp, "      aspin = 1.0d0\n");
        for(n = 0; n < nextern; n++) {
            pa  = ndname(n);
            p   = abs(pa);
            il  = ndiline(n, 0);
            tp  = iltable[il].iltype;
            Discard cmindx(momstr, tp, p);
    
            fprintf(redfp, "\n");
            fprintf(redfp, "*     %d: %s %s mass=%s\n",
                    n+1, eliorf(ndiorf(n)), pvename(ndiorf(n),pa),
                    ptable[p]->mass);
    
            Discard cmindx(momstr1, tp, p);
            fprintf(redfp, "      jhs(%d) = 0\n", n+1);
            fprintf(redfp, "      jhe(%d) = %s - 1\n", n+1, momstr);
            if(ndiorf(n) == ELINITIAL) {
                fprintf(redfp, "      aspin = aspin/");
                fprintf(redfp, "dble(jhe(%d) - jhs(%d)+1)\n",
                        n+1, n+1);
            }
        }
        if(cpol) {
            fprintf(redfp, "\n* Flag of cyclic polarization\n");
            for(n = 0; n < ndnodes(); n++) {
                if(ndtype(n) != NDEXTERN)
                    continue;
                if(cmphoton(n))
                    fprintf(redfp, "      jcpol(%d) = 0\n", n+1);
            }
        }
    } else {
        fprintf(redfp, "\n* Spin average\n");
        fprintf(redfp, "      aspin = 1.0d0");
        for(n = 0; n < nextern; n++) {
            nd  = ndnode(n);
            if(nd->nvkind == ELINITIAL) {
                p = abs(ndname(n));
                fprintf(redfp, "/%d.0d0", pvsdeg(p));
            }
        }
    }
*/
    fprintf(redfp, "\n");
}
