/********************************************************************
 *  File : cmgrf.c
 *
 *  Common part of source code generation.
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "common.h"
#include "config.h"
#include "dbgh.h"
#include "ptable.h"
#include "grdata.h"
#include "gensc.h"
#include "version.h"

/*------------------------------------------------------------------*/
/* Entry points */
Public  Void    cmsepar    ARG((int));
Public  Void    cmopen     ARG((int,string));
Public  int     cmdfcpl    ARG((int,int,int,int));
Public  bool    cmcoupl    ARG((vindex,int,int,string));
Public  int     cmttype    ARG((int));
Public  Void    cmmom      ARG((string,int,int,bool));
Public  Void    cmimom     ARG((string,int,int));
Public  int     cmindx     ARG((string,int,pindex));
Public  int     cmvector   ARG((pindex));
Public  bool    cmitype    ARG((int,int*,int*));
Public  bool    cmphoton   ARG((node));
Public  Void    cmmkrnw    ARG((NoArg));
Public  unsigned int  cmntype    ARG((node));

/*------------------------------------------------------------------*/
/* External references */
/* path.c */
Extern  string  pnfname    ARG((string,string,string,string));

/*------------------------------------------------------------------*/
/* Private functions */
Private int     cmndf      ARG((int));

/*------------------------------------------------------------------*/
/* Local variable */

/*==================================================================*/
Public  Void    cmsepar(c)
int     c;
/*------------------------------------------------------------------*
 */
{
    int l;

    putc('*', redfp);
    for(l = 0; l < 71; l++)
        putc(c, redfp);
    putc(EOL, redfp);
}

/*==================================================================*/
Public  Void    cmopen(progt, fname)
int     progt;
string  fname;
/*------------------------------------------------------------------*
 * Open file
 */
{
    static bool ost=False;
    string cb, ce, ca = NULL;
    int i;

    switch(progt) {
        case ROFORM :
        case ROFORM1 :
            ca = "#-";
            cb = "*";
            ce = "*";
            break;
        case ROFORT :
            cb = "*";
            ce = "*";
            break;
        case ROC :
            cb = "/*";
            ce = "*/";
            break;
        case ROMAKE :
            cb = "#";
            ce = "#";
            break;
        case ROSH :
            cb = "#";
            ce = "#";
            ca = "#!/bin/sh";
            break;
        case RONQS :
            cb = "##";
            ce = "##";
            ca = "##";
            break;
        case RORED :
            ca = "%";
            cb = "%";
            ce = "%";
            break;
        default :
            ca = "unknown language";
            cb = "<??";
            ce = "??>";
            break;
    }
    if(ost) {
        fclose(redfp);
    }

    if((redfp = fopen(fname, "w")) == NULL) {
        fprintf(Stderr, "reopen : cannot open form file : %s\n",
                fname);
        exit(1);
    }

    ost = True;

    if(ca != NULL) {
        fprintf(redfp, "%s\n", ca);
    }
    fprintf(redfp, "%s File %s ", cb, fname);
    fprintf(redfp, "generated by GRACE ");
    if(GRCVER2 == 0) {
        fprintf(redfp, "Ver.%2d.%02d",GRCVER0, GRCVER1);
    } else {
        fprintf(redfp, "Ver.%2d.%02d(%02d)",GRCVER0, GRCVER1, GRCVER2);
    }
    fprintf(redfp, "%+24s\n", GRCDATE);
    fprintf(redfp, "%s \n", ce);
    fprintf(redfp, "%s          ", cb);
    if(lang == ROFORT)
        fprintf(redfp, "Fortran");
    else if(lang == ROFORM || lang == ROFORM1)
        fprintf(redfp, "Form");
    else if(lang == RORED )
        fprintf(redfp, "REDUCE");
    else if(lang == ROREDA)
        fprintf(redfp, "REDUCE (for Amplitude)");
    fprintf(redfp, " source code generator\n");
    fprintf(redfp, "%s", ce);
    fprintf(redfp, "     (c)copyright 1990-1996 Minami-Tateya Group,");
    fprintf(redfp, " Japan\n");
    fprintf(redfp, "%s", ce);
    for(i=1;i<72;i++) fprintf(redfp, "-"); fprintf(redfp, "\n");
}

/*==================================================================*/
Public  int     cmdfcpl(sp, nl, lp, ct)
int     sp;
int     lp;
int     nl;
int     ct;
/*------------------------------------------------------------------*
 *  The number of array elements of coupling constant
 *
 *  Arguments
 *    sp : see cmttype
 *    nl : the number of legs
 *    lp : loop
 *    ct : V_NOTYPE, V_V, V_A, V_VminusA or V_VplusA
 *------------------------------------------------------------------*/
{
    if(lp == 0) {
        if(sp == 1) {  /* fermion */
            if(ct == V_NOTYPE || lang == ROFORT)
/*
            if(ct == V_NOTYPE)
*/
                return 2;
            else
                return 0;
        } else {
            return 0;
        }
    } else {
        if(sp == 1) {  /* fermion */
            if(nl == 2)    /* ff */
                return 4;
            else
                return 2;
        } else if(sp == 2) {  /* with ghost */
            return 2;
        } else if(sp == 3) {
            return 2;
        } else {
            return 0;
        }
    }
}

/*==================================================================*/
Public  bool    cmcoupl(v, ind, lp, str)
vindex  v;
int     ind;
int     lp;
string  str;
/*------------------------------------------------------------------*
 *  Construct the name and arguments of a vertex.
 *
 *  Args:
 *    v    :  input : vertex in the vertex table.
 *    ind  :  input : 
 *    lp   :  input : loop of the vertex
 *    str  : output : 
 *------------------------------------------------------------------*/
{
    int  out1, out2, ct, st, df;
    char name[16];

    st = True;
    ct = vtable[v]->ctype;

    if(lp == 0) {
        sprintf(name, "c%s", vtable[v]->fname);
    } else {
        sprintf(name, "c%d%s", lp, vtable[v]->fname);
    }

    if(ind == 0) {
        sprintf(str, "%s", name);

    /* ind = -2 : zxxx */
    } else if(ind == -2) {
        sprintf(str, "z%s", name+1);

    /* ind = -1 : cxxx(n) */
    } else if(ind == -1) {
        df = cmdfcpl(cmttype(v), vtable[v]->nvlegs, lp, ct);
        if(df < 2)
            sprintf(str, "%-6s   ", name);
        else
            sprintf(str, "%-6s(%d)", name, df);

    } else if(lp > 0 || ct == V_NOTYPE) {
        if(lang == ROFORM)
            sprintf(str, "%s(%d)", name, ind);
        else if(lang == ROFORM1)
            sprintf(str, "[%s(%d)]", name, ind);
        else
            sprintf(str, "%s(h%d)", name, ind);

    } else if(ct == V_V) {
        sprintf(str, "%s", name);
        if(ind != 1) {
            st = False;
        }

    } else if(ct == V_A) {
        if(ind == 1) {
            sprintf(str, "-%s", name);
            st = False;
        } else {
            sprintf(str, "%s", name);
        }
    } else if(ct == V_VminusA) {
        if(ind == 1) {
            sprintf(str, "%s", name);
        } else {
            sprintf(str, "0");
            st = False;
        }
    } else if(ct == V_VplusA) {
        if(ind == 1) {
            sprintf(str, "0");
            st = False;
        } else {
            sprintf(str, "%s", name);
        }
    }

    return st;
}

/*=================================================================*/
Public  int     cmttype(n)
int     n;
/*------------------------------------------------------------------*
 * returns the type of vertex n of vertex table in the defined model.
 * return value
 *     -1 : particle with illegal spin
 *      0 : without fermion nor ghost
 *      1 : with fermion
 *      2 : with ghost
 *      3 : vector self-energy
 *------------------------------------------------------------------*/
{
    int  s, lg, ng, nf, p;
    unsigned int rt;

    if(vtable[n]->nvlegs == 2) { /* self-energy part */
        p = abs(vtable[n]->vlegs[0]);
        s = ptable[p]->ptype;
        if(s == P_VECTOR)
            return 3;
    }

    ng = nf = 0;
    for(lg = 0; lg < vtable[n]->nvlegs; lg++) {
        p = abs(vtable[n]->vlegs[lg]);
        s = ptable[p]->ptype;

        if(s == P_UNDEF)
            return -1;
        else if(s == P_FERMION)
            nf++;
        else if(s == P_MAJORANA)        /* ??? */
            nf++;
        else if(s == P_GHOST)
            ng++; 
    }

    if(ng > 0)
        return 2;
    else if(nf > 0)
        return 1;
    else
        return 0;
}

/*=================================================================*/
Public  Void    cmmom(str, n, leg, sgn)
string  str;
node    n;
int     leg;
bool    sgn;
/*------------------------------------------------------------------*
 * Construct momentum which comes to a node.
 *------------------------------------------------------------------*/
{
    int  il, s;

    if(ndtype(n) == NDONSHELL)
        leg = 0;

    il = ndiline(n, leg);

    if(sgn && iltable[il].ilnodes[0] == n)
        s = -1;
    else
        s = 1;

    cmimom(str, il, s);
}

/*=================================================================*/
Public  Void    cmimom(str, il, sgn)
string  str;
int     il;
int     sgn;
/*------------------------------------------------------------------*
 * consturct the name of an internal line
 *------------------------------------------------------------------*/
{
    char c[3];
    int  itype;
    int  im, ii;

#if DBG
    if(dbg >= redbg)
        printf("cmimom(%p, %d, %d)\n", str, il, sgn);
#endif

    itype = iltable[il].iltype;

    switch (itype) {
        case ILEXTERN : 
            if(lang == ROFORT || lang == ROREDA)
                strcpy(c, "pe");
            else
                strcpy(c, "p");
            im = il;
            for(ii = 1; ii < nilines; ii++) {
                if(iltable[il].ilmlist[ii] != 0) {
                    im = ii;
                    break;
                }
            }
            break;
        case ILLOOP :
            if(lang == ROFORT)
                strcpy(c, "ke");
            else
                strcpy(c, "k");
            im = il + momoffs;
            break;
        case ILDLOOP :
            if(lang == ROFORT)
                strcpy(c, "le");
            else
                strcpy(c, "l");
            im = il + momoffs;
            break;
        case ILECONJG :
            if(lang == ROFORT)
                strcpy(c, "pc");
            else
                strcpy(c, "p");
            im = il + momoffs;
            break;
        case ILTREE :
            if(lang == ROFORT)
                strcpy(c, "pe");
            else
                strcpy(c, "q");
            im = il + momoffs;
            break;
        default :
            if(lang == ROFORT)
                strcpy(c, "??");
            else
                strcpy(c, "?");
            im = il + momoffs;
            break;
    }

    if(sgn < 0)
        sprintf(str, "-%s%d\0", c, im);
    else
        sprintf(str, "%s%d\0", c, im);

    return;
}

/*=================================================================*/
Public  int     cmindx(str, typ, prt)
string  str;
int     typ;
pindex  prt;
/*------------------------------------------------------------------*
 * Construct momentum which comes to a node.
 *------------------------------------------------------------------*/
{
    int ptype, pcode, c;

    ptype = ptable[prt]->ptype;
    switch(ptype) {
        case P_SCALAR :
            strcpy(str, "lscalr");
            return 1;
        case P_MAJORANA :
            if(typ == ILEXTERN) {
                strcpy(str, "lextrn");
                return 2;
            } else {
                strcpy(str, "lintrn");
                return 4;
            }
        case P_FERMION :
            if(typ == ILEXTERN) {
                strcpy(str, "lextrn");
                return 2;
            } else {
                strcpy(str, "lintrn");
                return 4;
            }
        case P_VECTOR :
            c = cmvector(prt);
            if(typ == ILEXTERN) {
                sprintf(str, "lepex%c", c);
                if(c == 'a')
                    return 2;
                 else
                    return 3;
            } else {
                sprintf(str, "lepin%c", c);
                return 4;
            }
        default :
            strcpy(str, "l????");
            return 0;
    }
}

/*=================================================================*/
Public  int     cmvector(prt)
pindex  prt;
/*------------------------------------------------------------------*
 * Which vector boson?
 *------------------------------------------------------------------*/
{
    switch (ptable[prt]->pcode) {
        case 1:
            return 'a';
        case 2:
            return 'w';
        case 4:
            return 'z';
        case 8:
            return 'g';
        case 31:
            return 'h';
        case 41:
            return '3';
        case 42 :
            return 'x';
        default:
            return '?';
    }
}

/*==================================================================*/
Public  bool    cmitype(il, fl1, fl2)
int     il;
int     *fl1, *fl2;
/*------------------------------------------------------------------*
 * Analyze type of an internal line.
 * It returns on a loop or not.
 *
 * Arguments:
 *    il  : internal line
 *    fl1, fl2 : output : fermion line number of nodes on both sides.
 *                        They are -1 for external particle,
 *                        and -2 for bosonic vertex.
 */
{
    int it;

    *fl1 = cmndf(iltable[il].ilnodes[0]);
    *fl2 = cmndf(iltable[il].ilnodes[1]);
    it = iltable[il].iltype;
    return (it == ILLOOP || it == ILDLOOP);
}

/*==================================================================*/
Private int     cmndf(nd)
int     nd;
/*------------------------------------------------------------------*
 * Returns the number of fermion line on which the node is located.
 * If the node is external particle, return value is -1 and
 * is -2 if the node is bosonic vertex.
 *
 * Arguments:
 *    nd : node.
 */
{
    int f, n;

    if(ndtype(nd) == NDEXTERN || ndtype(nd) == NDONSHELL)
        return -1;
    for(f = 0; f < nflines; f++)
        for(n = 0; n < fltable[f].length; n++)
            if(nd == fltable[f].ndlist[n])
                return f;
    return -2;
}

/*=================================================================*/
Public  unsigned int  cmntype(n)
node    n;
/*------------------------------------------------------------------*
 * return value
 *     -1 : particle with illegal spin
 *      0 : without fermion nor ghost
 *      1 : with fermion
 *      2 : with ghost
 *------------------------------------------------------------------*/
{
    int  lg, s;
    unsigned int rt;

    rt = 0;
    for(lg = 0; lg < ndlegs(n); lg++) {
        s = ptable[abs(ndparticle(n, lg))]->ptype;
        if(s == P_UNDEF)
            return -1;
        rt |= pmask[s];
    }

    return rt;
}

/*==================================================================*/
Public  bool    cmphoton(n)
node    n;
/*------------------------------------------------------------------*
 * check external particle "n" is whether massless vector or not.
 */
{
    pindex p;

    p   = abs(ndparticle(n, 0));
    if(ptable[p]->ptype != P_VECTOR)
        return False;
    return ptable[p]->massless;
}
/*==================================================================*/
Public  Void    cmmkrnw(NoArg)
/*------------------------------------------------------------------*
 * generate entry "renew" to the Makefile.
 */
{
    fprintf(redfp,"\nrenew: clean\n");
    fprintf(redfp,"\trm -rf *.f *.red *.frm *.h *.c out2f ");
    fprintf(redfp,"*.mk *.sh *.dat form.set *.fout \n");
}
