/***************************************************************
 *  File : ftgrf.c  prefix=ft
 *
 *  Functions for generating fortran source code.
 *
 *                                                 by Kaneko Toshiaki
 *
 *  1995/08/29 TI: Color Base;"CF"
 *  1995/07/13 TI:Supress generation of "smcpol.f".
 *             (add this subrouine in CHANEL library) 
 *  1995/05/30 add "cmfin.c".  See this file in detail
 *             ftincl should be called before cmbases
 *             ftmakes should be called after cmbases
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "common.h"
#include "dbgh.h"
#include "ptable.h"
#include "grdata.h"
#include "glue4.h"
#include "putidh.h"
#include "config.h"
#include "gensc.h"
#include "cfbase.h"

/* Entry points */
Public  Void    ftprocs  ARG((int*,int*));
Public  Void    ftinit   ARG((NoArg));
Public  Void    ftgraph  ARG((int,tproc*,tgraph*));
Public  Void    ftmain   ARG((int,int));
Public  Void    ftvldmp  ARG((NoArg));

/* External references */
/* dagrf.c */
Extern  Void    daclose      ARG((NoArg));
Extern  tproc  *dagetp       ARG((int));
Extern  tgraph *daread       ARG((int, int));
Extern  tgraph *dareadf      ARG((int, int));
Extern  tgraph *damult       ARG((int,int,int));
/* ergrf.c */
Extern  Void    error        ARG((string, string));
/* flgrf.c */
Extern  Void    flget        ARG((tgraph*));
Extern  bool    flmajo       ARG((NoArg));
Extern  bool    fldump       ARG((NoArg));
/* ilgrf.c */
Extern  Void    ildump     ARG((NoArg));
Extern  Void    iliconsv   ARG((NoArg));
Extern  Void    ilmom      ARG((NoArg));
/* rdptbl.c */
Extern  int     pvboson    ARG((pindex));
Extern  string  pvename    ARG((int,pindex));
Extern  int     pvgl3      ARG((node));
Extern  int     pvvnfind        ARG((int));
/* cmgrf.c */
Extern  Void    cmbases    ARG((NoArg));
Extern  Void    cmopen     ARG((int,string));
Extern  Void    cmsepar    ARG((int));
Public  int     cmdfcpl    ARG((int,int,int,int));
Extern  bool    cmcoupl    ARG((vindex,int,int,string));
Extern  Void    cmimom     ARG((string,int,int));
Extern  Void    cmmom      ARG((string,int,int,bool));
Extern  int     cmindx     ARG((string,int,pindex));
Extern  int     cmvector   ARG((pindex));
Extern  Void    cminclk    ARG((NoArg));
Extern  bool    cmphoton   ARG((nindex));
Extern  Void    cmmkrnw    ARG((NoArg));
/* cmfin.c */
Extern  Void    cmfobj     ARG((FILE*));
/* cfgrf.c */
Extern  Void    cfactor    ARG((long*, long*));

/* CF:begin */
/* cfconv.c */
Extern  Void    cfcbase      ARG((cftbase*));
Extern  int     cfcproj      ARG((cftbase*, cftgraph*));
Extern  Void    cfcbfree     ARG((cftbase*));
Extern  Void    cfcgfree     ARG((cftgraph*));
/* CF:end */

/* ftmakes_$(MACHINE).c */
Extern  Void    ftmakes  ARG((NoArg));

/* local variables */
Local FILE   *tfp1;

/* the current graph */
Local int    cnprc;
Local int    cngrf;

#define  MAXSSTR  256
Local char   sizestr[MAXSSTR];

Local int    reptype[PTYPEMAX];

#define NVARS  100
#define FT_INT     1
#define FT_REAL    2
#define FT_COMPLEX 3
#define FT_MAX     4
Local int vartype[FT_MAX];

#define MAXVLTBL    (MAXNODES*2)

Local struct {          /* table of local variables in fortran */
        int    vtype;
        string name;
        string dim;
        int    size;
    } vartbl[NVARS];
Local int nvars = 0;
Local int incl2r = 0;
Local int incl2i = 0;

typedef struct {        /* leg of a node */
        pindex p;       /* particle */
        int    t;       /* type of i-line */
        int    i;       /* internal line */
        int    m;       /* momentum number */
        int    d;       /* direction of momentum */
    } vlegt;

Local struct {
        vlegt *vl;
        int   size;
    } vltbl[MAXVLTBL];
Local int nvltbl = 0, nvlusd = 0;

#define csign(l)   ((l)<0 ? '-':'+')

#define EOFSYM     '\0'

Local   bool majorana;
/* S-chanel graph(1) or not (0) */
Local   bool isschan;
#define TCUT    (100)

/* local functions */
Private Void    ftgtree  ARG((NoArg));
Private Void    ftdenom  ARG((NoArg));
Private Void    ftincl   ARG((NoArg));
Private Void    ftampord ARG((NoArg));
Private Void    ftsmcpol ARG((NoArg));
Private Void    ftampsum ARG((NoArg));
Private Void    ftampsmc ARG((NoArg));
Private int     ftgmask  ARG((NoArg));

Private bool    ftmleg   ARG((node,vlegt*,int*,int*));
Private Void    ftmffv   ARG((node,vlegt*));
Private Void    ftmffs   ARG((node,vlegt*));
Private Void    ftmmfv   ARG((node,vlegt*));
Private Void    ftmmfs   ARG((node,vlegt*));
Private Void    ftmmmv   ARG((node,vlegt*));
Private Void    ftmmms   ARG((node,vlegt*));
Private Void    ftnffv   ARG((string,node,vlegt*));
Private Void    ftnffs   ARG((string,node,vlegt*));
Private Void    ftnvvv   ARG((node,vlegt*));
Private Void    ftnsvv   ARG((node,vlegt*));
Private Void    ftnssv   ARG((node,vlegt*));
Private Void    ftnsss   ARG((node,vlegt*));
Private Void    ftnvvvv  ARG((node,vlegt*));
Private Void    ftnssvv  ARG((node,vlegt*));
Private Void    ftnssss  ARG((node,vlegt*));
Private Void    ftvlts   ARG((NoArg));
Private Void    ftvlte   ARG((NoArg));
Private Void    ftvlcp   ARG((vlegt*,vlegt*,int,int));
Private Void    ftvlex   ARG((vlegt*,int,int));
Private Void    ftclpord ARG((NoArg));
Private Void    ftdprop  ARG((NoArg));
Private Void    ftvars   ARG((int));

Private Void    fthindx  ARG((node,vlegt*,int));
Private Void    ftncoupl ARG((node));
Private Void    ftmassl  ARG((node,vlegt*,int,int));
Private Void    ftcrans  ARG((node,vlegt*,int));
Private Void    ftmlist  ARG((node,vlegt*,int));

Private Void    ftbegin  ARG((NoArg));
Private Void    ftmconsv ARG((NoArg));
Private Void    ftnode   ARG((NoArg));
Private Void    ftconn   ARG((NoArg));
Private Void    ftiline  ARG((NoArg));
Private int     ftgaug   ARG((string,pindex));
Private Void    ftgend   ARG((NoArg));

Private Void    fttmpopen ARG((NoArg));
Private Void    fttmpcp   ARG((NoArg));
Private Void    fttmpclr  ARG((NoArg));
Private Void    fttmp     ARG((int,string,string,int));
Private Void    fttmdmp   ARG((NoArg));
Private Void    ftincw    ARG((NoArg));
Private Void    ftdmom    ARG((string,int,int));
Private unsigned int ftvtype  ARG((string,int,int*,int));
Private unsigned int ftnvtype ARG((string,node));
Private Void    ftscfact  ARG((NoArg));
Private Void    ftacfact  ARG((NoArg));

Private Void    ftoright  ARG((string));

/* Number of Internal Higss */
Private int     ftishiggs ARG((NoArg));
/* Number of Internal Gluon */
Private int     ftisgluon ARG((NoArg));
/* Check S-chanel or not */
Private bool    ftisschan ARG((NoArg));

#define fparm(y, nx)  (((y)+1)%(nx)+1)
#define ilpartl(l,lg)     \
        ndparticle(iltable[l].ilnodes[lg],iltable[l].illegs[lg])
#define ilpart(l)     abs(ilpartl(l, 0))
#define ndconj(n)     ((n) >= conjstart)

/*********************************************************************
 *    1. Common part.
 */
/*==================================================================*/
Public  Void    ftprocs(proct, procg)
int     proct[], procg[];
/*------------------------------------------------------------------*
 * 
 *
 * Arguments
 *    proct  : the table of process numbers.
 *             proct[l] is the process number of l-loop graphs.
 *    procg  : the table of the number of graphs in each process.
 *    dvprt  : include divergent parts or not.
 *------------------------------------------------------------------*/
{
    int       lp, g0;
    int       l0;
    char      prefix;
    long      ngllp;
/* CF */
    int       csgn;
    int       ibase = 0, iprt = 0;

    lang    = ROFORT;

    printf("vcode = %d ",vcode);

    if(maxloop > 0) {
        fprintf(Stderr, "*** This version supports only tree graphs\n");
        fprintf(Stderr, "    Current loop = %d\n", maxloop);
        exit(1);
    }

/* CF */
    if( ( colgr = (cftgraph *) malloc(sizeof(cftgraph))) == NULL ) {
         fprintf(Stderr, "*** colgr: memory allocation failed\n");
         exit(1);
    }
    if( ( cbase = (cftbase *) malloc(sizeof(cftbase))) == NULL ) {
         fprintf(Stderr, "*** cbase: memory allocation failed\n");
         exit(1);
    }

    for(l0 = 0; l0 <= maxloop; l0++) {
        reproct[l0] = proct[l0];
        reprocg[l0] = procg[l0];
    }

    /* names of graphs are:
     *   a1, a2, ...
     */
    prefix = 'a';

    incl2r = 0;
    incl2i = 0;

    /* repeat by loop */
    for(lp = 0; lp <= maxloop; lp++) {

        /* generate processes l0 x l1, l0 >= l1 */
        /* for(l0 = lp0min(lp); l0 <= lp0max(lp); l0++) { */ {
            l0 = lp0min(lp);

            /* first process */
            cnprc = proct[l0];       /* process number */
            cprocess = dagetp(cnprc);   /* process data */

            printf("%% +++ loop = %d (%d) +++\n", l0, maxloop);

            /* repeat graph with loop lp */
            for(g0 = 0; g0 < procg[l0]; g0++) {

                cngrf  = g0;
                cgraph = daread(cnprc, cngrf);

                /* initialize i-line table */
                ilmom();
                flget(cgraph);
                isschan = ftisschan();

                /* construct fermion lines */
                majorana = flmajo();

                /* the first graph */
                if(cngrf == 0) {

/* CF:begin */
                    if( colorv > 0 ) {
                      cfcbase(cbase);
                      cfbase(cbase,ibase,iprt);
#if DBG
                      cfbdump(cbase);
#endif
                      if( cbase->nbase == 1 && cbase->nexcolor == 0) {
                          printf("No Color Particle\n");
                      }
                    }
                    /* the first process */
                    if(lp == 0 && l0 == lp0min(lp)) {

                        /* initialization */
                        ftinit();
                    }
/* CF:end */
                }

                printf("graph %d", cngrf+1);
#if DBG
                if(dbg > 0) {
                    ildump();
                }
#endif

                if( colorv == 1 ) {

                if( !(cbase->nbase == 1 && cbase->nexcolor == 0) ) {

                csgn = cfcproj(cbase,colgr);

                cfproj(cbase, colgr, ibase, iprt);

                }

#if DBG
/*                cfgdump(colgr);*/
#endif
                }

                if( isschan ) 
                    printf(" S-chan ");
                else
                    printf(" NON-S-chan ");
                /* generate fortran code for a graph */
                  ftgraph(prefix, cprocess, cgraph);

                if( colorv == 1 ) {
                    if( !(cbase->nbase == 1 && cbase->nexcolor == 0) ) {
                        cfcgfree(colgr);
                    }
                }
                printf("\n");
            }

            ftmain(prefix, procg[l0]);

            if( colorv == 1 ) {
                if( !(cbase->nbase == 1 && cbase->nexcolor == 0) )
                     cfcbfree(cbase);
            }
            /* print message */
            prefix++;
        }
    }

    ftincw();

    /* create makefiles. */
    cprocess = dagetp(proct[0]);
    ftmakes();
/* CF */
    if( colorv == 0 ) {
        if(maxloop > 0 || cprocess->extcol > 0) {
           if(cprocess->qcd > 0)
              ftscfact();
           else
              ftacfact();
        }
    }

    daclose();
}

/*==================================================================*/
Public  Void    ftinit(NoArg)
/*------------------------------------------------------------------*
 * Initialize and generate fortran codes.
 *
 * Arguments:
 *    proc  : the process in which the graph blongs.
 *    graph : the graph to be converted.
 *------------------------------------------------------------------*/
{
    string p;
    int i;

    /* setup variables */
    nextern   = cprocess->oextern;
    nilines   = cgraph->nilines;
    conjstart = cgraph->conjsrt;

    momoffs = 0;
    legoffs = 0;

    /* create include files for fortran code */
    ftincl();

    /* userin and usrout */
    cmbases();

    ftampord();
/* CF */

    if( colorv == 0 ) {
        ftampsum();
    } else {
        ftampsmc();
    }

/*
    if(cpol)
      ftsmcpol();
*/
    
}

/*==================================================================*/
Private Void    ftincl(NoArg)
/*------------------------------------------------------------------*
 * generate "incl1.f"
 *------------------------------------------------------------------*/
{
    int p, s, l, i, n0, n1, j, nc, k, lp, lag, il, lg, tp;
    string m;
    node   nd;
    int    df,sizecp;

    cmopen(ROFORT, "incl1.f");

    fprintf(redfp,"      parameter (loutgo =  2, lincom =  1)\n");
    fprintf(redfp,"      parameter (lantip = -1, lprtcl =  1)\n");
    fprintf(redfp,"      parameter (lscalr =  1)\n");
    fprintf(redfp,"      parameter (lepexa =  2, lepexw =  3,");
    fprintf(redfp," lepexz =  3, lepexg =  2)\n");
    fprintf(redfp,"      parameter (lepina =  4, lepinw =  4,");
    fprintf(redfp," lepinz =  4, leping =  3)\n");
    fprintf(redfp,"      parameter (lextrn =  2, lintrn =  4)\n");

    fprintf(redfp,"\n* table of amplitudes\n");
    fprintf(redfp,"      parameter (ngraph =%d)\n", reprocg[0]);
    fprintf(redfp,"      parameter (nextn  =%d)\n", nextern);
/* CF */
    if( colorv == 1 ) {
      printf("nexcolor %d \n",cbase->nexcolor);
/*      if( cbase->nexcolor != 1 ) cbase->nexcolor=0;*/
      if( cbase->nexcolor == 1 ) {
          fprintf(redfp,"\n* number of color base\n");
          fprintf(redfp,"      parameter (ncbase =%d)\n", cbase->nbase);
      }
    }

    lag = 1;
    for(il = 1; il < nilines; il++) {
        if(iltable[il].iltype != ILEXTERN)
            continue;
        nd  = iltable[il].ilnodes[0];
        lg  = iltable[il].illegs[0];
        p   = abs(ndparticle(nd, lg));
        tp  = iltable[il].iltype;
        lag *= cmindx(momstr1, tp, p);
    }

    fprintf(redfp,"\n* number of all helicity states\n");
    fprintf(redfp,"      parameter (lag    =%d)\n", lag); /* ??? */
    fprintf(redfp,"      parameter (ngrpsq = ngraph*ngraph)\n", lag);
    fprintf(redfp,"      common /amslct/jselg(ngraph), jgraph, ");
    fprintf(redfp,"jhiggs,jgluon\n");
    fprintf(redfp,"\n");
/* CF */
    if( colorv == 0 ) {
        if( !vcode )
            fprintf(redfp,"      common /amgrph/ ag(0:lag-1,ngraph), ");
        else {
            fprintf(redfp,"      common /amgrph/ ag(nsize,0:lag-1,");
            fprintf(redfp,"ngraph), ");
        }
        if( !vcode )
            fprintf(redfp,"aprop(ngraph),\n");
        else
            fprintf(redfp,"aprop(nsize,ngraph),\n");
        if( !vcode )
            fprintf(redfp,"     &               ancp(ngraph), ");
        else
            fprintf(redfp,"     &               ancp(nsize,ngraph), ");
        if( !vcode )
            fprintf(redfp,"ansp(0:ngraph),\n");
        else
            fprintf(redfp,"ansp(nsize,0:ngraph)\n");
        fprintf(redfp,"     &               igraph(ngraph)");
        if(cprocess->extcol > 0)
           fprintf(redfp,",cf(ngraph,ngraph)");
        fprintf(redfp,"\n");

        fprintf(redfp,"      complex*16 ag, aprop\n");
        fprintf(redfp,"\n");
    } else if( colorv == 1 ) {
        if( !vcode )
            fprintf(redfp,"      common /amgrph/ agcwrk(0:lag-1),");
        else {
            fprintf(redfp,"      common /amgrph/ agcwrk(nsize,");
            fprintf(redfp,"0:lag-1),");
        }
        if( cbase->nbase > 1 ) {
          if( !vcode )
              fprintf(redfp,"agc(0:lag-1,0:ncbase-1),\n");
          else {
              fprintf(redfp,"\n     &              ");
              fprintf(redfp,"agc(nsize,0:lag-1,0:ncbase-1),\n");
          }
        } else {
          if( !vcode )
              fprintf(redfp,"agc(0:lag-1),\n");
          else
              fprintf(redfp,"agc(nsize,0:lag-1),\n");
        }
        if( !vcode )
            fprintf(redfp,"     &              aprop,");
        else
            fprintf(redfp,"     &              aprop(nsize),");
        if( !vcode )
            fprintf(redfp,"ancp(ngraph),ansp(0:ngraph)");
        else
            fprintf(redfp,"ancp(nsize,ngraph),ansp(nsize,0:ngraph)");
        if( cbase->nbase > 1 ) {
            fprintf(redfp,"\n     &             ");
            fprintf(redfp,",cfmtx(0:ncbase-1,0:ncbase-1)\n");
        } else {
            if( cbase->nexcolor == 1 ) {
               fprintf(redfp,"\n     &             ");
               fprintf(redfp,",cfmtx\n");
/*            } else if( cbase->nexcolor == 0 ) {*/
            } else {
               fprintf(redfp,"\n");
            }
        }
        fprintf(redfp,"      common /amgrpi/igraph(ngraph)\n");
        fprintf(redfp,"      complex*16 agc, agcwrk, aprop\n");
    }

    /* common /amcplc/... */
    for(i = 0; i < 2; i++) {
        if(i == 0)
            puinit(redfp,"      common /amcplc/",
                         "     &               ",
                         ",", ",", "", 72, 16);
        else
            puinit(redfp,"      complex*16     ",
                         "     &               ",
                         ",", ",", "", 72, 16);
        putid("     cqed");
        putid("     cqcd");
        for(lp = 0; lp < 1; lp++) {
            for(p = nvstart; p < nvtable; p++) {
                if(vtable[p]->loop > lp)
                    continue;    
                if(p != pvvnfind(p))
                    continue;
                Discard cmcoupl(p, -i, lp, momstr);
                sprintf(momstr3, "%9s", momstr);
                ftoright(momstr3);
                putid(momstr3);
            }
        }
        putend();
        putc(EOL, redfp);
    }
/* TI : count size of amcplc */
    sizecp = 2;
    lp = 0;
    for(p = nvstart; p < nvtable; p++) {
        if(vtable[p]->loop > lp)
            continue;    
        if(p != pvvnfind(p))
            continue;
        df = cmdfcpl(cmttype(p),vtable[p]->nvlegs,0,vtable[p]->ctype);
        if( df < 1 ) {
            sizecp++;
        } else {
            sizecp+=df;
        }
/*
        Discard cmcoupl(p, -1, lp, momstr);
        printf("%s %d%c",momstr,
        cmdfcpl(cmttype(p),vtable[p]->nvlegs,0,vtable[p]->ctype),
        (((p+1) % 5  == 0)?'\n':' '));
*/
    }
    printf("---- size of amcplc   %d \n",sizecp);

    /* common /amextr/... */
    fprintf(redfp,"\n");
    fprintf(redfp,"* Momenta of external particles\n");

    puinit(redfp,"      common /amextr/",
                  "     &               ",
                  ",", ",", "", 72, 16);
    if( !vcode ) {
        for(n0 = 1; n0 <= nextern; n0++) {
            sprintf(momstr,"pe%d(4)", n0);
            putid(momstr);
        }
        putid("prod(nextn, nextn)");
    } else {
        for(n0 = 1; n0 <= nextern; n0++) {
            sprintf(momstr,"pe%d(nsize,4)", n0);
            putid(momstr);
        }
        putid("prod(nsize, nextn, nextn)");
    }
    putend();
    putc(EOL, redfp);

    puinit(redfp,"      common /smextp/",
                  "     &               ",
                  ",", ",", "", 72, 16);

    for(nd = 0; nd < ndnodes(); nd++) {
        if(ndtype(nd) != NDEXTERN)
            continue;
        p = abs(ndparticle(nd, 0));
        if(ptable[p]->ptype == P_VECTOR) {
            sprintf(momstr, "ep%d", nd+1);
            putid(momstr);
            sprintf(momstr, "ew%d", nd+1);
            putid(momstr);
        } else if(ptable[p]->ptype == P_FERMION) {
            sprintf(momstr, "ps%d", nd+1);
            putid(momstr);
            sprintf(momstr, "ew%d", nd+1);
            putid(momstr);
            sprintf(momstr, "ce%d", nd+1);
            putid(momstr);
        } else if(ptable[p]->ptype == P_MAJORANA) {
            sprintf(momstr, "ps%d%s", nd+1, ftsuorv[1]);
            putid(momstr);
            sprintf(momstr, "ew%d%s", nd+1, ftsuorv[1]);
            putid(momstr);
            sprintf(momstr, "ce%d%s", nd+1, ftsuorv[1]);
            putid(momstr);

            sprintf(momstr, "ps%d%s", nd+1, ftsuorv[2]);
            putid(momstr);
            sprintf(momstr, "ew%d%s", nd+1, ftsuorv[2]);
            putid(momstr);
            sprintf(momstr, "ce%d%s", nd+1, ftsuorv[2]);
            putid(momstr);
        }
    }
    putend();
    fprintf(redfp,"\n");
    for(nd = 0; nd < ndnodes(); nd++) {
        if(ndtype(nd) != NDEXTERN)
            continue;
        p = abs(ndparticle(nd, 0));
        il  = ndiline(nd, 0);
        tp  = iltable[il].iltype;
        Discard cmindx(momstr, tp, p);
        if(ptable[p]->ptype == P_VECTOR) {
            if( !vcode )
              fprintf(redfp, "      real*8     ep%d(4,%s), ew%d(%s)\n",
                       nd+1, momstr, nd+1, momstr);
            else
              fprintf(redfp, "      real*8     ep%d(nsize,4,%s), ew%d(nsize,%s)\n",
                       nd+1, momstr, nd+1, momstr);
        } else if(ptable[p]->ptype == P_FERMION) {
            if( !vcode )
              fprintf(redfp, "      real*8     ps%d(4,2), ew%d(1)\n",
                    nd+1, nd+1);
            else
              fprintf(redfp, "      real*8     ps%d(nsize,4,2), ew%d(nsize,1)\n",
                    nd+1, nd+1);
            if( !vcode )
              fprintf(redfp, "      complex*16 ce%d(2,2)\n", nd+1);
            else
              fprintf(redfp, "      complex*16 ce%d(nsize,2,2)\n", nd+1);
        } else if(ptable[p]->ptype == P_MAJORANA) {
            if( !vcode )
              fprintf(redfp, "      real*8     ps%d%s(4,2), ew%d%s(1)\n",
                    nd+1, ftsuorv[1], nd+1, ftsuorv[1]);
            else
              fprintf(redfp, "      real*8     ps%d%s(nsize,4,2), ew%d%s(nsize,1)\n",
                    nd+1, ftsuorv[1], nd+1, ftsuorv[1]);
            if( !vcode )
              fprintf(redfp, "      complex*16 ce%d%s(2,2)\n",
                    nd+1, ftsuorv[1]);
            else
              fprintf(redfp, "      complex*16 ce%d%s(nsize,2,2)\n",
                    nd+1, ftsuorv[1]);

            if( !vcode )
              fprintf(redfp, "      real*8     ps%d%s(4,2), ew%d%s(1)\n",
                    nd+1, ftsuorv[2], nd+1, ftsuorv[2]);
            else
              fprintf(redfp, "      real*8     ps%d%s(nsize,4,2), ew%d%s(nsize,1)\n",
                    nd+1, ftsuorv[2], nd+1, ftsuorv[2]);
            if( !vcode )
              fprintf(redfp, "      complex*16 ce%d%s(2,2)\n",
                    nd+1, ftsuorv[2]);
            else
              fprintf(redfp, "      complex*16 ce%d%s(nsize,2,2)\n",
                    nd+1, ftsuorv[2]);
        }
    }
    fprintf(redfp,"\n");

    fprintf(redfp,"* Switch of gauge parameters\n");
    fprintf(redfp,"      common /smgaus/igau00,igauab,igauwb,igauzb,");
    fprintf(redfp,"igaugl\n");
    fprintf(redfp,"      common /smgaug/agauge(0:4)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"* Normalization\n");
    fprintf(redfp,"      common /smdbgg/fknorm,fkcall,nkcall\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"* Calculated table of amplitudes\n");
    fprintf(redfp,"      common /smatbl/av, lt, indexg\n");
    if( !vcode )
      fprintf(redfp,"      complex*16 av(0:lag-1)\n");
    else
      fprintf(redfp,"      complex*16 av(nsize,0:lag-1)\n");
    fprintf(redfp,"      integer    lt(0:nextn), indexg(nextn)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"* Spin average\n");
    fprintf(redfp,"      common /amspin/aspin, ");
    if(cpol)
        fprintf(redfp,"jhs(nextn), jhe(nextn), jcpol(nextn)\n");
    else
        fprintf(redfp,"jhs(nextn), jhe(nextn)\n");
    fprintf(redfp,"\n");
    if( jqedps ) {
        fprintf(redfp,"*     QEDPS \n");
        fprintf(redfp,"      common /amoptq/jqedps\n");
    }

    cminclk();
}

/*********************************************************************
 *    2. For each process
 */
/*==================================================================*/
Public  Void    ftmain(prefix, gn0)
int     prefix;
int     gn0;
/*------------------------------------------------------------------*
 * Generate amptbl.f
 *------------------------------------------------------------------*/
{
    node nd;
    tnode  *ndc;
    int  i, il, tp, s, n0, ioro, g0, gaumask, c, cn;
    pindex p;
    int  numhiggs,numgluon;

#if DBG
    if(dbg >= redbg)
        printf("ftmain: begin (%c, %d, %d)\n", prefix, cnprc, gn0);
#endif

    cmopen(ROFORT, "amptbl.f");

    cmsepar('*');
    if( !vcode )
        fprintf(redfp, "      subroutine amptbl\n");
    else
        fprintf(redfp, "      subroutine amptbl(nsampl)\n");
    fprintf(redfp, "      implicit real*8(a-h,o-z)\n\n");
    if( vcode )
        fprintf(redfp, "      include 'inclvs.f'\n");
    fprintf(redfp, "      include 'incl1.f'\n");
    fprintf(redfp, "      include 'incl2.f'\n");
    fprintf(redfp, "      include 'inclk.f'\n");
    cmsepar('-');

    fprintf(redfp, "      jgraph = 0\n");

    fprintf(redfp, "\n* External lines\n");
    for(nd = 0; nd < ndnodes(); nd++) {
        if(ndtype(nd) != NDEXTERN)
            continue;
        p   = ndparticle(nd, 0);
        if(p > 0) {
            ioro = -1;
        } else {
            p = -p;
            ioro = 1;
        }

        if(ptable[p]->ptype == P_VECTOR) {
            il  = ndiline(nd, 0);
            tp  = iltable[il].iltype;
            Discard cmindx(momstr, tp, p);
            if( !vcode )
                fprintf(redfp, "      call smextv(%s,%s,pe%d,",
                        momstr,ptable[p]->mass, nd+1);
            else
                fprintf(redfp, "      call smextv(nsampl,%s,%s,pe%d,",
                        momstr,ptable[p]->mass, nd+1);
            ftgaug(momstr1, p);
            fprintf(redfp, "ep%d,ew%d,%s)\n", nd+1,nd+1,momstr1);

        } else if(ptable[p]->ptype == P_FERMION) {
            ndc  = ndnode(nd);
            if(ndc->nvkind == ELINITIAL) {
                s = ioro;
            } else {
                s = -ioro;
            }
            if( !vcode )
                fprintf(redfp, "      call smextf(%s,%s,pe%d,",
                       ((ioro > 0)?"lincom":"loutgo"),       
                       ptable[p]->mass, nd+1);
            else
                fprintf(redfp, "      call smextf(nsampl,%s,%s,pe%d,",
                       ((ioro > 0)?"lincom":"loutgo"),       
                       ptable[p]->mass, nd+1);
            fprintf(redfp, "ps%d,ce%d)\n", nd+1,nd+1);
            if( !vcode )
                fprintf(redfp, "      ew%d(1) = %s\n",
                        nd+1,((s > 0)?"lprtcl":"lantip"));
            else {
                fprintf(redfp, "      do %3d j = 1, nsampl\n",100+nd);
                fprintf(redfp, "         ew%d(j,1) = %s\n",
                        nd+1,((s > 0)?"lprtcl":"lantip"));
                fprintf(redfp, "  %3d continue\n",100+nd);
            }
        } else if(ptable[p]->ptype == P_MAJORANA) {
            ndc  = ndnode(nd);
            if(ndc->nvkind == ELINITIAL) {
                s = 1;
            } else {
                s = -1;
            }
            if( !vcode )
                fprintf(redfp, "      call smextf(%s,%s,pe%d,",
                       ((s > 0)?"lincom":"loutgo"),       
                       ptable[p]->mass, nd+1);
            else
                fprintf(redfp, "      call smextf(nsampl,%s,%s,pe%d,",
                       ((s > 0)?"lincom":"loutgo"),       
                       ptable[p]->mass, nd+1);
            fprintf(redfp, "ps%d%s,ce%d%s)\n",
                           nd+1,ftsuorv[1],nd+1,ftsuorv[1]);
            if( !vcode )
                fprintf(redfp, "      ew%d%s(1) = lprtcl\n",
                        nd+1,ftsuorv[1]);
            else {
                fprintf(redfp, "      do %3d j = 1, nsampl\n",100+nd);
                fprintf(redfp, "         ew%d%s(j,1) = lprtcl\n",
                        nd+1,ftsuorv[1]);
                fprintf(redfp, "  %3d continue\n",100+nd);
            }

            if( !vcode )
                fprintf(redfp, "      call smextf(%s,%s,pe%d%,",
                        ((s < 0)?"lincom":"loutgo"),       
                        ptable[p]->mass, nd+1);
            else
                fprintf(redfp, "      call smextf(nsampl,%s,%s,pe%d%,",
                        ((s < 0)?"lincom":"loutgo"),       
                        ptable[p]->mass, nd+1);
            fprintf(redfp, "ps%d%s,ce%d%s)\n",
                           nd+1,ftsuorv[2],nd+1,ftsuorv[2]);
            if( !vcode )
                fprintf(redfp, "      ew%d%s(1) = lantip\n",
                        nd+1,ftsuorv[2]);
            else {
                fprintf(redfp, "      do %3d j = 1, nsampl\n",100+nd);
                fprintf(redfp, "         ew%d%s(j,1) = lantip\n",
                        nd+1,ftsuorv[2]);
                fprintf(redfp, "  %3d continue\n",100+nd);
            }
        }
    }
    fprintf(redfp, "\n");
/* CF */
    if( colorv == 1 ) {
    if( cbase->nbase > 1 ) {
    fprintf(redfp, "\n* Buffer clear for color-based-amplitude\n");
    fprintf(redfp,"      do 200 ic = 0 , ncbase-1\n");
    fprintf(redfp,"      do 200 ih = 0 , lag-1\n");
    if( !vcode )
        fprintf(redfp,"         agc(ih,ic) = (0.0d0,0.0d0)\n");
    else {
        fprintf(redfp,"      do 200 j = 1 , nsampl\n");
        fprintf(redfp,"         agc(j,ih,ic) = (0.0d0,0.0d0)\n");
    }
    fprintf(redfp,"  200 continue\n\n");
    } else {
    fprintf(redfp, "\n* Buffer clear for amplitude\n");
    fprintf(redfp,"      do 200 ih = 0 , lag-1\n");
    if( !vcode )
        fprintf(redfp,"         agc(ih) = (0.0d0,0.0d0)\n");
    else {
        fprintf(redfp,"      do 200 j = 1 , nsampl\n");
        fprintf(redfp,"         agc(j,ih) = (0.0d0,0.0d0)\n");
    }
    fprintf(redfp,"  200 continue\n\n");
    }
    }


    cn = (gn0 + TCUT - 1)/TCUT;

    for(c = 0; c < cn; c++) {
        if( !vcode )
            fprintf(redfp, "      call ampt%d\n", c);
        else
            fprintf(redfp, "      call ampt%d(nsampl)\n", c);
    }

    fprintf(redfp, "\n");
    fprintf(redfp, "      return\n");
    fprintf(redfp, "      end\n");

    c = 0;

    /* repeat graph with loop lp */
    for(g0 = 0; g0 < gn0; g0++) {
        if(g0%TCUT == 0) {
            if(g0 > 0) {
                fprintf(redfp, "      return\n");
                fprintf(redfp, "      end\n");
            }
            sprintf(momstr, "ampt%d.f", c);
            cmopen(ROFORT, momstr);
            cmsepar('*');
            if( !vcode )
                fprintf(redfp, "      subroutine ampt%d\n", c);
            else
                fprintf(redfp, "      subroutine ampt%d(nsampl)\n", c);
            fprintf(redfp, "      implicit real*8(a-h,o-z)\n\n");
            if( vcode )
                fprintf(redfp, "      include 'inclvs.f'\n");
            fprintf(redfp, "      include 'incl1.f'\n");
            fprintf(redfp, "      include 'incl2.f'\n");
            fprintf(redfp, "      include 'inclk.f'\n");
            cmsepar('-');

            c++;
        }
        n0 = g0 + 1;

        cgraph = daread(cnprc, g0);

        /* initialize i-line table */
        ilmom();

        numhiggs = ftishiggs();
        if( numhiggs > 0 ) {
            fprintf(redfp, "      if(jhiggs .ne. 0) then\n");
        }

        numgluon = ftisgluon();
        if( numgluon > 0 ) {
            fprintf(redfp, "      if(jgluon .ne. 0) then\n");
        }

        gaumask = ftgmask();
        for(i = 1; i <= 4; i++) {
            if(gaumask & gmask[i]) {
                fprintf(redfp, "      if(igau%cb .ne. 0) then\n", 
                        gchr[i]);
            }
        }
        fprintf(redfp, "      if(jselg(%d) .ne. 0) then\n", n0);
        fprintf(redfp, "        jgraph = jgraph + 1\n");
        fprintf(redfp, "        igraph(jgraph) = %d\n", n0);
        if( !vcode )
            fprintf(redfp, "        call am%d\n", n0);
        else
            fprintf(redfp, "        call am%d(nsampl)\n", n0);
        fprintf(redfp, "      endif\n");
        for(i = 4; i >= 1; i--) {
            if(gaumask & gmask[i]) {
                fprintf(redfp, "      endif\n");
            }
        }
        if( numgluon > 0 ) {
            fprintf(redfp, "      endif\n");
        }
        if( numhiggs > 0 ) {
            fprintf(redfp, "      endif\n");
        }
        fprintf(redfp, "\n");
    }

    fprintf(redfp, "      return\n");
    fprintf(redfp, "      end\n");
}

/*==================================================================*/
Private int     ftgmask(NoArg)
/*------------------------------------------------------------------*
 * generate REDUCE code for one graph g0.
 *------------------------------------------------------------------*/
{
    pindex p;
    int    il, msk, ig;

    msk = 0;
    for(il = 1; il < cgraph->nilines; il++) {
        p = ilpart(il);
        if(ptable[p]->ptype != P_SCALAR)
            continue;
        if((ig = ftgaug(momstr, p)) > 0)
            msk |= gmask[ig];
    }
    return msk;
}

/*********************************************************************
 *    3. For each graph
 */
/*==================================================================*/
Public  Void    ftgraph(prefix, proc, graph)
int     prefix;
tproc   *proc;
tgraph  *graph;
/*------------------------------------------------------------------*
 * generate REDUCE code for one graph cngrf.
 *------------------------------------------------------------------*/
{
    int y;

#if DBG
    if(dbg >= redbg)
        printf("ftgraph: begin (%d)\n", cngrf+1);
#endif

    /* initialize variable */
    cprocess  = proc;
    cgraph    = graph;
    nilines   = cgraph->nilines;

    conjstart = graph->conjsrt;

    /* file name */
    sprintf(outname, "%c%d.f\0", prefix, cngrf+1);

    fttmpopen();
    fttmpclr();

    ftgtree();

    /* open file */
    cmopen(ROFORT, outname);
    ftbegin();

    ftvars(cngrf);

    fttmpcp();

    ftgend();

    /* increase offset of local variables */
/* back 
    momoffs += cgraph->nilines;
    legoffs += cprocess->maxlegs;
 */
    fclose(redfp);

#if DBG
    if(dbg >= redbg)
        printf("ftgraph: end (%d)\n", cngrf+1);
#endif
}

/*==================================================================*/
Private Void    fttmpopen(NoArg)
/*------------------------------------------------------------------*
 * Open tempolary files.
 *------------------------------------------------------------------*/
{
    static int opn = False;

    if(opn) {
        rewind(tfp1);
        return;
    }

    opn = True;

    /* open temporary file */
#ifdef SONY_NEWS
    tfp1 = fopen("tmp1", "w+"));
#else
    tfp1 = tmpfile();
#endif

    if(tfp1 == NULL || tfp1 == NULL) {
        error("fttmpopen", "temporary file cannot open");
        exit(1);
    }

}
/*==================================================================*/
Private Void    fttmpcp(NoArg)
/*------------------------------------------------------------------*
 * Copy tempolary files.
 *------------------------------------------------------------------*/
{
    int c;

    putc(EOFSYM, tfp1);

    fflush(tfp1);

    rewind(tfp1);

    while((c = getc(tfp1)) != EOFSYM) {
        putc(c, redfp);
    }
}
/*==================================================================*/
Private Void    ftgtree(NoArg)
/*------------------------------------------------------------------*
 * tree graph
 *------------------------------------------------------------------*/
{
    string p;
    int    i, y, x, c, cc, il, cesize;

#if DBG
    if(dbg >= redbg)
        printf("ftgtree: begin\n");
#endif

    /* momentum conservation */
    ftmconsv();

    /* denominators of propagators */
    ftdenom();

    /* internal lines */
    ftiline();

    /* initialize working table of vertices */
    ftvlts();
/*
prdump();
*/

    /* vertices */
    ftnode();

    /* connect vertices */
    ftconn();

    /* denominator of propagator */
    ftdprop();

    /* call ampord */
    ftclpord();

    /* clear work */
    ftvlte();

#if DBG
    if(dbg >= redbg)
        printf("ftgtree: end\n");
#endif
}

/*==================================================================*/
Private Void    ftbegin(NoArg)
/*------------------------------------------------------------------*
 * Output begining message to REDUCE code.
 *------------------------------------------------------------------*/
{
#if DBG
    if(dbg >= redbg)
        printf("ftbegin: begin\n");
#endif

    if(cgraph->subid == 0)
        fprintf(redfp, "*             Graph No. %d\n",
                cgraph->grid);
    else
        fprintf(redfp, "*             Graph No. %d - %d\n",
                cgraph->grid, cgraph->subid);
    if(cngrf != cgraph->grid)
        fprintf(redfp, "*         Generated No. %d\n", cngrf);
    cmsepar('*');
}

/*==================================================================*/
Private Void    ftmconsv(NoArg)
/*------------------------------------------------------------------*
 *  Momentum conservation of internal momenta.
 *------------------------------------------------------------------*/
{
    int    il, ii, ce, y, x;
    tleg   *lg;
    node   nd;
    int    leg;

#if DBG
    if(dbg >= redbg)
        printf("ftmconsv: begin\n");
#endif

    iliconsv();

    fprintf(tfp1, "* Internal momenta\n");
    fprintf(tfp1, "      do 100 i = 1, 4\n");
    if( vcode )
        fprintf(tfp1, "      do 100 j = 1, nsampl\n");

    for(il = 1; il < nilines; il++) {
        if(iltable[il].iltype == ILEXTERN)
            continue;
        else if(iltable[il].iltype == ILLOOP)
            continue;

        cmimom(momstr, il, 1);
        if( !vcode )
            fprintf(tfp1, "        %s(i) =", momstr);
        else
            fprintf(tfp1, "        %s(j,i) =", momstr);
        if( !vcode )
            fttmp(FT_REAL, momstr, "4", 4);
        else
            fttmp(FT_REAL, momstr, "nsize,4", 4);

        for(ii = 1; ii < nilines; ii++) {
            if(ii == kloop)
                continue;
            else if((ce = iltable[il].ilmlist[ii]) == 0)
                continue;
            cmmom(momstr, iltable[ii].ilnodes[0], 
                  iltable[ii].illegs[0], False);
            if(ce == -1) {
                if( !vcode )
                    fprintf(tfp1, " - %s(i)", momstr);
                else
                    fprintf(tfp1, " - %s(j,i)", momstr);
            } else if(ce == 1) {
                if( !vcode )
                    fprintf(tfp1, " + %s(i)", momstr);
                else
                    fprintf(tfp1, " + %s(j,i)", momstr);
            } else if(ce > 0) {
                if( !vcode )
                    fprintf(tfp1, " + %d*%s(i)", ce, momstr);
                else
                    fprintf(tfp1, " + %d*%s(j,i)", ce, momstr);
            } else if(ce < 0) {
                if( !vcode )
                    fprintf(tfp1, " - %d*%s(i)", abs(ce), momstr);
                else
                    fprintf(tfp1, " - %d*%s(j,i)", abs(ce), momstr);
            }
        }
        fprintf(tfp1, "\n");
    }
    fprintf(tfp1, "  100 continue\n");

#if DBG
    if(dbg >= redbg)
        printf("ftmconsv: end\n");
#endif
}

/*==================================================================*/
Private Void    ftdenom(NoArg)
/*------------------------------------------------------------------*
 *  Clear temp. var table
 *------------------------------------------------------------------*/
{
    int il, e1, e2, ce1, ce2, s, c, pc;
    pindex p, pl;

    fprintf(tfp1, "\n* Denominators of propagators\n");
    if(nilines < 1) {
        return;
    }
    if( colorv == 0 ) {
        if( !vcode )
            fprintf(tfp1, "      aprop(jgraph) = 1.0d0\n");
        else {
            fprintf(tfp1, "      do 200 j = 1, nsampl\n");
            fprintf(tfp1, "         aprop(j,jgraph) = 1.0d0\n");
            fprintf(tfp1, "  200 continue\n");
        }
    } else {
        if( !vcode )
            fprintf(tfp1, "      aprop         = 1.0d0\n");
        else {
            fprintf(tfp1, "      do 200 j = 1, nsampl\n");
            fprintf(tfp1, "         aprop(j) = 1.0d0\n");
            fprintf(tfp1, "  200 continue\n");
        }
    }

    if(nilines < 1) {
        fprintf(tfp1, "* %d propagators\n", nilines);
        return;
    }
    for(il = 1; il < nilines; il++) {
        if(iltable[il].iltype == ILEXTERN)
            continue;
        else if(iltable[il].iltype == ILLOOP)
            continue;

        /* denominator of the propagator for i-line il */
        if( vcode )
            fprintf(tfp1, "      do %3d j = 1, nsampl\n",200+il);
        ftdmom(momstr, il, 1);
        puinit(tfp1, "      ", "     &      ","","","", 70, 16);
        putid(momstr);
        if( vcode )
            putid("(j)");
        putid(" = ");
        if( !vcode )
            fttmp(FT_REAL, momstr, NULL, 1);
        else
            fttmp(FT_REAL, momstr, "nsize", 1);

        /* inner products of momenta */
        for(e1 = 0; e1 < nilines; e1++) {
            if(e1 == kloop)
                continue;
            else if((ce1 = iltable[il].ilmlist[e1]) == 0)
                continue;

            for(e2 = e1+1; e2 < nilines; e2++) {
                if(e2 == kloop)
                    continue;
                else if((ce2 = iltable[il].ilmlist[e2]) == 0)
                    continue;

                ce2 *= ce1;
                if(ce2 > 0) {
                    s = ce2;
                    c = '+';
                } else {
                    s = -ce2;
                    c = '-';
                }
                sprintf(momstr1, " %c ", c);
                putid(momstr1);
                s *= 2;
                if(s != 1) {
                    sprintf(momstr1, "%d.0d0*", s);
                    putid(momstr1);
                }
                if( !vcode )
                    sprintf(momstr1, "prod(%d,%d)", e1, e2);
                else
                    sprintf(momstr1, "prod(j,%d,%d)", e1, e2);
                putid(momstr1);
            }
        }

        /* mass terms */
        pl = ilpart(il);
        pc = 0;
        for(e1 = 1; e1 < nilines; e1++) {
            if(e1 == kloop)
                continue;
            else if((ce1 = iltable[il].ilmlist[e1]) == 0)
                continue;

            p = ilpart(e1);
            if(p == pl)
                pc++;
            else {
                sprintf(momstr1, " + %s**2", ptable[p]->mass);
                putid(momstr1);
            }
        }
        if(pc != 0) {
            if(pc > 0) {
                s = pc;
                c = '+';
            } else {
                s = -pc;
                c = '-';
            }
            if(s == 1)
                sprintf(momstr1, " %c ", c);
            else
                sprintf(momstr1, " %c %d.0d0*", c, s);
            putid(momstr1);
            sprintf(momstr1, "%s**2", ptable[pl]->mass);
            putid(momstr1);
        }
        putend();
        fprintf(tfp1, "  %3d continue\n",200+il);

        /* product of denominators */
        if( colorv == 0 ) {
            if( !vcode )
                fprintf(tfp1, "      call smprpd(aprop(jgraph),%s,",
                    momstr);
            else {
                fprintf(tfp1, "      call smprpd(nsampl,");
                fprintf(tfp1, "aprop(1,jgraph),%s,",momstr);
            }
        } else {
            if( !vcode )
                fprintf(tfp1, "      call smprpd(aprop,%s,",momstr);
            else
                fprintf(tfp1, "      call smprpd(nsampl,aprop,%s,",
                momstr);
        }
        p = ilpart(il);
        if(ptable[p]->ptype == P_SCALAR) {
            ftgaug(momstr1, p);
            if(momstr1[4] != '?')
                fprintf(tfp1, "agauge(%s)*", momstr1);
        }
        fprintf(tfp1, "%s**2,",   ptable[p]->mass);
        if(ptable[p]->width[0] == '0')
            fprintf(tfp1, "0.0d0)\n");
        else
            fprintf(tfp1, "%s*%s)\n", ptable[p]->mass,
                                      ptable[p]->width);
    }
}

/*==================================================================*/
Private Void    fttmpclr(NoArg)
/*------------------------------------------------------------------*
 *  Clear temp. var table
 *------------------------------------------------------------------*/
{
    int i;

    for(i = 0; i < nvars; i++) {
        if(vartbl[i].name != NULL)
            free(vartbl[i].name);
        if(vartbl[i].dim != NULL)
            free(vartbl[i].dim);
    }
    nvars = 0;
    vartype[FT_INT]     = 0; 
    vartype[FT_REAL]    = 0; 
    vartype[FT_COMPLEX] = 0; 

}

/*==================================================================*/
Private Void    fttmp(vtype, name, dim, size)
int    vtype;
string name;
string dim;
int    size;
/*------------------------------------------------------------------*
 *  Save temporary variables
 *------------------------------------------------------------------*/
{
    if(nvars >= NVARS) {
        fprintf(Stderr, "fttmp: var-table overflow\n");
        exit(1);
    }
    vartype[vtype]++;
    vartbl[nvars].vtype = vtype;
    vartbl[nvars].name  = (string) strdup(name);
    if(dim == NULL)
        vartbl[nvars].dim   = NULL;
    else
        vartbl[nvars].dim   = (string) strdup(dim);
    vartbl[nvars].size  = size;
    nvars++;

}

/*==================================================================*/
Private Void    fttmdmp(NoArg)
/*------------------------------------------------------------------*
 *  Dump temporary variables
 *------------------------------------------------------------------*/
{
    int v, s, c, ts = 0;

    printf("\nVairbles : compex %d, real %d, int %d\n",
           vartype[FT_COMPLEX], vartype[FT_REAL],
           vartype[FT_INT]);
    for(v = 0; v < nvars; v++) {
        if(vartbl[v].vtype == FT_COMPLEX) {
            c = 'C';
            s = 16*vartbl[v].size;
            ts += s;
        } else if(vartbl[v].vtype == FT_REAL) {
            c = 'R';
            s = 8*vartbl[v].size;
            ts += s;
        } else if(vartbl[v].vtype == FT_INT) {
            c = 'I';
            s = 4*vartbl[v].size;
            ts += s;
        } else {
            c = '?';
            s = vartbl[v].size;
            ts += s;
        }
        printf("%c %6d(%4d) %6s", c,s,vartbl[v].size,vartbl[v].name);
        if(vartbl[v].dim != NULL) 
            printf("(%s)", vartbl[v].dim);
        printf("\n");
    }
    printf("  %6d\n\n", ts);
}

/*==================================================================*/
Private Void    ftvars(g)
int     g;
/*------------------------------------------------------------------*
 *  Save temporary variables
 *------------------------------------------------------------------*/
{
    long int lsizer = 0, lsizei=0;
    int  v, ldim, c, tp, typ, p, p0;
    int i;
    string str;

#if DBG
    if(dbg >= redbg)
        fttmdmp();
#endif

    if( !vcode )
        fprintf(redfp, "      subroutine am%d\n", g+1);
    else
        fprintf(redfp, "      subroutine am%d(nsampl)\n", g+1);
    fprintf(redfp, "      implicit real*8(a-h,o-z)\n\n");
    if( vcode )
        fprintf(redfp, "      include 'inclvs.f'\n");
    fprintf(redfp, "      include 'incl1.f'\n");
    fprintf(redfp, "      include 'inclk.f'\n");
    cmsepar('-');

    if(vartype[FT_COMPLEX] + vartype[FT_REAL] > 0) {
        puinit(redfp, "      common /amwork/",
                      "     &               ",
                      ",", ",", "", 64, 16);
        for(v = 0; v < nvars; v++) {
            if(vartbl[v].vtype != FT_COMPLEX)
                continue;
            putid(vartbl[v].name);
            lsizer += 16*vartbl[v].size;
        }
        for(v = 0; v < nvars; v++) {
            if(vartbl[v].vtype != FT_REAL)
                continue;
            putid(vartbl[v].name);
            lsizer += 8*vartbl[v].size;
        }
        putend();
    }

    if(vartype[FT_INT] > 0) {
        puinit(redfp, "      common /amwori/",
                      "     &          ",
                      ",", ",", "", 64, 16);
        for(v = 0; v < nvars; v++) {
            if(vartbl[v].vtype != FT_INT)
                continue;
            putid(vartbl[v].name);
            lsizei += 4*vartbl[v].size;
        }
        putend();
    }

    fprintf(redfp,"*     %ld + %ld bytes used\n\n",
            lsizer, lsizei);
    incl2r = max(incl2r, lsizer);
    incl2i = max(incl2i, lsizei);

    for(typ = FT_INT; typ < FT_COMPLEX; typ++) {
        if(vartype[typ] > 0) {
            if(typ == FT_COMPLEX) {
                puinit(redfp, "      complex*16 ",
                              "     &           ",
                              ",", ",", "", 64, 16);
            } else if(typ == FT_REAL) {
                puinit(redfp, "      real*8     ",
                              "     &           ",
                              ",", ",", "", 64, 16);
            } else if(typ == FT_INT) {
                puinit(redfp, "      integer    ",
                              "     &           ",
                              ",", ",", "", 64, 16);
            } else {
                fprintf(Stderr, "*** illegal variable type = %d\n",
                                typ);
                exit(1);
            }
            for(v = 0; v < nvars; v++) {
                if(vartbl[v].vtype != typ)
                    continue;
                if(vartbl[v].dim == NULL)
                    ldim = 0;
                else
                    ldim = strlen(vartbl[v].dim);
                if(strlen(vartbl[v].name) + ldim + 1
                   >= MAXSSTR) {
                    fprintf(Stderr, "ftvar: too long string\n");
                    exit(1);
                }
                str = sizestr;
                strcpy(str, vartbl[v].name);
                str += strlen(vartbl[v].name);
                if(vartbl[v].dim != NULL) {
                    *str++ = '(';
                    strcpy(str, vartbl[v].dim);
                    str += strlen(vartbl[v].dim);
                    *str++ = ')';
                    *str   = '\0';
                }
                putid(sizestr);
            }
            putend();
        }
    }

    for(typ = FT_COMPLEX; typ < FT_MAX; typ++) {
        if(vartype[typ] > 0) {
            for(v = 0; v < nvars; v++) {
                if(vartbl[v].vtype != typ)
                    continue;

                if(vartbl[v].dim == NULL)
                    ldim = 0;
                else
                    ldim = strlen(vartbl[v].dim);

                if(typ == FT_COMPLEX) {
                    puinit(redfp, "      complex*16 ",
                                  "     &           ",
                                  "", "", "", 70, 16);
                } else if(typ == FT_REAL) {
                    puinit(redfp, "      real*8     ",
                                  "     &           ",
                                  "", "", "", 70, 16);
                } else if(typ == FT_INT) {
                    puinit(redfp, "      integer    ",
                                  "     &           ",
                                  "", "", "", 70, 16);
                } else {
                    fprintf(Stderr, "*** illegal variable type = %d\n",
                                    typ);
                    exit(1);
                }
                putid(vartbl[v].name);
                if(vartbl[v].dim != NULL) {
                    putid("(");
                    p = 0;
                    while(p < ldim) {
                        p0 = p;
                        p++;
                        while(p < ldim && isalnum(vartbl[v].dim[p]))
                            p++;
                        tp = vartbl[v].dim[p];
                        vartbl[v].dim[p] = '\0';
                        putid(vartbl[v].dim + p0);
                        vartbl[v].dim[p] = tp;
                    }
                    putid(")");
                }
                putend();
            }
        }
    }
/* CF  */
    if( colorv == 1 ) {
        fprintf(redfp,"      complex*16 atmp\n");
        if( cbase->nbase > 1 ) {
        fprintf(redfp,"      real*8     cwgt(0:%d)\n",cbase->nbase-1);
/* For KK-COLOR-BASE for AUG95-version
        fprintf(redfp,"      data       cwgt /");
        for(i=0;i<cbase->nbase;i++) {
          fprintf(redfp,"%fd0%c",colgr->wgt[i],
                 (i==cbase->nbase-1)?'/':',');
        }
*/
        fprintf(redfp,"\n");
        }
    }

    cmsepar('-');
}

/*==================================================================*/
Private Void    ftincw(NoArg)
/*------------------------------------------------------------------*
 *  Output incl2.f
 *------------------------------------------------------------------*/
{
    cmopen(ROFORT, "incl2.f");

    if( !vcode ) 
        fprintf(redfp, "      common /amwork/rdmi(%d)\n", incl2r/8); /* ??? */
    else
        fprintf(redfp, "      common /amwork/rdmi(nsize,%d)\n", incl2r/8); /* ??? */
    fprintf(redfp, "      common /amwori/idmi(%d)\n", incl2i/4);
}

/*==================================================================*/
Private Void    ftconn(NoArg)
/*------------------------------------------------------------------*
 * Connect vertices by i-line.
 *------------------------------------------------------------------*/
{
    node n, nc, no, nn;
    int  nv, size, il, c;
    int  l, lg, lo, ln;
    vlegt *vl;

    fprintf(tfp1, "\n");
    while(True) {
        if(nvltbl <= nvlusd + 1)
            return;

        /* find a node "nc" to be connected */
        nc = -1;
        for(n = 0; n < nvltbl; n++) {
            if(vltbl[n].size <= 0)
                continue;
            nc = n;
            break;
        }
        if(nc < 0)
            return;

        /* find a leg of node "nc" to be connected */
        size = vltbl[nc].size;
        vl   = vltbl[nc].vl;
        for(l = 0; l < size; l++) {
            if(vl[l].t != ILEXTERN) {
                lg = l;
                break;
            }
        }
        il   = vl[lg].i;

        /* find a node and its node to be connected with (nc,lg) */
        for(n = 0; n < nvltbl; n++) {
            if(vltbl[n].size <= 0)
                continue;
            if(n == nc)
                continue;
            for(l = 0; l < vltbl[n].size; l++) {
                if(vltbl[n].vl[l].i == il) {
                    lo = l;
                    no = n;
                    goto found;
                }
            }
        }
        fprintf(Stderr,"ftconn : inconsistent connection\n");
        fprintf(Stderr,"nc = %d, lg = %d, il = %d\n", nc, lg, il);
        ftvldmp();
        exit(1);

        found:;
        /* connect (nc, lg) with (no, lo) */
        if(nvltbl >= MAXVLTBL - 1) {
            fprintf(Stderr, "ftconn: too many vertices\n");
            ftvldmp();
            exit(1);
        }
        nn = nvltbl;
        vltbl[nn].size = vltbl[nc].size + vltbl[no].size - 2;
        if((vltbl[nn].vl = 
          (vlegt*) malloc(vltbl[nn].size*sizeof(vlegt)))
          == NULL) {
            fprintf(Stderr, "ftconn: malloc failed\n");
            ftvldmp();
            exit(1);
        }

        ln = 0;
        for(l = 0; l < vltbl[nc].size; l++) {
            if(l == lg)
                continue;
            ftvlcp(vltbl[nn].vl, vltbl[nc].vl, ln, l);
            ln++;
        }
        for(l = 0; l < vltbl[no].size; l++) {
            if(l == lo)
                continue;
            ftvlcp(vltbl[nn].vl, vltbl[no].vl, ln, l);
            ln++;
        }
        vltbl[nc].size *= -1;
        vltbl[no].size *= -1;
        nvltbl++;
        nvlusd += 2;
        c = ptypesym[ptable[abs(ilpart(il))]->ptype];

        /* for majorana */
        if(c == 'm')
            c = 'f';
        fprintf(tfp1, "      call smcon%c(", c);
        if( vcode )
            fprintf(tfp1, "nsampl,");

        if(c == 's') {
            fprintf(tfp1, "lt%d,lt%d,%d,%d,av%d,av%d,",
                    nc, no, lg+1, lo+1, nc, no);
        } else {
            fprintf(tfp1, "lt%d,lt%d,%d,%d,ew%d,av%d,av%d,",
                    nc, no, lg+1, lo+1, vltbl[nc].vl[lg].i,
                    nc, no);
        }
        ftcrans(nn, vltbl[nn].vl, vltbl[nn].size);
        fprintf(tfp1, ")\n");
    }
}

/*==================================================================*/
Private Void    ftclpord(NoArg)
/*------------------------------------------------------------------*
 * Connect vertices by i-line.
 *------------------------------------------------------------------*/
{
    int n, nc, l, i;

    /* find a node "nc" to be connected */
    nc = -1;
    for(n = 0; n < nvltbl; n++) {
        if(vltbl[n].size <= 0)
            continue;
        nc = n;
        break;
    }
    if(nc < 0) {
        fprintf(Stderr,"ftampord : inconsistent connection\n");
        ftvldmp();
        exit(1);
    }
    fprintf(tfp1,"\n");
    for(l = 0; l < vltbl[nc].size; l++)
        fprintf(tfp1,"      indexg(%d) = %d\n",
                l+1, vltbl[nc].vl[l].i);
    if(cpol) {
        for(l = 0; l < vltbl[nc].size; l++) {
            n = vltbl[nc].vl[l].i;
            if(cmphoton(n-1)) {
                fprintf(tfp1,"\n      if(jcpol(%d).ne.0) ", n);
                if( !vcode )
                    fprintf(tfp1,"call smcpol(%d, lt, av)\n", l+1);
                else
                    fprintf(tfp1,"call smcpol(nsampl, %d, lt, av)\n", l+1);
            }
        }
    }
    fprintf(tfp1,"\n      ");
/*  CF  */
    if( colorv == 0 ) {
      if( !vcode )
          fprintf(tfp1,"call ampord(lt, av, indexg, ag(0,jgraph))\n");
      else
          fprintf(tfp1,"call ampord(nsampl, lt, av, indexg, ag(1,0,jgraph))\n");
    } else {
      if( !vcode )
          fprintf(tfp1,"call ampord(lt, av, indexg, agcwrk)\n\n");
      else
          fprintf(tfp1,"call ampord(nsampl, lt, av, indexg, agcwrk)\n\n");
      if( !vcode )
          fprintf(tfp1,"      ancp(jgraph) = 0.0d0\n");
      else {
          fprintf(tfp1,"      do 400 j = 1 , nsampl\n");
          fprintf(tfp1,"         ancp(j,jgraph) = 0.0d0\n");
          fprintf(tfp1,"  400 continue\n");
      }
      fprintf(tfp1,"*     nbase = %d\n",cbase->nbase);

      if( cbase->nbase == 1 ) {
        if( cbase->nexcolor == 1 ) {
/*           fprintf(tfp1,"      cwgt = %fd0\n",colgr->wgt[0]);*/


           if( (colgr->wgt[0]).den == 1 ) {
              fprintf(tfp1,"      cwgt = %d.d0\n",
                  (colgr->wgt[0]).num);
           } else {
              fprintf(tfp1,"      cwgt = %d.d0/%d.d0\n",
                  (colgr->wgt[0]).num,(colgr->wgt[0]).den);
           }


           fprintf(tfp1,"      do 500 ih = 0 , lag-1\n");
           if( !vcode ) {
               fprintf(tfp1,"         atmp    = cwgt*agcwrk(ih)*aprop\n");
               if( jcolmb && isschan )
                   fprintf(tfp1,"         agc(ih) = agc(ih) + colmbf*atmp\n");
               else
                   fprintf(tfp1,"         agc(ih) = agc(ih) + atmp\n");
           } else {
               fprintf(tfp1,"      do 500 j = 1 , nsampl\n");
               fprintf(tfp1,"         atmp    = cwgt*agcwrk(j,ih)*aprop(j)\n");
               if( jcolmb && isschan )
                   fprintf(tfp1,"         agc(j,ih) = agc(j,ih) + colmbf*atmp\n");
               else
                   fprintf(tfp1,"         agc(j,ih) = agc(j,ih) + atmp\n");
           }
/*        } else if( cbase->nexcolor == 0 ) {*/
        } else {
           fprintf(tfp1,"      do 500 ih = 0 , lag-1\n");
           if( !vcode ) {
               fprintf(tfp1,"         atmp    = agcwrk(ih)*aprop\n");
               if( jcolmb && isschan )
                   fprintf(tfp1,"         agc(ih) = agc(ih) + colmbf*atmp\n");
               else
                   fprintf(tfp1,"         agc(ih) = agc(ih) + atmp\n");
           } else {
               fprintf(tfp1,"      do 500 j = 1 , nsampl\n");
               fprintf(tfp1,"         atmp    = agcwrk(j,ih)*aprop(j)\n");
               if( jcolmb && isschan )
                   fprintf(tfp1,"         agc(j,ih) = agc(j,ih) + colmbf*atmp\n");
               else
                   fprintf(tfp1,"         agc(j,ih) = agc(j,ih) + atmp\n");
           }
        }
      } else {
        for(i=0;i<cbase->nbase;i++) {
           if( (colgr->wgt[i]).den == 1 ) {
              fprintf(tfp1,"      cwgt(%d) = %d.d0\n",
                  i,(colgr->wgt[i]).num);
           } else {
              fprintf(tfp1,"      cwgt(%d) = %d.d0/%d.d0\n",
                  i,(colgr->wgt[i]).num,(colgr->wgt[i]).den);
           }
        }
        fprintf(tfp1,"      do 500 ic = 0 , ncbase-1\n");
        fprintf(tfp1,"      do 500 ih = 0 , lag-1\n");
        if( !vcode ) {
            fprintf(tfp1,"         atmp       =  cwgt(ic)*agcwrk(ih)*aprop\n");
            if( jcolmb && isschan )
                fprintf(tfp1,"         agc(ih,ic) = agc(ih,ic) + colmbf*atmp\n");
            else
                fprintf(tfp1,"         agc(ih,ic) = agc(ih,ic) + atmp\n");
        } else {
            fprintf(tfp1,"      do 500 j = 1 , nsampl\n");
            fprintf(tfp1,"         atmp       =  cwgt(ic)*agcwrk(j,ih)*aprop(j)\n");
            if( jcolmb && isschan )
                fprintf(tfp1,"         agc(j,ih,ic) = agc(j,ih,ic) + colmbf*atmp\n");
            else
                fprintf(tfp1,"         agc(j,ih,ic) = agc(j,ih,ic) + atmp\n");
        }
      }
      if( !vcode )
          fprintf(tfp1,"         ancp(jgraph) = ancp(jgraph) + atmp*conjg(atmp)\n");
      else
          fprintf(tfp1,"         ancp(j,jgraph) = ancp(j,jgraph) + atmp*conjg(atmp)\n");
      fprintf(tfp1,"  500 continue\n");
    }
}

/*==================================================================*/
Private Void    ftdprop(NoArg)
/*------------------------------------------------------------------*
 * Invert the numerator of the propagators.
 *------------------------------------------------------------------*/
{
    int sgn, sym;
    long num, den;

    sym = cgraph->sfactor;
    if(sym < 0) {
        fprintf(tfp1, "\n      sym = - ");
        sym = - sym;
    } else {
        fprintf(tfp1, "\n      sym = + ");
    }
    if(sym == 1)
        fprintf(tfp1, "1.0d0\n");
    else
        fprintf(tfp1, "1.0d0/%d.0.d0\n", sym);

    if(cprocess->extcol <= 0) {
        cfactor(&num, &den);
        cgraph = dareadf(cnprc, cngrf);    /* reset the current graph */
        ilmom();
    
        if(den < 0) {
            den *= -1;
            sgn = -1;
        } else {
            sgn = 1;
        }
        if(num < 0) {
            num *= -1;
            sgn *= -1;
        }
        fprintf(tfp1, "      cf  = ");
        if(sgn < 0)
            fprintf(tfp1, "- ");
        else
            fprintf(tfp1, "+ ");
        if(den == 1)
            fprintf(tfp1, "%ld.0d0\n", num);
        else
            fprintf(tfp1, "%ld.0d0/%ld.0d0\n", num, den);
        if( colorv == 0 ) {
            if( !vcode )
                fprintf(tfp1, "      aprop(jgraph) = cf*sym/aprop(jgraph)\n");
            else {
                fprintf(tfp1, "      do 300 j = 1, nsampl\n");
                fprintf(tfp1, "         aprop(j,jgraph) = cf*sym/");
                fprintf(tfp1, "aprop(j,jgraph)\n");
                fprintf(tfp1, "  300 continue\n");
            }
        } else {
            if( !vcode )
                fprintf(tfp1, "      aprop         = cf*sym/aprop\n");
            else {
                fprintf(tfp1, "      do 300 j = 1, nsampl\n");
                fprintf(tfp1, "         aprop(j)  = cf*sym/aprop(j)\n");
                fprintf(tfp1, "  300 continue\n");
            }
        }
    } else {
        if( colorv == 0 ) {
            if( !vcode )
                fprintf(tfp1, "      aprop(jgraph) = sym/aprop(jgraph)\n");
            else {
                fprintf(tfp1, "      do 300 j = 1, nsampl\n");
                fprintf(tfp1, "         aprop(j,jgraph)  = sym/aprop(j,jgraph)\n");
                fprintf(tfp1, "  300 continue\n");
            }
        } else {
            if( !vcode )
                fprintf(tfp1, "      aprop         = sym/aprop\n");
            else {
                fprintf(tfp1, "      do 300 j = 1, nsampl\n");
                fprintf(tfp1, "         aprop(j)  = sym/aprop(j)\n");
                fprintf(tfp1, "  300 continue\n");
            }
        }
    }
    
}

/*********************************************************************
 *    4. End of output
 */

/*********************************************************************
 *    5. Other files
 */
/*==================================================================*/
Private Void    ftampord(NoArg)
/*------------------------------------------------------------------*
 *------------------------------------------------------------------*/
{
    cmopen(ROFORT, "ampord.f");
    cmsepar('*');
    if( !vcode ) 
        fprintf(redfp,"      subroutine ampord(lvt, avt, iord, at)\n");
    else {
        fprintf(redfp,"      subroutine ampord(nsampl, lvt, avt, ");
        fprintf(redfp,"iord, at)\n");
    }
    fprintf(redfp,"      implicit real*8(a-h,o-z)\n");

    if( vcode )
        fprintf(redfp,"      include 'inclvs.f'\n");
    fprintf(redfp,"      include 'incl1.f'\n");
    fprintf(redfp,"      include 'inclk.f'\n");
    fprintf(redfp,"      integer    lvt(0:nextn), iord(nextn)\n");
    if( !vcode ) 
        fprintf(redfp,"      complex*16 avt(0:lag-1), at(0:lag-1)\n");
    else {
        fprintf(redfp,"      complex*16 avt(nsize,0:lag-1), ");
        fprintf(redfp,"at(nsize,0:lag-1)\n");
    }
    fprintf(redfp,"*\n");
    fprintf(redfp,"      integer j(nextn), jv(nextn)\n");
    cmsepar('-');

    fprintf(redfp,"      ibas  = 1\n");
    fprintf(redfp,"      do 10 i = 1, nextn\n");
    fprintf(redfp,"        jv(iord(i)) = ibas\n");
    fprintf(redfp,"        ibas    = ibas*lvt(i)\n");
    fprintf(redfp,"        j(i)    = jhs(i)\n");
    fprintf(redfp,"   10 continue\n");
    fprintf(redfp,"\n");

    fprintf(redfp,"      it = 0\n");
    fprintf(redfp,"  100 continue\n");
    fprintf(redfp,"        iv = 0\n");
    fprintf(redfp,"        do 110 i = 1, nextn\n");
    fprintf(redfp,"          iv = iv +   jv(i)*j(i)\n");
    fprintf(redfp,"  110   continue\n");
    fprintf(redfp,"\n");

    if( !vcode ) 
        fprintf(redfp,"        at(it) = avt(iv)\n");
    else {
        fprintf(redfp,"        do 115 jj = 1, nsampl\n");
        fprintf(redfp,"           at(jj,it) = avt(jj,iv)\n");
        fprintf(redfp,"  115   continue\n");
    }
    fprintf(redfp,"        it = it + 1\n");
    fprintf(redfp,"\n");

    fprintf(redfp,"        ii = 1\n");
    fprintf(redfp,"  120   continue\n");
    fprintf(redfp,"          j(ii) = j(ii) + 1\n");
    fprintf(redfp,"          if(j(ii).gt.jhe(ii)) then\n");
    fprintf(redfp,"            j(ii) = jhs(ii)\n");
    fprintf(redfp,"            ii = ii + 1\n");
    fprintf(redfp,"            if(ii.le.nextn) then\n");
    fprintf(redfp,"              goto 120\n");
    fprintf(redfp,"            else\n");
    fprintf(redfp,"              goto 190\n");
    fprintf(redfp,"            endif\n");
    fprintf(redfp,"          endif\n");
    fprintf(redfp,"        goto 100\n");
    fprintf(redfp,"  190 continue\n");
    fprintf(redfp,"      return\n");
    fprintf(redfp,"      end\n");
    fprintf(redfp,"\n");
}

/*==================================================================*/
Private Void    ftsmcpol(NoArg)
/*------------------------------------------------------------------*
 *------------------------------------------------------------------*/
{
    cmopen(ROFORT, "smcpol.f");
    cmsepar('*');
    fprintf(redfp,"      subroutine smcpol(la, lvt, avt)\n");
    fprintf(redfp,"      implicit real*8(a-h,o-z)\n");
    fprintf(redfp,"      include 'incl1.f'\n");
    fprintf(redfp,"      include 'inclk.f'\n");
    fprintf(redfp,"      integer    lvt(0:nextn)\n");
    fprintf(redfp,"      complex*16 avt(0:lag-1)\n");
    fprintf(redfp,"*\n");
    fprintf(redfp,"      integer j(nextn), jv(nextn)\n");
    fprintf(redfp,"      complex*16 zi, zr, ztv, ztw\n");
    fprintf(redfp,"      data ifst/0/\n");
    fprintf(redfp,"      save ifst, zr, zi\n");
    cmsepar('-');
    fprintf(redfp,"      if(ifst.eq.0) then\n");
    fprintf(redfp,"        ifst = 1\n");
    fprintf(redfp,"        r2 = 1.0d0/sqrt(2.0d0)\n");
    fprintf(redfp,"        zr = dcmplx(1.0d0,0.0d0)*r2\n");
    fprintf(redfp,"        zi = dcmplx(0.0d0,1.0d0)*r2\n");
    fprintf(redfp,"      endif\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"      ja = la\n");
    fprintf(redfp,"      if(lt(ja) .ne. 2) then\n");
    fprintf(redfp,"        write(*,*) 'smcpol: cannot calculate ");
    fprintf(redfp,"circlar polarization'\n");
    fprintf(redfp,"        write(*,*) 'lt = ', lt\n");
    fprintf(redfp,"        write(*,*) 'lt(',ja,') = ', lt(ja), ' <> ");
    fprintf(redfp,"2'\n");
    fprintf(redfp,"        stop\n");
    fprintf(redfp,"      endif\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"      ibas  = 1\n");
    fprintf(redfp,"      do 10 i = 1, nextn\n");
    fprintf(redfp,"        jv(i) = ibas\n");
    fprintf(redfp,"        ibas  = ibas*lvt(i)\n");
    fprintf(redfp,"        j(i)  = 0\n");
    fprintf(redfp,"   10 continue\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"  100 continue\n");
    fprintf(redfp,"        iv = 0\n");
    fprintf(redfp,"        do 110 i = 1, nextn\n");
    fprintf(redfp,"          iv = iv +   jv(i)*j(i)\n");
    fprintf(redfp,"  110   continue\n");
    fprintf(redfp,"        iw = iv + jv(ja)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"        ztv =   zr*avt(iv) + zi*avt(iw)\n");
    fprintf(redfp,"        ztw =   zr*avt(iv) - zi*avt(iw)\n");
    fprintf(redfp,"        avt(iv) = ztv\n");
    fprintf(redfp,"        avt(iw) = ztw\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"c       write(*,*) 'smcpol:', j\n");
    fprintf(redfp,"c       write(*,*) 'smcpol:', iv, iw, avt(iv)\n");
    fprintf(redfp,"c       write(*,*) 'smcpol:', iw, iv, avt(iw)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"        ii = 1\n");
    fprintf(redfp,"  120   continue\n");
    fprintf(redfp,"          if(ii .eq. ja) then\n");
    fprintf(redfp,"            ii = ii + 1\n");
    fprintf(redfp,"            if(ii.gt.nextn) then\n");
    fprintf(redfp,"              goto 190\n");
    fprintf(redfp,"            endif\n");
    fprintf(redfp,"          endif\n");
    fprintf(redfp,"          j(ii) = j(ii) + 1\n");
    fprintf(redfp,"          if(j(ii).ge.lvt(ii)) then\n");
    fprintf(redfp,"            j(ii) = 0\n");
    fprintf(redfp,"            ii = ii + 1\n");
    fprintf(redfp,"            if(ii.le.nextn) then\n");
    fprintf(redfp,"              goto 120\n");
    fprintf(redfp,"            else\n");
    fprintf(redfp,"              goto 190\n");
    fprintf(redfp,"            endif\n");
    fprintf(redfp,"          endif\n");
    fprintf(redfp,"        goto 100\n");
    fprintf(redfp,"  190 continue\n");
    fprintf(redfp,"      return\n");
    fprintf(redfp,"      end\n");
    fprintf(redfp,"\n");
}

/*==================================================================*/
Private Void    ftampsum(NoArg)
/*------------------------------------------------------------------*
 *------------------------------------------------------------------*/
{
    cmopen(ROFORT, "ampsum.f");
    cmsepar('*');
    if( !vcode ) 
        fprintf(redfp,"      subroutine ampsum(ans)\n");
    else
        fprintf(redfp,"      subroutine ampsum(nsampl,ans)\n");
    fprintf(redfp,"      implicit real*8(a-h,o-z)\n");
    if( vcode )
        fprintf(redfp,"      include 'inclvs.f'\n");
    fprintf(redfp,"      include 'incl1.f'\n");
    fprintf(redfp,"      include 'inclk.f'\n");
    if( !vcode ) {
        fprintf(redfp,"      real*8     ans\n");
        fprintf(redfp,"      complex*16 anc, amp, ampp(ngraph)\n");
    } else {
        fprintf(redfp,"      real*8     ans(nsize)\n");
        fprintf(redfp,"      complex*16 anc(nsize), amp(nsize),");
        fprintf(redfp," ampp(ngraph)\n");
    }
    cmsepar('-');

    fprintf(redfp,"      jsum = 1\n");
    fprintf(redfp,"      do 10 ie = 1, nextn\n");
    fprintf(redfp,"        jsum = jsum*(jhe(ie) - jhs(ie) + 1)\n");
    fprintf(redfp,"   10 continue\n");
    fprintf(redfp,"\n");

    if( !vcode ) {
        fprintf(redfp,"      anc = 0\n");
    } else {
        fprintf(redfp,"      do 15 j = 1, nsampl\n");
        fprintf(redfp,"         anc(j) = 0\n");
        fprintf(redfp,"   15 continue\n");
    }
    fprintf(redfp,"      do 20 ig = 1, jgraph\n");
    if( !vcode ) {
        fprintf(redfp,"         ancp(ig) = 0.0d0\n");
    } else {
        fprintf(redfp,"      do 20 j = 1, nsampl\n");
        fprintf(redfp,"         ancp(j,ig) = 0.0d0\n");
    }
    fprintf(redfp,"   20   continue\n");
    fprintf(redfp,"\n");

    fprintf(redfp,"*     sum over helicities\n");
    fprintf(redfp,"      do 100 ind = 0, jsum - 1\n");
    fprintf(redfp,"\n");

    fprintf(redfp,"*       sum over graphs\n");
    if(!qcd) {
       if( !vcode ) {
           fprintf(redfp,"        amp  = 0.0d0\n");
       } else {
           fprintf(redfp,"        do 110 j = 1, nsampl\n");
           fprintf(redfp,"           amp(j)  = 0.0d0\n");
           fprintf(redfp,"  110   continue\n");
       }
    }
    fprintf(redfp,"        do 200 ig = 1, jgraph\n");
    if( !vcode ) {
        fprintf(redfp,"          ampp(ig) = aprop(ig)*ag(ind,ig)\n");
    } else {
        fprintf(redfp,"        do 200 j = 1, nsampl\n");
        fprintf(redfp,"           ampp(j,ig) = aprop(j,ig)*");
        fprintf(redfp,"ag(j,ind,ig)\n");
    }
    if(!qcd) {
       if( !vcode )
           fprintf(redfp,"          amp = amp + ampp(ig)\n");
       else
           fprintf(redfp,"          amp(j) = amp(j) + ampp(j,ig)\n");
    }
    fprintf(redfp,"  200   continue\n");
    fprintf(redfp,"\n");

    fprintf(redfp,"        do 300 ig = 1, jgraph\n");
    if( !vcode ) {
        fprintf(redfp,"          ancp(ig) = ancp(ig) + ampp(ig)");
        fprintf(redfp,"*conjg(ampp(ig))\n");
    } else {
        fprintf(redfp,"        do 300 j = 1, nsampl\n");
        fprintf(redfp,"          ancp(j,ig) = ancp(j,ig) + ampp(j,ig)");
        fprintf(redfp,"*conjg(ampp(ig))\n");
    }
    fprintf(redfp,"  300   continue\n");

    if(cprocess->extcol) {
        fprintf(redfp,"        do 400 ig = 1, jgraph\n");
        fprintf(redfp,"        do 400 jg = 1, jgraph\n");
        if( !vcode ) {
            fprintf(redfp,"          anc = anc +");
            fprintf(redfp,"  cf(igraph(ig),igraph(jg))\n");
            fprintf(redfp,"     &               ");
            fprintf(redfp,"*ampp(ig)*conjg(ampp(jg))\n");
        } else {
            fprintf(redfp,"        do 400 j = 1, nsampl\n");
            fprintf(redfp,"          anc(j) = anc(j) +");
            fprintf(redfp,"  cf(igraph(ig),igraph(jg))\n");
            fprintf(redfp,"     &               ");
            fprintf(redfp,"*ampp(j,ig)*conjg(ampp(j,jg))\n");
        }
        fprintf(redfp,"  400 continue\n");
    } else {
        if( !vcode ) {
            fprintf(redfp,"      anc = anc + amp*conjg(amp)\n");
        } else {
            fprintf(redfp,"        do 400 j = 1, nsampl\n");
            fprintf(redfp,"           anc(j) = anc(j) + ");
            fprintf(redfp,"amp(j)*conjg(amp(j))\n");
            fprintf(redfp,"  400 continue\n");
        }
    }
    fprintf(redfp,"  100 continue\n");
    if( !vcode ) {
        fprintf(redfp,"      ans = anc\n");
    } else {
        fprintf(redfp,"      do 500 j = 1, nsampl\n");
        fprintf(redfp,"         ans(j) = anc(j)\n");
        fprintf(redfp,"  500 continue\n");
    }
    fprintf(redfp,"\n");

    fprintf(redfp,"      return\n");
    fprintf(redfp,"      end\n");
    fprintf(redfp,"\n");
}

/*==================================================================*/
Private Void    ftampsmc(NoArg)
/*------------------------------------------------------------------*
 ampsum subroutine for CF = 1 version
 *------------------------------------------------------------------*/
{
    int ic1,ic2;
    cmopen(ROFORT, "ampsum.f");
    cmsepar('*');
    fprintf(redfp,"*     CF version \n");
    cmsepar('*');
    if( !vcode )
        fprintf(redfp,"      subroutine ampsum(ans)\n");
    else
        fprintf(redfp,"      subroutine ampsum(nsampl,ans)\n");
    fprintf(redfp,"      implicit real*8(a-h,o-z)\n");
    if( vcode )
        fprintf(redfp,"      include 'inclvs.f'\n");
    fprintf(redfp,"      include 'incl1.f'\n");
    fprintf(redfp,"      include 'inclk.f'\n");
    if( !vcode ) {
        fprintf(redfp,"      real*8     ans\n");
        fprintf(redfp,"      complex*16 anc\n");
    } else {
        fprintf(redfp,"      real*8     ans(nsize)\n");
        fprintf(redfp,"      complex*16 anc(nsize)\n");
    }
    cmsepar('-');

    fprintf(redfp,"      jsum = 1\n");
    fprintf(redfp,"      do 10 ie = 1, nextn\n");
    fprintf(redfp,"        jsum = jsum*(jhe(ie) - jhs(ie) + 1)\n");
    fprintf(redfp,"   10 continue\n");
    fprintf(redfp,"\n");

    if( !vcode ) {
        fprintf(redfp,"      anc = 0\n");
    } else {
        fprintf(redfp,"      do 15 j = 1, nsampl\n");
        fprintf(redfp,"         anc(j) = 0\n");
        fprintf(redfp,"   15 continue\n");
    }
    fprintf(redfp,"\n");

    fprintf(redfp,"*     sum over helicities\n");
    fprintf(redfp,"      do 100 ind = 0, jsum - 1\n");
    fprintf(redfp,"\n");

    fprintf(redfp,"*       sum over color bases\n");
    if( !vcode ) {
        fprintf(redfp,"        anc = anc ");
    } else {
        fprintf(redfp,"        do 100 j = 1, nsampl\n");
        fprintf(redfp,"          anc(j) = anc(j) ");
    }
 
    if( cbase->nbase > 1 ) { 
      for(ic1=0;ic1<cbase->nbase;ic1++) {
        for(ic2=0;ic2<cbase->nbase;ic2++) {
            if( ic1+ic2 != 0 )  fprintf(redfp,"     .              ");
            if( !vcode ) {
                fprintf(redfp," + cfmtx(%d,%d)*agc(ind,%d)*",ic1,ic2,ic1);
                fprintf(redfp,"conjg(agc(ind,%d))\n",ic2);
            } else {
                fprintf(redfp," + cfmtx(%d,%d)*agc(j,ind,%d)*",ic1,ic2,ic1);
                fprintf(redfp,"conjg(agc(j,ind,%d))\n",ic2);
            }
        }
      }
    } else if( cbase->nbase == 1 ) {
        if( !vcode ) {
            fprintf(redfp," + agc(ind)*conjg(agc(ind))\n");
        } else {
            fprintf(redfp," + agc(j,ind)*conjg(agc(j,ind))\n");
        }
    }
    fprintf(redfp,"  100 continue\n");
    if( cbase->nbase > 1 ) {
        if( !vcode ) {
            fprintf(redfp,"      ans = anc\n");
        } else {
            fprintf(redfp,"      do 200 j = 1, nsampl\n");
            fprintf(redfp,"         ans(j) = anc(j)\n");
            fprintf(redfp,"  200 continue\n");
        }
    } else if( cbase->nbase == 1 ) {
        if( cbase->nexcolor == 0 ) {
            if( !vcode ) {
                fprintf(redfp,"      ans = anc\n");
            } else {
                fprintf(redfp,"      do 200 j = 1, nsampl\n");
                fprintf(redfp,"         ans(j) = anc(j)\n");
                fprintf(redfp,"  200 continue\n");
            }
        } else if( cbase->nexcolor == 1 ) {
            if( !vcode ) {
                fprintf(redfp,"      ans = cfmtx*anc\n");
            } else {
                fprintf(redfp,"      do 200 j = 1, nsampl\n");
                fprintf(redfp,"         ans(j) = cfmtx*anc(j)\n");
                fprintf(redfp,"  200 continue\n");
            }
        }
    }
    fprintf(redfp,"\n");

    fprintf(redfp,"      return\n");
    fprintf(redfp,"      end\n");
    fprintf(redfp,"\n");
}

/*********************************************************************
 *    6. Tools
 */
/*==================================================================*/
Private Void    ftgend(NoArg)
/*------------------------------------------------------------------*
 * end of REDUCE file.
 *------------------------------------------------------------------*/
{
    fprintf(redfp, "\n");
    fprintf(redfp, "      return\n");
    fprintf(redfp, "      end\n");
}

/*==================================================================*/
Private Void    ftvlts(NoArg)
/*------------------------------------------------------------------*
 * Initialize vltbl
 *------------------------------------------------------------------*/
{
    node    n;


    nvltbl = cgraph->nnodes;
    nvlusd = nextern;
    for(n = 0; n < nvltbl; n++) {
        /* null node */
        if(ndnode(n) == NULL) {
            vltbl[n].size = 0; 
        } else if(ndtype(n) != NDTREEV) {
            vltbl[n].size = 0; 
        } else {
            vltbl[n].size = ndlegs(n);
            if((vltbl[n].vl = 
              (vlegt*) malloc(vltbl[n].size*sizeof(vlegt)))
              == NULL) {
                fprintf(Stderr, "ftvlts: malloc failed\n");
                exit(1);
            }
            ftmlist(n, vltbl[n].vl, vltbl[n].size);
        }
    }
}

/*==================================================================*/
Private Void    ftvlcp(vl1, vl2, l1, l2)
vlegt   *vl1, *vl2;
int     l1, l2;
/*------------------------------------------------------------------*
 * Copy vl2[l2] to vl1[l1]
 *------------------------------------------------------------------*/
{
    vl1[l1].p = vl2[l2].p;
    vl1[l1].t = vl2[l2].t;
    vl1[l1].i = vl2[l2].i;
    vl1[l1].m = vl2[l2].m;
    vl1[l1].d = vl2[l2].d;
}

/*==================================================================*/
Private Void    ftvlex(vl, l1, l2)
vlegt   *vl;
int     l1, l2;
/*------------------------------------------------------------------*
 * Exchange l1 and l2 elements of vl
 *------------------------------------------------------------------*/
{
    pindex p;
    int    t, i, m, d;

    p        = vl[l1].p;
    vl[l1].p = vl[l2].p;
    vl[l2].p = p;

    t        = vl[l1].t;
    vl[l1].t = vl[l2].t;
    vl[l2].t = t;

    i        = vl[l1].i;
    vl[l1].i = vl[l2].i;
    vl[l2].i = i;

    m        = vl[l1].m;
    vl[l1].m = vl[l2].m;
    vl[l2].m = m;

    d        = vl[l1].d;
    vl[l1].d = vl[l2].d;
    vl[l2].d = d;
}

/*==================================================================*/
Public  Void    ftvldmp(NoArg)
/*------------------------------------------------------------------*
 * Dump vltbl
 *------------------------------------------------------------------*/
{
    int n, l;

    fprintf(Stderr, "\nDump vltbl: nvltbl = %d, nvlusd = %d\n",
                    nvltbl, nvlusd);
    if(nvltbl < 0)
        return;
    for(n = 0; n < nvltbl; n++) {
        fprintf(Stderr, "%4d, size = %4d\n", n, vltbl[n].size);
        for(l = 0; l < vltbl[n].size; l++) {
            fprintf(Stderr, "p=%4d, t=%4d, i=%4d, m=%4d, d=%4d\n",
                    vltbl[n].vl[l].p, vltbl[n].vl[l].t,
                    vltbl[n].vl[l].i, vltbl[n].vl[l].m,
                    vltbl[n].vl[l].d);
        }
    }
}

/*==================================================================*/
Private Void    ftvlte(NoArg)
/*------------------------------------------------------------------*
 * Clear vltbl
 *------------------------------------------------------------------*/
{
    node    n;

    for(n = 0; n < nvltbl; n++) {
        if(vltbl[n].size != 0)
            free(vltbl[n].vl);
    }
    nvltbl = 0;
}

/*==================================================================*/
Private Void    ftnode(NoArg)
/*------------------------------------------------------------------*
 * Output node.
 *------------------------------------------------------------------*/
{
    int    nnodes = cgraph->nnodes, ttype;
    node    n;
    pindex p, vf;
    vindex vt;
    int    s, l, d, a, tp, nlg;
    tnode  *nd;
    tleg   *lg;
    vlegt  *vl;
    bool   ok;


    fprintf(tfp1, "\n* Vertices (%d)\n", cgraph->nnodes);

    nvltbl = nnodes;
    for(n = 0; n < nnodes; n++) {
        /* null node */
        if(ndnode(n) == NULL) {
            fprintf(tfp1, "      null%d", n);
        }

        /* name of the vertex */
        if(ndtype(n) != NDTREEV) {
        } else {
            vl  = vltbl[n].vl;
            nlg = vltbl[n].size;
#if DBG
            fprintf(tfp1, "\n");
            for(l = 0; l < nlg; l++) {
                nd  = ndnode(n);
                lg  = ndleg(n, l);
                p   = abs(ndparticle(n, l));
                cmmom(momstr, n, l, True);
    
                fprintf(tfp1, "*     %d(%d): %c pe%d %s\n",
                    n, l, csign(vl[l].d), vl[l].m, 
                    pvename(nd->nvkind, vl[l].p));
            }
#endif
            vt = ndvertex(n);
            if(vt == VNULL) {
                fprintf(tfp1, "       vnull(%d)", n);
                continue;
            }

            Discard ftnvtype(momstr, n);

            ok = True;
            if(strcmp(momstr, "ffv") == 0) {
                if(majorana)
                    ftmffv(n, vl);
                else
                    ftnffv("ffv", n, vl);
            } else if(strcmp(momstr, "sff") == 0) {
                if(majorana)
                    ftmffs(n, vl);
                else
                    ftnffs("ffs", n, vl);
            } else if(strcmp(momstr, "vvv") == 0) {
                ftnvvv(n, vl);
            } else if(strcmp(momstr, "svv") == 0) {
                ftnsvv(n, vl);
            } else if(strcmp(momstr, "ssv") == 0) {
                ftnssv(n, vl);
            } else if(strcmp(momstr, "sss") == 0) {
                ftnsss(n, vl);
            } else if(strcmp(momstr, "vvvv") == 0) {
                ftnvvvv(n, vl);
            } else if(strcmp(momstr, "ssvv") == 0) {
                ftnssvv(n, vl);
            } else if(strcmp(momstr, "ssss") == 0) {
                ftnssss(n, vl);
            } else if(strcmp(momstr, "smf") == 0) {
                if(majorana)
                    ftmmfs(n, vl);
                else
                    ok = False;
            } else if(strcmp(momstr, "mfv") == 0) {
                if(majorana)
                    ftmmfv(n, vl);
                else
                    ok = False;
            } else if(strcmp(momstr, "smm") == 0) {
                if(majorana)
                    ftmmms(n, vl);
                else
                    ok = False;
            } else if(strcmp(momstr, "mmv") == 0) {
                if(majorana)
                    ftmmmv(n, vl);
                else
                    ok = False;
            } else {
                ok = False;
            }
            if(!ok)
                fprintf(tfp1, "      call ??%s(%d)??\n", momstr, n);
        }
    }
}

/*==================================================================*/
Private Void    ftmffv(n, vl)
node    n;
vlegt   vl[];
/*------------------------------------------------------------------*
 * Output ffv vertex in case with majorana fermion.
 *------------------------------------------------------------------*/
{
    pindex p0, p1;

    p0 = ndparticle(n, 0);
    p1 = ndparticle(n, 1);

    if(p0 < 0) {
        if(p1 > 0) {
            ftnffv("ffv", n, vl);
            return;
        }
    } else {
        if(p1 < 0) {
            ftnffv("ffvt", n, vl);
            return;
        }
    }
    error("ftmffv", "Illegal fermion line");
    fprintf(Stderr, "*** n = %d ***\n", n);
    ftvldmp();
    exit(1);
}

/*==================================================================*/
Private Void    ftmffs(n, vl)
node    n;
vlegt   vl[];
/*------------------------------------------------------------------*
 * Output ffv vertex in case with majorana fermion.
 *------------------------------------------------------------------*/
{
    pindex p0, p1;

    p0 = ndparticle(n, 1);
    p1 = ndparticle(n, 2);

    if(p0 < 0) {
        if(p1 > 0) {
            ftnffs("ffs", n, vl);
            return;
        }
    } else {
        if(p1 < 0) {
            ftnffs("ffst", n, vl);
            return;
        }
    }
    error("ftmffs", "Illegal fermion line");
    fprintf(Stderr, "*** n = %d ***\n", n);
    ftvldmp();
    exit(1);
}

/*==================================================================*/
Private bool    ftmleg(n, vl, m, f)
node    n;
vlegt   vl[];
int     *m, *f;
/*------------------------------------------------------------------*
 * Find leg of majorana and dirac fermions.
 *------------------------------------------------------------------*/
{
    int l, err = 0;

    *m = *f = -1;
    for(l = 0; l < 3; l++) {
        if(ptable[vl[l].p]->ptype == P_MAJORANA) {
            if(*m < 0)
                *m = l;
            else
                err |= 1;
        } else if(ptable[vl[l].p]->ptype == P_FERMION) {
            if(*f < 0)
                *f = l;
            else
                err |= 2;
        }
    }
    if(*m < 0) {
        error("flmleg", "No Majorana particle");
        return False;
    }
    if(*f < 0) {
        error("flmleg", "No Dirac particle");
        return False;
    }
    if(err & 1) {
        error("flmleg", "No too many Majorana particle");
        return False;
    }
    if(err & 2) {
        error("flmleg", "No too many Dirac particle");
        return False;
    }
    return True;
}

/*==================================================================*/
Private Void    ftmmfv(n, vl)
node    n;
vlegt   vl[];
/*------------------------------------------------------------------*
 * Output ffv vertex in case with majorana fermion.
 *------------------------------------------------------------------*/
{
    int m, f;
    pindex p0, p1;

    if(ftmleg(n, vl, &m, &f)) {
        p0 = ndparticle(n, m);
        p1 = ndparticle(n, f);

        if(m == 0 && p0 < 0) {
            if(f == 1) {
                if(p1 > 0) {
                    ftnffv("ffv", n, vl);
                    return;
                } else {
                    ftnffv("ffcv", n, vl);
                    return;
                }
            }
        } else if(m == 1 && p0 > 0) {
            if(f == 0) {
                if(p1 < 0) {
                    ftnffv("ffv", n, vl);
                    return;
                } else {
                    ftnffv("ffvc", n, vl);
                    return;
                }
            }
        }
    }
    error("ftmmfv", "Illegal fermion line");
    fprintf(Stderr, "*** n = %d ***\n", n);
    ftvldmp();
    exit(1);
}

/*==================================================================*/
Private Void    ftmmfs(n, vl)
node    n;
vlegt   vl[];
/*------------------------------------------------------------------*
 * Output ffs vertex in case with majorana fermion.
 *------------------------------------------------------------------*/
{
    int m, f;
    pindex p0, p1;

    if(ftmleg(n, vl, &m, &f)) {
        p0 = ndparticle(n, m);
        p1 = ndparticle(n, f);

        if(m == 1 && p0 < 0) {
            if(f == 2) {
                if(p1 > 0) {
                    ftnffs("ffs", n, vl);
                    return;
                } else {
                    ftnffs("ffcs", n, vl);
                    return;
                }
            }
        } else if(m == 2 && p0 > 0) {
            if(f == 1) {
                if(p1 < 0) {
                    ftnffs("ffs", n, vl);
                    return;
                } else {
                    ftnffs("ffsc", n, vl);
                    return;
                }
            }
        }
    }
    error("ftmmfs", "Illegal fermion line");
    fprintf(Stderr, "*** n = %d ***\n", n);
    ftvldmp();
    exit(1);
}

/*==================================================================*/
Private Void    ftmmmv(n, vl)
node    n;
vlegt   vl[];
/*------------------------------------------------------------------*
 * Output ffv vertex in case with majorana fermion.
 *------------------------------------------------------------------*/
{
    pindex p0, p1;

    p0 = ndparticle(n, 0);
    p1 = ndparticle(n, 1);

    if(p0 < 0) {
        if(p1 > 0) {
            ftnffv("ffv", n, vl);
                return;
            }
    }
    error("ftmmmv", "Illegal fermion line");
    fprintf(Stderr, "*** n = %d ***\n", n);
    ftvldmp();
    exit(1);
}

/*==================================================================*/
Private Void    ftmmms(n, vl)
node    n;
vlegt   vl[];
/*------------------------------------------------------------------*
 * Output ffv vertex in case with majorana fermion.
 *------------------------------------------------------------------*/
{
    pindex p0, p1;

    p0 = ndparticle(n, 1);
    p1 = ndparticle(n, 2);

    if(p0 < 0) {
        if(p1 > 0) {
            ftnffs("ffs", n, vl);
                return;
            }
    }
    error("ftmmms", "Illegal fermion line");
    fprintf(Stderr, "*** n = %d ***\n", n);
    ftvldmp();
    exit(1);
}

/*==================================================================*/
Private int     ftuorv(vl, l, sgn)
vlegt   vl[];
int     l;
int     sgn;
/*------------------------------------------------------------------*
 * Output ffv vertex.
 *------------------------------------------------------------------*/
{
    int uorv;    

    uorv = 0;
    if(ptable[vl[l].p]->ptype == P_MAJORANA) {
        if(vl[l].t == ILEXTERN) {
            if(sgn*vl[l].d > 0)
                uorv = 2;
            else
                uorv = 1;
        }
    }
    return uorv;
}

/*==================================================================*/
Private Void    ftnffv(name, n, vl)
string  name;
node    n;
vlegt   vl[];
/*------------------------------------------------------------------*
 * Output ffv vertex.
 *    leg : 0=f-bar, 1=f, 2=v
 *------------------------------------------------------------------*/
{
    int    uorv[2];

    ftvlex(vl, 0, 1);
    uorv[0] = ftuorv(vl, 0, -1);
    uorv[1] = ftuorv(vl, 1,  1);
    
    fprintf(tfp1, "      call sm%s(", name);
    if( vcode )
        fprintf(tfp1, "nsampl,");
    fthindx(n, vl, 3);                             /* l1, l2, lv, */
    fprintf(tfp1, "ew%d%s,ew%d%s,",                /* ew1, ew2, */
            vl[0].m, ftsuorv[uorv[0]], vl[1].m, ftsuorv[uorv[1]]);
    ftmassl(n, vl, 0, 1);                          /* am1, am2, */
    ftncoupl(n);                                   /* cpl, */
    fprintf(tfp1, "\n     &           ");
    fprintf(tfp1, "ce%d%s,ce%d%s,",                /* ce1, ce2, */
            vl[0].m, ftsuorv[uorv[0]], vl[1].m, ftsuorv[uorv[1]]);
    fprintf(tfp1, "ps%d%s,ps%d%s,",                /* ps1, ps2, */
            vl[0].m, ftsuorv[uorv[0]], vl[1].m, ftsuorv[uorv[1]]);
    fprintf(tfp1, "ep%d,", vl[2].m);               /* ep */
    ftcrans(n, vl, 3);                             /* lt, av */
    fprintf(tfp1, ")\n");
}

/*==================================================================*/
Private Void    ftnffs(name, n, vl)
string  name;
node    n;
vlegt   vl[];
/*------------------------------------------------------------------*
 * Output ffs vertex.
 *    leg : 0=f-bar, 1=f, 2=s    ???
 *------------------------------------------------------------------*/
{
    int    uorv[2];

    ftvlex(vl, 0, 2);
    uorv[0] = ftuorv(vl, 0, -1);
    uorv[1] = ftuorv(vl, 1,  1);
    
    fprintf(tfp1, "      call sm%s(", name);
    if( vcode )
        fprintf(tfp1, "nsampl,");
    fthindx(n, vl, 3);                             /* l1, l2, lv, */
    fprintf(tfp1, "ew%d%s,ew%d%s,",                /* ew1, ew2, */
            vl[0].m, ftsuorv[uorv[0]], vl[1].m, ftsuorv[uorv[1]]);
    ftmassl(n, vl, 0, 1);                          /* am1, am2, */
    ftncoupl(n);                                   /* cpl, */
    fprintf(tfp1, "\n     &           ");
    fprintf(tfp1, "ce%d%s,ce%d%s,",                /* ce1, ce2, */
            vl[0].m, ftsuorv[uorv[0]], vl[1].m, ftsuorv[uorv[1]]);
    fprintf(tfp1, "ps%d%s,ps%d%s,",                /* ps1, ps2, */
            vl[0].m, ftsuorv[uorv[0]], vl[1].m, ftsuorv[uorv[1]]);
    ftcrans(n, vl, 3);                  /* lt, av */
    fprintf(tfp1, ")\n");
}

/*==================================================================*/
Private Void    ftnvvv(n, vl)
node    n;
vlegt   vl[];
/*------------------------------------------------------------------*
 * Output vvv vertex.
 *------------------------------------------------------------------*/
{
    int gl3 = pvgl3(n);

    if(gl3) {
        fprintf(tfp1, "      call smggg(");
        if( vcode )
            fprintf(tfp1, "nsampl,");
        fthindx(n, vl, 3);                        /* l1, l2, lv, */
        fprintf(tfp1, "%c1,%c1,%c1,", 
            csign(vl[0].d), csign(vl[1].d), csign(vl[2].d));
        ftncoupl(n);                                   /* cpl, */
        fprintf(tfp1, "\n     &           ");
        fprintf(tfp1, "pe%d,pe%d,pe%d,", vl[0].m, vl[1].m, vl[2].m);
        fprintf(tfp1, "ep%d,ep%d,ep%d,", vl[0].m, vl[1].m, vl[2].m);
        ftcrans(n, vl, 3);                  /* lt, av */
        fprintf(tfp1, ")\n");
    } else {
        if( jano3v )
            fprintf(tfp1, "      if( jano3v .eq. 0 ) then\n");
        fprintf(tfp1, "         call smvvv(");
        if( vcode )
            fprintf(tfp1, "nsampl,");
        fthindx(n, vl, 3);                        /* l1, l2, lv, */
        fprintf(tfp1, "%c1,%c1,%c1,", 
            csign(vl[0].d), csign(vl[1].d), csign(vl[2].d));
        ftncoupl(n);                                   /* cpl, */
        fprintf(tfp1, "\n     &              ");
        fprintf(tfp1, "pe%d,pe%d,pe%d,", vl[0].m, vl[1].m, vl[2].m);
        fprintf(tfp1, "ep%d,ep%d,ep%d,", vl[0].m, vl[1].m, vl[2].m);
        ftcrans(n, vl, 3);                  /* lt, av */
        fprintf(tfp1, ")\n");
        if( jano3v ) {
            fprintf(tfp1, "      else\n");
            fprintf(tfp1, "         call smvww(");
            if( vcode )
                fprintf(tfp1, "nsampl,");
            fthindx(n, vl, 3);                      /* l1, l2, lv, */
            fprintf(tfp1, "%c1,%c1,%c1,", 
                csign(vl[0].d), csign(vl[1].d), csign(vl[2].d));
            ftncoupl(n);                                 /* cpl, */
            fprintf(tfp1, "\n     &              ");
            fprintf(tfp1, "pe%d,pe%d,pe%d,", vl[0].m, vl[1].m, vl[2].m);
/* CHECK ordering of vww: 2nd,3rd are 'w' */
            if( ptable[vl[1].p]->kfcode != 24 || 
                ptable[vl[2].p]->kfcode != 24  ) {
                printf("Invalid ordering vww %s %s %s \n",
                ptable[vl[0].p]->mass, ptable[vl[1].p]->mass,
                ptable[vl[2].p]->mass);
                printf("Please send a mail to tishika@minami.kek.jp\n");
                exit(1);
            }
            if( ptable[vl[0].p]->kfcode == 22 ) {
                fprintf(tfp1, "1.0d0,andka,anlma,%s**2,",
                ptable[vl[1].p]->mass);
            } else if( ptable[vl[0].p]->kfcode == 23 ) {
                fprintf(tfp1, "1.0d0,andkz,anlmz,%s**2,",
                ptable[vl[1].p]->mass);
            } else {
                printf("Invalid vector partilce vww %s\n",
                ptable[vl[0].p]->mass);
                printf("Please send a mail to tishika@minami.kek.jp\n");
                exit(1);
            }
            fprintf(tfp1, "\n     &              ");
            fprintf(tfp1, "ep%d,ep%d,ep%d,", vl[0].m, vl[1].m, vl[2].m);
/*        ftcrans(n, vl, 3);                *//* lt, av */
/*  TI 95/11/21 */
            fprintf(tfp1, "lt%d,av%d",n,n);
            fprintf(tfp1, ")\n");
            fprintf(tfp1, "      endif");
        }
    }
}

/*==================================================================*/
Private Void    ftnsvv(n, vl)
node    n;
vlegt   vl[];
/*------------------------------------------------------------------*
 * Output svv vertex.
 *------------------------------------------------------------------*/
{
    fprintf(tfp1, "      call smsvv(");
    if( vcode )
        fprintf(tfp1, "nsampl,");
    fthindx(n, vl, 3);                             /* l2, l3 */
    ftncoupl(n);                                   /* cpl, */
    fprintf(tfp1, "ep%d,ep%d,", vl[1].m, vl[2].m);
    ftcrans(n, vl, 3);                                     /* lt, av */
    fprintf(tfp1, ")\n");
}

/*==================================================================*/
Private Void    ftnssv(n, vl)
node    n;
vlegt   vl[];
/*------------------------------------------------------------------*
 * Output ssv vertex.
 *------------------------------------------------------------------*/
{
    fprintf(tfp1, "      call smssv(");
    if( vcode )
        fprintf(tfp1, "nsampl,");
    fthindx(n, vl, 3);                             /* lv, */
    fprintf(tfp1, "%c1,%c1,", csign(vl[0].d), csign(vl[1].d));
    ftncoupl(n);                                   /* cpl, */
    fprintf(tfp1, "pe%d,pe%d,", vl[0].m, vl[1].m);
    fprintf(tfp1, "ep%d,", vl[2].m);
    ftcrans(n, vl, 3);                                     /* lt, av */
    fprintf(tfp1, ")\n");
}

/*==================================================================*/
Private Void    ftnsss(n, vl)
node    n;
vlegt   vl[];
/*------------------------------------------------------------------*
 * Output ssv vertex.
 *------------------------------------------------------------------*/
{
    fprintf(tfp1, "      call smsss(");
    if( vcode )
        fprintf(tfp1, "nsampl,");
    ftncoupl(n);                                   /* cpl, */
    fthindx(n, vl, 3);                             /* lv, */
    ftcrans(n, vl, 3);                                     /* lt, av */
    fprintf(tfp1, ")\n");
}

/*==================================================================*/
Private Void    ftnvvvv(n, vl)
node    n;
vlegt   vl[];
/*------------------------------------------------------------------*
 * Output vvvv vertex.
 *------------------------------------------------------------------*/
{
    int gl4 = ndgl4(n);

    if(gl4 == 0) 
        fprintf(tfp1, "      call smvvvv(");
    else if(1 <= gl4 && gl4 <= 3)
        fprintf(tfp1, "      call smgggg(");
    else {
        error("ftnvvvv", "illegal ndgl4");
        fprintf(Stderr, "n = %d, gl4=%d\n", n, gl4);
        erfail();
    }
    if( vcode )
        fprintf(tfp1, "nsampl,");
    fthindx(n, vl, 4);                           /* l1, l2, l3, l4 */
    ftncoupl(n);                                 /* cpl, */
    fprintf(tfp1, "\n     &           ");
    fprintf(tfp1, "ep%d,ep%d,ep%d,ep%d,",
            vl[glue4leg[gl4][0]].m, vl[glue4leg[gl4][1]].m, 
            vl[glue4leg[gl4][2]].m, vl[glue4leg[gl4][3]].m);
    ftcrans(n, vl, 4);                                   /* lt, av */
    fprintf(tfp1, ")\n");
}

/*==================================================================*/
Private Void    ftnssvv(n, vl)
node    n;
vlegt   vl[];
/*------------------------------------------------------------------*
 * Output ssss vertex.
 *------------------------------------------------------------------*/
{
    fprintf(tfp1, "      call smssvv(");
    if( vcode )
        fprintf(tfp1, "nsampl,");
    fthindx(n, vl, 4);                           /* l3, l4 */
    ftncoupl(n);                                 /* cpl, */
    fprintf(tfp1, "ep%d,ep%d,", vl[2].m, vl[3].m);
    ftcrans(n, vl, 4);                                   /* lt, av */
    fprintf(tfp1, ")\n");
}

/*==================================================================*/
Private Void    ftnssss(n, vl)
node    n;
vlegt   vl[];
/*------------------------------------------------------------------*
 * Output ssss vertex.
 *------------------------------------------------------------------*/
{
    fprintf(tfp1, "      call smssss(");
    if( vcode )
        fprintf(tfp1, "nsampl,");
    ftncoupl(n);                                 /* cpl, */
    ftcrans(n, vl, 4);                                   /* lt, av */
    fprintf(tfp1, ")\n");
}

/*==================================================================*/
Private Void    fthindx(n, vl, size)
node    n;
vlegt   vl[];
int     size;
/*------------------------------------------------------------------*
 * Output degree of freedom of helicity
 *------------------------------------------------------------------*/
{
    int l, tp;

    for(l = 0; l < size; l++) {
        if(ptable[vl[l].p]->ptype == P_SCALAR)
            continue;
        tp  = iltable[vl[l].i].iltype;
        Discard cmindx(momstr1, tp, vl[l].p);
        fprintf(tfp1, "%s,", momstr1);
    }
}

/*==================================================================*/
Private Void    ftncoupl(n)
node    n;
/*------------------------------------------------------------------*
 * Output coupling constant
 *------------------------------------------------------------------*/
{
    Discard cmcoupl(ndvertex(n), 0, ndloop(n), momstr);
    fprintf(tfp1, "%s,", momstr);
}

/*==================================================================*/
Private Void    ftmassl(n, vl, bgn, end)
node    n;
vlegt   vl[];
int     bgn, end;
/*------------------------------------------------------------------*
 * Output the list of mass
 *------------------------------------------------------------------*/
{
    int l;

    for(l = bgn; l <= end; l++)
        fprintf(tfp1, "%s,", ptable[vl[l].p]->mass);
}


/*==================================================================*/
Private Void    ftcrans(n, vl, size)
node    n;
vlegt   vl[];
int     size;
/*------------------------------------------------------------------*
 * Output amplitude table variable.
 *------------------------------------------------------------------*/
{
    long int lsize;
    int l, lstr, ll;
    char *str;
    char vwork[MAXPATHLEN];

    if(nvltbl - nvlusd == 1) {
        fprintf(tfp1, "lt,av");
        return;
    }

    sprintf(momstr1, "lt%d", n);
    sprintf(momstr2, "0:%d", size);
    fprintf(tfp1, "%s,", momstr1);
    fttmp(FT_INT, momstr1, momstr2, size+1);

    lstr = 0;
    str = sizestr;
    for(l = 0, lsize = 1; l < size; l++) {
        if(lstr + 9 >= 256) {
            fprintf(Stderr, "ftcrans: too long string\n");
            str[1] = '\0';
            fprintf(Stderr, "<%s>\n", sizestr);
            exit(1);
        }
        lsize *= cmindx(str, vl[l].t, vl[l].p);
        ll = strlen(str);
        str += ll;
        str[0] = '*';
        str++;
        lstr += ll + 1;
    }
    str--;
    str[0] = '\0';

    sprintf(momstr1, "av%d", n);
    sprintf(momstr2, "%ld", lsize);
    fprintf(tfp1, "%s", momstr1);
    if( !vcode )
        fttmp(FT_COMPLEX, momstr1, sizestr, ((int)lsize));
    else {
        strcpy(vwork,"nsize,");
        strcat(vwork,sizestr);
        fttmp(FT_COMPLEX, momstr1, vwork, ((int)lsize));
    }
}

/*==================================================================*/
Private Void    ftmlist(n, vl, size)
node    n;
vlegt   vl[];
int     size;
/*------------------------------------------------------------------*
 * Output list of coupling constants.
 *------------------------------------------------------------------*/
{
    int l, il;

    /* momentum and polarization */
    for(l = 0; l < size; l++) {

        /* particle */
        vl[l].p = abs(ndparticle(n,l));

        /* internal line */
        vl[l].i = il = ndiline(n, l);

        /* direction */
        if(iltable[il].ilnodes[0] == n)
            vl[l].d = -1;
        else
            vl[l].d = 1;

        /* momentum number */
        if((vl[l].t = iltable[il].iltype) == ILEXTERN)
            vl[l].m = il;
        else
            vl[l].m = il + momoffs;
    }
}

/*==================================================================*/
Private Void    ftiline(NoArg)
/*------------------------------------------------------------------*
 * output an internal line
 *------------------------------------------------------------------*/
{
    int    il, s, lg, tp, sz, c;
    node   n;
    pindex p;
    string st;
    char vwork[MAXPATHLEN];

    if(nilines < 1) {
        fprintf(tfp1, "* %d internal momenta\n", nilines);
        return;
    }
        
    fprintf(tfp1, "\n* Internal momenta\n");
    for(il = 1; il < nilines; il++) {
        if(iltable[il].iltype == ILEXTERN)
            continue;
        else if(iltable[il].iltype == ILLOOP)
            continue;

        n   = iltable[il].ilnodes[0];
        lg  = iltable[il].illegs[0];
        p   = abs(ndparticle(n, lg));
        s   = ptable[p]->ptype;
        tp  = iltable[il].iltype;
        c   = cmvector(p);
    
        if(s == P_VECTOR) {        /* vector */
            cmimom(momstr, il, 1);
            sz = cmindx(momstr1, tp, p);
            st = momstr + 2;
            ftgaug(momstr2, p);

            if(jtgamm && c == 'a') {
/*                fprintf(tfp1, "      if(jtgamm.eq.0) then\n");*/
                if( !vcode ) {
                    fprintf(tfp1, "      if(jtgamm .eq. 0 .or. ");
                    fprintf(tfp1, "vm%s .gt. 0.0d0 ) then\n",st);
                    fprintf(tfp1, "          call smintv(%s,%s,",
                        momstr1, ptable[p]->mass);
                } else {
                    fprintf(tfp1, "      if(jtgamm .eq. 0 ) then\n");
                    fprintf(tfp1, "          call smintv(nsampl,%s,%s,",
                        momstr1, ptable[p]->mass);
                }
                fprintf(tfp1, "%s,ep%s,ew%s,vm%s,%s)\n",
                        momstr, st, st, st, momstr2);
                fprintf(tfp1, "      else\n");
                if( !vcode )
                    fprintf(tfp1, "          call smintp(%s,%s,",
                        momstr1, ptable[p]->mass);
                else
                    fprintf(tfp1, "          call smintp(nsampl,%s,%s,",
                        momstr1, ptable[p]->mass);
                fprintf(tfp1, "%s,ep%s,ew%s,vm%s,%s)\n",
                        momstr, st, st, st, momstr2);
                fprintf(tfp1, "      endif\n");
            } else {
                if( !vcode )
                    fprintf(tfp1, "      call smintv(%s,%s,",
                            momstr1, ptable[p]->mass);
                else
                    fprintf(tfp1, "      call smintv(nsampl,%s,%s,",
                            momstr1, ptable[p]->mass);
                fprintf(tfp1, "%s,ep%s,ew%s,vm%s,%s)\n",
                        momstr, st, st, st, momstr2);
            }
            momstr[0] = 'e';
            momstr[1] = 'w';
            if( !vcode )
                fttmp(FT_REAL, momstr, momstr1, sz);
            else {
                strcpy(vwork,"nsize,");
                strcat(vwork,momstr1);
                fttmp(FT_REAL, momstr, vwork, sz);
            }
            momstr[0] = 'e';
            momstr[1] = 'p';
            if( !vcode )
                sprintf(momstr2, "4,%s", momstr1);
            else
                sprintf(momstr2, "nsize,4,%s", momstr1);
            fttmp(FT_REAL, momstr, momstr2, 4*sz);

        } else if(!pvboson(p)) {   /* fermion */
            cmimom(momstr, il, 1);
            sz = cmindx(momstr1, tp, p);
            st = momstr + 2;

            if( !vcode )
                fprintf(tfp1, "      call smintf(%s,",
                        ptable[p]->mass);
            else
                fprintf(tfp1, "      call smintf(nsampl,%s,",
                        ptable[p]->mass);
            fprintf(tfp1, "%s, vm%s, ew%s, ps%s, ce%s)\n",
                    momstr, st, st, st, st);
            momstr[0] = 'e';
            momstr[1] = 'w';
            if( !vcode )
                fttmp(FT_REAL, momstr, "2", 2);
            else
                fttmp(FT_REAL, momstr, "nsize,2", 2);
            momstr[0] = 'p';
            momstr[1] = 's';
            if( !vcode )
                fttmp(FT_REAL, momstr, "4,3", 12);
            else
                fttmp(FT_REAL, momstr, "nsize,4,3", 12);
            momstr[0] = 'c';
            momstr[1] = 'e';
            if( !vcode )
                fttmp(FT_COMPLEX, momstr, "2,4", 8);
            else
                fttmp(FT_COMPLEX, momstr, "nsize,2,4", 8);

        } else if(s == P_SCALAR) {
            fprintf(tfp1, "*     Scalar internal line i%4.4d (%s)\n",
                il, ptable[p]->pname);

        } else {
            fprintf(tfp1, "*     ?????? internal line i%4.4d (%s)\n",
                il, ptable[p]->pname);
        }
    }
}

/*=================================================================*/
Private int     ftgaug(str, prt)
string  str;
pindex  prt;
/*------------------------------------------------------------------*
 * Construct momentum which comes to a node.
 *------------------------------------------------------------------*/
{
    int ptype, pcode, c;

    ptype = ptable[prt]->ptype;
    switch(ptype) {
        case P_SCALAR :
            c = cmvector(prt);
            if(c == 'x')
                c = 'w';
            else if(c == '3')
                c = 'z';
            else
                c = '?';
            break;
        case P_VECTOR :
            c = cmvector(prt);
            break;
        default :
            c = '?';    
            break;
    }
    sprintf(str, "igau%cb", c);
    if(c == 'a')
        return 1;
    else if(c == 'w')
        return 2;
    else if(c == 'z')
        return 3;
    else if(c == 'g')
        return 4;
    else
        return -1;
}

/*=================================================================*/
Private Void    ftdmom(str, il, sgn)
string  str;
int     il;
int     sgn;
/*------------------------------------------------------------------*
 * consturct the name of an internal line
 *------------------------------------------------------------------*/
{
    char c[3];
    int  itype;
    int  im, ii;

#if DBG
    if(dbg >= redbg)
        printf("ftdmom(%p, %d, %d)\n", str, il, sgn);
#endif

    itype = iltable[il].iltype;

    switch (itype) {
        case ILEXTERN : 
            strcpy(c, "??");
            im = il;
            for(ii = 1; ii < nilines; ii++) {
                if(iltable[il].ilmlist[ii] != 0) {
                    im = ii;
                    break;
                }
            }
            break;
        case ILTREE :
            strcpy(c, "vm");
            im = il + momoffs;
            break;
        default :
            strcpy(c, "vm");
            im = il + momoffs;
            break;
    }

    if(sgn < 0)
        sprintf(str, "-%s%d\0", c, im);
    else
        sprintf(str, "%s%d\0", c, im);

    return;
}

/*=================================================================*/
Private unsigned int ftvtype(str, len, ptype, n)
string  str;
int     len;
int     *ptype;
int     n;
/*------------------------------------------------------------------*
 * Construct the name of vertex.
 *------------------------------------------------------------------*/
{
    int  s, t, i, l = 0, rt;
    bool ok = True;

    rt = 0;
    for(t = 0; t < PTYPEMAX; t++) {
        for(i = 0; i < ptype[t]; i++) {
            *str++ = ptypesym[t];
            l++;
            rt |= pmask[t];
        }
    }

    *str = '\0';

    if(len != l)
        return -1;
    return rt;
}

/*=================================================================*/
Private unsigned int ftnvtype(str, n)
string  str;
node    n;
/*------------------------------------------------------------------*
 * construct vertex name to "str". (predicate)
 * return value
 *     -1 : particle with illegal spin
 *      0 : without fermion nor ghost
 *      1 : with fermion
 *      2 : with ghost
 *------------------------------------------------------------------*/
{
    int  lg;

    for(lg = 0; lg < PTYPEMAX; lg++)
        reptype[lg] = 0;
    for(lg = 0; lg < ndlegs(n); lg++) {
        reptype[ptable[abs(ndparticle(n, lg))]->ptype]++;
    }

    return ftvtype(str, ndlegs(n), reptype, n);
}

/*==================================================================*/
Private Void    ftscfact(NoArg)
/*------------------------------------------------------------------*
 * Calculate color factor
 */
{
    long      num, den;
    int       lp;
    int       mprc, g0, g1;
    int       p0, p1, g1s, l0, l1;
    tgraph   *graph;
    tproc    *proc, *proc0, *proc1;

    /* repeat by loop */
    for(lp = 0; lp <= maxloop; lp++) {

        /* generate processes l0 x l1, l0 >= l1 */
        for(l0 = lp0min(lp); l0 <= lp0max(lp); l0++) {

            /* first process */
            p0 = reproct[l0];       /* process number */
            proc0 = dagetp(p0);   /* process data */

            /* the second process */
            l1 = lp - l0;         /* loop */
            p1 = reproct[l1];       /* process number */
            proc1 = dagetp(p1);   /* process data */

            printf("Color factors : %d x %d\n", l0, l1);

            /* multiply process data */
            mprc = damproc(p0, p1);
            proc = dagetp(mprc);

            sprintf(momstr,"cf%dx%d.dat", p0, p1);
            cmopen(ROFORT, momstr);

            /* repeat graph with loop lp */
            for(g0 = 0; g0 < reprocg[l0]; g0++) {

                /* if two process are same, g0 >= g1 */
                if(p0 == p1)
                    g1s = g0;
                else
                    g1s = 0;

                /* repeat graph with loop l1 */
/*
                for(g1 = g1s; g1 < reprocg[l1]; g1++) {
 */
                for(g1 = 0; g1 < reprocg[l1]; g1++) {

                    /* multiply graphs */
                    graph = damult(mprc, g0, g1);

                    /* initialize i-line table */
                    ilmom();

                    cfactor(&num, &den);
                    fprintf(redfp, " %d %d %ld %ld\n", 
                           g0+1, g1+1, num, den);
#if DBG
                    if(dbg > 0) {
                        ildump();
                    }
#endif
                }
            }
        }
    }
    fclose(redfp);
}


/*==================================================================*/
Private Void    ftacfact(NoArg)
/*------------------------------------------------------------------*
 * Calculate color factor
 */
{
    long      num, den;
    int       lp;
    int       g0;
    int       p0, l0;
    node      nd;
    tgraph   *graph;
    tproc    *proc;
    int       con[MAXNODES], col[MAXNODES];

    /* repeat by loop */
    for(lp = 0; lp <= maxloop; lp++) {

        /* generate processes l0 */
        for(l0 = lp0min(lp); l0 <= lp0max(lp); l0++) {

            /* first process */
            p0 = reproct[l0];       /* process number */
            proc = dagetp(p0);   /* process data */

            printf("Color factors : %d\n", l0);

            sprintf(momstr,"cf%d.dat", p0);
            cmopen(ROFORT, momstr);

            /* repeat graph with loop lp */
            for(g0 = 0; g0 < reprocg[l0]; g0++) {

                cgraph = daread(p0, g0);

                /* initialize i-line table */
                ilmom();
                flget(cgraph);

                cfamp(&num, &den, con, col);
                fprintf(redfp, " %d %ld %ld ", g0+1, num, den);
                for(nd = 0; nd < cprocess->nextern; nd++)
                    fprintf(redfp, " %d %d", con[nd]+1, col[nd]);
                fprintf(redfp, "\n");
            
#if DBG
                if(dbg > 0) {
                    ildump();
                }
#endif
            }
        }
    }

    cmopen(ROFORT, "cfcalc.f");
    fprintf(redfp,"      subroutine cfcalc\n");
    fprintf(redfp,"      implicit real*8(a-h,o-z)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"      include 'incl1.f'\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"      integer jcon(nextn,ngraph), jcol(nextn,");
    fprintf(redfp,"ngraph+1), nn(ngraph)\n");
    cmsepar('-');
    fprintf(redfp,"\n");
    fprintf(redfp,"* read color factor\n");
    fprintf(redfp,"c     write(*,*) 'nextn = ', nextn\n");
    fprintf(redfp,"      open(22, file='cf0.dat', status='old')\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"      do 140 i = 1, 5\n");
    fprintf(redfp,"        read(22,*)\n");
    fprintf(redfp,"  140 continue\n");
    fprintf(redfp,"      do 100 i = 1, ngraph\n");
    fprintf(redfp,"        read(22,*) ig, nn(i), nd, (jcon(k,i), ");
    fprintf(redfp,"jcol(k,i), k = 1, nextn)\n");
    fprintf(redfp,"c       write(6,600) 'read:', ig, nn(i), nd,\n");
    fprintf(redfp,"c    &               (jcon(k,i), jcol(k,i), k = ");
    fprintf(redfp,"1, nextn)\n");
    fprintf(redfp,"        if(ig.ne.i) then\n");
    fprintf(redfp,"          write(*,*) 'illegal file of color ");
    fprintf(redfp,"factors (cf0.dat)' \n");
    fprintf(redfp,"          write(6,600) ' ', ig, nn(i), nd,\n");
    fprintf(redfp,"     &                 (jcon(k,i), jcol(k,i), k =");
    fprintf(redfp," 1, nextn)\n");
    fprintf(redfp,"          stop\n");
    fprintf(redfp,"        else if(nd.ne.1) then\n");
    fprintf(redfp,"          write(*,*) 'illegal denominator (cf0.");
    fprintf(redfp,"dat)' \n");
    fprintf(redfp,"          write(6,600) ' ', ig, nn(i), nd,\n");
    fprintf(redfp,"     &                 (jcon(k,i), jcol(k,i), k =");
    fprintf(redfp," 1, nextn)\n");
    fprintf(redfp,"          stop\n");
    fprintf(redfp,"        endif\n");
    fprintf(redfp,"  100 continue\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"      close(22)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"*   calculate cf(i,jj)\n");
    fprintf(redfp,"      do 200 i = 1, ngraph\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"        do 210 jj = i, ngraph\n");
    fprintf(redfp,"          if(i .eq. jj) then\n");
    fprintf(redfp,"            j = ngraph + 1\n");
    fprintf(redfp,"            do 260 k = 1, nextn\n");
    fprintf(redfp,"                jcol(k, j) = jcol(k, jj)\n");
    fprintf(redfp,"  260       continue\n");
    fprintf(redfp,"          else\n");
    fprintf(redfp,"            j = jj\n");
    fprintf(redfp,"          endif\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"c         write(*,600) 'mult1(con)',i, (jcon(k, ");
    fprintf(redfp,"i), k=1,nextn)\n");
    fprintf(redfp,"c         write(*,600) 'mult1(col)',i, (jcol(k, ");
    fprintf(redfp,"i), k=1,nextn)\n");
    fprintf(redfp,"c         write(*,600) 'mult2(con)',jj, (jcon(k,");
    fprintf(redfp,"jj), k=1,nextn)\n");
    fprintf(redfp,"c         write(*,600) 'mult2(col)',jj, (jcol(k, ");
    fprintf(redfp,"j), k=1,nextn)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"          icf = nn(i)*nn(jj)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"          do 220 n = 1, nextn\n");
    fprintf(redfp,"c           write(*,*) 'node = ', n, ', jcol =',");
    fprintf(redfp,"jcol(n, i), \n");
    fprintf(redfp,"c    &                 ' icf=', icf\n");
    fprintf(redfp,"            if(jcol(n, i) .le. 1) goto 220\n");
    fprintf(redfp,"            inx = n\n");
    fprintf(redfp,"            icf = icf * jcol(n, i)\n");
    fprintf(redfp,"            do 240 k = 1, nextn\n");
    fprintf(redfp,"*             * check (i, inx)\n");
    fprintf(redfp,"c             write(*,*) 'mult(1-0) ',i, jj, k, ");
    fprintf(redfp,"inx\n");
    fprintf(redfp,"              if(inx.lt.1 .or. jcol(inx, i) .lt. ");
    fprintf(redfp,"1) goto 900\n");
    fprintf(redfp,"              jcol(inx, i) = - jcol(inx, i)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"*             * move in i\n");
    fprintf(redfp,"              inx = jcon(inx, i)\n");
    fprintf(redfp,"c             write(*,*) 'mult(1-1) ',i, jj, k, ");
    fprintf(redfp,"inx\n");
    fprintf(redfp,"              if(inx.lt.1 .or. jcol(inx, i) .lt. ");
    fprintf(redfp,"1) goto 900\n");
    fprintf(redfp,"              jcol(inx, i) = - jcol(inx, i)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"*             * move to j\n");
    fprintf(redfp,"c             write(*,*) 'mult(2-0) ',i, jj, k, ");
    fprintf(redfp,"inx\n");
    fprintf(redfp,"              if(jcol(inx, j) .lt. 1) goto 900\n");
    fprintf(redfp,"              jcol(inx, j) = - jcol(inx, j)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"*             * move in j\n");
    fprintf(redfp,"              inx = jcon(inx, jj)\n");
    fprintf(redfp,"c             write(*,*) 'mult(2-1) ',i, jj, k, ");
    fprintf(redfp,"inx\n");
    fprintf(redfp,"              if(inx.lt.1 .or. jcol(inx, j) .lt. ");
    fprintf(redfp,"1) goto 900\n");
    fprintf(redfp,"              jcol(inx, j) = - jcol(inx, j)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"              if(inx .eq. n) goto 230\n");
    fprintf(redfp,"  240       continue\n");
    fprintf(redfp,"            goto 910\n");
    fprintf(redfp,"  230       continue\n");
    fprintf(redfp,"  220     continue\n");
    fprintf(redfp,"          do 250 k = 1, nextn\n");
    fprintf(redfp,"            if(jcol(k, i) .lt. 0) jcol(k, i) = - ");
    fprintf(redfp,"jcol(k, i)\n");
    fprintf(redfp,"            if(jcol(k, j) .lt. 0) jcol(k, j) = - ");
    fprintf(redfp,"jcol(k, j)\n");
    fprintf(redfp,"  250     continue\n");
    fprintf(redfp,"          cf(i, jj) = icf\n");
    fprintf(redfp,"          cf(jj, i) = icf\n");
    fprintf(redfp,"c         write(*,*) 'cf:', i, jj, icf, cf(i, jj)");
    fprintf(redfp,"\n");
    fprintf(redfp,"  210   continue\n");
    fprintf(redfp,"  200 continue\n");
    fprintf(redfp,"  600 format((' ',a,16i4))\n");
    fprintf(redfp,"      return\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"  900 continue\n");
    fprintf(redfp,"      write(*,*) 'cfcalc : inconsistent'\n");
    fprintf(redfp,"      stop\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"  910 continue\n");
    fprintf(redfp,"      write(*,*) 'cfcalc : not reach to the ");
    fprintf(redfp,"start'\n");
    fprintf(redfp,"      end\n");
    fprintf(redfp,"\n");
}

/*==================================================================*/
Private Void    ftoright(str)
string  str;
/*------------------------------------------------------------------*
 *   move spaces to the left.
 */
{
    int  s, t, len;

    t = len = strlen(str);
    for(s = len-1; s >= 0; s--)
        if(str[s] != ' ')
            str[--t] = str[s];
    for(s = t-1; s >= 0; s--)
        str[s] = ' ';
}
/*==================================================================*/
Private int    ftishiggs(NoArg)
/*------------------------------------------------------------------*
 *  Higgs appears or not in internal lines. (return number of Higgs)
 *  called in ftmain                                 (1995/08/06:TI)
 *  (Warning: this treatment depends on model definition file.)
 *------------------------------------------------------------------*/
{
    int il, higgs;
    pindex p;

    higgs = 0;

    for(il = 1; il < nilines; il++) {
        if(iltable[il].iltype == ILEXTERN)
            continue;
        else if(iltable[il].iltype == ILLOOP)
            continue;

        p = ilpart(il);
        if( strcmp(ptable[p]->mass,"amh") == 0 ) higgs++;
    }
    return higgs;
}
Private int    ftisgluon(NoArg)
/*------------------------------------------------------------------*
 *  Gluon appears or not in internal lines. (return number of Gluons)
 *  called in ftmain                                 (1995/11/18:TI)
 *  (Warning: this treatment depends on model definition file.)
 *------------------------------------------------------------------*/
{
    int il, gluon;
    pindex p;

    gluon = 0;

    for(il = 1; il < nilines; il++) {
        if(iltable[il].iltype == ILEXTERN)
            continue;
        else if(iltable[il].iltype == ILLOOP)
            continue;

        p = ilpart(il);
        if( strcmp(ptable[p]->mass,"amg") == 0 ) gluon++;
    }
    return gluon;
}
#define fltnam(typ)  (((typ)==FLOPEN)?"open  ": \
                      ((typ)==FLCLOSED)?"closed":"undef ")
/*==================================================================*/
Private bool   ftisschan(NoArg)
/*------------------------------------------------------------------*
 *  S-chanel or not.
 *  called in ftmain                                 (1995/11/14:TI)
 *------------------------------------------------------------------*/
{
    int f, n, scheck;
    bool isschan;

    isschan = False;
/*    printf("\n%% +++ %d fermion lines.\n", nflines);*/
    for(f = 0; f < nflines; f++) {
/*
        printf("%% type=%s, length=%4d : ",
               fltnam(fltable[f].type), fltable[f].length);
*/
        scheck = 0;
        for(n = 0; n < fltable[f].length; n++) {
/*
            printf("[%d, %d] ",
                   tovert(fltable[f].ndlist[n]),
                   fltable[f].lglist[n]);
*/
            if(fltable[f].lglist[n] == NDEXTERN ) {
                if( ndiorf(tovert(fltable[f].ndlist[n]))==ELINITIAL){
/*                    printf(" Initial ");*/
                    scheck++;
                }
            }
        }
/*        putchar(EOL);*/
        if( scheck == 2 ) { 
            isschan = True;
            return isschan;
        }
    }
    return isschan;
}
