/*********************************************************************
 *  File : genfort.c
 *
 *  Entry for testing particle table
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <stdio.h>

#define  DEFGLOBAL
#define  MAIN_GENRED
#include "common.h"
#include "config.h"
#include "dbgh.h"
#include "ptable.h"
#include "grdata.h"
#include "tokenh.h"
#include "args.h"
#include "gensc.h"
#include "glue4.h"
#include "version.h"
#include "cfbase.h"

/* Entry points */
Public  int    main          ARG((int, string*));

/* External refernces */
/* dagrf.c */
Extern  int     dasave       ARG((string,int*,int*));
/* ftgrf.c */
Extern  Void    ftprocs      ARG((int*,int*));
/* ilgrf.c */
Extern  Void    ildump       ARG((NoArg));
/* rdptbl.c */
Extern  Void    pvdump       ARG((NoArg));
Extern  Void    pvhelp       ARG((FILE*));
/* rdgrf.c */
Extern  Void    rgdump       ARG((tproc*, tgraph*));

/* Local functions */
Private int     grcarg       ARG((int, char**));
Private Void    grhelp       ARG((string));

/* Local variables */
Local   char    fgname[] = "out.grf";
Local   string  fngrf = NULL;
Local   int     djun  = 0;
Local   int     ddiv  = 0;
Local   int     dmina = 0;
Local   int     dhlp  = 0;
Local   int     dvect = 0;
Local   int     dlund = 0;
Local   int     dcol  = 0;
Local   int     dsusy = 0;
/* bases 50 or 100 (0/1) */
Local   int     dbas  = 0;
#if DBG
Local   int     ddbg;
#endif

Local   topt    options[] = {
    {"j", ARFLAG, (int*) &djun},
    {"f", ARFLAG, (int*) &ddiv},
    {"m", ARFLAG, (int*) &dmina},
    {"v", ARFLAG, (int*) &dvect},
    {"l", ARFLAG, (int*) &dlund},
    {"h", ARFLAG, (int*) &dhlp},
    {"?", ARFLAG, (int*) &dhlp},
    {"c", ARINT,  (int*) &dcol},
    {"b", ARINT,  (int*) &dbas},
    {"s", ARFLAG, (int*) &dsusy},
#if DBG
    {"d", ARINT,  (int*) &ddbg},
#endif
};

Local int noptions = sizeof(options)/sizeof(topt);

Local   int     proct[MAXLOOP];
Local   int     procg[MAXLOOP];

/*==================================================================*/
Public  int    main(argc, argv)
int     argc;
string  *argv;
/*--------------------------------------------------------------------
 *  Main routine for fortran source generation.
 */
{
    int       err, p0;
    int       prcl[MAXPROCS], prcg[MAXPROCS];

    grcmsg("Fortran code generator");

#if DBG
    /* debugging flags for modules.
     * if(dbg >= xxdbg) then enable debugging
     */
    trace   = 0;
    dbg = 0;
    dadbg = 1;  /* debug flag for dagrf.c */
    fldbg = 2;  /* debug flag for flgrf.c */
    ildbg = 1;  /* debug flag for ilgrf.c */
    pmdbg = 2;  /* debug flag for matchv.c */
    rddbg = 2;  /* debug flag for token.c */
    rgdbg = 2;  /* debug flag for rdgrf.c */
    pvdbg = 2;  /* debug flag for rdptbl.c */
    redbg = 1;  /* debug flag for rdptbl.c */
#endif

    testfv = True;
    cpol   = True;

    /* analyze arguments */
    argc = grcarg(argc, argv);
    if(argc != 1) {
        grhelp(argv[0]);
        exit(0);
    }

    /* read graphs.
     *    nprocs : the number of processes.
     *    prcl[p] : the number of loops of p-th process
     *    prct[p] : the number of graphs of p-th process
     */
    nprocs = dasave(fngrf, prcl, prcg);

#if DBG
    if(dbg > 0)
        pvdump();
#endif

    /* rearrange the process table.
     *    proct[l] : the process number with l loops.
     *    procg[l] : the number of graphs with l loops.
     */
    for(p0 = 0; p0 < MAXLOOP; p0++)
        proct[p0] = -1;

    maxloop = 0;
    for(p0 = 0; p0 < nprocs; p0++) {
        maxloop = max(maxloop, prcl[p0]);
        if(prcl[p0] >= MAXLOOP) {
            fprintf(Stderr, "Too large loop: proc=%d, loop=%d\n",
                    p0, prcl[p0]);
            exit(1);
        }
        if(proct[prcl[p0]] >= 0) {
            fprintf(Stderr, "Duplicated loop : proc = %d, %d\n",
                    proct[prcl[p0]], p0);
            exit(1);
        }
        proct[prcl[p0]] = p0;
        procg[prcl[p0]] = prcg[p0];
    }

    /* check the table */
    err = 0;
    for(p0 = 0; p0 <= maxloop; p0++) {
        printf("loop = %d,   process = %d,    graphs = %d\n",
               p0, proct[p0], procg[p0]);
        if(proct[p0] < 0) {
            fprintf(Stderr, "*** Loop=%d is missing\n", p0);
            err++;
        }
        if(err > 0)
            exit(1);
    }

    /* Two or more loop graphs are not suppored */
    if(maxloop > 2) {
        fprintf(Stderr, "*** This program does not support ");
        fprintf(Stderr, "%d loop graphs\n", 2);
        exit(0);
    }

    /* create fortran code */
    ftprocs(proct, procg);

    return 0;
}

/*==================================================================*/
Private int    grcarg(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
 */
{
    string cmd, s;
    int    pos;

    cmd     = argv[0];
    divpart = True;
    junpei  = False;
    colorv  = 0;
    jtgamm  = False;
    jano3v  = False;
    jqmas0  = False;
    jcolmb  = False;
    jqcdcr  = False;
    jetset  = False;
    jqedps  = False;
    susylib = False;

    argc = aropt(argc, argv, noptions, options);

    if(dhlp != 0) {
        grhelp(cmd);
        exit(0);
    }

    switch (ddiv) {
        case ARPLUS:
            divpart = True;
            break;
        case ARMINUS:
            divpart = False;
            break;
    }

    switch (dmina) {
        case ARPLUS:
            minami = True;
            break;
        case ARMINUS:
            minami = True;
            break;
    }

    switch (dvect) {
        case ARPLUS:
            vcode = True;
            break;
        case ARMINUS:
            vcode = True;
            break;
    }

    switch (djun) {
        case ARPLUS:
            junpei = True;
            break;
        case ARMINUS:
            junpei = True;
            break;
    }

    switch (dlund) {
        case ARPLUS:
            jetset = True;
            break;
        case ARMINUS:
            jetset = True;
            break;
    }

    switch (dsusy) {
        case ARPLUS:
            susylib = True;
            break;
        case ARMINUS:
            susylib = True;
            break;
    }

    if(dcol > 0) {
        colorv   = dcol;
    }
    if(dbas > 0) {
        basesv   = dbas;
    }
    if( minami ) {
        printf(" Minami-Tateya Tentative fortran code \n");
        jtgamm = True;
        jano3v = True;
        jqmas0 = True;
        jcolmb = True;
        jqcdcr = True;
        jqedps = True;
        basesv   = 1;
        colorv = 1;
    }
    if( colorv > 0 )
        printf(" color version %d (Kato-san inside)\n",colorv);
    if( dbas > 0 )
        printf(" BASES/SPRING 100\n");

    /* input process file name */
    pos = 1;
    if((fngrf = arnext(argc, argv, &pos)) != NULL) {
        pos--;
        argc = ardel(argc, argv, pos, 1);
    } else {
        fngrf = fgname;
    }

#if DBG
    /* debugging flags */
    dbg   = 0;

    if(ddbg > 0) {
        dbg   = ddbg;
    }
    if(ddbg < 0) {
        dbgst = -ddbg;
    }
#endif

    return argc;
}

/*==================================================================*/
Private Void   grhelp(cmd)
string  cmd;
/*--------------------------------------------------------------------
*/
{
    printf("Syntax %s [options] [Graph_File_name]\n",
           cmd);
    printf("\nOption\n");
    printf("    -c [0|1]            ");
    printf(": version for new color fact.\n");
    printf("    -j                  ");
    printf(": version for J.Fujimoto\n");
    printf("    -l                  ");
    printf(": Interface to JETSET\n");
    printf("    -f                  ");
    printf(": finite part only\n");
    printf("    -v                  ");
    printf(": generate vector code\n");
    printf("    -b [0|1]            ");
    printf(": BASES/SPRING 5.1 or hundred\n");
    printf("    -m                  ");
    printf(": Minami-Tateya Tentative version code\n");
    printf("  (jtgamm,jano3v,.... are supported.)\n");
#if DBG
    printf("    -d int              ");
    printf(": debug leve \n");
#endif
    printf("    -h                  ");
    printf(": this info\n");
    printf("\nDefault file name is:\n");
    printf("     Graph_File_name = %s\n", fgname);
}

/*==================================================================*/
Public  Void   prdump(NoArg)
/*--------------------------------------------------------------------
    Dump.
*/
{
    rgdump(cprocess, cgraph);
    ildump();
    fldump();
    ftvldmp();
}
