/*********************************************************************
 *    File :  grdata.h
 *
 *    data of graph file.
 *
 *                                               by Kaneko Toshiaki
 */

#include "ptable.h"

/*===================================================================
 *    File names
 */
Global string   fnmdl;  /* model */

/*-------------------------------------------------------------------
 *    The parameters of the current graph
 */

#define MAXELINE  10    /* The maximum number of external lines */
#define MAXLOOP   10    /* The maximum number of loops */
#define MAXNODES  (2*MAXLOOP + MAXELINE - 2)
                /* The maximum number of nodes */
#define MAXLGTBL  500
#define MAXLEGS  MAXELINE*3
                /* The maximum number of edges of nodes */

typedef struct {
    int     lgnext;        /* next node */
    int     lgpart;        /* particle */
    int     lgiline;       /* internal line number */
    int     lgwork;        /* work space */
} tleg;

typedef struct {
    int          nvtype;       /* type of the node */
    int          nvkind;       /* iorf of e-line or kind of vertex */
    int          nvloop;       /* the number of loops in the node */
    int          nvlegs;       /* the number of legs */
    int          nvgl4;        /* index of 4 gluon vertex, which is
                                  used when it is expaned to 3 gluon
                                  vertex. 
                                  0 : not 4 gluon vertex
                                  1 : (0, 1)--(2, 3)
                                  2 : (0, 2)--(3, 1)
                                  3 : (0, 3)--(1, 2)
                                  4 : 4 gluon vertex but not expanded */
    unsigned int basel;
} tnode;

typedef int node;         /* index of node */

typedef struct {
    int          grid;          /* graph number in the file */
    int          subid;         /* sub-graph number */
    int          nnodes;        /* the number of nodes */
    int          nilines;       /* The number of internal lines.
                                 * if table of i-lines are not ready,
                                 * value of this var = -1  */
    int          conjsrt;       /* start of conjugate node */
    long         sfactor;       /* symmetric factor */
    long         gtype;         /* graph type */
    int          ngl4;          /* the number of 4 gluon vertex */
    unsigned int basen;
    unsigned int offsl;
} tgraph;

#define MAXPROCS   5

/* value of mult in tproc */
#define PC_NORMAL  0
#define PC_MULT    1
#define PC_PROJ    2

typedef struct {
    int     morder[MAXMORD];
    int     procid;     /* Process id */
    int     graphs;     /* The number of generated graphs.  */
    int     nextern;    /* The number of external particles.  */
    int     oextern;    /* The number of original external particles.
                         * valid only for squared process.  */
    int     loop;       /* The number of loops.  */
    int     order;      /* The order of coupling constants.  */
    int     qcd;        /* The order of qcd coupling constants.  */
    int     extcol;     /* The number of colored external particles. */
    int     bgraph;     /* The number of graphs in the buffer */
    int     maxnodes;   /* maximum possible number of nodes */
    int     maxlegs;    /* maximum possible number of legs */
    int     proc1;      /* the first process of multiplication */
    int     proc2;      /* the second process of multiplication */
    int     msize;      /* memory size for one graph */
    bool    expand;     /* expand looped parts  */
    bool    onepi;      /* one paritcle irreducibility */
    bool    gm2;        /* one paritcle irreducibility */
    bool    undefp;     /* including undefined particle */
    int     mult;       /* mltiplied graphs or not */
    string  kinem;      /* path of kinematics */
} tproc;

/* maxnodes = nextern +   loop; (without Majorana) */
/* maxlegs  = nextern + 3*loop; (without Majorana) */
/* msize    = (sizeof tgraph) + (sizeof tnode)*maxnodes 
              + (sizeof tleg)*maxlegs */

/*-------------------------------------------------------------------
 * List of Fermion lines
 *-----------------------------------------------------------------*/
#define MAXFLINES 20

#define FLUNDEF  0
#define FLOPEN   1
#define FLCLOSED 2
typedef struct {
    int    type;
    int    length;
    node  *ndlist;
    int   *lglist;
} tfline;

/*-------------------------------------------------------------------
 * internal line.
 *-----------------------------------------------------------------*/
/* value of .iltype */
#define ILDEL    -1   /* deleted */
#define ILEXTERN  0   /* external particle */
#define ILLOOP    1   /* loop momentum */
#define ILDLOOP   2   /* dependent on loop momenta */
#define ILECONJG  3   /* conjugated external momentum */
#define ILTREE    4   /* linear conb. of ext. mom. */

typedef int tmom;

typedef struct {
    int    iltype;       /* type of the internal line */
    int    ilname;       /* print name of internal line */
    int    ilblock;      /* block id */
    int    ilnodes[2];   /* nodes at endpoints */
    int    illegs[2];    /* legs of nodes connect to this line */
    tmom  *ilmlist;      /* momentum conservation */
} tiline;

/*-------------------------------------------------------------------
 *  the current graph
 */

Global  int       nprocess;   /* The current output process number */
Global  tproc    *cprocess;

Global  int       ngraph;     /* The current graph number (read) */
Global  tgraph   *cgraph;

Global  int       gntype;     /* The type of graph */

Global  tfline  fltable[MAXFLINES];    /* table of fermion lines */
Global  int     nflines;
 
Global  tiline  *iltable;              /* table of internal lines */

/* the table of Feynman parameters */
Global  int     *ilxtable; /* parameter to node */
Global  int     ilxnumb;
/*-------------------------------------------------------------------
 *  Table of nodes.
 */

#define ELINITIAL  (-1)
#define ELFINAL      1
#define eliorf(x) (((x) == ELINITIAL) ? "initial" : "final")

/* value of tnode.vdtype */
#define NDDEL       -1        /* deleted */
#define NDEXTERN     0        /* external line */
#define NDTREEV      1        /* tree vertex */
#define NDLOOPEDV    2        /* looped vertex */
#define NDONSHELL    3        /* on-shell line */
#define NDPROJ       4        /* projection operator */
#define NDNULL    9999        /* null node index ???= nptable ???*/

/* access function of nodes and legs */
#define nenodes(g)         (g)->nnodes
#define nengl4(g)          (g)->ngl4
#define nenode(g,n)        ((tnode *)(((char*)(g))+(g)->basen)+(n))
#define netype(g,n)        nenode(g,n)->nvtype
#define nelegs(g,n)        nenode(g,n)->nvlegs
#define neloop(g,n)        nenode(g,n)->nvloop
#define nevertex(g,n)      nenode(g,n)->nvkind
#define neiorf(g,n)        nenode(g,n)->nvkind 
#define negl4(g,n)         nenode(g,n)->nvgl4
#define nebasel(g,n)       nenode(g,n)->basel 
#define neleg(g,n,p)       ((tleg *)(((char*)(g))+nebasel(g,n))+(p))
#define neparticle(g,n,p)  neleg(g,n,p)->lgpart
#define nenextn(g,n,p)     neleg(g,n,p)->lgnext
#define neiline(g,n,p)     neleg(g,n,p)->lgiline
#define neilwk(g,n,p)      neleg(g,n,p)->lgwork
#define nename(g,e)        neparticle(g,e,0)

#define ndnodes()          nenodes(cgraph)
#define ndngl4()           nengl4(cgraph)
#define ndnode(n)          nenode(cgraph, n)
#define ndtype(n)          netype(cgraph, n)
#define ndlegs(n)          nelegs(cgraph, n)
#define ndloop(n)          neloop(cgraph, n)
#define ndvertex(n)        nevertex(cgraph, n)
#define ndiorf(n)          neiorf(cgraph, n)
#define ndgl4(n)           negl4(cgraph, n)
#define ndbasel(n)         nebasel(cgraph, n)
#define ndleg(n,p)         neleg(cgraph, n, p)
#define ndparticle(n,p)    neparticle(cgraph, n, p)
#define ndnextn(n,p)       nenextn(cgraph, n, p)
#define ndiline(n,p)       neiline(cgraph, n, p)
#define ndilwk(n,p)        neilwk(cgraph, n, p)
#define ndname(e)          nename(cgraph, e)

/* direction of internal line.
 *      +1 : outgoing from the node (leg=1)
 *      -1 : ingoing to the node    (leg=0)
 */
#define ildir(lg)          (2*(lg) - 1)
/*==================================================================*/
/* Standard macros
 */
/* Macros to convert node number.
 *
 *  1. tonode : macro function to covert e-line or vertex number
 *              to node number.
 *  2. tovert : macro function to convert node number to
 *              e-line or vertex number.
 */
/*
 * #define tonode(x)   (((x)<0) ? (-(x)-1) : ((x)-1+cprocess->nextern))
 * #define tovert(x)   (((x) < cprocess->nextern) ? \
 *                      (-(x)-1) : ((x)+1-cprocess->nextern))
*/
#define tonode(x)   (x)
#define tovert(x)   (x)

#define boolstr(x)  ((x) ? "Yes" : "No")

#ifndef min
#define min(x, y)   (((x) < (y)) ?  (x) : (y))
#define max(x, y)   (((x) > (y)) ?  (x) : (y))
#endif

#ifndef abs
#define abs(x)      (((x) <  0 ) ? -(x) : (x))
#endif

#ifndef sign
#define sign(x)      (((x) <  0 ) ? (-1) : (1))
#endif

/* Redefine Stderr to stdout */
#ifdef  Stderr
#undef  Stderr
#endif
#define Stderr stdout

