/*********************************************************************
 *  File : ilgrf.c
 *
 *  Tools for internal lines.
 *
 *      Analysis and alocation of momentum conservation.
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "common.h"
#include "dbgh.h"
#include "ptable.h"
#include "grdata.h"

#ifdef OLD
#undef OLD
#endif

/* Entry points */
Public  int     ilmake       ARG((bool));
Public  Void    ildump       ARG((NoArg));
Public  int     ilnextl      ARG((node,int));
Public  Void    ilmom        ARG((NoArg));
Public  Void    iliconsv     ARG((NoArg));

/* External references */
Extern  Void    erfail       ARG((NoArg));
Extern  Void    error        ARG((string,string));
Extern  Void    flget        ARG((tgraph*));
Extern  Void    flmdir       ARG((NoArg));
Extern  bool    flmajo       ARG((NoArg));

/* Local functions */
Private Void    ilalloc      ARG((int));
Private Void    ilbicon      ARG((NoArg));
Private int     ilbinit      ARG((NoArg));
Private Void    ilbisrch     ARG((node,node,int,int));
Private Void    ilexchg      ARG((NoArg));
Private Void    ilswap       ARG((int,bool));

/* local variables */
Local int  siltbl  = 0;
Local int  nilines = 0;
Local int  bicount;
Local int  bivdef[MAXNODES];
Local int  bivlow[MAXNODES];
Local int  *mconsv;
Local int  *loopm;    /* list of loop momenta */
Local int  nloopm;

#define bidef(v)   bivdef[v]
#define bilow(v)   bivlow[v]
#define BINEW      (MAXNODES + 1)

/*==================================================================*/
Public  int     ilmake(mode)
bool    mode;
/*--------------------------------------------------------------------
 *    Construct internal line table for the current graph.
 *      items "ilnodes" and "illegs" are set here.
 * 
 *    if(mode)
 *        (set ndnextn(*,*) )
 *------------------------------------------------------------------*/
{
    node nd;
    int  lg, il, ip, niln;

#if DBG
    if(dbg >= ildbg)
        printf("ilmake(%d)\n", mode);
#endif

    /* size of the table */
    niln = 0;
    for(nd = 0; nd < ndnodes(); nd++) {
        for(lg = 0; lg < ndlegs(nd); lg++) {
            niln = max(niln, ndiline(nd, lg));
        }
    }
    niln++;

    ilalloc(niln);
    /* clear the internal line table */
    for(il = 0; il < niln; il++) {
        iltable[il].ilnodes[0] = -1;
        iltable[il].ilnodes[1] = -1;
    }

    /* set ilnodes and illegs */
    for(nd = 0; nd < ndnodes(); nd++) {
        for(lg = 0; lg < ndlegs(nd); lg++) {
            if(mode)
                ndnextn(nd, lg) = 0;

            il = ndiline(nd, lg);
            /* 0 -->(particle)--> 1 */
            if(!cprocess->gm2 && ndparticle(nd, lg) < 0) {
                ip = 0;
                if(iltable[il].ilnodes[0] >= 0) {
                    error("ilmake", "too many legs for an i-line");
                    fprintf(Stderr,"il = %d\n", il);
                    erfail();
                }
            } else {
                ip = 1;
                if(iltable[il].ilnodes[1] >= 0) {
                    if(iltable[il].ilnodes[0] >= 0) {
                        error("ilmake", "too many legs for an i-line");
                        fprintf(Stderr,"il = %d\n", il);
                        erfail();
                    }
                    ip = 0;
                }
            }
            iltable[il].ilnodes[ip] = nd;
            iltable[il].illegs[ip]  = lg;
        }
    }

    /* check table.  iltbl[0] is not used.*/
    for(il = 1; il < niln; il++) {
        for(ip = 0; ip < 2; ip++) {
            if(iltable[il].ilnodes[ip] < 0) {
                error("ilmake", "too few legs for an i-line");
                fprintf(Stderr,"il = %d\n", il);
                erfail();
            }

            if(mode)
                ndnextn(iltable[il].ilnodes[ip],iltable[il].illegs[ip])
                    = iltable[il].ilnodes[1-ip];
        }
    }

    cgraph->nilines = niln;

    /* construct fermion line and arrange direction of majorana */
    flget(cgraph);
    if(flmajo())
        flmdir();

#if DBG
    if(dbg >= ildbg)
        printf("ilmake(%d) : end\n", mode);
#endif

    return niln;
}

/*==================================================================*/
Private Void    ilalloc(niln)
int     niln;
/*--------------------------------------------------------------------
 * allocate memory for the internal line table
 *------------------------------------------------------------------*/
{
    int il;

#if DBG
    if(dbg >= ildbg)
        printf("ilalloc(%d)\n", niln);
#endif

    if(niln < 1) {
        error("ilalloc", "0 i-lines");
        erfail();
    }

    /* allocate memory for the internal line table */
    if(siltbl < niln) {
        if(siltbl > 0) {
            free((AnyP)mconsv);
            free((AnyP)loopm);
            free((AnyP)ilxtable);
            for(il = 0; il < siltbl; il++)
                free((AnyP)iltable[il].ilmlist);
            free((AnyP)iltable);
        }
    
        if((iltable = (tiline*) malloc(niln*sizeof(tiline)))
             == NULL) {
             error("ilalloc", "memory allocation failed.");
             erfail();
        }
        for(il = 0; il < niln; il++) {
            if((iltable[il].ilmlist 
                 = (tmom*) malloc(niln*sizeof(tmom))) == NULL) {
                 error("ilalloc", "memory allocation failed.");
                 erfail();
            }
        }
        if((mconsv = (int*) malloc(niln*sizeof(int))) == NULL) {
             error("ilalloc", "memory allocation failed.");
             erfail();
        }
        if((loopm = (tmom*) malloc(niln*sizeof(tmom))) == NULL) {
             error("ilalloc", "memory allocation failed.");
             erfail();
        }
        if((ilxtable = (int*) malloc(niln*sizeof(int))) == NULL) {
             error("ilalloc", "memory allocation failed.");
             erfail();
        }
        siltbl = niln;
    }
#if DBG
    if(dbg >= ildbg)
        printf("ilalloc(%d): end\n", niln);
#endif
}

/*==================================================================*/
Public  Void    ildump(NoArg)
/*--------------------------------------------------------------------
 *------------------------------------------------------------------*/
{
    int il, ip, ii, ce;
    tiline *ti;

    nilines = cgraph->nilines;

    printf("\n nilines = %d, siltbl = %d\n", nilines, siltbl);

    if(nilines < 0)
        return;

    for(il = 1; il < min(nilines, siltbl); il++) {
        ti = iltable + il;
        printf("%2d  t:%d x:%d [",
               il, ti->iltype, ti->ilname);
        for(ip = 0; ip < 2; ip++)
            printf(" %2d(%d)", ti->ilnodes[ip], ti->illegs[ip]);
        printf("]: ");
        for(ii = 1; ii < min(nilines, siltbl); ii++) {
            if((ce = ti->ilmlist[ii]) == -1)
                printf("-%d", ii);
            else if(ce == 1)
                printf("+%d", ii);
            else if(ce > 0)
                printf("+%d*%d", ce, ii);
            else if(ce < 0)
                printf("%d*%d", ce, ii);
        }
        printf("\n");
    }

    printf("ilxnumb = %d\n", ilxnumb);
    for(il = 0; il < ilxnumb;  il++)
        printf(" %d", ilxtable[il]);
    printf("\n");
}

/*==================================================================*/
Public  int     ilnextl(n, l)
node    n;
int     l;
/*--------------------------------------------------------------------
 * determine momentum of internal lines.
 *------------------------------------------------------------------*/
{
    int il, j;
    node nv;

    nv = ndnextn(n, l);
    il = ndiline(n, l);
    j = 0;
    if(iltable[il].ilnodes[j] != nv)
        j = 1;
    return iltable[il].illegs[j];
}

/*==================================================================*/
Public  Void    ilmom(NoArg)
/*--------------------------------------------------------------------
 * determine momentum of internal lines.
 *------------------------------------------------------------------*/
{
#if DBG
    if(dbg>=ildbg)
        printf("ilmom: begin\n");
#endif

    if(cprocess->loop < 0) {
        error("ilmom", "illegal number of loops.");
        erfail();
    }

    if(!cprocess->expand) {
        error("ilmom", "not expanded graph.");
        erfail();
    }

    /* determine loop momenta */
    if(cgraph->nilines < 0)
        cgraph->nilines = ilmake(False);

    nilines = cgraph->nilines;

    ilbicon();

    ilexchg();

#if DBG
    if(dbg>=ildbg)
        printf("ilmom: end\n");
#endif
}

/*==================================================================*/
Private Void    ilbicon(NoArg)
/*--------------------------------------------------------------------
 *------------------------------------------------------------------*/
{
    int nd;

#if DBG
    if(dbg>=ildbg)
        printf("ilbicon: begin (loop=%d, expand=%d)\n",
                cprocess->loop, cprocess->expand);
#endif

    nd = ilbinit();

    ilbisrch(nd, -1, -1, 0);

#if DBG
    if(dbg>=ildbg)
        printf("ilbicon: end\n");
#endif

    /* No check of the anoswer in the test version */
    return;
}


/*==================================================================*/
Private int     ilbinit(NoArg)
/*--------------------------------------------------------------------
 *------------------------------------------------------------------*/
{
    node v, eo, ee;
    int il, im, il0, il1, dir, l;

    bicount = 0;
    nloopm  = 0;
    ilxnumb = 0;

    for(v = 0; v < ndnodes(); v++) {
        bidef(v) = BINEW;
        bilow(v) = BINEW;
    }

#if DBG
    if(dbg >= ildbg)
        printf("ilbinit: nilines = %d\n", nilines);
#endif
    for(il = 0; il < nilines; il++) {
        iltable[il].iltype = ILTREE;
        iltable[il].ilname = 0;
        for(im = 0; im < nilines; im++)
            iltable[il].ilmlist[im] = 0;
        loopm[il] = 0;
    }

    for(il = 0; il < nilines; il++)
        mconsv[il] = 0;

    ee = eo = -1;
    for(v = ndnodes() - 1; v >= 0; v--) {
        if(ndtype(v) == NDEXTERN) {
            eo = v;
            il0 = ndiline(v, 0);
            iltable[il0].iltype = ILEXTERN;
            dir = - ndiorf(v);
            if(iltable[il0].ilnodes[1] == v)
                dir = - dir;
            if(dir < 0)
                ilswap(il0, False);
            iltable[il0].ilmlist[il0] = 1;

            if(ndiorf(v) == ELINITIAL)
                mconsv[il0] = 1;
            else
                mconsv[il0] = -1;

        } else if(ndtype(v) == NDONSHELL) {
            eo = v;
            il0 = ndiline(v, 0);
            il1 = ndiline(v, 1);
            iltable[il0].iltype = ILEXTERN;
            iltable[il1].iltype = ILEXTERN;
            dir = - ndiorf(v);
            if(iltable[il0].ilnodes[1] == v)
                dir = - dir;
            if(dir < 0) {
                ilswap(il0, False);
            }
            dir = - ndiorf(v);
            if(iltable[il1].ilnodes[1] == v)
                dir = - dir;
            if(dir < 0) {
                ilswap(il1, False);
            }
            iltable[il0].ilmlist[il0] = 1;
            iltable[il1].ilmlist[il0] = 1;

            if(ndiorf(v) == ELINITIAL) { /* ??? */
                mconsv[il0] = 1;
                mconsv[il1] = 1;
            } else {
                mconsv[il0] = -1;
                mconsv[il1] = -1;
            }

        } else if(ndtype(v) == NDPROJ) {
            eo = v;
            for(l = 0; l < ndlegs(v); l++) {
                il0 = ndiline(v, l);
                iltable[il0].iltype = ILEXTERN;
                dir = - ndiorf(v);
                if(iltable[il0].ilnodes[1] == v)
                    dir = - dir;
                if(dir < 0)
                    ilswap(il0, False);
                iltable[il0].ilmlist[il0] = 1;

                if(ndiorf(v) == ELINITIAL)
                    mconsv[il0] = 1;
                else
                    mconsv[il0] = -1;
            }
        }
    }

#if DBG
    if(dbg >= ildbg) {
        for(il = 0; il < nilines; il++)
            printf("monsv[%d]=%d\n", il, mconsv[il]);

        printf("ilbinit: end\n");
    }
#endif
    if(ee != -1)
        return ee;
    else if(eo != -1)
        return eo;

    error("ilbinit", "no external particles");
    return 0;
}

/*==================================================================*/
Private Void    ilbisrch(v, ov, il, ilg)
node    v, ov;
int     il, ilg;
/*--------------------------------------------------------------------
   Search graph to fined bridge.

   Arguments:
       v   : input  : node to visit.
       ov  : input  : the last node.
       il  : input  : i-line number connecting ov and v
       ilg : input  : iltable[il].ilnodes[ilg] = v

   Returns
       whether the node v is in a loop or not.
 *------------------------------------------------------------------*/
{
    node w;
    int lg, dir, dirn, iln, ilgn, i;

#if DBG
    if(dbg >= ildbg)
        printf("*dbg* ilbisrch : begin %d -> %d\n", ov, v);
#endif

    dir = ildir(ilg);   /* incoming to the node v */

    /* visit node "v" */
    bicount++;
    bidef(v)  = bicount;
    bilow(v)  = bicount;

    /* visit sons */
    for(lg = 0; lg < ndlegs(v); lg++) {
        if(ndiline(v, lg) == il) /* back to ov */
            continue;

        w = ndnextn(v, lg);

        iln = ndiline(v, lg);
        if(iltable[iln].ilnodes[0] == v)
            ilgn = 1;
        else
            ilgn = 0;
        dirn = ildir(ilgn);  /* incoming to the node w */

        /* self loop */
        if(w == v) {
            iltable[iln].iltype = ILLOOP;
            iltable[iln].ilmlist[iln] = 1;

            /* Feynman parameter */
            ilxtable[ilxnumb] = iln;
            iltable[iln].ilname = ++ilxnumb;
            loopm[nloopm++] = iln;
            continue;

        /* external particle */
        } else if(ndtype(w) == NDEXTERN) {

        /* squared external particle */
        } else if(ndtype(w) == NDONSHELL) {    

        /* projection operator */
        } else if(ndtype(w) == NDPROJ) {    

        /* edge (v, w) is a back edge */
        } else if(bidef(w) != BINEW) {  
            if(bidef(v) > bidef(w)) {
                iltable[iln].iltype = ILLOOP;

                /* incoming to v */
                iltable[iln].ilmlist[iln] = dirn;

                /* Feynman parameter */
                ilxtable[ilxnumb] = iln;
                iltable[iln].ilname = ++ilxnumb;
                loopm[nloopm++] = iln;
            }
            bilow(v)  = min(bilow(v), bidef(w));

        } else {    /* new vertex */
            ilbisrch(w, v, iln, ilgn);
            bilow(v)  = min(bilow(v), bilow(w));
        }

        if(il > 0 && iltable[il].iltype != ILEXTERN) {
            for(i = 1; i < nilines; i++) {
                /* direction = - (vo->v)*(w->v) */
                iltable[il].ilmlist[i]
                    += dir*dirn * iltable[iln].ilmlist[i];
#if DBG
                if(dbg>=ildbg)
                    printf("%d->%d: %d * %d(%d,%d)\n",
                           v,w,dir*dirn,iltable[iln].ilmlist[i],iln,i);
#endif
            }
        }
#if DBG
        if(dbg >= ildbg) {
            printf("%d -> %d -> %d : %d\n", ov, v, w, 
                   iltable[iln].iltype);
        }
#endif

    } /* of for */

    /* Feynman parameter */
    if(ov >= 0 && bilow(v) <= bidef(ov)) {
        ilxtable[ilxnumb]  = il;
        iltable[il].iltype = ILDLOOP;
        iltable[il].ilname = ++ilxnumb;
    }

#if DBG
    if(dbg >= ildbg) {
        if(ov >= 0)
            printf("%d(%d) ->(%d)-> %d(%d)\n", 
                   ov, bidef(ov), il, v, bilow(v));
    }
#endif
}

/*==================================================================*/
Public  Void    iliconsv(NoArg)
/*--------------------------------------------------------------------
 * Momentum conservation of tree internal line.
 *------------------------------------------------------------------*/
{
    int il, ii, ij, ce, cs, tst = 0;
    int nextern = cprocess->nextern;
    bool cont;

    cont = True;
    while(cont) {
#if DBG
        if(++tst > max(nilines*nilines, 2)) {
            fprintf(Stderr, "*** program loops at iliconsv\n");
            ildump();
            exit(1);
        }
#endif
        cont = False;
        for(il = 0; il < nilines; il++) {
            if(iltable[il].iltype == ILEXTERN)
                continue;
            else if(iltable[il].iltype == ILLOOP)
                continue;
    
            for(ii = 0; ii < nilines; ii++) {
                if(iltable[ii].iltype == ILEXTERN)
                    continue;
                else if(iltable[ii].iltype == ILLOOP)
                    continue;
                else if((ce = iltable[ii].ilmlist[il]) == 0)
                    continue;
#if DBG
                printf("iliconsv: il = %d, ii = %d\n", il, ii);
                ildump();
#endif
                iltable[ii].ilmlist[il] = 0;
                for(ij = 0; ij < nilines; ij++)
                    iltable[ii].ilmlist[ij] += 
                        ce*iltable[il].ilmlist[ij];
                cont = True;
            }
        }
    }

    for(il = 1; il < nilines; il++) {
        if(iltable[il].iltype == ILEXTERN)
            continue;
        else if(iltable[il].iltype == ILLOOP)
            continue;
        ce = 0;
        cs = 0;
        for(ii = 0; ii < nilines; ii++) {
            if(iltable[il].ilmlist[ii] != 0) {
                cs++;
                ce = ii;
            }
        }
        if(cs > nextern/2) {
            cs = iltable[il].ilmlist[ce]*mconsv[ce];
            for(ii = 0; ii < nilines; ii++) {
                iltable[il].ilmlist[ii] -= cs*mconsv[ii];
            }
        }
    }
}

/*==================================================================*/
Private Void    ilexchg(NoArg)
/*--------------------------------------------------------------------
 * Change direction of ilines.
 *------------------------------------------------------------------*/
{
    int il, is, i, iv, p;

#if DBG
    if(dbg >= ildbg)
        printf("ilexchg\n");
#endif
    for(i = 0; i < nloopm; i++) {
        il = loopm[i];
        if(iltable[il].ilmlist[il] < 0) {
            ilswap(il, True);
        }
    }
    for(il = 1; il < nilines; il++) {
#ifdef OLD
        if(iltable[il].iltype == ILLOOP) {
        } else if(iltable[il].iltype == ILDLOOP) {
            for(i = 0; i < nloopm; i++) {
                if((is = iltable[il].ilmlist[loopm[i]]) < 0) {
                    ilswap(il, True);
                    break;
                } else if(is > 0) {
                    break;
                }
            }
        } else if(iltable[il].iltype != ILEXTERN) {
            is = 0;
            for(i = 1; i < nilines; i++) {
                if((iv = iltable[il].ilmlist[i]) > 0)
                    is++;
                else if(iv < 0)
                    is--;
            }
            if(is < 0) {
                ilswap(il, True);
            }
        }
#else
        if(iltable[il].iltype == ILLOOP) {
        } else if(iltable[il].iltype == ILDLOOP) {
            for(i = 0; i < nloopm; i++) {
                if((is = iltable[il].ilmlist[loopm[i]]) < 0) {
                    ilswap(il, True);
                    break;
                } else if(is > 0) {
                    break;
                }
            }
        } else if(iltable[il].iltype == ILTREE) {
            p = ndparticle(iltable[il].ilnodes[1],
                           iltable[il].illegs[1]);
            if(p < 0) {
                ilswap(il, True);
            }
        }
#endif
    }
#if DBG
    if(dbg >= ildbg)
        printf("ilexchg: end\n");
#endif
}

/*==================================================================*/
Private Void    ilswap(il, mode)
int     il;
bool    mode;
/*--------------------------------------------------------------------
 *------------------------------------------------------------------*/
{
    int tn, tl, i;

#if DBG
    if(dbg >= ildbg)
        printf("ilswap(%d, %d)\n", il, mode);
#endif
    if(cprocess->gm2) {
        return;
    }

    tn = iltable[il].ilnodes[0];
    iltable[il].ilnodes[0] = iltable[il].ilnodes[1];
    iltable[il].ilnodes[1] = tn;

    tl = iltable[il].illegs[0];
    iltable[il].illegs[0] = iltable[il].illegs[1];
    iltable[il].illegs[1] = tl;

    if(!mode)
        return;
    for(i = 1; i < nilines; i++) {
        iltable[il].ilmlist[ i] *= -1;
    }

#if DBG
    if(dbg >= ildbg)
        printf("ilswap: end\n");
#endif
}
