/*********************************************************************
 *  File :  ptable.h
 *
 *  Table of particles
 *
 *                                              by Kaneko Toshiaki
 */

#ifndef PNULL

#ifndef DBG
#define DBG 2
Global int dbg;
Global int pvdbg;
Global int rddbg;
#endif
/*===================================================================
 *  Indices of tables for model definition.
 */
typedef int     pindex;     /* index in the particle table */
                            /* +(particle) or -(antiparticle) */
#define PNULL   0           /* undefined partile as a value of pindex */

typedef int     nindex;     /* index in the particle name table */

typedef int     vindex;         /* index in the vertex table */
#define VNULL   0               /* undefined vertex */
#define VSELF   -1              /* undefined vertex */

/*===================================================================
 *  Table of particle definitions.
 */
/* #define LPTABLE 100 */              /* size of particle table */
                                 /* see ptable.h */

/*===================================================================
 *  Table of particle names.
 */

/* #define LNTABLE (2*LPTABLE) */ /* size of table of particle name */

/*===================================================================
 *  Table of definitions of coupling constants.
 */

#define MAXMORD  3      /* number of kinds of order */

/*===================================================================
 *  Table of particle definitions.
 */
/* The value of ptype */
#define P_UNDEF    0
#define P_SCALAR   1
#define P_MAJORANA 2
#define P_FERMION  3
#define P_GHOST    4
#define P_VECTOR   5
#define PTYPEMAX   6

#define M_UNDEF    1
#define M_SCALAR   (1 << P_SCALAR)
#define M_MAJORANA (1 << P_MAJORANA)
#define M_FERMION  (1 << P_FERMION)
#define M_GHOST    (1 << P_GHOST)
#define M_VECTOR   (1 << P_VECTOR)
#define M_BOSON    (M_SCALAR | M_VECTOR)

#ifdef DEFGLOBAL
 Global char *ptypesym  = "?smfxv";
 Global string  pvtnam[PTYPEMAX] =
       {"undef", "scalar", "majorana", "fermion", "ghost", "vector"};
 Global unsigned int pmask[PTYPEMAX] =
       {M_UNDEF, M_SCALAR, M_MAJORANA, M_FERMION, M_GHOST, M_VECTOR};
#else
 Global char *ptypesym;
 Global string  pvtnam[PTYPEMAX];
 Global unsigned int pmask[PTYPEMAX];
#endif

typedef struct {            /* particle table */
        string  psym[2];    /* print name */
        string  mass;       /* fortran name for mass */
        string  width;      /* fortran name for width */
        string  mvalue;     /* value of mass */
        string  wvalue;     /* value of width */
        string  gname[3];   /* name for graphic output */
        nindex  pname[2];   /* name of the [particle, anti-p] */
        int     pcode;      /* particle code */
        int     kfcode;     /* KF particle code */
        int     nvlist[2];  /* length of vlist */
        int     antip;      /* (antiparticle == particle) ? 1 : -1 */
        int     charge3;    /* 3*spin */
        int     ptype;      /* type of the particle */
        int     color;      /* color multiplet */
        bool    massless;   /* massless or massive */
        bool    furry;      /* flag for Furry's theorem */
        vindex *vlist[2];   /* list of vertices */
}       tptable;
#define LPTABLE 100              /* size of particle table */
Global tptable *ptable[LPTABLE]; /* table of particles */
Global int      nptable;         /* the number of particles in ptable */

/*===================================================================
 *  Table of particle names.
 */
typedef struct {
        char   *name;           /* name */
        pindex  pindx;          /* position in the ptable ??? */
}       tntable;

#define LNTABLE (2*LPTABLE)     /* size of table of particle name */
Global tntable  ntable[LNTABLE];  /* table of particle names */
Global int      nntable;          /* the number of data in the table */

#define N_MASS      0
#define N_WIDTH     1
#define N_SIZE      2

Global string   mtable[N_SIZE][LPTABLE];
Global int      nmtable[N_SIZE];

/*===================================================================
 *  Table of definitions of coupling constants.
 */

/*  MAXMORD : number of kinds of order, defined in "grch.h" */
Global char    *mordtbl[MAXMORD];       /* name of coupling const */
Global int      nmordtbl;               /* # of data in the table */

/*-------------------------------------------------------------------
 *  Table of vertix definitions.
 */

#define MAXTL     4 /* The maximum number of legs of tree vertex.*/
#define MINTL     3 /* The minimum number of legs of tree vertex.*/

Global int maxtl, mintl;  /* Defined in model file */

#define VTTREE    0    /* tree vertex */
#define VTLOOP    1    /* vertex with loop */
#define VTGENCT   2    /* generated counter term */

#define V_NOTYPE  0
#define V_V       1
#define V_A       2
#define V_VminusA 3
#define V_VplusA  4

#define MAXFVAL   2 /* the number of fvalues */
#define V_PUNDEF  0 /* phase */
#define V_REAL    1
#define V_IMAG    2
#define V_CMPLX   3

/* Vertex table */
typedef struct {
        int     vtype;           /* type of the vertex */
        int     nvlegs;          /* the number of legs */
        int     morder[MAXMORD]; /* table of coupling constants */
        int     loop;            /* loop in the vertex */
        int     ctype;           /* type of coupling */
        string  fname;           /* name for fortran */
        pindex *vlegs;           /* list of particles */
        int     phase;           /* phase of coupling constant */
        string  fvalue[MAXFVAL]; /* value of coupling constant */
}       tvtable;

#define LVTABLE 500              /* size of vertex table */

Global int      nvtable;         /* the number of defined vertices */
Global int      nvmodel;         /* vetices defined in the model */
Global tvtable *vtable[LVTABLE];        /* table of vertices */
Global int      nvstart;         /* star of the table */
Global bool     majop;           /* majorana is defined or not */

Global string   ctable[LVTABLE];
Global int      vnamdef[LVTABLE];
Global int      nctable;

Global bool     testfv;          /* test FValue */

#endif
