/*********************************************************************
 *  File : rational.c
 *
 *  Arithmetics of rational numbers
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "common.h"
#include "rational.h"

/* Entry points */

/* External references */

/* local functions */
Private long      rtgcd        ARG((long, long));

/*==================================================================*/
Public  rational rtnew(num, den)
long    num, den;
/*--------------------------------------------------------------------
 *    Create a rational number.  They are initialized to num/den.
 */
{
    int i;
    rational  p;

    if((p = (rational) malloc(sizeof(ratio))) == NULL) {
        fprintf(Stderr, "*** rtnew : memory allocation failed\n");
        exit(1);
    }
    p->num = num;
    p->den = den;

    return p;
}

/*==================================================================*/
Public  Void     rtfree(p)
rational p;
/*--------------------------------------------------------------------
 *    Free rational number.
 */
{
    free(p);
}

/*==================================================================*/
Public  Void     rtset(p, num, den)
rational p;
long     num, den;
/*--------------------------------------------------------------------
 *    Set value to p = num/den.
 */
{
    p->num = num;
    p->den = den;
}

/*==================================================================*/
Private long      rtgcd(x, y)
long    x, y;
/*--------------------------------------------------------------------
 *    Calculate greatest common diviser of x and y.
 */
{
    long m;

    x = (x < 0) ? -x : x;
    y = (y < 0) ? -y : y;

    while(y != 0) {
        m = x % y;
        x = y;
        y = m;
    }
    return x;
}

/*==================================================================*/
Private Void      rtnorm(p)
rational p;
/*--------------------------------------------------------------------
 *    normalize rational number.
 */
{
    long gcd;

    if(p->num == 0){
        p->den = 1L;
        return;
    } else if(p->den == 0L) {
        fprintf(Stderr, "rtnorm : 0 div\n");
        exit(1);
    }

    if(p->den < 0){
        p->num = - p->num;
        p->den = - p->den;
    }
    gcd = rtgcd(p->num, p->den);

    p->num = p->num/gcd;
    p->den = p->den/gcd;
}

/*==================================================================*/
Public  Void      rtadd(p, q)
rational p, q;
/*--------------------------------------------------------------------
 *    p += q;
 */
{
    p->num  = p->num * q->den + q->num * p->den;
    p->den *= q->den;

    rtnorm(p);
}

/*==================================================================*/
Public  Void      rtadds(p, q)
rational p;
long      q;
/*--------------------------------------------------------------------
 *    add numerator q/1 to p.
 */
{
    p->num += q * p->den;

    rtnorm(p);
}

/*==================================================================*/
Public  Void      rtmul(p, q)
rational p, q;
/*--------------------------------------------------------------------
 *    p *= q;
 */
{
    p->num *= q->num;
    p->den *= q->den;

    rtnorm(p);
}

/*==================================================================*/
Public  Void      rtmuls(p, q, r)
rational p;
long     q, r;
/*--------------------------------------------------------------------
 *    multiply numerator q/1 to p.
 */
{
    p->num *= q;
    p->den *= r;

    rtnorm(p);
}
/*==================================================================*/
Public  Void      rtdiv(p, q)
rational p, q;
/*--------------------------------------------------------------------
 *    p *= q;
 */
{
    p->num *= q->den;
    p->den *= q->num;

    rtnorm(p);
}

/*==================================================================*/
Public  Void      rtdivs(p, q)
rational p;
long      q;
/*--------------------------------------------------------------------
 *    multiply numerator 1/q to p.
 */
{
    p->den *= q;

    rtnorm(p);
}

/*==================================================================*/
Public  Void      rtprint(p)
rational p;
/*--------------------------------------------------------------------
 *    Print rational number.
 */
{
    printf("%ld/%ld", p->num, p->den);
}
