/*********************************************************************
 * File : rdptbl.c (prefix = pv)
 *
 * read particle table
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "common.h"
#include "config.h"
#include "dbgh.h"
#include "ptable.h"
#include "grdata.h"
#include "tokenh.h"

#ifndef sign
#define sign(x)      (((x) <  0 ) ? (-1) : (1))
#endif

#define CHANEL
#define OLDNAME

#define N3        3    /* fundamental representation of color group */
#define N8 (N3*N3 - 1) /* adjoint representation of color group */

/*====================================================================
        Entry points
 */

/*      Initialization */
Public  Void    pvfname         ARG((string,string,int));
Public  Void    pvfenam         ARG((string,string,string,int));
Public  Void    pvread          ARG((string));

/*      Find particle name and return particle number */
Public  pindex  pvfind          ARG((string));

/*      Returns name of pindex */
Public  string  pvname          ARG((pindex));
Public  string  pvename         ARG((int,pindex));
Public  string  pvgname         ARG((pindex,int));

/*      Print particle and vertex table */
Public  Void    pvprint         ARG((string));
Public  Void    pvhelp          ARG((FILE*));
Public  Void    pvdump          ARG((NoArg));
Public  Void    pvldump         ARG((FILE*, int, vindex[]));
Public  Void    pvprtex         ARG((string));
Public  Void    pvprnam         ARG((string));
Public  Void    pvprgnam        ARG((string));

Public  int     pvboson         ARG((pindex));
Public  int     pvfermi         ARG((pindex));
Public  int     pvsdeg          ARG((pindex));
Public  string  pvnptype        ARG((pindex));
Public  vindex  pvaddct         ARG((vindex,int*));
Public  int     pvvnfind        ARG((int));

Public  int     pvgl3           ARG((int));
Public  int     pvgl4           ARG((int));

/*  Global variables

        ptable  : table of particles.
        vtable  : table of vertices.

/*------------------------------------------------------------------*/

/* External references */
/* ergrf.c */
Extern  Void    erdbg        ARG((string, string, int));
Public  string pnfname       ARG((string,string,string,string));


/* Local functions */
Private int     pvpenter     ARG((string,string,string,string));
Private nindex  pvfindnam    ARG((string,int,int));
Private string  pvmname      ARG((string,int));
Private string  pvvname      ARG((string,int,int*));
Private bool    pvcpcpl      ARG((int,int));
Private int     pvnenter     ARG((int,string*));
Private Void    pvendtbl     ARG((NoArg));
Private Void    pvscale      ARG((int));
Private bool    pvparticle   ARG((NoArg));
Private bool    pvpnam       ARG((int, char*, char*));
Private bool    pvertex      ARG((NoArg));
Private bool    pvfort       ARG((string, int));
Private string  pvstrdup     ARG((bool, string));
Private int     pvprtxp      ARG((NoArg));
Private int     pvprtv1      ARG((NoArg));
Private int     pvprtv2      ARG((int));
Private string  pvctname     ARG((string,int*,int));
Private int     pvlord       ARG((pindex,pindex));
Private bool    pvfermiv     ARG((int));

#define PVNEWN  0
#define PVOLDN  1

Local   FILE    *prpf;
Local   FILE    *pvpf;
Local   int     pvind;
Local   char    pvfnam[MAXPATHLEN];

#define MAXVAL 64

#define MAXCMD 32
Local char   cmd[MAXCMD], argstr[MAXCMD];

/* help message */
Local string syntax[] = {
"Model definition file",
"  Syntax :",
"",
"\t file   ::= [ order-definition | particle-definition ",
"\t\t | vertex-definition ]* \"Mend\" \";\".",
"",
"\t order-definition ::=",
"\t\t \"Order={\" coupling-name [ \",\" coupling-name ]* \"};\".",
"",
"\t particle-definition ::= particle [ anti-particle | mass |",
"\t\t width | ptype | charge | color | pcode | kfcode | massless | furry ]*",
"\t\t \"Pend;\".",
"\t particle ::= \"Particle=\" particle-name ",
"\t\t [ \"[\" print-name \"]\" ] \";\".",
"\t anti-particle ::= \"Antiparticle=\" particle-name ",
"\t\t [ \"[\" print-name \"]\" ] \";\".",
"\t mass   ::= \"Mass=\"   ( mass-name  | \"0\" ) \";\".",
"\t width  ::= \"Width=\"  ( width-name | \"0\" ) \";\".",
"\t ptype  ::= \"PType=\"  ( Scalar | Ferion | Ghost | Vector) \";\".",
"\t charge ::= \"Charge=\" rational-number \";\".",
"\t color  ::= \"Color=\"  integer-number  \";\".",
"\t pcode  ::= \"PCode=\"  integer-number  \";\".",
"\t kfcode ::= \"KFCode=\"  integer-number  \";\".",
"\t massless ::= \"Massless \";\".",
"\t furry  ::= \"Furry=\"  ( \"Yes\" | \"No\" )  \";\".",
"",
"\t vertex-definition ::= p-list [ v-order | v-name ]* \"Vend;\".",
"\t p-list  ::= \"Vertex={\" [particle-name]+ \"};\".",
"\t v-order ::= coupling-name \"=\" integer-number \";\".",
"\t v-name  ::= \"Fname=\" vertex-name \";\"." };

/*********************************************************************
  Read model definition file.
*/
/*==================================================================*/
Public  Void    pvfname(fname, fin, len)
string  fname;             /* normalized name */
string  fin;               /* name of input file */
/*--------------------------------------------------------------------
 *  Normalize input model file name
 */
{
    string fullname;

    if((fullname = pnfname("GRCMODEL", modelpath, "", fin)) == NULL) {
        fprintf(Stderr, "*** pvfname : file not found %s\n", fin);
        exit(1);
    }
    if(strlen(fullname) >= len) {
        fprintf(Stderr, "*** pvfname : too long path name %d\n", len);
        exit(1);
    }
    strcpy(fname, fullname);
}

/*==================================================================*/
Public  Void    pvfenam(fname, fin, ext, maxlen)
string  fname;             /* normalized name */
string  fin;               /* name of input file */
string  ext;               /* name of input file */
/*--------------------------------------------------------------------
 *  Change extension part ".mdl" to ext.
 */
{
    int ln;

    ln = strlen(fin);
    if(ln >= 5 && strcmp(fin+ln-4, ".mdl") == 0)
        ln -= 4;
    if(ln+4 >= maxlen) {
        fprintf(Stderr, "*** pvfenam : length of fname is too small\n");
        exit(1);
    }
    strncpy(fname, fin, ln);
    strcat(fname, ".fin");
}

/*==================================================================*/
Public  Void    pvread(fname)
string  fname;             /* name of input file */
/*--------------------------------------------------------------------
    Read particle and vertex table.

    In the current version, this is test one.
--------------------------------------------------------------------*/
{
    int arg;
    int i, n, nv, j;
    string vp[MAXTL];
    bool err;

#if DBG
    erdbg("pvread", "begin", pvdbg);
#endif

    strcpy(pvfnam, fname);

    if((prpf = fopen(pvfnam, "r")) == NULL) {
        fprintf(Stderr, "\n*err* pvread: Model file open err,\n");
        fprintf(Stderr, "    File name = %s\n", pvfnam);
        exit(1);
    }

#if DBG
    if(dbg >= pvdbg) {
        printf("pvread: model = \"%s\"\n", pvfnam);
        rdprint(pvind);
    }
#endif

    /* initialize token package */
#if DBG
    if(dbg >= pvdbg) {
        rdinit(prpf, RDNOCHANGE, stdout);
    } else {
#endif
        rdinit(prpf, RDNOCHANGE, (FILE*) NULL);
#if DBG
    }
#endif

    nntable  = 0; /* table of particle names */
    nptable  = 0; /* table of particles */
    nvtable  = 0; /* table of vertices */
    nmordtbl = 0; /* table of coupling constants */
    nctable  = 0; /* table of names of coupling const of vertex */
    for(i = 0; i < N_SIZE; i++)
        nmtable[i] = 0;  /* table of names of mass and width */
    majop    = False;
    maxtl    = MINTL;
    mintl    = MAXTL;

    /* dummy particle */
    if(pvpenter("undef", "undef", "?", "?") != PNULL) {
        fprintf(Stderr, "\n*err* pvread: illegal particle table.\n");
        exit(1);
    }
    ptable[0]->mass     = "0";
    ptable[0]->width    = "0";
    ptable[0]->mvalue   = NULL;
    ptable[0]->wvalue   = NULL;
    ptable[0]->ptype    = P_UNDEF;
    ptable[0]->charge3  = 0;
    ptable[0]->color    = 1;
    ptable[0]->pcode    = 0;
    ptable[0]->kfcode   = 0;
    ptable[0]->furry    = False;
    ptable[0]->massless = False;


    /* dummy vertices ??? necessary ??? */
    for(i = 0; i < MAXTL; i++)
        vp[i] = "undef";
    for(n = MINTL; n <= MAXTL; n++) {
        nv = pvnenter(n, vp);
        vtable[nv]->vtype     = VTTREE;
        vtable[nv]->morder[0] = n - 2;
        vtable[nv]->loop      = 0;
        for (i = 1; i < nmordtbl; i++)
            vtable[nv]->morder[i] = 0;
        vtable[nv]->fname = "?";
        vtable[nv]->ctype = V_NOTYPE;
        vtable[nv]->phase = V_PUNDEF;
        for(j = 0; j < MAXFVAL; j++)
            vtable[nv]->fvalue[j] = NULL;
    }
    nvstart = nv + 1; /* read definition is >= nvstart */

    /* read "order= ... ", "particle= ...", "vertex= ..." or
            "medn;".
    */
    err = False;
    while (True) {
        pvind = rdstat();

        for(i = 0; i < MAXCMD - 1; i++)
            cmd[i] = tolower(rdid(pvind)[i]);
#if DBG
       if(dbg >= pvdbg) {
            erdbg("pvread", "statement :", pvdbg);
            rdprint(pvind);
        }
#endif

        if(rdtype(pvind) == RDEOF) {
            fprintf(Stderr, "\n*err* pvread : EOF before Mend;\n");
            err = True;
            break;

        }else if(rdtype(pvind) != RDID) {
            fprintf(Stderr, "\n*err* pvread : Id is expected\n");
            rddump(pvind);
            err = True;

        /* end of the file */
        } else if(strcmp(cmd, "mend") == 0) {
#if DBG
            erdbg("pvread", "id = mend", pvdbg);
#endif
            break;

        /* definition of coupling constants */
        } else if(strcmp(cmd, "order") == 0) {
#if DBG
            erdbg("pvread", "id = order", pvdbg);
#endif

            /* do not redefine */
            if(nmordtbl != 0) {
                fprintf(Stderr, "\n*err* pvread : ");
                fprintf(Stderr, "Order cannot be redefined.\n");
                rddump(pvind);
                err = True;

            /* definition */
            } else {
                for (arg = rdarg(pvind); rdtype(arg) == RDID;
                                         arg = rdarg(arg)) {
                    if(nmordtbl < MAXMORD-1) {
                        mordtbl[nmordtbl++] = pvstrdup(True, rdid(arg));
                    } else {
                        fprintf(Stderr, "\n*err* pvread : ");
                        fprintf(Stderr, "order table ovberflow.\n");
                        err = True;
                    }
                } /* for arg */
            }

        /* definition of particles */
        } else if (strcmp(cmd, "particle") == 0) {
#if DBG
            erdbg("pvread", "id = particel", pvdbg);
#endif
            err = !pvparticle() || err;

        /* definition of vertices */
        } else if (strcmp(cmd, "vertex") == 0) {
#if DBG
            erdbg("pvread", "id = vertex", pvdbg);
#endif
            /* vertex should be defined after particles */
            if (nptable < 2) {
                fprintf(Stderr, "\n*err* pvread : ");
                fprintf(Stderr, "vertex without particle\n");
                rddump(pvind);
                err = True;
            }

            /* vertex should be defined after order */
            if (nmordtbl < 1) {
                fprintf(Stderr, "\n*err* pvread : ");
                fprintf(Stderr, "vertex without coupling const def.\n");
                rddump(pvind);
                err = True;
            }
            err = !pvertex() || err;

        } else {
            fprintf(Stderr, "\n*err* pvread : syntax\n");
            rddump(pvind);
            err = True;
        }
    }

    rdend();
    fclose(prpf);

#if DBG
    erdbg("pvread", "end", pvdbg);
#endif

    if(err) {
        rdprint(pvind);
        fprintf(Stderr, "\n*err* pvread: stop by error.\n");
        exit(1);
    }

    pvendtbl();

    nvmodel = nvtable;

#if DBG
    if(dbg >= pvdbg) {
        pvdump();
    }
#endif
}

/*==================================================================*/
Private bool    pvparticle(NoArg)
/*--------------------------------------------------------------------
    read particle upto "pend;"
--------------------------------------------------------------------*/
{
    int arg, np, ngn;
    char pnam[MAXLID], apnam[MAXLID], psym[MAXLID], apsym[MAXLID];
    char mass[7], width[7], mvalue[MAXVAL], wvalue[MAXVAL];
    char *gname[3];
    int  charge[2], color, value, pcode, kfcode, ptype, massless;
    int  i;
    bool furry, err;

#if DBG
    erdbg("pvparticle", "begin", pvdbg);
#endif

    /* get particle name and print-name */
    arg = rdarg(pvind);
    err = !pvpnam(arg, pnam, psym);

    /* variables to be set */
    apnam[0]  = '\0';      /* mark for undef */
    mass[0]   = '\0';      /* mark for undef */
    width[0]  = '\0';      /* mark for undef */
    mvalue[0] = '\0';      /* mark for undef */
    wvalue[0] = '\0';      /* mark for undef */
    ptype     = P_UNDEF;   /* mark for undef */
    charge[1] = 0;         /* mark for undef */
    color     = 1;         /* default */
    pcode     = 0;         /* default */
    kfcode    = 0;         /* default */
    furry     = False;     /* default */
    massless  = False;     /* default */
    gname[0]  = NULL;
    gname[1]  = NULL;
    gname[2]  = NULL;

    /* read "antiparticle=...", "charge=...", "color=..." or "pend;"
    */
    while (True) {
        pvind = rdstat();

        for(i = 0; i < MAXCMD - 1; i++)
            cmd[i] = tolower(rdid(pvind)[i]);

        if (rdtype(pvind) == RDEOF) {
            fprintf(Stderr, "\n*err* pvparticle : EOF\n");
            err = True;
            break;

        } else if (rdtype(pvind) != RDID) {
            fprintf(Stderr, "\n*err* pvparticle : illegal data\n");
            rddump(pvind);
            err = True;

        /* end of particle definition */
        } else if (strcmp(cmd, "pend") == 0) {
            break;

        /* antiparticle = name [ print-name ] */
        } else if (strcmp(cmd, "antiparticle") == 0) {
            arg = rdarg(pvind);
            err = !pvpnam(arg, apnam, apsym) || err;

        /* gname = {common-name [, particle-name, anti-paritcle-name]}; */
        } else if (strcmp(cmd, "gname") == 0) {
            for (arg = rdarg(pvind), ngn = 0; arg >= 0;
                     arg = rdarg(arg), ngn++) {
                if(ngn < 3) {
                    gname[ngn] = pvstrdup(False, rdid(arg));
                }
            }
            if(ngn > 3) {
                fprintf(Stderr, "\n*err* pvparticle : illgal gname\n");
                rddump(pvind);
                err = True;
            } 

        /* ptype = name; */
        } else if (strcmp(cmd, "ptype") == 0) {
            arg = rdarg(pvind);
            if (rdtype(arg) != RDID) {
                fprintf(Stderr, "\n*err* pvparticle : illgal ptype\n");
                rddump(pvind);
                err = True;
            } else {
                for(ptype = PTYPEMAX-1; ptype > P_UNDEF; ptype--) {
                    for(i = 0; i < MAXCMD - 1; i++)
                        argstr[i] = tolower(rdid(arg)[i]);
                    if(strcmp(argstr, pvtnam[ptype]) == 0)
                        break;
                }
                if(ptype <= P_UNDEF) {
                    fprintf(Stderr, "\n*err* pvparticle :");
                    fprintf(Stderr, " illgal ptype (not defined)\n");
                    rddump(pvind);
                    err = True;
                }
            }

        /* charge = rational-number; */
        } else if (strcmp(cmd, "charge") == 0) {
            arg = rdarg(pvind);
            if ((rdtype(arg) != RDRNUM && rdtype(arg) != RDINT)) {
                fprintf(Stderr, "\n*err* pvparticle: illgal charge.\n");
                rddump(pvind);
                err = True;
            } else {
                charge[0] = (int)rdrnum(arg, 0);
                charge[1] = (int)rdrnum(arg, 1);
            }

        /* pcode = integer; */
        } else if (strcmp(cmd, "pcode") == 0) {
            arg = rdarg(pvind);
            if (rdtype(arg) != RDINT) {
                fprintf(Stderr, "\n*err* pvparticle :");
                fprintf(Stderr, " illgal pcode (not int)\n");
                rddump(pvind);
                err = True;
            } else {
                pcode = (int)rdint(arg);
            }

        /* kfcode = integer; */
        } else if (strcmp(cmd, "kfcode") == 0) {
            arg = rdarg(pvind);
            if (rdtype(arg) != RDINT) {
                fprintf(Stderr, "\n*err* pvparticle :");
                fprintf(Stderr, " illgal kfcode (not int)\n");
                rddump(pvind);
                err = True;
            } else {
                kfcode = (int)rdint(arg);
            }

        /* color = integer; */
        } else if (strcmp(cmd, "color") == 0) {
            arg = rdarg(pvind);
            if (rdtype(arg) != RDINT) {
                fprintf(Stderr, "\n*err* pvparticle :");
                fprintf(Stderr, " illgal color (not int)\n");
                rddump(pvind);
                err = True;
            } else {
                color = (int)rdint(arg);
            }

        /* furry = bool; */
        } else if (strcmp(cmd, "furry") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* pvparticle :");
                fprintf(Stderr, " furry=yes/no\n");
                rddump(pvind);
                err = True;
            } else {
                furry = value;
            }

        /* massive; */
        } else if (strcmp(cmd, "massive") == 0) {
            massless = False;

        /* massless; */
        } else if (strcmp(cmd, "massless") == 0) {
            massless = True;

        /* mass = fortran-name; */
        } else if (strcmp(cmd, "mass") == 0) {
            arg = rdarg(pvind);

            /* name */
            if (rdtype(arg) == RDID) {
                strcpy(mass, rdid(arg));
                if(!pvfort(mass, 6)) {
                    fprintf(Stderr, "\n*err* pvparticle: mass should ");
                    fprintf(Stderr, " be fortran id.\n");
                    rddump(pvind);
                    err = True;
                }

            /* 0 mass */
            } else if (rdtype(arg) == RDINT &&
                      ((int)rdint(arg)) == 0) {
                mass[0] = '0';
                mass[1] = '\0';

            } else {
                fprintf(Stderr, "\n*err* pvparticle: mass should ");
                fprintf(Stderr, " be fortran id.\n");
                rddump(pvind);
                err = True;
            }

        /* width = fortran-name; */
        } else if (strcmp(cmd, "width") == 0) {
            arg = rdarg(pvind);

            /* fortran name */
            if (rdtype(arg) == RDID) {
                strcpy(width, rdid(arg));
                if(!pvfort(width, 6)) {
                    fprintf(Stderr, "\n*err* pvparticle: width should ");
                    fprintf(Stderr, " be fortran id.\n");
                    rddump(pvind);
                    err = True;
                }

            /* 0 width */
            } else if (rdtype(arg) == RDINT && 
                      ((int)rdint(arg)) == 0) {
                width[0] = '0';
                width[1] = '\0';
            } else {
                fprintf(Stderr, "\n*err* pvparticle: width should ");
                fprintf(Stderr, " be fortran id.\n");
                rddump(pvind);
                err = True;
            }

        /* mvalue = string; */
        } else if (strcmp(cmd, "mvalue") == 0) {
            arg = rdarg(pvind);

            /* fortran name */
            if (rdtype(arg) == RDSTRING) {
                if(strlen(mvalue) >= MAXVAL) {
                    fprintf(Stderr, "\n*err* pvparticle: mvalue ");
                    fprintf(Stderr, " too long.\n");
                    rddump(pvind);
                    err = True;
                } else {
                    strcpy(mvalue, rdid(arg));
                }

            } else {
                fprintf(Stderr, "\n*err* pvparticle: mvalue should ");
                fprintf(Stderr, " be string.\n");
                rddump(pvind);
                err = True;
            }


        /* wvalue = string; */
        } else if (strcmp(cmd, "wvalue") == 0) {
            arg = rdarg(pvind);

            /* fortran name */
            if (rdtype(arg) == RDSTRING) {
                if(strlen(wvalue) >= MAXVAL) {
                    fprintf(Stderr, "\n*err* pvparticle: wvalue ");
                    fprintf(Stderr, " too long.\n");
                    rddump(pvind);
                    err = True;
                } else {
                    strcpy(wvalue, rdid(arg));
                }

            } else {
                fprintf(Stderr, "\n*err* pvparticle: wvalue should ");
                fprintf(Stderr, " be string.\n");
                rddump(pvind);
                err = True;
            }

        /* unknown statement */
        } else {
            fprintf(Stderr, "\n*err* pvparticle : illegal data\n");
            rddump(pvind);
            err = True;
        }
    }

    if(apnam[0] == '\0') {
        fprintf(Stderr, "\n*err* pvparticle : undefined anti-particle");
        fprintf(Stderr, " of %s\n", pnam);
        err = True;
    }
    if(mass[0] == '\0') {
        fprintf(Stderr, "\n*err* pvparticle : undefined mass");
        fprintf(Stderr, " of %s\n", pnam);
        err = True;
    }
    if(width[0] == '\0') {
        fprintf(Stderr, "\n*err* pvparticle : undefined width");
        fprintf(Stderr, " of %s\n", pnam);
        err = True;
    }
    if(charge[1] == 0) {
        fprintf(Stderr, "\n*err* pvparticle : undefined charge");
        fprintf(Stderr, " of %s\n", pnam);
        err = True;
    }
#if DBG
    erdbg("pvparticle", "end", pvdbg);
#endif

    if(err) {
        rdprint(pvind);
        return False;
    }

    /* save particle and antiparticle */
    np = pvpenter(pnam, apnam, psym, apsym);

    if(ptype == P_MAJORANA)
        majop = True;

    ptable[np]->mass     = pvmname(mass, N_MASS);
    ptable[np]->width    = pvmname(width, N_WIDTH);
    ptable[np]->charge3  = ((int) 3*charge[0])/charge[1];
    ptable[np]->pcode    = pcode;
    ptable[np]->kfcode   = kfcode;
    ptable[np]->ptype    = ptype;
    ptable[np]->color    = color;
    ptable[np]->furry    = furry;
    ptable[np]->massless = massless;

    if(mvalue[0] == '\0')
        ptable[np]->mvalue    = NULL;
    else
        ptable[np]->mvalue    = pvstrdup(False, mvalue);

    if(wvalue[0] == '\0')
        ptable[np]->wvalue    = NULL;
    else
        ptable[np]->wvalue    = pvstrdup(False, wvalue);

    if(gname[0] == NULL) {
        gname[0] = psym;
        gname[1] = psym;
        gname[2] = apsym;
    } else {
        if(gname[1] == NULL) {
            gname[1] = gname[0];
        }
        if(gname[2] == NULL) {
            gname[2] = gname[0];
        }
    }
    ptable[np]->gname[0] = gname[0];
    ptable[np]->gname[1] = gname[1];
    ptable[np]->gname[2] = gname[2];

    return True;
}

/*==================================================================*/
Private bool    pvpnam(arg, pnam, psym)
int     arg;
char    *pnam;
char    *psym;
/*--------------------------------------------------------------------
    read particle name and abbraviation.

    particle-name [ "[" print-name "]" ] 
--------------------------------------------------------------------*/
{
    int opt;

    if(rdtype(arg) != RDID) {
        fprintf(Stderr, "\n*err* pvparticle : illegal particle name\n");
        rddump(arg);
        rdprint(pvind);
        return False;

    } else if(strlen(rdid(arg)) >= MAXLID - 1) {
        fprintf(Stderr, "\n*err* pvpnam : too long name\n");
        rddump(arg);
        rdprint(pvind);
        return False;

    } else {
        strcpy(pnam, rdid(arg));
        if(rdtype(opt = rdopt(arg)) == RDSTRING) {
            if(strlen(rdid(opt)) >= MAXLID - 1) {
                fprintf(Stderr, "\n*err* pvpnam : too long option\n");
                rddump(arg);
                rdprint(pvind);
                return False;
            } else {
                strcpy(psym, rdid(opt));
            }
        } else if(opt < 0) {
            psym[0] = '\0';
        } else {
            fprintf(Stderr, "\n*err* pvpnam: illegal print name ");
            fprintf(Stderr, "(should be string).\n");
            rddump(arg);
            rdprint(pvind);
            return False;
        }
    }
    return True;
}
/*==================================================================*/
Private int     pvpenter(pnam, anam, psym, asym)
string   pnam;
string   anam;
string   psym;
string   asym;
/*--------------------------------------------------------------------
    Enter particle to the table.
    In the ptable, items pname, aname, antip are stored.
--------------------------------------------------------------------*/
{
    int np = nptable;
    int i;

#if DBG
   if(dbg >= pvdbg) {
        printf("pventer: nptable = %d, (%s, %s, %s, %s)\n",
                nptable, pnam, anam, psym, asym);
        rdprint(pvind);
    }
#endif

    /* check table size */
    if(nptable >= LPTABLE) {
        fprintf(Stderr, "\n*err* pvpenter: particle table overflow\n");
        exit(1);
    }

    /* create space for new particle */
    nptable++;
    if((ptable[np] = (tptable*) malloc(sizeof(tptable))) == NULL) {
        fprintf(Stderr, "\n*err* pvpenter: memory allocation error\n");
        exit(1);
    }

    /* register particle name */
    ptable[np]->pname[0] = pvfindnam(pnam, np, PVNEWN);
    if(psym[0] == '\0')
        ptable[np]->psym[0] = ntable[ptable[np]->pname[0]].name;
    else
        ptable[np]->psym[0] = pvstrdup(False, psym);

    for(i = 0; i < MAXCMD - 1; i++)
        argstr[i] = tolower(anam[i]);

    if(strcmp(pnam, anam) == 0 || strcmp("particle", argstr) == 0) {
        ptable[np]->pname[1] = ptable[np]->pname[0];
        ptable[np]->psym[1]  = ptable[np]->psym[0];
        ptable[np]->antip = 1;
    } else {
        ptable[np]->pname[1] = pvfindnam(anam, -np, PVNEWN);
        if(asym[1] == '\0')
            ptable[np]->psym[1] = ntable[ptable[np]->pname[1]].name;
        else
            ptable[np]->psym[1] = pvstrdup(False, asym);
        ptable[np]->antip = -1;
    }

    return np;
}

/*==================================================================*/
Public  string   pvename(iorf, pc)
int     iorf;
pindex  pc;
/*--------------------------------------------------------------------
    returns name of pindex pc.
--------------------------------------------------------------------*/
{
    pindex pca;

    if(pc == PNULL) return "undef";

    pca = pc;
    if(iorf == ELINITIAL)
        pca *= ptable[abs(pc)]->antip;

    return pvname(pca);
}

/*==================================================================*/
Public  string   pvname(pc)
pindex  pc;
/*--------------------------------------------------------------------
    returns name of pindex pc.
--------------------------------------------------------------------*/
{
    nindex n;
    if(pc == PNULL) return "undef";
    n = (pc < 0) ? ptable[-pc]->pname[1] : ptable[pc]->pname[0];
    return ntable[n].name;
}

/*==================================================================*/
Public  string   pvgname(pc, gpa)
pindex  pc;
int     gpa;
/*--------------------------------------------------------------------
    returns gname of pindex pc.

    gpa = 0 : generic name
          1 : particle (pc > 0),      anti-aprticle (pc < 0)
          2 : anti-particle (pc > 0), particle (pc < 0)
--------------------------------------------------------------------*/
{
    if(pc == PNULL) return "undef";
    if(pc < 0) {
        pc = - pc;
        if(gpa != 0)
           gpa = 3 - gpa;
    }
    return ptable[pc]->gname[gpa];
}

/*==================================================================*/
Public pindex   pvfind(pnam)
string   pnam;
/*--------------------------------------------------------------------
    find pname from name table (inverse of pvname)
--------------------------------------------------------------------*/
{
    if(strcmp(pnam, "undef") == 0)
        return PNULL;
    else
        return ntable[pvfindnam(pnam, 0, PVOLDN)].pindx;
}

/*==================================================================*/
Private int     pvfindnam(pnam, pind, mode)
string   pnam;
int     pind;
int     mode;
/*--------------------------------------------------------------------
    find pname from name table and add it.

    pind : particle index to be added.

    if mode >= 0 : pname should be new one.
            <    : pname should be old one.
--------------------------------------------------------------------*/
{
    int p;

    /* check the same name in the table */
    for(p = 0; p < nntable; p++) {
        if(strcmp(ntable[p].name, pnam) == 0) {
            if(mode == PVNEWN) {
                fprintf(Stderr, "\n*err* pvfindnam: particle is ");
                fprintf(Stderr," already defined. name=%s.\n", pnam);
                exit(1);
            } else {
                return p;
            }
        }
    }

    /* not found */
    if(mode != PVNEWN) {
        fprintf(Stderr, "\n*err* pvfindnam: particle is not found\n");
        fprintf(Stderr,"name=%s.\n", pnam);
        exit(1);
    }

    /* check size to save */
    if(nntable >= LNTABLE) {
        fprintf(Stderr, "\n*err* pvfindnam: name table overflow\n");
        exit(1);
    }

    p = nntable++;
    if((ntable[p].name = pvstrdup(False, pnam)) == NULL) {
        fprintf(Stderr, "\n*err* pvfindnam: memory allocation error\n");
        exit(1);
    }
    ntable[p].pindx = pind;

    return p;
}

/*==================================================================*/
Private string  pvmname(nam, ntype)
string  nam;
int     ntype;
/*--------------------------------------------------------------------
    find pname from name table and add it.
--------------------------------------------------------------------*/
{
    int p;

    /* check the same name in the table */
    for(p = 0; p < nmtable[ntype]; p++) {
        if(strcmp(mtable[ntype][p], nam) == 0) {
            return mtable[ntype][p];
        }
    }

    /* check size to save */
    if(nmtable[ntype] >= LPTABLE) {
        fprintf(Stderr, "\n*err* pvmname: name table overflow\n");
        exit(1);
    }

    p = nmtable[ntype]++;
    if((mtable[ntype][p] = pvstrdup(True, nam)) == NULL) {
        fprintf(Stderr, "\n*err* pvmname: memory allocation error\n");
        exit(1);
    }

    return mtable[ntype][p];
}

/*==================================================================*/
Private string  pvvname(nam, vnew, vold)
string  nam;
int     vnew;
int    *vold ;
/*--------------------------------------------------------------------
--------------------------------------------------------------------*/
{
    int p;

    /* check the same name in the table */
    for(p = 0; p < nctable; p++) {
        if(strcmp(ctable[p], nam) == 0) {
            *vold = vnamdef[p];
            return ctable[p];
        }
    }

    /* check size to save */
    if(nctable >= LVTABLE) {
        fprintf(Stderr, "\n*err* pvvname: name table overflow\n");
        exit(1);
    }

    p = nctable++;
    if((ctable[p] = pvstrdup(False, nam)) == NULL) {
        fprintf(Stderr, "\n*err* pvvname: memory allocation error\n");
        exit(1);
    }

    vnamdef[p] = *vold = vnew;

    return ctable[p];
}


/*==================================================================*/
Public  int     pvvnfind(vn)
int     vn;
/*--------------------------------------------------------------------
--------------------------------------------------------------------*/
{
    int p;
    string nam = vtable[vn]->fname;

    /* check the same name in the table */
    for(p = 0; p < nctable; p++) {
        if(strcmp(ctable[p], nam) == 0) {
            return vnamdef[p];
        }
    }
    return -1;
}

/*==================================================================*/
Private bool    pvertex(NoArg)
/*--------------------------------------------------------------------
    read vertex.
--------------------------------------------------------------------*/
{
    string vp[MAXTL+1];
    string id;
    int    arg, nvp, nv, nc, i, ord, cg, p, c, lp, j, nvdef;
    bool   err = False;

#if DBG
    erdbg("pvertex", "begin", pvdbg);
#endif

    /* for list of particles */
    for (arg = rdarg(pvind), nvp = 0; arg >= 0;
             arg = rdarg(arg), nvp++) {

        /* too many particles in the list */
        if (nvp > MAXTL) {
            fprintf(Stderr, "\n*err* pvertex : too many legs\n");
            rddump(pvind);
            err = True;
            break;

        } else if (rdtype(arg) != RDID) {
            fprintf(Stderr, "\n*err* pvertex : illegal leg\n");
            rddump(pvind);
            err = True;
            break;

        /* save temporary */
        } else {
            vp[nvp] = rdid(arg);
        }
    }

    if(!err) {
        nv = pvnenter(nvp, vp);

        cg = 0;
        for(i = 0; i < nvp; i++) {
            p   = vtable[nv]->vlegs[i];
            c   = sign(p)*ptable[abs(p)]->charge3;
            cg += c;
        }

        if(cg != 0) {
            fprintf(Stderr, "\n*err* pvertex : charge not conserved\n");
            rddump(pvind);
            err = True;
        }

        for (i = 0; i < nmordtbl; i++)
            vtable[nv]->morder[i] = 0;
        vtable[nv]->fname = NULL;
        vtable[nv]->ctype = V_NOTYPE;
        vtable[nv]->phase = V_PUNDEF;
        for(j = 0; j < MAXFVAL; j++)
            vtable[nv]->fvalue[j] = NULL;
    }

    while (True) {
        pvind = rdstat();

        for(i = 0; i < MAXCMD - 1; i++)
            cmd[i] = tolower(rdid(pvind)[i]);

        if (rdtype(pvind) == RDEOF) {
            fprintf(Stderr, "\n*err* pvertex : EOF\n");
            rddump(pvind);
            err = True;
            break;

        } else if (rdtype(pvind) != RDID) {
            fprintf(Stderr, "\n*err* pvertex : illegal data\n");
            rddump(pvind);
            err = True;

        } else if (strcmp(cmd, "vend") == 0) {
            break;

        } else if (strcmp(cmd, "fname") == 0) {
            arg = rdarg(pvind);
            if (rdtype(arg) != RDID) {
                fprintf(Stderr, "\n*err* pvertex : illegal fname\n");
                rddump(pvind);
                err = True;
            } else if(!pvfort(rdid(arg), 4)) {
                fprintf(Stderr, "\n*err* pvertex: name should be ");
                fprintf(Stderr, "fortran name\n");
                rddump(pvind);
                err = True;
            } else if(!err) {
                vtable[nv]->fname  = pvvname(rdid(arg), nv, &nvdef);
            }

        } else if (strcmp(cmd, "ftype") == 0) {
            arg = rdarg(pvind);
            if (rdtype(arg) != RDSTRING) {
                fprintf(Stderr, "\n*err* pvertex : illegal ftype\n");
                rddump(pvind);
                err = True;
            } else {
                id = rdid(arg);
                for(i = 0; i < strlen(id); i++)
                    id[i] = toupper(id[i]);
                if(strcmp(id, "NON") == 0)
                    vtable[nv]->ctype = V_NOTYPE;
                else if(strcmp(id, "V") == 0)
                    vtable[nv]->ctype = V_V;
                else if(strcmp(id, "A") == 0)
                    vtable[nv]->ctype = V_A;
                else if(strcmp(id, "V-A") == 0)
                    vtable[nv]->ctype = V_VminusA;
                else if(strcmp(id, "V+A") == 0)
                    vtable[nv]->ctype = V_VplusA;
                else if(strcmp(id, "S") == 0)
                    vtable[nv]->ctype = V_V;
                else if(strcmp(id, "P") == 0)
                    vtable[nv]->ctype = V_A;
                else if(strcmp(id, "S-P") == 0)
                    vtable[nv]->ctype = V_VminusA;
                else if(strcmp(id, "S+P") == 0)
                    vtable[nv]->ctype = V_VplusA;
                else {
                    fprintf(Stderr, "\n*err* pvertex: illegal FType");
                    fprintf(Stderr, "%s\n", id);
                    rddump(pvind);
                    err = True;
                }
            }

        } else if (strcmp(cmd, "fvalue") == 0) {
            arg = rdarg(pvind);
            if (rdtype(arg) != RDID) {
                fprintf(Stderr, "\n*err* pvertex : illegal phase\n");
                rddump(pvind);
                err = True;
            } else {
                id = rdid(arg);
                for(i = 0; i < strlen(id); i++)
                    id[i] = tolower(id[i]);
                if(strcmp(id, "r") == 0)
                    vtable[nv]->phase = V_REAL;
                else if(strcmp(id, "i") == 0)
                    vtable[nv]->phase = V_IMAG;
                else if(strcmp(id, "c") == 0)
                    vtable[nv]->phase = V_CMPLX;
                else {
                    fprintf(Stderr, "\n*err* pvertex: illegal phase\n");
                    fprintf(Stderr, "%s\n", id);
                    rddump(pvind);
                    err = True;
                }
            }
            if(!err) {
                for (arg = rdarg(arg), j = 0; arg >= 0;
                     arg = rdarg(arg), j++) {
                    if(j >= MAXFVAL) {
                        fprintf(Stderr, "\n*err* pvertex : ");
                        fprintf(Stderr, "too many fvalues.\n");
                        rddump(pvind);
                        err = True;
                    } else if(rdtype(arg) != RDSTRING) {
                        fprintf(Stderr, "\n*err* pvertex :");
                        fprintf(Stderr, " illegal FValue\n");
                        fprintf(Stderr, "/%s/(%d)\n", 
                                rdid(arg), rdtype(arg));
                        rddump(pvind);
                        err = True;
                    } else {
                        vtable[nv]->fvalue[j] = pvstrdup(True, rdid(arg));
                    }
                }
            }


        /* coupling constants */
        } else {
            arg = -1;
            for (i = 0; i < nmordtbl; i++) {
                if(strcmp(mordtbl[i], cmd) == 0) {
                    arg = rdarg(pvind);
                    if(rdtype(arg) != RDINT) {
                        fprintf(Stderr, "\n*err* pvertex : ");
                        fprintf(Stderr, "illegal fname\n");
                        rddump(pvind);
                        err = True;
                    } else if (err) {
                        ;
                    } else if (vtable[nv]->morder[i] != 0) {
                        fprintf(Stderr, "\n*err* pvertex : ");
                        fprintf(Stderr, "duplicate coupling\n");
                        rddump(pvind);
                        err = True;
                    } else {
                        vtable[nv]->morder[i] = (int)rdint(arg);
                    }
                }
            }
            if (arg < 0) {
                fprintf(Stderr, "\n*err* pvertex : illegal data\n");
                rddump(arg);
                err = True;
            }
        }
    }
#if DBG
    erdbg("pvertex", "end", pvdbg);
#endif

    ord = 0;
    for (i = 0; i < nmordtbl; i++)
        ord += vtable[nv]->morder[i];

    lp = ord - nvp + 2;
    if (lp % 2 != 0 || lp < 0) {  /* change for counter terms ???? */
        fprintf(Stderr, "\n*err* pvertex : illegal order\n");
        fprintf(Stderr, "loop=%d.5\n", lp/2);
        err = True;
    } else {
        vtable[nv]->loop = lp / 2;
    }

    if (vtable[nv]->fname == NULL) {
        fprintf(Stderr,"\n*err* pvertex : Fortran name is not defined\n");
        err = True;
    }
    if(lp == 0) {
        maxtl = max(maxtl, nvp);
        mintl = min(mintl, nvp);
        vtable[nv]->vtype = VTTREE;
    } else {
        vtable[nv]->vtype = VTLOOP;
    }

    if(!err && nv != nvdef) {
        err = !pvcpcpl(nv, nvdef);
        if(err) {
            fprintf(Stderr,"\n*err* pvertex : Fortran name %s ",
                    vtable[nv]->fname);
            fprintf(Stderr,"defined for vertex %d is \n", nv);
            fprintf(Stderr,"inconsistent to the definition for %d.\n",
                    nvdef);
        }
    }

    if(pvfermiv(nv)) {
        if(vtable[nv]->ctype == V_NOTYPE)
            nc = 2;
        else
            nc = 1;
    } else if(vtable[nv]->ctype != V_NOTYPE) {
        fprintf(Stderr, "\n*err* pvertex : FType should not be ");
        fprintf(Stderr, "defined for boson vertex\n");
        err = True;
    } else {
        nc = 1;
    }

    if(!err && testfv) {
        for(i = 0; i < MAXFVAL; i++) {
            if(vtable[nv]->fvalue[i] == NULL) {
                if(i == 0) {
                    if(vtable[nv]->loop < 1) {
                        fprintf(Stderr, "\n*warning* pvertex : ");
                        fprintf(Stderr, "FValue for ");
                        fprintf(Stderr, "(FName = %s)", vtable[nv]->fname);
                        fprintf(Stderr, " is not set.");
                    }
                    break;
                } else if(i < nc) {
                    err = True;
                    break;
                }
            } else {
                if(i >= nc)
                    err = True;
            }
        }
        if(err) {
            fprintf(Stderr, "\n*err* pvertex : ");
            fprintf(Stderr, "(Fname = %s)", vtable[nv]->fname);
            fprintf(Stderr, "argument of FValue ");
            fprintf(Stderr, "should be \n\t{<type>");
            for(i = 0; i < nc; i++)
                fprintf(Stderr, ",\"<arg%d>\"",i+1);
            fprintf(Stderr, "};\twhere, <type> = R, I or C\n");
        }
    }

    if(err) {
        rdprint(pvind);
    }
    return ! err;
}

/*==================================================================*/
Private bool    pvcpcpl(nn, no)
int     nn, no;
/*------------------------------------------------------------------*
 * Check consistency of coupling constant.
 */
{
    bool    err = False;
    tvtable *vn, *vo;
    int     i, t1, t2;

    if(nn == no)
        return True;

    vn = vtable[nn];
    vo = vtable[no];

    err |= (vn->vtype  != vo->vtype);
    err |= (vn->nvlegs != vo->nvlegs);
    err |= (vn->loop   != vo->loop);
    for(i = 0; i < vn->nvlegs; i++) {
        t1 = ptable[abs(vn->vlegs[i])]->ptype;
        t2 = ptable[abs(vo->vlegs[i])]->ptype;
        if(t1 == P_MAJORANA)
            t1 = P_FERMION;
        if(t2 == P_MAJORANA)
            t2 = P_FERMION;
        err |= (t1 != t2);
    }

    if(vn->ctype == V_NOTYPE)
        vn->ctype = vo->ctype;
    else
        err |= (vn->ctype != vo->ctype);

    if(vn->phase == V_PUNDEF)
        vn->phase = vo->phase;
    else
        err |= (vn->phase != vo->phase);

    for(i = 0; i < MAXFVAL; i++) {
        if(vn->fvalue[i] == NULL)
            vn->fvalue[i] = vo->fvalue[i];
        else if(strcmp(vn->fvalue[i], vo->fvalue[i]) != 0)
            err |= True;
    }
    
    return !err;
}

/*==================================================================*/
Private int     pvlord(p1, p2)
pindex  p1, p2;
/*
 * compare type of two particles.
 * scalar < majorana < anti-fermion < fermion < anti-ghost 
 *        < ghost < vector
 */
{
    int    s1, s2, b1, b2;
    int    a1, a2, d;

    if(p1 >= 0) {
        a1  = 1;
    } else {
        p1  = -p1;
        a1  = -1;
    }
    s1  = ptable[p1]->ptype;
    if(s1 == P_FERMION || s1 == P_GHOST)
        b1 = False;
    else
        b1 = True;

    if(p2 >= 0) {
        a2  = 1;
    } else {
        p2  = -p2;
        a2  = -1;
    }
    s2  = ptable[p2]->ptype;
    if(s2 == P_FERMION || s2 == P_GHOST)
        b2 = False;
    else
        b2 = True;

    d   = s1 - s2;
    if(d != 0 || b1 || b2)
        return d;

    /* Fermion and check particle/anti-paritlce */
    return (a1 - a2);
}

/*==================================================================*/
Private int     pvnenter(n, vp)
int     n;
string  *vp;
/*--------------------------------------------------------------------
 *  Enter vertex.
 */
{
    int    i, nv;

#if DBG
    erdbg("pvnenter", "begin", pvdbg);
#endif

    if(nvtable >= LVTABLE) {
        fprintf(Stderr, "\n*err* pvnenter: vertex table overflow\n");
        exit(1);
    }

    /* create space for new vertex */
    nv = nvtable++;
    if((vtable[nv] = (tvtable*) malloc(sizeof(tvtable))) == NULL) {
        fprintf(Stderr, "\n*err* pvnenter: memory allocation ");
        fprintf(Stderr, "error(1)\n");
        exit(1);
    }
    if((vtable[nv]->vlegs = (pindex*)malloc(sizeof(pindex)*n))
        == NULL) {
        fprintf(Stderr, "\n*err* pvnenter: memory allocation ");
        fprintf(Stderr, "error(2)\n");
        exit(1);
    }

    vtable[nv]->nvlegs = n;

    if(vp == NULL)
        return nv;

    for(i = 0; i < n; i++) {
        vtable[nv]->vlegs[i] = pvfind(vp[i]);
#ifdef CHANEL
        if(i > 0 && pvlord(vtable[nv]->vlegs[i-1],
                           vtable[nv]->vlegs[i]) > 0) {
            fprintf(Stderr, "\n*err* pventer : wrong order of");
            fprintf(Stderr, " legs %s > %s\n",
                            pvname(vtable[nv]->vlegs[i-1]),
                            pvname(vtable[nv]->vlegs[i]));
            
            fprintf(Stderr, "Particles in a vertex should be");
            fprintf(Stderr, "arranged in the order of\n");
            fprintf(Stderr, "\tscalar, majorana, anti-fermion, ");
            fprintf(Stderr, "fermion, anti-ghost, ghost, vector\n");
            exit(1);
        }
#endif
    }

#if DBG
    erdbg("pvnenter", "end", pvdbg);
#endif

    return nv;
}

/*==================================================================*/
Private Void    pvendtbl(NoArg)
/*--------------------------------------------------------------------
    Now entering particle and vertices has finished.
    Set up other items of the table here.
--------------------------------------------------------------------*/
{
    int         p, v, l, vc[2], vl[2];
    vindex     *vt[2];

#if DBG
    erdbg("pvendtbl", "begin", pvdbg);
#endif

    /* work space */
    for(l = 0; l < 2; l++) {
        if((vt[l] = (vindex*)malloc(nvtable*sizeof(vindex))) == NULL) {
            fprintf(Stderr, "\n*err* pvendtbl: memory allocation ");
            fprintf(Stderr, "error(1)\n");
            exit(1);
        }
    }

    /* list of vertices using particle "p" */
    for(p = 0; p < nptable; p++) {
        vc[0] = vc[1] = 0;
        vl[0] = vl[1] = VNULL;
        for(v = 0; v < nvtable; v++) {
            for(l = 0; l < vtable[v]->nvlegs; l++) {
                if(v != vl[0] && vtable[v]->vlegs[l] == p) {
                    vt[0][vc[0]++] = v;
                    vl[0] = v;
                }
                if(v != vl[1] && vtable[v]->vlegs[l] == -p) {
                    vt[1][vc[1]++] = v;
                    vl[1] = v;
                }
            }
        }
#if DBG
        if(dbg >= pvdbg) {
            erdbg("pvendtbl", "1", pvdbg);
            printf(" p = %d vc[0] = %d, vc[1] = %d\n", p, vc[0], vc[1]);
        }
#endif
        ptable[p]->nvlist[0] = vc[0];
        if(vc[0] >= 1 && 
           (ptable[p]->vlist[0] = (vindex*)malloc(vc[0]*sizeof(vindex)))
            == NULL) {
            fprintf(Stderr, "\n*err* pvendtbl: memory allocation ");
            fprintf(Stderr, "error(2)\n");
            fprintf(Stderr, "size = %d\n", vc[0]*sizeof(vindex));
            exit(1);
        }
        for(v = 0; v < vc[0]; v++)
            ptable[p]->vlist[0][v] = vt[0][v];

        ptable[p]->nvlist[1] = vc[1];
        if(vc[1] >= 1 && 
           (ptable[p]->vlist[1] = (vindex*)malloc(vc[1]*sizeof(vindex)))
            == NULL) {
            fprintf(Stderr, "\n*err* pvendtbl: memory allocation ");
            fprintf(Stderr, "error(3)\n");
            fprintf(Stderr, "size = %d\n", vc[1]*sizeof(vindex));
            exit(1);
        }
        for(v = 0; v < vc[1]; v++)
            ptable[p]->vlist[1][v] = vt[1][v];
    }
    for(l = 0; l < 2; l++) {
        free((AnyP) vt[l]);
    }
#if DBG
    erdbg("pvendtbl", "end", pvdbg);
#endif
}

/*==================================================================*/
Private bool    pvfort(name, len)
string  name;
int     len;
/*--------------------------------------------------------------------
    Check name is fortran name.
--------------------------------------------------------------------*/
{
    int i;

    if(strlen(name) > len) {
        fprintf(Stderr, "\n*err* pvfort: name excees %d characters.\n",
                len);
        fprintf(Stderr, "      name='%s'\n", name);
        return False;
    }
    if(len == 6 && !isalpha(name[0])) {
        fprintf(Stderr, "\n*err* pvfort: first char of name should ");
        fprintf(Stderr, "be alphabet.\n");
        return False;
    }
    for(i = 1; i < 6; i++) {
        if(name[i] == '\0')
            break;
        if(!isalnum(name[i])) {
            fprintf(Stderr, 
                    "\n*err* pvfort : %s contains illegal char",
                     name);
            fprintf(Stderr, "(not alphabet nor decimal digit\n");
            return False;
        }
    }
    for(i = 0; i < strlen(name); i++)
        name[i] = tolower(name[i]);
    return True;   
}
/*********************************************************************
  Output
*/
/*==================================================================*/
Public  Void    pvprint(fn)
string  fn;
/*--------------------------------------------------------------------
    Dump tables.
--------------------------------------------------------------------*/
{
    int    p, v, j;
    int    mode = 0;

    if(strcmp(fn, "stdout") == 0) {
        pvpf = stdout;
        mode = 1;
    } else if((pvpf = fopen(fn, "w")) == NULL) {
        fprintf(Stderr, "\n*err* pvprint : open err %s\n", fn);
        return;
    }

    /* table of coupling constatns */
    pvscale('%');
    fprintf(pvpf, "Order={%s", mordtbl[0]);
    for (v = 1; v < nmordtbl; v++)
        fprintf(pvpf, ",%s", mordtbl[v]);
    fprintf(pvpf, "};\n");

    fprintf(pvpf, "%% maxtl = %d, mintl = %d\n", maxtl, mintl);

    pvscale('-');

    /* table of particles */
    for(p = 1; p < nptable; p++) {

        if(mode)
                fprintf(pvpf, "%d ", p);

        /* names and print-names of particle and antiparticle */
        fprintf(pvpf, "Particle=%s[\"%s\"]; ",
                ntable[ptable[p]->pname[0]].name,
                ptable[p]->psym[0]);
        fprintf(pvpf, "Antiparticle=%s[\"%s\"];\n",
                ntable[ptable[p]->pname[1]].name,
                ptable[p]->psym[1]);

        /* mass and width */
        fprintf(pvpf, "  Mass=%s; Width=%s; ",
                ptable[p]->mass, ptable[p]->width);

        /* mvalue */
        if(ptable[p]->mvalue != NULL)
            fprintf(pvpf, "MValue=%s; ", ptable[p]->mvalue);

        /* wvalue */
        if(ptable[p]->wvalue != NULL)
            fprintf(pvpf, "WValue=%s; ", ptable[p]->wvalue);

        /* ptype */
        fprintf(pvpf, "PType=%s; ", pvnptype(p));

        /* electric charge */
        if(ptable[p]->charge3%3 == 0)
            fprintf(pvpf, "Charge=%d; ", (int) ptable[p]->charge3/3);
        else
            fprintf(pvpf, "Charge=%d/3; ", (int) ptable[p]->charge3);

        /* color */
        fprintf(pvpf, "Color=%d; ", ptable[p]->color);

        /* pcode */
        fprintf(pvpf, "PCode=%d; ", ptable[p]->pcode);

        /* kfcode */
        fprintf(pvpf, "KFCode=%d; ", ptable[p]->kfcode);

        /* massless */
        if(ptable[p]->massless)
            fprintf(pvpf, "Massless; ");
        else
            fprintf(pvpf, "Massive; ");

        /* end of the particle */
        fprintf(pvpf,"Pend;\n");
    }

    /* vertices */
    pvscale('-');
    for(v = nvstart; v < nvmodel; v++) {

        if(mode)
                fprintf(pvpf, "%d ", v);

        /* list of particle */
        fprintf(pvpf, "Vertex={%s", pvname(vtable[v]->vlegs[0]));
        for(p = 1; p < vtable[v]->nvlegs; p++)
            fprintf(pvpf, ", %s", pvname(vtable[v]->vlegs[p]));
        fprintf(pvpf,"}; ");

        /* coupling constants */
        for(p = 0; p < nmordtbl; p++)
            fprintf(pvpf, "%s=%d; ",
                    mordtbl[p], vtable[v]->morder[p]);

        /* fortran name */
        fprintf(pvpf, "FName=%s; ", vtable[v]->fname);

        /* type of fermion coupling */
        if(vtable[v]->ctype == V_V)
            fprintf(pvpf, "FType=\"V\"; ");
        else if(vtable[v]->ctype == V_A)
            fprintf(pvpf, "FType=\"A\"; ");
        else if(vtable[v]->ctype == V_VminusA)
            fprintf(pvpf, "FType=\"V-A\"; ");
        else if(vtable[v]->ctype == V_VplusA)
            fprintf(pvpf, "FType=\"V+A\"; ");

        /* value of coupling constants */
        if(vtable[v]->phase != V_PUNDEF) {
            if(vtable[v]->phase == V_REAL) {
                fprintf(pvpf, "FValue={R");
            } else if(vtable[v]->phase == V_IMAG) {
                fprintf(pvpf, "FValue={I");
            } else if(vtable[v]->phase == V_CMPLX) {
                fprintf(pvpf, "FValue={C");
            } else {
                fprintf(pvpf, "FValue={?");
            }
            for(j = 0; j < MAXFVAL; j++) {
                if(vtable[v]->fvalue[j] == NULL)
                    break;
                else
                    fprintf(pvpf, ",\"%s\"", vtable[v]->fvalue[j]);
            }
            fprintf(pvpf, "}; ");
        }

        /* end of the vertex */
        fprintf(pvpf,"Vend;\n");
        fprintf(pvpf,"%%vtype=%d;\n",vtable[v]->vtype);
    }

    /* end of the file */
    fprintf(pvpf,"Mend;\n");

    if(pvpf != stdout)
        fclose(pvpf);
}

/*==================================================================*/
Public  Void    pvprnam(fn)
string  fn;
/*--------------------------------------------------------------------
    Dump tables.
--------------------------------------------------------------------*/
{
    int    p, v;

    if(strcmp(fn, "stdout") == 0) {
        pvpf = stdout;
    } else if((pvpf = fopen(fn, "w")) == NULL) {
        fprintf(Stderr, "\n*err* pvprint : open err %s\n", fn);
        return;
    }

    /* table of coupling constatns */
    pvscale('%');

    /* table of particles */
    for(p = 1; p < nptable; p++) {

        /* mass and width */
        fprintf(pvpf, "  Mass=%-8s; Width=%-8s; ",
                ptable[p]->mass, ptable[p]->width);

        /* names and print-names of particle and antiparticle */
        fprintf(pvpf, "Particle=%s; \n",
                ntable[ptable[p]->pname[0]].name);

    }

    /* vertices */
    pvscale('-');
    for(v = nvstart; v < nvtable; v++) {

        /* fortran name */
        fprintf(pvpf, "FName=%-8s; ", vtable[v]->fname);

        /* list of particle */
        fprintf(pvpf, "Vertex={%s", pvname(vtable[v]->vlegs[0]));
        for(p = 1; p < vtable[v]->nvlegs; p++)
            fprintf(pvpf, ", %s", pvname(vtable[v]->vlegs[p]));
        fprintf(pvpf,"}; \n");

    }

    if(pvpf != stdout)
        fclose(pvpf);
}

/*==================================================================*/
Public  Void    pvprgnam(fn)
string  fn;
/*--------------------------------------------------------------------
    Dump gname table.
--------------------------------------------------------------------*/
{
    int    p;

    if(strcmp(fn, "stdout") == 0) {
        pvpf = stdout;
    } else if((pvpf = fopen(fn, "w")) == NULL) {
        fprintf(Stderr, "\n*err* pvprint : open err %s\n", fn);
        return;
    }

    /* table of particles */
    for(p = 1; p < nptable; p++) {

        /* mass and width */
        fprintf(pvpf, "%-20s %-20s %-20s %-20s\n",
                ntable[ptable[p]->pname[0]].name, ptable[p]->gname[0],
                ptable[p]->gname[1], ptable[p]->gname[2]);
        if(ptable[p]->pname[0] != ptable[p]->pname[1])
            fprintf(pvpf, "%-20s %-20s %-20s %-20s\n",
                    ntable[ptable[p]->pname[1]].name, ptable[p]->gname[0],
                    ptable[p]->gname[2], ptable[p]->gname[1]);
    }

    if(pvpf != stdout)
        fclose(pvpf);
}

/*==================================================================*/
Public  Void    pvprtex(fn)
string  fn;
/*--------------------------------------------------------------------
    Dump tables.
--------------------------------------------------------------------*/
{
    int n;

    if(strcmp(fn, "stdout") == 0) {
        pvpf = stdout;
    } else if((pvpf = fopen(fn, "w")) == NULL) {
        fprintf(Stderr, "\n*err* pvprtex : open err %s\n", fn);
        return;
    }

    fprintf(pvpf, "\\documentstyle[12pt]{article}\n");
    fprintf(pvpf, "\\begin{document} {\\small");

    n = pvprtxp();
    n = pvprtv1();
    n = pvprtv2(n);

    fprintf(pvpf, "}\\end{document}\n");

    if(pvpf != stdout)
        fclose(pvpf);
}

/*==================================================================*/
Private int     pvprtxp(NoArg)
/*--------------------------------------------------------------------
--------------------------------------------------------------------*/
{
    int    p, v;

    /* table of coupling constatns */
    pvscale('%');

    fprintf(pvpf, "{\\bf Filename : }\n");
    fprintf(pvpf, "\"%s\"\n\n", fnmdl);

    fprintf(pvpf, "{\\bf Order : }\n");
    for (v = 0; v < nmordtbl; v++)
        fprintf(pvpf, "%s ", mordtbl[v]);
    fprintf(pvpf, "\n\n");

    /* table of particles */
    pvscale('%');
    fprintf(pvpf, "{\\bf Particles}\n\n");
    fprintf(pvpf, "{\\footnotesize \n");

    /* page 1. */
    fprintf(pvpf, "\\begin{tabular}{|r");
    for(p = 1; p < 8; p++)
        if(p >= 4) {
            fprintf(pvpf, "|r");
        } else {
            fprintf(pvpf, "|l");
        }
    fprintf(pvpf, "|}\n\\hline\\hline\n");

    fprintf(pvpf, "1");
    for(p = 1; p < 8; p++)
        fprintf(pvpf, "& %d", p+1);
    fprintf(pvpf, "\\\\ \n\\hline\n");

    for(p = 1; p < nptable; p++) {

        /* 1 : seq. num. */
        fprintf(pvpf, " %d ", p);  /* 1 */

        /* 2 : name of particle */
        fprintf(pvpf, " & %s ", ntable[ptable[p]->pname[0]].name);

        /* 3 : name of anti-particle */
        fprintf(pvpf, " & %s ", ntable[ptable[p]->pname[1]].name);

        /* 4 : ptype */
        fprintf(pvpf, " & %s ", pvnptype(p));

        /* 5 : electric charge */
        if(ptable[p]->charge3%3 == 0)
            fprintf(pvpf, " & %d ", (int) ptable[p]->charge3/3);
        else
            fprintf(pvpf, " & %d/3 ", (int) ptable[p]->charge3);

        /* 6 : color */
        fprintf(pvpf, " & %d ", ptable[p]->color);

        /* 7 : pcode */
        fprintf(pvpf, " & %d ", ptable[p]->pcode);

        /* 8 : kfcode */
        fprintf(pvpf, " & %d ", ptable[p]->kfcode);

        /* end of the particle */
        fprintf(pvpf,"\\\\ \n");
    }
    fprintf(pvpf, "\\hline\\hline\n");
    fprintf(pvpf,"\\end{tabular}\n\n");

    /* page 2. */
    fprintf(pvpf, "\\begin{tabular}{|r|l");
    for(p = 8; p < 13; p++)
        fprintf(pvpf, "|l");
    fprintf(pvpf, "|}\n\\hline\\hline\n");

    fprintf(pvpf, "1 & 2");
    for(p = 8; p < 13; p++)
        fprintf(pvpf, "& %d", p+1);
    fprintf(pvpf, "\\\\ \n\\hline\n");

    for(p = 1; p < nptable; p++) {

        /* 1 : seq. num. */
        fprintf(pvpf, " %d ", p);  /* 1 */

        /* 2 : name of particle */
        fprintf(pvpf, " & %s ", ntable[ptable[p]->pname[0]].name);

        /* 9 : massless */
        if(ptable[p]->massless)
            fprintf(pvpf, " & massless ");
        else
            fprintf(pvpf, " & massive  ");

        /* 10 : mass */
        fprintf(pvpf, " & {\\tt %s} ", ptable[p]->mass);

        /* 11 : mvalue */
        if(ptable[p]->mvalue != NULL)
            fprintf(pvpf, " & {\\tt %s} ", ptable[p]->mvalue);
        else
            fprintf(pvpf, " & -- ");

        /* 12 : width */
        fprintf(pvpf, " & {\\tt %s} ", ptable[p]->width);

        /* 13 : wvalue */
        if(ptable[p]->wvalue != NULL)
            fprintf(pvpf, " & {\\tt %s} ", ptable[p]->wvalue);
        else
            fprintf(pvpf, " & -- ");

        /* end of the particle */
        fprintf(pvpf,"\\\\ \n");
    }
    fprintf(pvpf, "\\hline\\hline\n");
    fprintf(pvpf,"\\end{tabular}\n\n");
    fprintf(pvpf,"\\begin{itemize}\n");
    fprintf(pvpf,"\\item[ 1] No.\n");
    fprintf(pvpf,"\\item[ 2] Name of particle\n");
    fprintf(pvpf,"\\item[ 3] Name of anti-particle\n");
    fprintf(pvpf,"\\item[ 4] Type\n");
    fprintf(pvpf,"\\item[ 5] Charge\n");
    fprintf(pvpf,"\\item[ 6] Color\n");
    fprintf(pvpf,"\\item[ 7] Particle code\n");
    fprintf(pvpf,"\\item[ 8] KF Particle code\n");
    fprintf(pvpf,"\\item[ 9] Massive or massless\n");
    fprintf(pvpf,"\\item[10] Mass\n");
    fprintf(pvpf,"\\item[11] Value of mass\n");
    fprintf(pvpf,"\\item[12] Width\n");
    fprintf(pvpf,"\\item[13] Value of width\n");
    fprintf(pvpf,"\\end{itemize}\n");
    fprintf(pvpf,"} %% footnotsize\n");
    fprintf(pvpf,"\\newpage\n");

    return 12;
}

/*==================================================================*/
Private int     pvprtv1(NoArg)
/*--------------------------------------------------------------------
--------------------------------------------------------------------*/
{
    int    p, v;

    /* vertices */
    pvscale('%');
    fprintf(pvpf, "{\\bf Vertices}\n\n");
    fprintf(pvpf, "{\\footnotesize \n");

    for(v = nvstart; v < nvtable; v++) {

        if((v - nvstart) % 40 == 0) {
            if(v != nvstart) {
                fprintf(pvpf, "\\hline\\hline \n");
                fprintf(pvpf,"\\end{tabular}\n\n");
            }

            fprintf(pvpf, "\\begin{tabular}{|r");
            for(p = 1; p < 6+nmordtbl; p++)
                fprintf(pvpf, "|l");
            fprintf(pvpf, "|}\n\\hline\\hline\n");

            fprintf(pvpf, "1");
            for(p = 1; p < 6+nmordtbl; p++)
                fprintf(pvpf, "& %d", p+1);
            fprintf(pvpf, "\\\\ \n\\hline\n");
        }

        /* 1 : No. */
        fprintf(pvpf, "%d ", v);

        /* 2 - 5 : list of particle */
        for(p = 0; p < 4; p++)
            if(p < vtable[v]->nvlegs)
                fprintf(pvpf, " & %s ", pvname(vtable[v]->vlegs[p]));
            else
                fprintf(pvpf, " & -- ");

        /* 6 - (5+nmordtbl) : coupling constants*/
        for(p = 0; p < nmordtbl; p++)
            fprintf(pvpf, " & %d ", vtable[v]->morder[p]);

        /* 6+nmordtbl : fortran name */
        fprintf(pvpf, " & {\\tt %s} ", vtable[v]->fname);

        /* end of the vertex */
        fprintf(pvpf,"\\\\ \n");
    }
    fprintf(pvpf, "\\hline\\hline \n");
    fprintf(pvpf,"\\end{tabular}\n\n");
    fprintf(pvpf,"\\begin{itemize}\n");

    fprintf(pvpf,"\\item[ 1] No.\n");
    fprintf(pvpf,"\\item[ 2 -- 5] list of particle \n");
    for(p = 0; p < nmordtbl; p++) {
        fprintf(pvpf,"\\item[ %d] ", p+6);
        fprintf(pvpf," order of coupling constnat : {\\tt %s}\n",
                mordtbl[p]);
    }
    fprintf(pvpf,"\\item[%d] Name of vertex\n", p+6);

    fprintf(pvpf,"\\end{itemize}\n");
    fprintf(pvpf,"} %% footnotsize\n");

    fprintf(pvpf,"\\newpage\n");

    return p+6;
}


/*==================================================================*/
Private int     pvprtv2(offs)
int     offs;
/*--------------------------------------------------------------------
--------------------------------------------------------------------*/
{
    int    p, v, j;

    /* vertices */
    pvscale('%');
    fprintf(pvpf, "{\\bf Vaules of coupling const}\n\n");
    fprintf(pvpf, "{\\footnotesize \n");

    for(v = nvstart; v < nvtable; v++) {

        if((v - nvstart) % 40 == 0) {
            if(v != nvstart) {
                fprintf(pvpf, "\\hline\\hline \n");
                fprintf(pvpf,"\\end{tabular}\n\n");
            }

            fprintf(pvpf, "\\begin{tabular}{|r");
            for(p = offs; p < offs+3+MAXFVAL; p++)
                fprintf(pvpf, "|l");
            fprintf(pvpf, "|}\n\\hline\\hline\n");

            fprintf(pvpf, "1");
            for(p = offs; p < offs+3+MAXFVAL; p++)
                fprintf(pvpf, "& %d", p+1);
            fprintf(pvpf, "\\\\ \n\\hline\n");
        }

        /* 1 : No. */
        fprintf(pvpf, "%d ", v);

        /* 2 : fortran name */
        fprintf(pvpf, " & {\\tt %s} ", vtable[v]->fname);

        /* 3 : type of fermion coupling */
        if(vtable[v]->ctype == V_V)
            fprintf(pvpf, " & V ");
        else if(vtable[v]->ctype == V_A)
            fprintf(pvpf, " & A ");
        else if(vtable[v]->ctype == V_VminusA)
            fprintf(pvpf, " & V-A ");
        else if(vtable[v]->ctype == V_VplusA)
            fprintf(pvpf, " & V+A ");
        else
            fprintf(pvpf, " & -- ");

        /* 4 : phase of fermion coupling */
        if(vtable[v]->phase == V_REAL)
            fprintf(pvpf, " & R ");
        else if(vtable[v]->phase == V_IMAG)
            fprintf(pvpf, " & I ");
        else if(vtable[v]->phase == V_CMPLX)
            fprintf(pvpf, " & C ");
        else
            fprintf(pvpf, " & -- ");

        /* 5 - (4+MAXFVAL) : coupling const. */
        for(j = 0; j < MAXFVAL; j++) {
            if(vtable[v]->phase != V_PUNDEF 
               && vtable[v]->fvalue[j] != NULL)
                fprintf(pvpf, " & {\\tt %s} ", vtable[v]->fvalue[j]);
            else
                fprintf(pvpf, " & -- ", vtable[v]->fvalue[j]);
        }

        /* end of the vertex */
        fprintf(pvpf,"\\\\ \n");
    }
    fprintf(pvpf, "\\hline\\hline \n");
    fprintf(pvpf,"\\end{tabular}\n\n");
    fprintf(pvpf,"\\begin{itemize}\n");
    fprintf(pvpf,"\\item[ 1] No.\n");
    fprintf(pvpf,"\\item[ %d] Name of vertex\n", offs+1);
    fprintf(pvpf,"\\item[ %d] Type of fermion vertex\n", offs+2);
    fprintf(pvpf,"\\item[ %d] Phase of coupling constants\n",
            offs+3);
    fprintf(pvpf,"\\item[ %d -- %d] Value of coupling constants\n",
           offs+3, offs+3+MAXFVAL);
    p += MAXFVAL;
    fprintf(pvpf,"\\end{itemize}\n");
    fprintf(pvpf,"} %% footnotsize\n");

    return p;
}

/*==================================================================*/
Public   Void    pvldump(out, len, lst)
FILE    *out;
int      len;
vindex   lst[];
/*--------------------------------------------------------------------
    Dump tables.
--------------------------------------------------------------------*/
{
    int i;

    fprintf(out, "{");
    for(i = 0; i < len; i++)
        fprintf(out, " %d", lst[i]);
    fprintf(out,"}\n");
}

/*==================================================================*/
Public  Void    pvdump(NoArg)
/*--------------------------------------------------------------------
    Dump tables.
--------------------------------------------------------------------*/
{
    int    p, v, l, j;

    fprintf(Stderr, "----- pvdump -- ");
    /* table of coupling constatns */
    fprintf(Stderr, "Order={%s", mordtbl[0]);
    for (v = 1; v < nmordtbl; v++)
        fprintf(Stderr, ",%s", mordtbl[v]);
    fprintf(Stderr, "};\n");

    fprintf(Stderr, "%% maxtl = %d, mintl = %d\n", maxtl, mintl);

    /* table of particles */
    for(p = 1; p < nptable; p++) {

        fprintf(Stderr, "%d ", p);

        /* names and print-names of particle and antiparticle */
        fprintf(Stderr, "Particle=%s[\"%s\"]; ",
                ntable[ptable[p]->pname[0]].name,
                ptable[p]->psym[0]);
        fprintf(Stderr, "Antiparticle=%s[\"%s\"];\n",
                ntable[ptable[p]->pname[1]].name,
                ptable[p]->psym[1]);

        /* mass and width */
        fprintf(Stderr, "  Mass=%s; Width=%s; ",
                ptable[p]->mass,
                ptable[p]->width);

        /* mvalue */
        if(ptable[p]->mvalue != NULL)
            fprintf(Stderr, "MValue=%s; ", ptable[p]->mvalue);

        /* wvalue */
        if(ptable[p]->wvalue != NULL)
            fprintf(Stderr, "WValue=%s; ", ptable[p]->wvalue);

        /* ptype */
        fprintf(Stderr, "PType=%s; ", pvnptype(p));

        /* electric charge */
        fprintf(Stderr, "Charge3=%d; ", (int) ptable[p]->charge3);

        /* color */
        fprintf(Stderr, "Color=%d; ", ptable[p]->color);

        /* pcode */
        fprintf(Stderr, "PCode=%d; ", ptable[p]->pcode);

        /* kfcode */
        fprintf(Stderr, "KFCode=%d; ", ptable[p]->kfcode);

        /* massless */
        if(ptable[p]->massless)
            fprintf(Stderr, "massless; ");
        else
            fprintf(Stderr, "massive; ");

        /* furry */
        fprintf(Stderr, "Furry=%d; ", ptable[p]->furry);

        fprintf(Stderr, "\n");

        /* nvlist and vlist */
        for(l = 0; l < 2; l++) {
            fprintf(Stderr, "  nvl[%d] = %d ", l, ptable[p]->nvlist[l]);
            pvldump(Stderr, ptable[p]->nvlist[l],
                            ptable[p]->vlist[l]);
        }
    }

    /* vertices */
    for(v = nvstart; v < nvtable; v++) {
        fprintf(Stderr, "%d ", v);

        /* list of particle */
        fprintf(Stderr, "Vertex={%s", pvname(vtable[v]->vlegs[0]));
        for(p = 1; p < vtable[v]->nvlegs; p++)
            fprintf(Stderr, ", %s", pvname(vtable[v]->vlegs[p]));
        fprintf(Stderr,"}; ");

        /* coupling constants */
        for(p = 0; p < nmordtbl; p++)
            fprintf(Stderr, "%s=%d; ", mordtbl[p],
                    vtable[v]->morder[p]);

        /* fortran name */
        fprintf(Stderr, "FName=%s; ", vtable[v]->fname);

        /* type of fermion coupling */
        if(vtable[v]->ctype == V_V)
            fprintf(Stderr, "FType=\"V\"; ");
        else if(vtable[v]->ctype == V_A)
            fprintf(Stderr, "FType=\"A\"; ");
        else if(vtable[v]->ctype == V_VminusA)
            fprintf(Stderr, "FType=\"V-A\"; ");
        else if(vtable[v]->ctype == V_VplusA)
            fprintf(Stderr, "FType=\"V+A\"; ");

        /* value of coupling constants */
        if(vtable[v]->phase != V_PUNDEF) {
            if(vtable[v]->phase == V_REAL) {
                fprintf(Stderr, "FValue={R");
            } else if(vtable[v]->phase == V_IMAG) {
                fprintf(Stderr, "FValue={I");
            } else if(vtable[v]->phase == V_CMPLX) {
                fprintf(Stderr, "FValue={C");
            } else {
                fprintf(Stderr, "FValue={?");
            }
            for(j = 0; j < MAXFVAL; j++) {
                if(vtable[v]->fvalue[j] == NULL)
                    break;
                else
                    fprintf(Stderr, ",\"%s\"", vtable[v]->fvalue[j]);
            }
            fprintf(Stderr, "}; ");
        }

        fprintf(Stderr, "\n");
        fprintf(Stderr, "%%vtype=%d;\n",vtable[v]->vtype);
    }

    /* end of the file */
    fprintf(Stderr,"---- end of pvdump ---\n");
}

/*==================================================================*/
Private Void    pvscale(ch)
int     ch;
/*--------------------------------------------------------------------
   print "%-------- ... \n" when ch = '-'
--------------------------------------------------------------------*/
{
    int i;

    putc('%', pvpf);
    for (i = 1; i < 40; i++)
        putc(ch, pvpf);
    putc(EOL, pvpf);
}

/*==================================================================*/
Public  Void    pvhelp(fp)
FILE    *fp;
/*--------------------------------------------------------------------
    Print help
--------------------------------------------------------------------*/
{
    int i, synlen;

    synlen = sizeof(syntax)/sizeof(syntax[0]);
    for(i = 0; i < synlen; i++)
        fprintf(fp, "%s\n", syntax[i]);
}

/*==================================================================*/
Private string  pvstrdup(conv, str)
bool    conv;
string  str;
/*--------------------------------------------------------------------
    duplicate string.
--------------------------------------------------------------------*/
{
    string s, p;

    s = (string) malloc((strlen(str)+1)*sizeof(char));
    if(s == NULL) {
        fprintf(Stderr, "\n*err* pvstrcup: memory allocation error\n");
        fprintf(Stderr, "for %s\n", str);
        exit(1);
    }

    /* convert to lower case characters */
    /* strcpy(s, str); */
    p = s;
    while(*str != '\0')
        if(conv)
            *p++ = tolower(*str++);
        else
            *p++ = *str++;
    *p = *str;
    

    return s;
}

/*==================================================================*/
Public  int     pvboson(p)
pindex  p;
/*--------------------------------------------------------------------
 * returns spin*2 of the particle
--------------------------------------------------------------------*/
{
    return (pmask[ptable[p]->ptype] & M_BOSON);
}

/*==================================================================*/
Public  int     pvfermi(p)
pindex  p;
/*--------------------------------------------------------------------
 * returns spin*2 of the particle
--------------------------------------------------------------------*/
{
    return (pmask[ptable[p]->ptype] & M_FERMION);
}

/*==================================================================*/
Public  int     pvsdeg(p)
pindex  p;
/*--------------------------------------------------------------------
 * returns spin*2 of the particle
--------------------------------------------------------------------*/
{
    bool m;
    m = ptable[p]->massless;

    switch (ptable[p]->ptype) {
    case P_SCALAR  : return 1;
    case P_MAJORANA: return 1;
    case P_FERMION : if(m) return 1; else return 2;
    case P_GHOST   : return 1;
    case P_VECTOR  : if(m) return 2; else return 3;
    default   : return -1;
    }
}

/*==================================================================*/
Public  string   pvnptype(p)
pindex  p;
/*--------------------------------------------------------------------
 * returns the name of particle type
--------------------------------------------------------------------*/
{
    return pvtnam[ptable[p]->ptype];
}

/*==================================================================*/
Public  vindex   pvaddct(v, mord)
vindex  v;
int     mord[];
/*--------------------------------------------------------------------
 * add counter term
 * The vertex is the higher order of vertex "v" with coupling
 * constants "mord".
--------------------------------------------------------------------*/
{
    vindex vct;
    int    nl, to, l, lp;

    /* check the number of loops */
    nl = vtable[v]->nvlegs;
    to = 0;
    for(l = 0; l < nmordtbl; l++) {
        to += mord[l];
    }
    lp = to - nl + 2;
    if(lp%2 != 0 || lp < 1) {
        fprintf(Stderr, "\n*err*pvaddct : invalid counter term\n");
        fprintf(Stderr, "   order = %d, nleg = %d\n", to, nl);
        exit(1);
    }
    lp /= 2;

    vct = pvnenter(nl, ((string*)NULL));

    vtable[vct]->vtype  = VTGENCT;
    vtable[vct]->nvlegs = nl;
    for(l = 0; l < nl; l++)
        vtable[vct]->vlegs = vtable[v]->vlegs;

    for(l = 0; l < nmordtbl; l++)
        vtable[vct]->morder[l] = mord[l];
    vtable[vct]->loop   = lp;
    vtable[vct]->ctype  = vtable[v]->ctype;
    vtable[vct]->fname  = pvctname(vtable[v]->fname, mord, lp);
    vtable[vct]->phase  = V_CMPLX;
    for(l = 0; l < MAXFVAL; l++)
        vtable[vct]->fvalue[l] = NULL;

    return vct;
}

/*==================================================================*/
Private string   pvctname(name, mord, lp)
string  name;
int    *mord;
int     lp;
/*--------------------------------------------------------------------
 * construct name of counter term from tree vertex name.
 */
{
    static char nstr[512];
    string str;
#ifdef OLDNAME

    sprintf(nstr, "%d%s", lp, name);
#else
    int    i;

    sprintf(nstr, "%s", name);

    str = nstr + strlen(nstr);
    for(i = 0; i < nmordtbl; i++) {
        sprintf(str, "%1d", mord[i]);
        str += strlen(str);
    }
#endif
    str = pvstrdup(False, nstr);

    return str;
}

/*==================================================================*/
Private bool    pvfermiv(nv)
int     nv;
{
    int i, tp;

    for(i = 0; i < vtable[nv]->nvlegs; i++) {
        tp = ptable[abs(vtable[nv]->vlegs[i])]->ptype;
        if(tp == P_FERMION || tp == P_MAJORANA)
            return True;
    }
    return False;
}


/*==================================================================*/
Public  int     pvgl3(nv)
int     nv;
/*--------------------------------------------------------------------
 *  
 */
{
    int i, tp;

    if(ndlegs(nv) != 3)
        return 0;

    for(i = 0; i < ndlegs(nv); i++) {
        if(ptable[abs(ndparticle(nv, i))]->color != N8)
            return 0;
    }
    return 1;
}
/*==================================================================*/
Public  int     pvgl4(nv)
int     nv;
/*--------------------------------------------------------------------
 *  
 */
{
    int i, tp;

    if(ndlegs(nv) != 4)
        return 0;

    for(i = 0; i < ndlegs(nv); i++) {
        if(ptable[abs(ndparticle(nv, i))]->color != N8)
            return 0;
    }
    return 4;
}
